Package: grpnet
Type: Package
Title: Group Elastic Net Regularized GLMs and GAMs
Version: 1.2
Date: 2026-04-29
Authors@R: person(c("Nathaniel", "E."), "Helwig", email = "helwig@umn.edu", role = c("aut", "cre"))
Description: Efficient algorithms for fitting generalized linear and additive models with group elastic net penalties as described in Helwig (2025) <doi:10.1080/10618600.2024.2362232>. Implements group LASSO, group MCP, and group SCAD with an optional group ridge penalty. Computes the regularization path for linear regression (gaussian), multivariate regression (multigaussian), smoothed support vector machines (svm1), squared support vector machines (svm2), logistic regression (binomial), proportional odds logistic regression (ordinal), multinomial logistic regression (multinomial), log-linear count regression (poisson and negative.binomial), and log-linear continuous regression (gamma and inverse gaussian). Supports default and formula methods for model specification, k-fold cross-validation for tuning the regularization parameters, and nonparametric regression via tensor product reproducing kernel (smoothing spline) basis function expansion. 
License: GPL (>= 2)
Encoding: UTF-8
Depends: R (>= 3.5.0)
NeedsCompilation: yes
Packaged: 2026-04-29 11:07:15 UTC; nate
Author: Nathaniel E. Helwig [aut, cre]
Maintainer: Nathaniel E. Helwig <helwig@umn.edu>
Repository: CRAN
Date/Publication: 2026-04-29 12:00:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-04-29 15:33:37 UTC; unix
Archs: grpnet.so.dSYM

useDynLib(grpnet, .registration = TRUE)

export(cv.compare, cv.grpnet, family.grpnet, grpnet,
       rk.model.matrix, rk, row.kronecker, 
       visualize.loss, visualize.penalty, visualize.shrink)

S3method(coef, cv.grpnet)
S3method(coef, grpnet)

S3method(cv.grpnet, default)
S3method(cv.grpnet, formula)

S3method(grpnet, default)
S3method(grpnet, formula)

S3method(plot, cv.grpnet)
S3method(plot, grpnet)

S3method(predict, cv.grpnet)
S3method(predict, grpnet)

S3method(print, coef.grpnet)
S3method(print, cv.grpnet)
S3method(print, grpnet)
S3method(print, summary.grpnet)
S3method(print, summary.cv.grpnet)

S3method(summary, cv.grpnet)
S3method(summary, grpnet)

importFrom("grDevices", "extendrange")
importFrom("graphics", "abline", "axis", "box", "legend", "lines", "mtext",
           "points", "segments")
importFrom("stats", "approx", "as.formula", "binomial", "coef", "gaussian",
           "glm", "model.matrix", "model.response", "poisson", "predict", 
           "quantile", "sd", "terms")
importFrom("parallel", "detectCores", "makeCluster", "parSapply", "stopCluster")
importFrom("utils", "packageVersion", "setTxtProgressBar", "txtProgressBar")

!   grpnet_ordinal.f90 - group elastic net (ordinal / polr)
!   Nathaniel E. Helwig (helwig@umn.edu)
!   Department of Psychology and School of Statistics
!   University of Minnesota
!   Date: 2025-12-12


! INPUTS/OUTPUTS
!   nobs = number of observations (N)
!   nvars = number of variables (P)
!   nresp = number of response classes for y (M)
!   x = grouped predictor matrix (N,P)
!       note: (x.1, x.2, ..., x.K) where x.k is (N,P.k) with sum(P.k) = P
!   y = response vector with entries in {1,...,M} (N)
!   w = observation weight vector (N)
!       note: if min(w) = max(w), then weights are ignored
!   off = offset vector (N)
!   ngrps = number of groups (K)
!   gsize = number of coefs in each group (K)
!           note: (P.1, P.2, ..., P.K) with sum(P.k) = P
!   pw = penalty weight vector (K)   (outputs xsdev)
!        note: set pw(k) = 0 to leave k-th group unpenalized
!   alpha = weight for L1 and L2 penalities (alpha = 1 for lasso/mcp/scad, alpha = 0 for ridge)
!           note: setting alpha between (0,1) gives elastic net
!   nlam = number of lambdas
!   lambda = sequence of lambda values
!            note: if max(lambda) <=0, then lambda.max is data dependent
!   lmr = lambda minimum ratio (lambda.min = lmr * lambda.max)
!         note: unless lambda is provided, lambda.max is data dependent
!   penid = penalty id: 1 = lasso, 2 = mcp, 3 = scad
!   gamma = additional hyper-parameter for mcd and scad penalities
!           note: gamma > 1 for mcp and gamma > 2 for scad
!   eps = convergence tolerance
!   maxit = maximum number of iterations
!   standardize = integer (0: no standardization, 1: yes standardization)
!   intercept = integer (0: no intercept, 1: yes intercept)
!   ibeta = output matrix of intercepts (nresp-1,nlam)
!   betas = output matrix of coefficients (nvars,nlam)
!   iters = output vector of iterations for each lambda (nlam)
!   nzgrps = number of non-zero groups for each lambda (nlam)
!   nzcoef = number of non-zero coefficients for each lambda (nlam)
!   edfs = effective degrees of freedom for each lambda (nlam)
!   devs = residual deviance for each lambda (nlam)
!   nulldev = null deviance


SUBROUTINE grpnet_ordinal(nobs, nvars, nresp, x, y, w, off, ngrps, gsize, pw, &
                          alpha, nlam, lambda, lmr, penid, gamma, eps, maxit, &
                          standardize, intercept, ibeta, betas, iters, &
                          nzgrps, nzcoef, edfs, devs, nulldev)

    IMPLICIT NONE

! --------------- ARGUMENTS --------------- !
    INTEGER nobs, nvars, nresp, y(nobs), ngrps, gsize(ngrps), nlam, penid, maxit
    INTEGER standardize, intercept, iters(nlam), nzgrps(nlam), nzcoef(nlam)
    DOUBLE PRECISION x(nobs, nvars), w(nobs), off(nobs), pw(ngrps)
    DOUBLE PRECISION alpha, lambda(nlam), lmr, gamma, eps, ibeta(nresp-1,nlam)
    DOUBLE PRECISION betas(nvars, nlam), edfs(nlam), devs(nlam), nulldev
! --------------- ARGUMENTS --------------- !


! --------------- LOCAL DEFINITIONS --------------- !
    INTEGER i, j, k, l, iter, violations, makelambda, weighted
    INTEGER active(ngrps), strong(ngrps), ia(ngrps), ib(ngrps), gid
    INTEGER yfreq(nresp)
    DOUBLE PRECISION ycumsum(nresp)
    DOUBLE PRECISION wmin, wmax, rnglam, minlam, maxlam, maxdif, macheps
    DOUBLE PRECISION r(nobs), difbeta(nvars), beta(nvars), zvec(nvars)
    DOUBLE PRECISION grad(nvars), gradnorm(ngrps), ctol, shrink, twolam
    DOUBLE PRECISION penone, pentwo, xmean(nvars), xsdev(ngrps), xev(ngrps)
    DOUBLE PRECISION znorm, bnorm, eta(nobs), mu(nobs), yprob(nresp-1)
    DOUBLE PRECISION shift(nresp-1), bigtheta(nresp+1), fit0(nobs)
    DOUBLE PRECISION fit1(nobs), Pi0(nobs), Pi1(nobs), xmb(nlam)
    DOUBLE PRECISION, ALLOCATABLE :: xtx(:,:)
! --------------- LOCAL DEFINITIONS --------------- !


! --------------- MACHINE EPSILON --------------- !
    macheps = EPSILON(eps)
! --------------- MACHINE EPSILON --------------- !


! --------------- CHECK WEIGHTS --------------- !
    wmin = MINVAL(w)
    wmax = MAXVAL(w)
    IF (wmax > wmin) THEN
        weighted = 1
        w = nobs * w / SUM(w)     ! normalize so SUM(w) = nobs
        w = SQRT(w)
        DO i=1,nobs
            x(i,:) = w(i) * x(i,:)
        END DO
    ELSE
        weighted = 0
        w = 1.0D0
    END IF
! --------------- CHECK WEIGHTS --------------- !


! --------------- GROUP INDICES --------------- !
    gid = 0
    DO k=1,ngrps
        ia(k) = gid + 1
        gid = gid + gsize(k)
        ib(k) = gid
    END DO
! --------------- GROUP INDICES --------------- !


! --------------- CENTER AND SCALE --------------- !
    IF (intercept == 1) THEN
        DO j=1,nvars
            xmean(j) = SUM(x(:,j)) / nobs
            x(:,j) = x(:,j) - xmean(j)
        END DO
    END IF
    xsdev = 1.0D0
    IF (standardize == 1) THEN
        DO k=1,ngrps
            xsdev(k) = SQRT( SUM(x(:, ia(k):ib(k))**2) / (nobs * gsize(k)) )
            IF (xsdev(k) > macheps) THEN
                x(:,ia(k):ib(k)) = x(:,ia(k):ib(k)) / xsdev(k)
            ELSE
                xsdev(k) = 1.0D0
            END IF
        END DO
    END IF
! --------------- CENTER AND SCALE --------------- !


! --------------- GET MAX EIGENVALUE --------------- !
    DO k=1,ngrps
        IF (gsize(k) == 1) THEN
            xev(k) = SUM(x(:,ia(k))**2) / nobs
        ELSE
            ALLOCATE(xtx(gsize(k), gsize(k)))
            xtx = MATMUL(TRANSPOSE(x(:,ia(k):ib(k))), x(:,ia(k):ib(k))) / nobs
            CALL grpnet_maxeigval(xtx, gsize(k), xev(k))
            DEALLOCATE(xtx)
        END IF
    END DO
    xev = xev / 2.0D0
! --------------- GET MAX EIGENVALUE --------------- !


! --------------- MISCELLANEOUS INITIALIZATIONS --------------- !
    maxlam = MAXVAL(lambda)
    makelambda = 0
    iter = 0
    strong = 0
    active = 0
    nzgrps = 0
    nzcoef = 0
    zvec = 0.0D0
    ibeta = 0.0D0
    beta = 0.0D0
    grad = 0.0D0
    gradnorm = 0.0D0
    twolam = 0.0D0
    devs = 0.0D0
    eta = off
    maxdif = 0.0D0
! --------------- MISCELLANEOUS INITIALIZATIONS --------------- !


! --------------- SHIFT INITIALIZATION --------------- !
    yfreq = 0
    ycumsum = 0
    yfreq(1) = COUNT(y == 1)
    ycumsum(1) = yfreq(1) + 0.0D0
    DO k=2,nresp
        yfreq(k) = COUNT(y == k)
        ycumsum(k) = ycumsum(k-1) + yfreq(k)
    END DO
    yprob = 1.0D0 - (ycumsum(1:(nresp-1)) / nobs)
    DO k=1,nresp-1
        IF (yprob(k) < 1.0E-6) THEN
            yprob(k) = 1.0E-6
        ELSE IF (yprob(k) > (1.0D0 - 1.0E-6)) THEN
            yprob(k) = 1.0D0 - 1.0E-6
        END IF
    END DO
    shift = LOG(yprob / (1.0D0 - yprob))
! --------------- SHIFT INITIALIZATION --------------- !


! --------------- RESID INITIALIZATION --------------- !
    bigtheta(1) = 100.0D0
    bigtheta(2:nresp) = shift
    bigtheta(nresp+1) = -100.0D0
    fit0 = eta + bigtheta(y)
    fit1 = eta + bigtheta(y+1)
    Pi0 = 1.0D0 / (1.0D0 + EXP(-fit0))
    Pi1 = 1.0D0 / (1.0D0 + EXP(-fit1))
    r = w * (1.0D0 - Pi0 - Pi1)
! --------------- RESID INITIALIZATION --------------- !


! --------------- GENERATE LAMBDA --------------- !
    IF (maxlam <= macheps) THEN

        makelambda = 1
        i = 1

        ! find unpenalized groups !
        DO k=1,ngrps
            IF (pw(k) <= macheps) THEN
                active(k) = 1
                nzgrps(i) = nzgrps(i) + 1
                nzcoef(i) = nzcoef(i) + gsize(k)
            END IF
        END DO
        ! find unpenalized groups !

        ! iterate until active coefficients converge !
        IF (nzgrps(i) > 0) THEN
            DO WHILE(iter < maxit)

                ! update iter and reset counters
                ctol = 0.0D0
                iter = iter + 1

                ! update shifts
                CALL grpnet_ordinal_int(nobs, nresp, eta, y, w**2, yfreq, shift, maxdif)
                ctol = MAX(maxdif, ctol)

                ! define r
                bigtheta(1) = 100.0D0
                bigtheta(2:nresp) = shift
                bigtheta(nresp+1) = -100.0D0
                fit0 = eta + bigtheta(y)
                fit1 = eta + bigtheta(y+1)
                Pi0 = 1.0D0 / (1.0D0 + EXP(-fit0))
                Pi1 = 1.0D0 / (1.0D0 + EXP(-fit1))
                r = w * (1.0D0 - Pi0 - Pi1)

                ! update active groups
                DO k=1,ngrps
                    IF(active(k) == 0) CYCLE
                    grad(ia(k):ib(k)) = MATMUL(r, x(:,ia(k):ib(k))) / nobs
                    zvec(ia(k):ib(k)) = beta(ia(k):ib(k)) + grad(ia(k):ib(k)) / xev(k)
                    difbeta(ia(k):ib(k)) = zvec(ia(k):ib(k)) - beta(ia(k):ib(k))
                    maxdif = MAXVAL( ABS(difbeta(ia(k):ib(k))) / (1.0D0 + ABS(beta(ia(k):ib(k)))) )
                    beta(ia(k):ib(k)) = beta(ia(k):ib(k)) + difbeta(ia(k):ib(k))
                    eta = eta + MATMUL(x(:,ia(k):ib(k)), difbeta(ia(k):ib(k))) / w
                    fit0 = eta + bigtheta(y)
                    fit1 = eta + bigtheta(y+1)
                    Pi0 = 1.0D0 / (1.0D0 + EXP(-fit0))
                    Pi1 = 1.0D0 / (1.0D0 + EXP(-fit1))
                    r = w * (1.0D0 - Pi0 - Pi1)
                    ctol = MAX(maxdif , ctol)
                END DO ! k=1,ngrps

                ! convergence check
                IF(ctol < eps) EXIT

            END DO ! WHILE(iter < maxit)

        ELSE

            ! intercept only
            DO WHILE(iter < maxit)
                ctol = 0.0D0
                iter = iter + 1
                CALL grpnet_ordinal_int(nobs, nresp, eta, y, w**2, yfreq, shift, maxdif)
                ctol = MAX(maxdif, ctol)
                IF(ctol < eps) EXIT
            END DO ! WHILE(iter < maxit)

            ! define r
            bigtheta(1) = 100.0D0
            bigtheta(2:nresp) = shift
            bigtheta(nresp+1) = -100.0D0
            fit0 = eta + bigtheta(y)
            fit1 = eta + bigtheta(y+1)
            Pi0 = 1.0D0 / (1.0D0 + EXP(-fit0))
            Pi1 = 1.0D0 / (1.0D0 + EXP(-fit1))
            r = w * (1.0D0 - Pi0 - Pi1)

        END IF ! (nzgrps(1) > 0)
        ! iterate until active coefficients converge !

        ! create lambda sequence !
        DO k=1,ngrps
            IF (pw(k) > macheps) THEN
                grad(ia(k):ib(k)) = MATMUL(r, x(:,ia(k):ib(k))) / nobs
                gradnorm(k) = SQRT(SUM(grad(ia(k):ib(k))**2)) / pw(k)
            END IF
        END DO
        IF (alpha > macheps) THEN
            maxlam = MAXVAL(gradnorm / alpha)
        ELSE
            maxlam = MAXVAL(gradnorm / 1.0E-3)
            makelambda = 0
        END IF
        minlam = lmr * maxlam
        lambda(i) = maxlam
        maxlam = LOG(maxlam)
        minlam = LOG(minlam)
        rnglam = maxlam - minlam
        DO k=2,nlam
            lambda(k) = EXP(maxlam - rnglam * (k - 1) / (nlam - 1))
        END DO
        ! create lambda sequence !

        ! calculate deviance !
        mu = Pi0 - Pi1
        devs(i) = -2.0D0 * SUM(w**2 * LOG(mu))
        ! calculate deviance !

        ! save results !
        ibeta(:,i) = shift
        betas(:,i) = beta
        iters(i) = iter
        nzgrps(i) = nzgrps(i) + intercept
        nzcoef(i) = nzcoef(i) + intercept * (nresp-1)
        edfs(i) = DBLE(nzcoef(i))
        ! save results !

    END IF
! --------------- GENERATE LAMBDA --------------- !


! --------------- ITERATIVE WORK --------------- !
    DO i=1,nlam

        ! initializations !
        IF (i == 1 .AND. makelambda == 1) CYCLE
        IF (i > 1) THEN
            shift = ibeta(:,i-1)
            beta = betas(:,i-1)
            twolam = alpha * (2.0D0 * lambda(i) - lambda(i-1))
        ELSE
            grad = MATMUL(r, x) / nobs
        END IF
        ! initializations !

        ! strong rule initialization !
        DO k=1,ngrps
            gradnorm(k) = SQRT(SUM(grad(ia(k):ib(k))**2))
            IF (gradnorm(k) + 1.0E-8 > pw(k) * twolam) THEN
                strong(k) = 1
            ELSE
                strong(k) = 0
            END IF
        END DO
        ! strong rule initialization !

        ! iterate until strong set converges !
        iter = 0
        DO WHILE(iter < maxit)

            ! iterate until active set converges !
            DO WHILE(iter < maxit)

                ! iterate until active coefficients converge !
                DO WHILE(iter < maxit)

                    ! update iter and reset counters
                    iter = iter + 1
                    nzgrps(i) = 0
                    edfs(i) = 0.0D0
                    ctol = 0.0D0

                    ! update shifts
                    CALL grpnet_ordinal_int(nobs, nresp, eta, y, w**2, yfreq, shift, maxdif)
                    ctol = MAX(maxdif, ctol)

                    ! define r
                    bigtheta(1) = 100.0D0
                    bigtheta(2:nresp) = shift
                    bigtheta(nresp+1) = -100.0D0
                    fit0 = eta + bigtheta(y)
                    fit1 = eta + bigtheta(y+1)
                    Pi0 = 1.0D0 / (1.0D0 + EXP(-fit0))
                    Pi1 = 1.0D0 / (1.0D0 + EXP(-fit1))
                    r = w * (1.0D0 - Pi0 - Pi1)

                    ! update active groups
                    DO k=1,ngrps
                        IF(active(k) == 0) CYCLE
                        penone = alpha * lambda(i) * pw(k) / xev(k)
                        pentwo = (1.0D0 - alpha) * lambda(i) * pw(k) / xev(k)
                        grad(ia(k):ib(k)) = MATMUL(r, x(:,ia(k):ib(k))) / nobs
                        zvec(ia(k):ib(k)) = beta(ia(k):ib(k)) + grad(ia(k):ib(k)) / xev(k)
                        znorm = SQRT(SUM(zvec(ia(k):ib(k))**2))
                        bnorm = SQRT(SUM(beta(ia(k):ib(k))**2))
                        CALL grpnet_penalty(znorm, penid, penone, pentwo, gamma, shrink)
                        IF(shrink == 0.0D0 .AND. bnorm == 0.0D0) CYCLE
                        difbeta(ia(k):ib(k)) = shrink * zvec(ia(k):ib(k)) - beta(ia(k):ib(k))
                        maxdif = MAXVAL( ABS(difbeta(ia(k):ib(k))) / (1.0D0 + ABS(beta(ia(k):ib(k)))) )
                        beta(ia(k):ib(k)) = beta(ia(k):ib(k)) + difbeta(ia(k):ib(k))
                        eta = eta + MATMUL(x(:,ia(k):ib(k)), difbeta(ia(k):ib(k))) / w
                        fit0 = eta + bigtheta(y)
                        fit1 = eta + bigtheta(y+1)
                        Pi0 = 1.0D0 / (1.0D0 + EXP(-fit0))
                        Pi1 = 1.0D0 / (1.0D0 + EXP(-fit1))
                        r = w * (1.0D0 - Pi0 - Pi1)
                        ctol = MAX(maxdif , ctol)
                        IF(shrink > 0.0D0) THEN
                            nzgrps(i) = nzgrps(i) + 1
                            edfs(i) = edfs(i) + gsize(k) * shrink
                        ENDIF
                    END DO ! k=1,ngrps

                    ! convergence check
                    IF(ctol < eps) EXIT

                END DO ! WHILE(iter < maxit) - inner
                ! iterate until active coefficients converge !

                ! check inactive groups in strong set !
                violations = 0
                DO k=1,ngrps
                    IF(strong(k) == 0 .OR. active(k) == 1) CYCLE
                    grad(ia(k):ib(k)) = MATMUL(r, x(:,ia(k):ib(k))) / nobs
                    gradnorm(k) = SQRT(SUM(grad(ia(k):ib(k))**2))
                    IF (gradnorm(k) > alpha * lambda(i) * pw(k)) THEN
                        active(k) = 1
                        violations = violations + 1
                    END IF
                END DO ! k=1,ngrps
                IF(violations == 0) EXIT
                ! check inactive groups in strong set !

            END DO ! WHILE(iter < maxit) - middle
            ! iterate until active set converges !

            ! check groups in weak set !
            violations = 0
            DO k=1,ngrps
                IF (strong(k) == 1) CYCLE
                grad(ia(k):ib(k)) = MATMUL(r, x(:,ia(k):ib(k))) / nobs
                gradnorm(k) = SQRT(SUM(grad(ia(k):ib(k))**2))
                IF (gradnorm(k) + 1e-8 > alpha * lambda(i) * pw(k)) THEN
                    strong(k) = 1
                    active(k) = 1
                    violations = violations + 1
                END IF
            END DO
            IF(violations == 0) EXIT
            ! check groups in weak set !

        END DO ! WHILE(iter < maxit) - outer
        ! iterate until strong set converges !

        ! calculate nzcoef !
        DO k=1,ngrps
            IF(active(k) == 0) CYCLE
            DO l=1,gsize(k)
                IF (ABS(beta(ia(k) + l - 1)) > macheps) THEN
                    nzcoef(i) = nzcoef(i) + 1
                END IF
            END DO
        END DO
        ! calculate nzcoef !

        ! calculate deviance !
        mu = Pi0 - Pi1
        devs(i) = -2.0D0 * SUM(w**2 * LOG(mu))
        ! calculate deviance !

        ! save results !
        ibeta(:,i) = shift
        betas(:,i) = beta
        iters(i) = iter
        nzgrps(i) = nzgrps(i) + intercept
        nzcoef(i) = nzcoef(i) + intercept
        edfs(i) = edfs(i) + DBLE(intercept)
        ! save results !

    END DO
! --------------- ITERATIVE WORK --------------- !


! --------------- POST PROCESSING --------------- !
    IF (standardize == 1) THEN
        DO k=1,ngrps
            betas(ia(k):ib(k),:) = betas(ia(k):ib(k),:) / xsdev(k)
        END DO
    END IF
    IF (intercept == 1) THEN
        xmb = MATMUL(xmean, betas)
        DO i=1,nlam
            ibeta(:,i) = ibeta(:,i) - xmb(i)
        END DO
    END IF
    IF(makelambda == 1 .AND. nzgrps(1) == 0) THEN
        shift = ibeta(:,1)
    ELSE
        shift = log(yprob / (1 - yprob))
        eta = 0.0D0
        DO WHILE(iter < maxit)
            ctol = 0.0D0
            CALL grpnet_ordinal_int(nobs, nresp, eta, y, w**2, yfreq, shift, maxdif)
            ctol = MAX(maxdif, ctol)
            IF(ctol < eps) EXIT
        END DO ! WHILE(iter < maxit)
    END IF
    bigtheta(1) = 100.0D0
    bigtheta(2:nresp) = shift
    bigtheta(nresp+1) = -100.0D0
    fit0 = bigtheta(y)
    fit1 = bigtheta(y+1)
    Pi0 = 1.0D0 / (1.0D0 + EXP(-fit0))
    Pi1 = 1.0D0 / (1.0D0 + EXP(-fit1))
    mu = Pi0 - Pi1
    nulldev = -2.0D0 * SUM(w**2 * LOG(mu))
    pw = xsdev
! --------------- POST PROCESSING --------------- !

END SUBROUTINE


SUBROUTINE grpnet_ordinal_int(nobs, nresp, eta, y, w, yfreq, theta, maxdif)
    IMPLICIT NONE

! --------------- ARGUMENTS --------------- !
    INTEGER nobs, nresp, y(nobs), yfreq(nresp)
    DOUBLE PRECISION eta(nobs), w(nobs), theta(nresp-1), maxdif
! --------------- ARGUMENTS --------------- !

! --------------- LOCAL DEFINITIONS --------------- !
    INTEGER i, j, yj(nobs), yjm1(nobs), n
    INTEGER index(nresp-1), umin(nresp-1), vmax(nresp-1)
    DOUBLE PRECISION bigtheta(nresp+1), grad(nresp-1), avec(nresp-1)
    DOUBLE PRECISION bvec(nresp-2), fit0(nobs), fit1(nobs), mu(nobs)
    DOUBLE PRECISION Pi0(nobs), Pi1(nobs), u0(nobs), u1(nobs)
    DOUBLE PRECISION dtop(nresp), dbot(nresp), dtheta(nresp)
    DOUBLE PRECISION vj(nobs), va(nobs), vb(nobs), ivec(nresp-1)
    DOUBLE PRECISION u(nresp-1), v(nresp-1), d(nresp-1), delta(nresp-1)
! --------------- LOCAL DEFINITIONS --------------- !

! --------------- GRADIENT AND HESSIAN --------------- !
    ! initialize bigtheta
    bigtheta(1) = 100.0D0
    bigtheta(2:nresp) = theta
    bigtheta(nresp+1) = -100.0D0

    ! initialize grad, avec, and bvec
    grad = 0.0D0
    avec = 0.0D0
    bvec = 0.0D0

    ! define shift + linear predictor (for y and y+1)
    fit0 = eta + bigtheta(y)
    fit1 = eta + bigtheta(y+1)

    ! evaluate gradient / hessian components
    Pi0 = 1.0D0 / (1.0D0 + EXP(-fit0))
    Pi1 = 1.0D0 / (1.0D0 + EXP(-fit1))
    u0 = Pi0 * (1.0D0 - Pi0)
    u1 = Pi1 * (1.0D0 - Pi1)
    mu = Pi0 - Pi1

    ! update bvec (off-diagonals)
    dtop = EXP(bigtheta(1:nresp) + bigtheta(2:(nresp+1)))
    dbot = ( EXP(bigtheta(1:nresp)) - EXP(bigtheta(2:(nresp+1))) )**2
    dtheta = dtop / dbot
    bvec = (yfreq(2:(nresp-1)) * dtheta(2:(nresp-1))) / nobs
    
    ! update other shift gradient and hessian components
    DO i=1,nresp-1
        j = i + 1
        yj = MERGE(1, 0, y == j)
        yjm1 = MERGE(1, 0, y == (j-1))
        grad(i) = SUM( w * (u0 * yj - u1 * yjm1) / mu ) / nobs
        vj = 1.0D0 / (1.0D0 + EXP(-(eta + bigtheta(j))))
        vj = (1.0D0 - vj)**2 - (1.0D0 - vj)
        va = vj - dtheta(j-1)
        vb = vj - dtheta(j)
        avec(i) = -SUM( w * (yjm1 * va + yj * vb) ) / nobs
    END DO
! --------------- GRADIENT AND HESSIAN --------------- !

! --------------- INVSYMTRIDIAG --------------- !
    ! initializations
    n = nresp - 1
    u = 0.0D0
    v = 0.0D0
    d = 0.0D0
    delta = 0.0D0

    ! update d
    d(n) = avec(n)
    DO i=n-1,1,-1
        d(i) = avec(i) - bvec(i)**2 / d(i+1)
    END DO

    ! update v
    v(1) = 1 / d(1)
    DO i=2,n
        v(i) = PRODUCT(bvec(1:(i-1)) / d(1:(i-1))) / d(i)
    END DO

    ! update delta
    delta(1) = avec(1)
    DO i=1,n-1
        delta(i+1) = avec(i+1) - bvec(i)**2 / delta(i)
    END DO

    ! update u
    u(n) = 1 / (delta(n) * v(n))
    DO i=1,n-1
        u(i) = PRODUCT(bvec(i:(n-1)) / delta(i:(n-1))) / (delta(n) * v(n))
    END DO
! --------------- INVSYMTRIDIAG --------------- !

! --------------- NEWTON-RAPHSON UPDATE --------------- !
    ! calculate MATMUL(solve(hess), grad)
    index = (/ (i, i=1,n) /)
    ivec = 0.0D0
    DO i=1,n
        umin = index
        vmax = index
        umin(i:n) = i
        vmax(1:i) = i
        ivec(i) = SUM( (u(umin) * v(vmax)) * grad )
    END DO

    ! update maxdif and theta
    maxdif = MAXVAL( ABS(ivec) / (1.0D0 + ABS(theta)) )
    theta = theta + ivec
! --------------- NEWTON-RAPHSON UPDATE --------------- !
    
END SUBROUTINE

visualize.penalty <- 
  function(x = seq(-5, 5, length.out = 1001), 
           penalty = c("LASSO", "MCP", "SCAD"), 
           alpha = 1, 
           lambda = 1, 
           gamma = 4, 
           derivative = FALSE,
           plot = TRUE,
           subtitle = TRUE,
           legend = TRUE,
           location = ifelse(derivative, "bottom", "top"),
           ...){
    # plot grpnet penalties
    # Nathaniel E. Helwig (helwig@umn.edu)
    # 2025-06-03
           
    
    
    #########***#########   INITIAL CHECKS   #########***#########
    
    # check x
    xorig <- sort(as.numeric(x))
    x <- abs(xorig)
    nobs <- length(x)
    
    # check penalty
    penalties <- c("LASSO", "MCP", "SCAD")
    penalty <- sort(unique(as.character(penalty)))
    pid <- pmatch(penalty, penalties)
    if(any(is.na(pid))) stop(paste("Input 'penalty' is invalid. Must be some subset of\n", 
                                   paste(penalties, collapse = ", ")))
    penalty <- penalties[pid]
    npen <- length(penalty)
    
    # check alpha
    alpha <- as.numeric(alpha[1])
    if(alpha < 0 | alpha > 1) stop("Input 'alpha' must satisfy:  0 <= alpha <= 1")
    
    # check lambda
    lambda <- as.numeric(lambda[1])
    if(lambda < 0) stop("Input 'lambda' must satisfy:  lambda >= 0")
    
    # check gamma
    gamma <- as.numeric(gamma[1])
    if(any(penalty == "SCAD")){
      if(gamma <= 2) stop("Input 'gamma' must satisfy: gamma > 2 (SCAD)")
    } else if(any(penalty == "MCP")){
      if(gamma <= 1) stop("Input 'gamma' must satisfy: gamma > 1 (MCP)")
    } 
    
    # check derivative
    derivative <- as.logical(derivative[1])
    if(!any(derivative == c(TRUE, FALSE))) stop("Input 'derivative' must be TRUE/FALSE")
    
    
    
    #########***#########   EVALUATE PENALTY   #########***#########
    
    # initialize matrix to hold results
    res <- matrix(0.0, nrow = nobs, ncol = npen)
    colnames(res) <- penalty
    
    # loop through penalties
    for(j in 1:npen){
      
      if(derivative){
        if(penalty[j] == "LASSO"){
          res[,j] <- lambda
        } else if (penalty[j] == "MCP"){
          id <- (x <= gamma * lambda)
          res[id,j] <- lambda - x[id] / gamma
          res[!id,j] <- 0
        } else if (penalty[j] == "SCAD"){
          id <- which(x <= lambda)
          if(length(id) > 0) res[id,j] <- lambda
          id <- which((x > lambda) & (x <= gamma * lambda))
          if(length(id) > 0) res[id,j] <- (gamma * lambda - x[id]) / (gamma - 1)
          id <- which(x >= gamma * lambda)
          if(length(id) > 0) res[id,j] <- 0
        }
      } else {
        if(penalty[j] == "LASSO"){
          res[,j] <- lambda * x
        } else if (penalty[j] == "MCP"){
          id <- (x <= gamma * lambda)
          res[id,j] <- lambda * x[id] - x[id]^2 / (2 * gamma)
          res[!id,j] <- gamma * lambda^2 / 2
        } else if (penalty[j] == "SCAD"){
          id <- which(x <= lambda)
          if(length(id) > 0) res[id,j] <- lambda * x[id]
          id <- which((x > lambda) & (x <= gamma * lambda))
          if(length(id) > 0) res[id,j] <- (gamma * lambda * x[id] - 0.5 * (x[id]^2 + lambda^2)) / (gamma - 1)
          id <- which(x >= gamma * lambda)
          if(length(id) > 0) res[id,j] <- lambda^2 * (gamma + 1) / 2
        }
      } # end if(derivative)
      
      # add ridge component?
      if(alpha < 1){
        if(derivative){
          res <- alpha * res + (1-alpha) * abs(x)
        } else {
          res <- alpha * res + (1-alpha) * x^2 / 2
        }
      }
      
    } # end for(j in 1:npen)
    
    # return results?
    if(!plot){
      return(as.data.frame(cbind(x = xorig, res))) 
    }
    
    
    
    #########***#########   COLLECT ARGS   #########***#########
      
    # collect ...
    args <- list(...)
    
    # check args$xlim
    if(is.null(args$xlim)) args$xlim <- extendrange(xorig)
    
    # check args$ylim
    if(is.null(args$ylim)) args$ylim <- extendrange(res)
    
    # check args$lty
    if(is.null(args$lty)) args$lty <- c(1, 2, 4)
    
    # check args$lwd
    if(is.null(args$lwd)) args$lwd <- c(2, 2, 2)
    
    # check args$col
    if(is.null(args$col)) args$col <- c("black", "blue", "red")
    
    # check args$xlab
    if(is.null(args$xlab)) args$xlab <- expression(beta)
    
    # check args$ylab
    if(is.null(args$ylab)) args$ylab <- ifelse(derivative,
                                               expression(italic(P)*"'" * (beta)),
                                               expression(italic(P)(beta)))
    # check args$main
    if(is.null(args$main)) args$main <- ifelse(derivative, "penalty derivatives", "penalty functions")
    
    
    
    #########***#########   PLOT PENALTY   #########***#########
    
    # plot first penalty
    plot(xorig, res[,1], type = "l", lty = args$lty[1], col = args$col[1], 
         lwd = args$lwd[1], xlim = args$xlim, ylim = args$ylim, 
         xlab = args$xlab, ylab = args$ylab, main = args$main)
    
    if(npen > 1L){
      for(j in 2:npen){
        lines(xorig, res[,j], lty = args$lty[j], col = args$col[j], lwd = args$lwd[j])
      }
    }
    
    # subtitle?
    if(subtitle){
      subtit <- bquote(alpha == .(alpha) ~ "    " ~ gamma == .(gamma) ~ "    " ~ lambda == .(lambda))
      mtext(as.expression(subtit), side = 3)
    }
    
    # legend?
    if(legend){
      legend(location, legend = penalty, 
             col = args$col, lty = args$lty, lwd = args$lwd, bty = "n")
    }
    
    
  } # end visualize.penalty

# Some code (for arguments and outputs) is re-purposed from the R package
# "glmnet" (Hastie et al., 2010) https://cran.r-project.org/package=glmnet

grpnet.default <-
  function(x,
           y,
           group,
           family = c("gaussian", "multigaussian", 
                      "svm1", "svm2", "logit",
                      "binomial", "multinomial", "ordinal",
                      "poisson", "negative.binomial", 
                      "Gamma", "inverse.gaussian"),
           weights = NULL,
           offset = NULL,
           alpha = 1,
           nlambda = 100,
           lambda.min.ratio = ifelse(nobs < nvars, 0.05, 0.0001),
           lambda = NULL,
           penalty.factor = NULL,
           penalty = c("LASSO", "MCP", "SCAD"),
           gamma = 4,
           theta = 1,
           standardized = !orthogonalized,
           orthogonalized = TRUE,
           intercept = TRUE,
           thresh = 1e-04,
           maxit = 1e05,
           proglang = c("Fortran", "R"),
           standardize.response = FALSE,
           keep.data = TRUE,
           ...){
    # group elastic net regularized regression (default)
    # Nathaniel E. Helwig (helwig@umn.edu)
    # Updated: 2025-04-07
    
    
    ######***######   INITIAL CHECKS   ######***######
    
    ### start clock
    tic <- proc.time()[3]
    
    ### get call
    grpnet.call <- match.call()
    
    ### check x and y
    x <- as.matrix(x) + 0.0
    nobs <- nrow(x)
    nvars <- ncol(x)
    ny <- if(is.matrix(y)) nrow(y) else length(y)
    if(ny != nobs) stop("Inputs 'x' and 'y' must satisfy:\nnrow(x) == length(y)  or  nrow(x) == nrow(y)")
    
    ### keep.data?
    if(keep.data){
      data <- list(x = x, y = y)
    } else {
      data <- NULL
    }
    
    ### x names
    xnames <- colnames(x)
    if(is.null(xnames)) xnames <- paste0("x", 1:nvars)
    
    ### check group
    if(missing(group)) {
      group <- ingroup <- 1:nvars
      gsize <- rep(1L, nvars)
      names(gsize) <- 1:nvars
      ngrps <- nvars
    } else {
      if(length(group) != nvars) stop("Inputs 'x' and 'group' must satisfy:  ncol(x) == length(group)")
      ingroup <- group
      group <- as.factor(group)
      ngrps <- nlevels(group)
      if(ngrps > nvars) stop("Inputs 'x' and 'group' must satisfy:  ncol(x) >= nlevels(group)")
      gsize <- table(group)
      group <- as.integer(group)
    }
    gnames <- names(gsize)
    
    ### check family
    family <- as.character(family[1])
    if(family == "mgaussian" | family == "mvn") family <- "multigaussian"
    if(family == "hsvm") family <- "svm1"
    if(family == "sqsvm") family <- "svm2"
    if(family == "polr") family <- "ordinal"
    families <- c("gaussian", "multigaussian", "svm1", "svm2", "logit", "binomial", "multinomial", "ordinal", "poisson", "negative.binomial", "Gamma", "inverse.gaussian")
    family <- pmatch(family, families)
    if(is.na(family)) stop("'family' not recognized")
    family <- families[family]
    family <- family.grpnet(family, theta)
    
    ### check weights
    if(is.null(weights)){
      noweights = TRUE
      weights <- wsqrt <- rep(1.0, nobs)
    } else {
      noweights = FALSE
      weights <- as.numeric(weights)
      if(length(weights) != nobs) stop("Inputs 'y' and 'weights' must satisfy:  length(y) == length(weights)")
      if(any(weights < 0)) stop("Input 'weights' must be non-negative")
      weights <- weights / mean(weights)
      wsqrt <- sqrt(weights)
    }
    
    ### check y
    ylev <- NULL
    if(family$family == "gaussian"){
      y <- as.numeric(y)
    } else if(family$family == "multigaussian"){
      y <- as.matrix(y) + 0.0
      if(!is.numeric(y[1,1])) stop("Input 'y' must be a numeric matrix when family = 'multigaussian'.")
      nresp <- ncol(y)
      ylev <- colnames(y)
      if(is.null(ylev)){
        ylev <- paste0("y", 1:nresp)
        colnames(y) <- ylev
      }
      if(nresp == 1L){
        y <- as.numeric(y)
        family <- gaussian()
      }
    } else if(family$family %in% c("svm1", "svm2", "logit")){
      if(is.character(y)) y <- as.factor(y)
      if(is.factor(y)){
        ylev <- levels(y)
        y <- ifelse(y == ylev[1], -1.0, 1.0)
      } else if(is.matrix(y)) {
        if(ncol(y) != 2L | any(y < 0)) stop("Input 'y' must be a matrix (# success, # failure) when family %in% c('svm1', 'svm2', 'logit')")
        ytotal <- rowSums(y)
        weights <- weights * ytotal
        wsqrt <- sqrt(weights)
        y <- 2 * (y[,1] / ytotal) - 1
      } else {
        y <- as.numeric(y)
        if(any(y < -1) | any(y > 1)) stop("Input 'y' must contain values between -1 and 1 when family %in% c('svm1', 'svm2', 'logit')" )
      }
      if(is.null(ylev)) ylev <- c(-1, 1)
    } else if(family$family == "binomial"){
      if(is.character(y)) y <- as.factor(y)
      if(is.factor(y)){
        #if(nlevels(y) != 2L) stop("Input 'y' must be (coercible into) a factor with two levels when family = 'binomial'.")
        ylev <- levels(y)
        y <- ifelse(y == ylev[1], 0.0, 1.0)
      } else if(is.matrix(y)) {
        if(ncol(y) != 2L | any(y < 0)) stop("Input 'y' must be a matrix (# success, # failure) when family = 'binomial'")
        ytotal <- rowSums(y)
        weights <- weights * ytotal
        wsqrt <- sqrt(weights)
        y <- y[,1] / ytotal
      } else {
        y <- as.numeric(y)
        if(any(y < 0) | any(y > 1)) stop("Input 'y' must contain values between 0 and 1 when family = 'binomial'" )
      }
      if(is.null(ylev)) ylev <- c(0, 1)
    } else if(family$family == "multinomial"){
      if(is.character(y)) y <- as.factor(y)
      yorig <- y
      if(is.factor(y)){
        yfac <- y
        yint <- as.integer(y)
        ylev <- levels(y)
        nlev <- nlevels(y)
        if(nlev < 3L) stop("Input 'y' must be (coercible into) a factor with 3 or more levels.")
        y <- matrix(0.0, nrow = nobs, ncol = nlev)
        colnames(y) <- ylev
        for(i in 1:nobs) y[i,yint[i]] <- 1.0
      } else if(is.matrix(y)){
        nlev <- ncol(y)
        ylev <- colnames(y)
        if(is.null(ylev)){
          ylev <- paste0("y", 1:nlev)
          colnames(y) <- ylev
        }
        y <- t(apply(y, 1, as.integer))
        if(min(y) < 0) stop("Input 'y' must be a matrix of counts (i.e, non-negative integers)")
        yrowsum <- rowSums(y)
        y <- y / yrowsum + 0.0
        weights <- weights * yrowsum
        wsqrt <- sqrt(weights)
      } else {
        stop("Invalid data input: 'y' must be a factor or matrix for when family = 'multinomial'.")
      }
    } else if(family$family == "ordinal"){
      y <- factor(y, ordered = TRUE)
      ylev <- levels(y)
      nlev <- nlevels(y)
      y <- as.integer(y)
      intercept <- TRUE
    } else if(family$family == "poisson") {
      y <- as.integer(y)
      if(any(y < 0)) stop("Input 'y' must contain non-negative integers when family = 'poisson'.")
    } else if(family$family == "negative.binomial") {
      y <- as.integer(y)
      if(any(y < 0)) stop("Input 'y' must contain non-negative integers when family = 'negative.binomial'.")
    } else if(family$family == "Gamma") {
      y <- as.numeric(y)
      if(any(y <= 0)) stop("Input 'y' must contain non-negative numerics when family = 'Gamma'.")
    } else if(family$family == "inverse.gaussian") {
      y <- as.numeric(y)
      if(any(y <= 0)) stop("Input 'y' must contain non-negative numerics when family = 'inverse.gaussian'.")
    } else {
      warning("need to add checks here...")
    }
    
    ### check offset
    if(is.null(offset)){
      include.offset <- FALSE
      offset <- rep(0.0, nobs)
      if(family$family == "multigaussian") offset <- matrix(offset, nrow = nobs, ncol = nresp)
      if(family$family == "multinomial") offset <- matrix(offset, nrow = nobs, ncol = nlev)
    } else {
      include.offset <- TRUE
      if(family$family == "multigaussian"){
        offset <- as.matrix(offset) + 0.0
        if(nrow(offset) != nobs) stop("Inputs 'x' and 'offset' must satisfy:  nrow(x) == nrow(offset)")
        if(ncol(offset) == 1L){
          offset <- matrix(offset, nrow = nobs, ncol = nresp)
        } else {
          if(ncol(offset) != nresp) stop("Input 'offset' must be a vector of length nobs\n or a matrix of dimension nobs x ncol(y)")
        }
      } else if(family$family == "multinomial"){
        offset <- as.matrix(offset) + 0.0
        if(nrow(offset) != nobs) stop("Inputs 'x' and 'offset' must satisfy:  nrow(x) == nrow(offset)")
        if(ncol(offset) == 1L){
          offset <- matrix(offset, nrow = nobs, ncol = nlev)
        } else {
          if(ncol(offset) != nlev) stop("Input 'offset' must be a vector of length nobs\n or a matrix of dimension nobs x nlev")
        }
      } else {
        offset <- as.numeric(offset)
        if(length(offset) != nobs) stop("Inputs 'y' and 'offset' must satisfy:  length(y) == length(offset)")
      }
    }
    
    ### check alpha
    alpha <- as.numeric(alpha[1])
    if(alpha < 0 | alpha > 1) stop("Input 'alpha' must satisfy:  0 <= alpha <= 1")
    
    ### check lambda and nlambda
    if(is.null(lambda)){
      nolam <- TRUE
      nlambda <- as.integer(nlambda[1])
      if(nlambda <= 0) stop("Input 'nlambda' must be a positive integer")
      lambda <- rep(0.0, nlambda)
    } else {
      nolam <- FALSE
      lambda <- as.numeric(lambda)
      if(length(lambda) > 1L) lambda <- sort(lambda, decreasing = TRUE)
      if(any(lambda < 0)) stop("Input 'lambda' must contain non-negative values")
      nlambda <- length(lambda)
    } # end if(is.null(lambda))
    
    ### check lambda.min.ratio
    lambda.min.ratio <- as.numeric(lambda.min.ratio[1])
    if(lambda.min.ratio <= 0 | lambda.min.ratio >= 1) stop("Input 'lambda.min.ratio' must satisfy:  0 < lambda.min.ratio < 1")
    
    ### check penalty.factor
    if(is.null(penalty.factor)){
      penalty.factor <- sqrt(gsize)
    } else {
      penalty.factor <- as.numeric(penalty.factor)
      if(length(penalty.factor) != ngrps) stop("Inputs 'group' and 'penalty.factor' must satisfy:  nlevels(group) == length(penalty.factor)")
      if(any(penalty.factor < 0)) stop("Input 'penalty.factor' must be non-negative")
    }
    
    ### check penalty
    penalty <- penalty[1]
    if(is.integer(penalty)){
      if(penalty < 1L || penalty > 3L) stop("Invalid 'penalty': must be 1 (LASSO), 2 (MCP), or 3 (SCAD)")
    } else {
      penalty <- toupper(as.character(penalty))
      penalty <- pmatch(penalty, c("LASSO", "MCP", "SCAD"))
      if(is.na(penalty)) stop("Invalid 'penalty': must be 'LASSO', 'MCP', or 'SCAD'")
    }
    
    ### check gamma
    gamma <- as.numeric(gamma[1])
    if(penalty == "MCP"){
      if(gamma <= 1L) stop("Need gamma > 1 when penalty = 'MCP'")
    } else if(penalty == "SCAD"){
      if(gamma <= 2L) stop("Need gamma > 2 when penalty = 'SCAD'")
    }
    
    ### standardized
    standardized <- as.logical(standardized[1])
    if(!any(standardized == c(TRUE, FALSE))) stop("Input 'standardized' must be TRUE or FALSE")
    
    ### orthogonalized
    orthogonalized <- as.logical(orthogonalized[1])
    if(!any(orthogonalized == c(TRUE, FALSE))) stop("Input 'orthogonalized' must be TRUE or FALSE")
    
    ### check intercept
    intercept <- as.logical(intercept[1])
    if(!any(intercept == c(TRUE, FALSE))) stop("Input 'intercept' must be TRUE or FALSE")
    if(intercept) gnames <- c("(Intercept)", gnames)
    
    ### check thresh
    thresh <- as.numeric(thresh[1])
    if(thresh <= 0) stop("Input 'thresh' must be a positive numeric")
    
    ### check maxit
    maxit <- as.integer(maxit[1])
    if(maxit <= 0) stop("Input 'maxit' must be a positive integer")
    
    ### check proglang
    proglangs <- c("Fortran", "R")
    proglang <- as.character(proglang[1])
    proglang <- pmatch(toupper(proglang), toupper(proglangs))
    if(is.na(proglang)){
      stop("Input 'proglang' must be 'Fortran' or 'R'.")
    }
    proglang <- proglangs[proglang]
    
    ### check standardize.response
    if(family$family == "multigaussian"){
      standardize.response <- as.logical(standardize.response[1])
      if(standardize.response){
        ysd <- apply(y, 2, sd)
        idx <- which(ysd == 0)
        if(length(idx) > 0){
          ysd[idx] <- 1.0
        }
        y <- y %*% diag(1/ysd)
      }
    }
    
    
    ######***######   WORK   ######***######
    
    ### orthogonalized?
    if(orthogonalized){
      xproj <- vector("list", ngrps)
      for(k in 1:ngrps){
        id <- which(group == k)
        nk <- length(id)
        xtemp <- wsqrt * x[,id]
        if(nk == 1L){
          xksd <- sqrt(mean((xtemp - mean(xtemp))^2))
          if(xksd <= sqrt(.Machine$double.eps)) xksd <- 1
          xproj[[k]] <- 1 / xksd
          x[,id] <- x[,id] * xproj[[k]]
        } else {
          xeig <- eigen(crossprod(xtemp - matrix(colMeans(xtemp), nobs, nk, byrow = TRUE))/nobs, symmetric = TRUE)
          xrank <- sum(xeig$values > xeig$values[1] * nk * .Machine$double.eps)
          if(xrank == 0L){
            xproj[[k]] <- diag(nk)
          } else {
            xproj[[k]] <- xeig$vectors[,1:xrank,drop=FALSE] %*% diag(1/sqrt(xeig$values[1:xrank]), xrank, xrank)
            if(xrank < nk){
              xproj[[k]] <- cbind(xproj[[k]], matrix(0.0, nk, nk - xrank))
            }
          }
          x[,id] <- x[,id] %*% xproj[[k]]
        }
      }
    } # end if(orthogonalized)
    
    ### check family
    if(family$family == "gaussian"){
      
      ## call fortran or R code
      if(proglang == "Fortran"){
        res <- .Fortran("grpnet_gaussian",
                        nobs = nobs,
                        nvars = nvars,
                        x = x,
                        y = y,
                        w = weights,
                        off = offset,
                        ngrps = ngrps,
                        gsize = gsize, 
                        pw = penalty.factor,
                        alpha = alpha,
                        nlam = nlambda,
                        lambda = lambda,
                        lmr = lambda.min.ratio, 
                        penid = penalty,
                        gamma = gamma,
                        eps = thresh,
                        maxit = maxit,
                        standardize = as.integer(standardized),
                        intercept = as.integer(intercept),
                        ibeta = rep(0.0, nlambda),
                        betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                        iters = rep(0L, nlambda),
                        nzgrps = rep(0L, nlambda),
                        nzcoef = rep(0L, nlambda),
                        edfs = rep(0.0, nlambda),
                        devs = rep(0.0, nlambda),
                        nulldev = 0.0)
      } else {
        res <- R_grpnet_gaussian(nobs = nobs,
                                 nvars = nvars,
                                 x = x,
                                 y = y,
                                 w = weights,
                                 off = offset,
                                 ngrps = ngrps,
                                 gsize = gsize, 
                                 pw = penalty.factor,
                                 alpha = alpha,
                                 nlam = nlambda,
                                 lambda = lambda,
                                 lmr = lambda.min.ratio, 
                                 penid = penalty,
                                 gamma = gamma,
                                 eps = thresh,
                                 maxit = maxit,
                                 standardize = as.integer(standardized),
                                 intercept = as.integer(intercept),
                                 ibeta = rep(0.0, nlambda),
                                 betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                                 iters = rep(0L, nlambda),
                                 nzgrps = rep(0L, nlambda),
                                 nzcoef = rep(0L, nlambda),
                                 edfs = rep(0.0, nlambda),
                                 devs = rep(0.0, nlambda),
                                 nulldev = 0.0)
        
      } # end if(proglang == "Fortran")
      
    } else if(family$family == "multigaussian"){
      
      ## call fortran or R code
      if(proglang == "Fortran"){
        res <- .Fortran("grpnet_multigaus",
                        nobs = nobs,
                        nvars = nvars,
                        nresp = nresp,
                        x = x,
                        y = y,
                        w = weights,
                        off = offset,
                        ngrps = ngrps,
                        gsize = gsize, 
                        pw = penalty.factor,
                        alpha = alpha,
                        nlam = nlambda,
                        lambda = lambda,
                        lmr = lambda.min.ratio, 
                        penid = penalty,
                        gamma = gamma,
                        eps = thresh,
                        maxit = maxit,
                        standardize = as.integer(standardized),
                        intercept = as.integer(intercept),
                        ibeta = matrix(0.0, nrow = nresp, ncol = nlambda),
                        betas = array(0.0, dim = c(nvars, nresp, nlambda)),
                        iters = rep(0L, nlambda),
                        nzgrps = rep(0L, nlambda),
                        nzcoef = rep(0L, nlambda),
                        edfs = rep(0.0, nlambda),
                        devs = rep(0.0, nlambda),
                        nulldev = 0.0)
      } else {
        res <- R_grpnet_multigaus(nobs = nobs,
                                  nvars = nvars,
                                  nresp = nresp,
                                  x = x,
                                  y = y,
                                  w = weights,
                                  off = offset,
                                  ngrps = ngrps,
                                  gsize = gsize, 
                                  pw = penalty.factor,
                                  alpha = alpha,
                                  nlam = nlambda,
                                  lambda = lambda,
                                  lmr = lambda.min.ratio, 
                                  penid = penalty,
                                  gamma = gamma,
                                  eps = thresh,
                                  maxit = maxit,
                                  standardize = as.integer(standardized),
                                  intercept = as.integer(intercept),
                                  ibeta = matrix(0.0, nrow = nresp, ncol = nlambda),
                                  betas = array(0.0, dim = c(nvars, nresp, nlambda)),
                                  iters = rep(0L, nlambda),
                                  nzgrps = rep(0L, nlambda),
                                  nzcoef = rep(0L, nlambda),
                                  edfs = rep(0.0, nlambda),
                                  devs = rep(0.0, nlambda),
                                  nulldev = 0.0)
      } # end if(proglang == "Fortran")
      
      ## post-process coefs
      rownames(res$ibeta) <- ylev
      colnames(res$ibeta) <- paste0("s", 1:nlambda)
      betas <- vector("list", nresp)
      names(betas) <- ylev
      for(j in 1:nresp) {
        betas[[j]] <- res$betas[,j,]
        rownames(betas[[j]]) <- xnames
        colnames(betas[[j]]) <- paste0("s", 1:nlambda)
      }
      res$betas <- betas
      
    } else if(family$family == "svm1"){
      
      ## call fortran or R code
      if(proglang == "Fortran"){
        
        res <- .Fortran("grpnet_svm1",
                        nobs = nobs,
                        nvars = nvars,
                        x = x,
                        y = y,
                        w = weights,
                        off = offset,
                        ngrps = ngrps,
                        gsize = gsize, 
                        pw = penalty.factor,
                        alpha = alpha,
                        nlam = nlambda,
                        lambda = lambda,
                        lmr = lambda.min.ratio, 
                        penid = penalty,
                        gamma = gamma,
                        eps = thresh,
                        maxit = maxit,
                        standardize = as.integer(standardized),
                        intercept = as.integer(intercept),
                        ibeta = rep(0.0, nlambda),
                        betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                        iters = rep(0L, nlambda),
                        nzgrps = rep(0L, nlambda),
                        nzcoef = rep(0L, nlambda),
                        edfs = rep(0.0, nlambda),
                        devs = rep(0.0, nlambda),
                        nulldev = 0.0,
                        theta = theta)
        
      } else {
        
        res <- R_grpnet_svm1(nobs = nobs,
                             nvars = nvars,
                             x = x,
                             y = y,
                             w = weights,
                             off = offset,
                             ngrps = ngrps,
                             gsize = gsize, 
                             pw = penalty.factor,
                             alpha = alpha,
                             nlam = nlambda,
                             lambda = lambda,
                             lmr = lambda.min.ratio, 
                             penid = penalty,
                             gamma = gamma,
                             eps = thresh,
                             maxit = maxit,
                             standardize = as.integer(standardized),
                             intercept = as.integer(intercept),
                             ibeta = rep(0.0, nlambda),
                             betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                             iters = rep(0L, nlambda),
                             nzgrps = rep(0L, nlambda),
                             nzcoef = rep(0L, nlambda),
                             edfs = rep(0.0, nlambda),
                             devs = rep(0.0, nlambda),
                             nulldev = 0.0,
                             theta = theta)
      } # end if(proglang == "Fortran")
      
    } else if(family$family == "svm2"){
      
      ## call fortran or R code
      if(proglang == "Fortran"){
        
        res <- .Fortran("grpnet_svm2",
                        nobs = nobs,
                        nvars = nvars,
                        x = x,
                        y = y,
                        w = weights,
                        off = offset,
                        ngrps = ngrps,
                        gsize = gsize, 
                        pw = penalty.factor,
                        alpha = alpha,
                        nlam = nlambda,
                        lambda = lambda,
                        lmr = lambda.min.ratio, 
                        penid = penalty,
                        gamma = gamma,
                        eps = thresh,
                        maxit = maxit,
                        standardize = as.integer(standardized),
                        intercept = as.integer(intercept),
                        ibeta = rep(0.0, nlambda),
                        betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                        iters = rep(0L, nlambda),
                        nzgrps = rep(0L, nlambda),
                        nzcoef = rep(0L, nlambda),
                        edfs = rep(0.0, nlambda),
                        devs = rep(0.0, nlambda),
                        nulldev = 0.0)
        
      } else {
        
        res <- R_grpnet_svm2(nobs = nobs,
                             nvars = nvars,
                             x = x,
                             y = y,
                             w = weights,
                             off = offset,
                             ngrps = ngrps,
                             gsize = gsize, 
                             pw = penalty.factor,
                             alpha = alpha,
                             nlam = nlambda,
                             lambda = lambda,
                             lmr = lambda.min.ratio, 
                             penid = penalty,
                             gamma = gamma,
                             eps = thresh,
                             maxit = maxit,
                             standardize = as.integer(standardized),
                             intercept = as.integer(intercept),
                             ibeta = rep(0.0, nlambda),
                             betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                             iters = rep(0L, nlambda),
                             nzgrps = rep(0L, nlambda),
                             nzcoef = rep(0L, nlambda),
                             edfs = rep(0.0, nlambda),
                             devs = rep(0.0, nlambda),
                             nulldev = 0.0)
      } # end if(proglang == "Fortran")
      
    } else if(family$family == "logit"){
      
      ## call fortran or R code
      if(proglang == "Fortran"){
        
        res <- .Fortran("grpnet_logit",
                        nobs = nobs,
                        nvars = nvars,
                        x = x,
                        y = y,
                        w = weights,
                        off = offset,
                        ngrps = ngrps,
                        gsize = gsize, 
                        pw = penalty.factor,
                        alpha = alpha,
                        nlam = nlambda,
                        lambda = lambda,
                        lmr = lambda.min.ratio, 
                        penid = penalty,
                        gamma = gamma,
                        eps = thresh,
                        maxit = maxit,
                        standardize = as.integer(standardized),
                        intercept = as.integer(intercept),
                        ibeta = rep(0.0, nlambda),
                        betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                        iters = rep(0L, nlambda),
                        nzgrps = rep(0L, nlambda),
                        nzcoef = rep(0L, nlambda),
                        edfs = rep(0.0, nlambda),
                        devs = rep(0.0, nlambda),
                        nulldev = 0.0)
        
      } else {
        
        res <- R_grpnet_logit(nobs = nobs,
                              nvars = nvars,
                              x = x,
                              y = y,
                              w = weights,
                              off = offset,
                              ngrps = ngrps,
                              gsize = gsize, 
                              pw = penalty.factor,
                              alpha = alpha,
                              nlam = nlambda,
                              lambda = lambda,
                              lmr = lambda.min.ratio, 
                              penid = penalty,
                              gamma = gamma,
                              eps = thresh,
                              maxit = maxit,
                              standardize = as.integer(standardized),
                              intercept = as.integer(intercept),
                              ibeta = rep(0.0, nlambda),
                              betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                              iters = rep(0L, nlambda),
                              nzgrps = rep(0L, nlambda),
                              nzcoef = rep(0L, nlambda),
                              edfs = rep(0.0, nlambda),
                              devs = rep(0.0, nlambda),
                              nulldev = 0.0)
      } # end if(proglang == "Fortran")
      
    } else if(family$family == "binomial"){
      
      ## call fortran or R code
      if(proglang == "Fortran"){
        res <- .Fortran("grpnet_binomial",
                        nobs = nobs,
                        nvars = nvars,
                        x = x,
                        y = y,
                        w = weights,
                        off = offset,
                        ngrps = ngrps,
                        gsize = gsize, 
                        pw = penalty.factor,
                        alpha = alpha,
                        nlam = nlambda,
                        lambda = lambda,
                        lmr = lambda.min.ratio, 
                        penid = penalty,
                        gamma = gamma,
                        eps = thresh,
                        maxit = maxit,
                        standardize = as.integer(standardized),
                        intercept = as.integer(intercept),
                        ibeta = rep(0.0, nlambda),
                        betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                        iters = rep(0L, nlambda),
                        nzgrps = rep(0L, nlambda),
                        nzcoef = rep(0L, nlambda),
                        edfs = rep(0.0, nlambda),
                        devs = rep(0.0, nlambda),
                        nulldev = 0.0)
      } else {
        res <- R_grpnet_binomial(nobs = nobs,
                                 nvars = nvars,
                                 x = x,
                                 y = y,
                                 w = weights,
                                 off = offset,
                                 ngrps = ngrps,
                                 gsize = gsize, 
                                 pw = penalty.factor,
                                 alpha = alpha,
                                 nlam = nlambda,
                                 lambda = lambda,
                                 lmr = lambda.min.ratio, 
                                 penid = penalty,
                                 gamma = gamma,
                                 eps = thresh,
                                 maxit = maxit,
                                 standardize = as.integer(standardized),
                                 intercept = as.integer(intercept),
                                 ibeta = rep(0.0, nlambda),
                                 betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                                 iters = rep(0L, nlambda),
                                 nzgrps = rep(0L, nlambda),
                                 nzcoef = rep(0L, nlambda),
                                 edfs = rep(0.0, nlambda),
                                 devs = rep(0.0, nlambda),
                                 nulldev = 0.0)
      } # end if(proglang == "Fortran")
      
    } else if(family$family == "multinomial"){
      
      ## call fortran or R code
      if(proglang == "Fortran"){
        res <- .Fortran("grpnet_multinom",
                        nobs = nobs,
                        nvars = nvars,
                        nresp = nlev,
                        x = x,
                        y = y,
                        w = weights,
                        off = offset,
                        ngrps = ngrps,
                        gsize = gsize, 
                        pw = penalty.factor,
                        alpha = alpha,
                        nlam = nlambda,
                        lambda = lambda,
                        lmr = lambda.min.ratio, 
                        penid = penalty,
                        gamma = gamma,
                        eps = thresh,
                        maxit = maxit,
                        standardize = as.integer(standardized),
                        intercept = as.integer(intercept),
                        ibeta = matrix(0.0, nrow = nlev, ncol = nlambda),
                        betas = array(0.0, dim = c(nvars, nlev, nlambda)),
                        iters = rep(0L, nlambda),
                        nzgrps = rep(0L, nlambda),
                        nzcoef = rep(0L, nlambda),
                        edfs = rep(0.0, nlambda),
                        devs = rep(0.0, nlambda),
                        nulldev = 0.0)
      } else {
        res <- R_grpnet_multinom(nobs = nobs,
                                 nvars = nvars,
                                 nresp = nlev,
                                 x = x,
                                 y = y,
                                 w = weights,
                                 off = offset,
                                 ngrps = ngrps,
                                 gsize = gsize, 
                                 pw = penalty.factor,
                                 alpha = alpha,
                                 nlam = nlambda,
                                 lambda = lambda,
                                 lmr = lambda.min.ratio, 
                                 penid = penalty,
                                 gamma = gamma,
                                 eps = thresh,
                                 maxit = maxit,
                                 standardize = as.integer(standardized),
                                 intercept = as.integer(intercept),
                                 ibeta = matrix(0.0, nrow = nlev, ncol = nlambda),
                                 betas = array(0.0, dim = c(nvars, nlev, nlambda)),
                                 iters = rep(0L, nlambda),
                                 nzgrps = rep(0L, nlambda),
                                 nzcoef = rep(0L, nlambda),
                                 edfs = rep(0.0, nlambda),
                                 devs = rep(0.0, nlambda),
                                 nulldev = 0.0)
      } # end if(proglang == "Fortran")
      
      ## post-process coefs
      rownames(res$ibeta) <- ylev
      colnames(res$ibeta) <- paste0("s", 1:nlambda)
      betas <- vector("list", nlev)
      names(betas) <- ylev
      for(j in 1:nlev) {
        betas[[j]] <- res$betas[,j,]
        rownames(betas[[j]]) <- xnames
        colnames(betas[[j]]) <- paste0("s", 1:nlambda)
      }
      res$betas <- betas
      
    } else if(family$family == "ordinal"){
      
      ## call fortran or R code
      if(proglang == "Fortran"){
        res <- .Fortran("grpnet_ordinal",
                        nobs = nobs,
                        nvars = nvars,
                        nresp = nlev,
                        x = x,
                        y = y,
                        w = weights,
                        off = offset,
                        ngrps = ngrps,
                        gsize = gsize, 
                        pw = penalty.factor,
                        alpha = alpha,
                        nlam = nlambda,
                        lambda = lambda,
                        lmr = lambda.min.ratio, 
                        penid = penalty,
                        gamma = gamma,
                        eps = thresh,
                        maxit = maxit,
                        standardize = as.integer(standardized),
                        intercept = as.integer(intercept),
                        ibeta = matrix(0.0, nrow = nlev - 1L, ncol = nlambda),
                        betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                        iters = rep(0L, nlambda),
                        nzgrps = rep(0L, nlambda),
                        nzcoef = rep(0L, nlambda),
                        edfs = rep(0.0, nlambda),
                        devs = rep(0.0, nlambda),
                        nulldev = 0.0)
      } else {
        res <- R_grpnet_ordinal(nobs = nobs,
                                nvars = nvars,
                                nresp = nlev,
                                x = x,
                                y = y,
                                w = weights,
                                off = offset,
                                ngrps = ngrps,
                                gsize = gsize, 
                                pw = penalty.factor,
                                alpha = alpha,
                                nlam = nlambda,
                                lambda = lambda,
                                lmr = lambda.min.ratio, 
                                penid = penalty,
                                gamma = gamma,
                                eps = thresh,
                                maxit = maxit,
                                standardize = as.integer(standardized),
                                intercept = as.integer(intercept),
                                ibeta = matrix(0.0, nrow = nlev - 1L, ncol = nlambda),
                                betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                                iters = rep(0L, nlambda),
                                nzgrps = rep(0L, nlambda),
                                nzcoef = rep(0L, nlambda),
                                edfs = rep(0.0, nlambda),
                                devs = rep(0.0, nlambda),
                                nulldev = 0.0)
      } # end if(proglang == "Fortran")
      
      rownames(res$ibeta) <- paste0("y>=", ylev[-1])
      colnames(res$ibeta) <- paste0("s", 1:nlambda)
      
    } else if(family$family == "poisson"){
    
      ## call fortran or R code
      if(proglang == "Fortran"){
        res <- .Fortran("grpnet_poisson",
                        nobs = nobs,
                        nvars = nvars,
                        x = x,
                        y = as.numeric(y),
                        w = weights,
                        off = offset,
                        ngrps = ngrps,
                        gsize = gsize, 
                        pw = penalty.factor,
                        alpha = alpha,
                        nlam = nlambda,
                        lambda = lambda,
                        lmr = lambda.min.ratio, 
                        penid = penalty,
                        gamma = gamma,
                        eps = thresh,
                        maxit = maxit,
                        standardize = as.integer(standardized),
                        intercept = as.integer(intercept),
                        ibeta = rep(0.0, nlambda),
                        betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                        iters = rep(0L, nlambda),
                        nzgrps = rep(0L, nlambda),
                        nzcoef = rep(0L, nlambda),
                        edfs = rep(0.0, nlambda),
                        devs = rep(0.0, nlambda),
                        nulldev = 0.0)
      } else {
        res <- R_grpnet_poisson(nobs = nobs,
                                nvars = nvars,
                                x = x,
                                y = as.numeric(y),
                                w = weights,
                                off = offset,
                                ngrps = ngrps,
                                gsize = gsize, 
                                pw = penalty.factor,
                                alpha = alpha,
                                nlam = nlambda,
                                lambda = lambda,
                                lmr = lambda.min.ratio, 
                                penid = penalty,
                                gamma = gamma,
                                eps = thresh,
                                maxit = maxit,
                                standardize = as.integer(standardized),
                                intercept = as.integer(intercept),
                                ibeta = rep(0.0, nlambda),
                                betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                                iters = rep(0L, nlambda),
                                nzgrps = rep(0L, nlambda),
                                nzcoef = rep(0L, nlambda),
                                edfs = rep(0.0, nlambda),
                                devs = rep(0.0, nlambda),
                                nulldev = 0.0)
      } # end if(proglang == "Fortran")
      
    } else if(family$family == "negative.binomial"){
      
      ## call fortran or R code
      if(proglang == "Fortran"){
        res <- .Fortran("grpnet_negbin",
                        nobs = nobs,
                        nvars = nvars,
                        x = x,
                        y = as.numeric(y),
                        w = weights,
                        off = offset,
                        ngrps = ngrps,
                        gsize = gsize, 
                        pw = penalty.factor,
                        alpha = alpha,
                        nlam = nlambda,
                        lambda = lambda,
                        lmr = lambda.min.ratio, 
                        penid = penalty,
                        gamma = gamma,
                        eps = thresh,
                        maxit = maxit,
                        standardize = as.integer(standardized),
                        intercept = as.integer(intercept),
                        ibeta = rep(0.0, nlambda),
                        betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                        iters = rep(0L, nlambda),
                        nzgrps = rep(0L, nlambda),
                        nzcoef = rep(0L, nlambda),
                        edfs = rep(0.0, nlambda),
                        devs = rep(0.0, nlambda),
                        nulldev = 0.0,
                        theta = theta)
      } else {
        res <- R_grpnet_negbin(nobs = nobs,
                               nvars = nvars,
                               x = x,
                               y = as.numeric(y),
                               w = weights,
                               off = offset,
                               ngrps = ngrps,
                               gsize = gsize, 
                               pw = penalty.factor,
                               alpha = alpha,
                               nlam = nlambda,
                               lambda = lambda,
                               lmr = lambda.min.ratio, 
                               penid = penalty,
                               gamma = gamma,
                               eps = thresh,
                               maxit = maxit,
                               standardize = as.integer(standardized),
                               intercept = as.integer(intercept),
                               ibeta = rep(0.0, nlambda),
                               betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                               iters = rep(0L, nlambda),
                               nzgrps = rep(0L, nlambda),
                               nzcoef = rep(0L, nlambda),
                               edfs = rep(0.0, nlambda),
                               devs = rep(0.0, nlambda),
                               nulldev = 0.0,
                               theta = theta)
      } # end if(proglang == "Fortran")
      
    } else if(family$family == "Gamma"){
      
      ## call fortran or R code
      if(proglang == "Fortran"){
        res <- .Fortran("grpnet_gamma",
                        nobs = nobs,
                        nvars = nvars,
                        x = x,
                        y = y,
                        w = weights,
                        off = offset,
                        ngrps = ngrps,
                        gsize = gsize, 
                        pw = penalty.factor,
                        alpha = alpha,
                        nlam = nlambda,
                        lambda = lambda,
                        lmr = lambda.min.ratio, 
                        penid = penalty,
                        gamma = gamma,
                        eps = thresh,
                        maxit = maxit,
                        standardize = as.integer(standardized),
                        intercept = as.integer(intercept),
                        ibeta = rep(0.0, nlambda),
                        betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                        iters = rep(0L, nlambda),
                        nzgrps = rep(0L, nlambda),
                        nzcoef = rep(0L, nlambda),
                        edfs = rep(0.0, nlambda),
                        devs = rep(0.0, nlambda),
                        nulldev = 0.0)
      } else {
        res <- R_grpnet_gamma(nobs = nobs,
                              nvars = nvars,
                              x = x,
                              y = y,
                              w = weights,
                              off = offset,
                              ngrps = ngrps,
                              gsize = gsize, 
                              pw = penalty.factor,
                              alpha = alpha,
                              nlam = nlambda,
                              lambda = lambda,
                              lmr = lambda.min.ratio, 
                              penid = penalty,
                              gamma = gamma,
                              eps = thresh,
                              maxit = maxit,
                              standardize = as.integer(standardized),
                              intercept = as.integer(intercept),
                              ibeta = rep(0.0, nlambda),
                              betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                              iters = rep(0L, nlambda),
                              nzgrps = rep(0L, nlambda),
                              nzcoef = rep(0L, nlambda),
                              edfs = rep(0.0, nlambda),
                              devs = rep(0.0, nlambda),
                              nulldev = 0.0)
      } # end if(proglang == "Fortran")
      
    } else if(family$family == "inverse.gaussian"){
      
      ## call fortran or R code
      if(proglang == "Fortran"){
        res <- .Fortran("grpnet_invgaus",
                        nobs = nobs,
                        nvars = nvars,
                        x = x,
                        y = y,
                        w = weights,
                        off = offset,
                        ngrps = ngrps,
                        gsize = gsize, 
                        pw = penalty.factor,
                        alpha = alpha,
                        nlam = nlambda,
                        lambda = lambda,
                        lmr = lambda.min.ratio, 
                        penid = penalty,
                        gamma = gamma,
                        eps = thresh,
                        maxit = maxit,
                        standardize = as.integer(standardized),
                        intercept = as.integer(intercept),
                        ibeta = rep(0.0, nlambda),
                        betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                        iters = rep(0L, nlambda),
                        nzgrps = rep(0L, nlambda),
                        nzcoef = rep(0L, nlambda),
                        edfs = rep(0.0, nlambda),
                        devs = rep(0.0, nlambda),
                        nulldev = 0.0)
      } else {
        res <- R_grpnet_invgaus(nobs = nobs,
                                nvars = nvars,
                                x = x,
                                y = y,
                                w = weights,
                                off = offset,
                                ngrps = ngrps,
                                gsize = gsize, 
                                pw = penalty.factor,
                                alpha = alpha,
                                nlam = nlambda,
                                lambda = lambda,
                                lmr = lambda.min.ratio, 
                                penid = penalty,
                                gamma = gamma,
                                eps = thresh,
                                maxit = maxit,
                                standardize = as.integer(standardized),
                                intercept = as.integer(intercept),
                                ibeta = rep(0.0, nlambda),
                                betas = matrix(0.0, nrow = nvars, ncol = nlambda),
                                iters = rep(0L, nlambda),
                                nzgrps = rep(0L, nlambda),
                                nzcoef = rep(0L, nlambda),
                                edfs = rep(0.0, nlambda),
                                devs = rep(0.0, nlambda),
                                nulldev = 0.0)
      } # end if(proglang == "Fortran")
      
    } # end if(family$family == "gaussian")
    
    
    ######***######   POST-PROCESSING   ######***######
    
    ### orthognalize?
    if(orthogonalized){
      if(family$family == "multigaussian"){
        for(k in 1:ngrps){
          id <- which(group == k)
          nk <- length(id)
          if(nk == 1L){
            for(l in 1:nresp){
              res$betas[[l]][id,] <- res$betas[[l]][id,] * xproj[[k]]
            }
          } else {
            for(l in 1:nresp){
              res$betas[[l]][id,] <- xproj[[k]] %*% res$betas[[l]][id,]
            }
          }
        }
      } else if(family$family == "multinomial"){
        for(k in 1:ngrps){
          id <- which(group == k)
          nk <- length(id)
          if(nk == 1L){
            for(l in 1:nlev){
              res$betas[[l]][id,] <- res$betas[[l]][id,] * xproj[[k]]
            }
          } else {
            for(l in 1:nlev){
              res$betas[[l]][id,] <- xproj[[k]] %*% res$betas[[l]][id,]
            }
          }
        }
      } else {
        for(k in 1:ngrps){
          id <- which(group == k)
          nk <- length(id)
          if(nk == 1L){
            res$betas[id,] <- res$betas[id,] * xproj[[k]]
          } else {
            res$betas[id,] <- xproj[[k]] %*% res$betas[id,]
          }
        }
      } # end if(family == "multinomial")
    } # end if(orthogonalized)
    
    ## unstandardize response (if applicable)
    if(family$family == "multigaussian" && standardize.response){
      for(j in 1:nresp){
        res$ibeta[j,] <- res$ibeta[j,] * ysd[j]
        res$betas[[j]] <- res$betas[[j]] * ysd[j]
      }
    }
    
    ## name coefficients
    if(!(family$family %in% c("multigaussian", "multinomial"))){
      rownames(res$betas) <- xnames
      colnames(res$betas) <- paste0("s", 1:nlambda)
    }
    
    ## collect arguments
    args <- list(penalty.factor = penalty.factor,
                 penalty = penalty,
                 gamma = gamma,
                 theta = theta,
                 standardized = standardized,
                 orthogonalized = orthogonalized,
                 intercept = intercept,
                 thresh = thresh,
                 maxit = maxit,
                 proglang = proglang,
                 standardize.response = standardize.response)
    
    ## intercept
    if(intercept){
      ingroup <- c(rep(0, ifelse(family$family == "ordinal", nlev-1, 1)), ingroup)
      ngrps <- ngrps + 1L
      thenames <- names(res$pw)
      res$pw <- c(0, res$pw)
      names(res$pw) <- c("(Intercept)", thenames) 
    }
    
    ## end clock
    toc <- proc.time()[3]
    
    ## collect results
    res <- list(call = grpnet.call,
                a0 = res$ibeta, 
                beta = res$betas, 
                alpha = alpha,
                lambda = res$lambda,
                family = family,
                dev.ratio = 1 - res$devs / res$nulldev,
                nulldev = res$nulldev,
                df = res$edfs,
                nzgrp = res$nzgrps,
                nzcoef = res$nzcoef,
                xsd = res$pw,
                ylev = ylev,
                nobs = nobs,
                group = ingroup,
                ngroups = ngrps,
                npasses = res$iters, 
                time = toc - tic,
                offset = include.offset,
                args = args,
                term.labels = gnames,
                data = data)
    
    ### return results
    class(res) <- "grpnet"
    return(res)
    
  } # end grpnet.default

R_grpnet_svm1 <-
  function(nobs, nvars, x, y, w, off, ngrps, gsize, pw, alpha, 
           nlam, lambda, lmr, penid, gamma, eps, maxit,
           standardize, intercept, ibeta, betas, iters,
           nzgrps, nzcoef, edfs, devs, nulldev, theta){
    # grpnet_svm1.f90 translation to R
    # Nathaniel E. Helwig (helwig@umn.edu)
    # Updated: 2026-04-09
    
    
    # ! --------------- LOCAL DEFINITIONS --------------- ! #
    ia <- ib <- rep(0L, ngrps)
    xmean <- rep(0.0, nvars)
    xev <- rep(0.0, ngrps)
    difbeta <- rep(0.0, nvars)
    # ! --------------- LOCAL DEFINITIONS --------------- ! #
    
    
    # ! --------------- MACHINE EPSILON --------------- ! #
    macheps <- .Machine$double.eps
    # ! --------------- MACHINE EPSILON --------------- ! #
    
    
    # ! --------------- CHECK WEIGHTS --------------- ! #
    wmin <- min(w)
    wmax <- max(w)
    if(wmax > wmin){
      w <- nobs * w / sum(w)     # ! normalize so SUM(w) = nobs
      w <- sqrt(w)
      yo <- y / w
      y <- w * y
      off <- w * off
      for(i in 1:nobs){
        x[i,] <- w[i] * x[i,]
      }
    } else {
      w <- rep(1.0, nobs)
      yo <- y
    }
    # ! --------------- CHECK WEIGHTS --------------- ! #
    
    
    # ! --------------- GROUP INDICES --------------- ! #
    gid <- 0L
    for(k in 1:ngrps){
      ia[k] <- gid + 1L
      gid <- gid + gsize[k]
      ib[k] <- gid
    }
    # ! --------------- GROUP INDICES --------------- ! #
    
    
    # ! --------------- CENTER AND SCALE --------------- ! #
    if(intercept == 1L){
      for(j in 1:nvars){
        xmean[j] <- sum(x[,j]) / nobs
        x[,j] <- x[,j] - xmean[j]
      }
    }
    xsdev <- rep(1.0, ngrps)
    if(standardize == 1L){
      for(k in 1:ngrps){
        xsdev[k] <- sqrt( sum(x[,ia[k]:ib[k]]^2) / (nobs * gsize[k]) )
        if(xsdev[k] > macheps){
          x[,ia[k]:ib[k]] <- x[,ia[k]:ib[k]] / xsdev[k]
        } else {
          xsdev[k] <- 1.0
        }
      }
    }
    # ! --------------- CENTER AND SCALE --------------- ! #
    
    
    # ! --------------- GET MAX EIGENVALUE --------------- ! #
    for(k in 1:ngrps){
      if(gsize[k] == 1L){
        xev[k] <- sum(x[,ia[k]]^2) / nobs
      } else {
        xtx <- crossprod(x[,ia[k]:ib[k]]) / nobs
        xev[k] <- R_grpnet_maxeigval(A = xtx, N = gsize[k])
      }
    }
    xev <- xev * (2.0 / theta)
    # ! --------------- GET MAX EIGENVALUE --------------- ! #
    
    
    # ! --------------- MISCELLANEOUS INITIALIZATIONS --------------- ! #
    maxlam <- max(lambda)
    makelambda <- 0L
    iter <- 0L
    active <- strong <- rep(0L, ngrps)
    nzgrps <- nzcoef <- rep(0L, nlam)
    ibeta <- rep(0.0, nlam)
    beta <- zvec <- grad <- rep(0.0, nvars)
    gradnorm <- rep(0.0, ngrps)
    twolam <- 0.0
    devs <- rep(0.0, nlam)
    r <- y - off
    ry <- r * yo
    id <- which(ry < theta)
    gr <- y
    gr[id] <- y[id] * pmax(0, ry[id]) / theta
    # ! --------------- MISCELLANEOUS INITIALIZATIONS --------------- ! #
    
    
    # ! --------------- GENERATE LAMBDA --------------- ! #
    if(maxlam <= macheps){
      
      makelambda <- 1L
      i <- 1L
      
      # ! find unpenalized groups ! # 
      for(k in 1:ngrps){
        if(pw[k] <= macheps){
          active[k] <- 1L
          nzgrps[i] <- nzgrps[i] + 1L
          nzcoef[i] <- nzcoef[i] + gsize[k]
        }
      }
      # ! find unpenalized groups ! #
      
      # ! iterate until active coefficients converge ! #
      if(nzgrps[i] > 0L){
        
        while(iter < maxit){
          
          # ! update iter and reset counters
          ctol <- 0.0
          iter <- iter + 1L
          
          # ! update active groups
          for(k in 1:ngrps){
            if(active[k] == 0L) next
            grad[ia[k]:ib[k]] <- crossprod(x[,ia[k]:ib[k]], gr) / nobs
            zvec[ia[k]:ib[k]] <- beta[ia[k]:ib[k]] + grad[ia[k]:ib[k]] / xev[k]
            difbeta[ia[k]:ib[k]] <- zvec[ia[k]:ib[k]] - beta[ia[k]:ib[k]]
            maxdif <- max( abs(difbeta[ia[k]:ib[k]]) / (1.0 + abs(beta[ia[k]:ib[k]])) )
            beta[ia[k]:ib[k]] <- beta[ia[k]:ib[k]] + difbeta[ia[k]:ib[k]]
            r <- r - x[,ia[k]:ib[k]] %*% difbeta[ia[k]:ib[k]]
            ry <- r * yo
            id <- which(ry < theta)
            gr <- y
            gr[id] <- y[id] * pmax(0, ry[id]) / theta
            ctol <- max(maxdif, ctol)
          } # ! k=1,ngrps
          
          # ! update intercept
          if(intercept == 1L){
            difibeta <- (sum(gr * w) / nobs) * (theta / 2)
            maxdif <- abs(difibeta) / (1.0 + abs(ibeta[i]))
            ibeta[i] <- ibeta[i] + difibeta
            r <- r - w * difibeta
            ry <- r * yo
            id <- which(ry < theta)
            gr <- y
            gr[id] <- y[id] * pmax(0, ry[id]) / theta
            ctol <- max(maxdif, ctol)
          } # if(intercept == 1L)
          
          # ! convergence check
          if(ctol < eps) break
          
        } # ! WHILE(iter < maxit)
        
      } else {
        
        while(iter < maxit){
          
          # ! update iter and reset counters
          ctol <- 0.0
          iter <- iter + 1L
          
          # ! intercept only
          if(intercept == 1L){
            difibeta <- (sum(gr * w) / nobs) * (theta / 2)
            maxdif <- abs(difibeta) / (1.0 + abs(ibeta[i]))
            ibeta[i] <- ibeta[i] + difibeta
            r <- r - w * difibeta
            ry <- r * yo
            id <- which(ry < theta)
            gr <- y
            gr[id] <- y[id] * pmax(0, ry[id]) / theta
            ctol <- max(maxdif, ctol)
          }
          
          # ! convergence check
          if(ctol < eps) break
          
        }  # ! WHILE(iter < maxit)
        
      } # ! (nzgrps[i] > 0)
      # ! iterate until active coefficients converge ! #
      
      # ! create lambda sequence ! #
      for(k in 1:ngrps){
        if(pw[k] > macheps){
          grad[ia[k]:ib[k]] <- crossprod(x[,ia[k]:ib[k]], gr) / nobs
          gradnorm[k] <- sqrt( sum(grad[ia[k]:ib[k]]^2) ) / pw[k]
        }
      }
      if(alpha > macheps){
        maxlam <- max(gradnorm / alpha)
      } else {
        maxlam <- max(gradnorm / 1e-3)
        makelambda <- 0L
      }
      minlam <- lmr * maxlam
      lambda[i] <- maxlam
      maxlam <- log(maxlam)
      minlam <- log(minlam)
      rnglam <- maxlam - minlam
      for(k in 2:nlam){
        lambda[k] <- exp( maxlam - rnglam * (k - 1) / (nlam - 1) )
      }
      # ! create lambda sequence ! #
      
      # ! calculate deviance ! #
      devs[i] <- R_grpnet_svm1_dev(nobs, y, r, w^2, theta)
      # ! calculate deviance ! #
      
      # ! save results ! # 
      betas[,i] <- beta
      iters[i] <- iter
      nzgrps[i] <- nzgrps[i] + intercept
      nzcoef[i] <- nzcoef[i] + intercept
      edfs[i] <- as.numeric(nzcoef[i])
      # ! save results ! # 
      
    }
    # ! --------------- GENERATE LAMBDA --------------- ! #
    
    
    # ! --------------- ITERATIVE WORK --------------- ! #
    for(i in 1:nlam){
      
      # ! initializations ! # 
      if(i == 1L && makelambda == 1L) next
      if(i > 1L){
        ibeta[i] <- ibeta[i-1]
        beta <- betas[,i-1]
        twolam <- alpha * (2.0 * lambda[i] - lambda[i-1])
      } else {
        grad <- crossprod(x, gr) / nobs
      }
      # ! initializations ! # 
      
      # ! strong rule initialization ! #
      for(k in 1:ngrps){
        gradnorm[k] <- sqrt(sum(grad[ia[k]:ib[k]]^2))
        if(gradnorm[k] + 1e-8 > pw[k] * twolam){
          strong[k] <- 1L
        } else {
          strong[k] <- 0L
        }
      }
      # ! strong rule initialization ! #
      
      # ! iterate until strong set converges ! #
      iter <- 0L
      while(iter < maxit){
        
        # ! iterate until active set converges ! # 
        while(iter < maxit){
          
          # ! iterate until active coefficients converge ! #
          while(iter < maxit){
            
            # ! update iter and reset counters
            iter <- iter + 1L
            nzgrps[i] <- 0L
            edfs[i] <- 0.0
            ctol <- 0.0
            
            # ! update active groups
            for(k in 1:ngrps){
              if(active[k] == 0L) next
              penone <- alpha * lambda[i] * pw[k] / xev[k]
              pentwo <- (1 - alpha) * lambda[i] * pw[k] / xev[k]
              grad[ia[k]:ib[k]] <- crossprod(x[,ia[k]:ib[k]], gr) / nobs
              zvec[ia[k]:ib[k]] <- beta[ia[k]:ib[k]] + grad[ia[k]:ib[k]] / xev[k]
              znorm <- sqrt(sum(zvec[ia[k]:ib[k]]^2))
              bnorm <- sqrt(sum(beta[ia[k]:ib[k]]^2))
              shrink <- R_grpnet_penalty(znorm, penid, penone, pentwo, gamma)
              if(shrink == 0.0 && bnorm == 0.0) next
              difbeta[ia[k]:ib[k]] <- shrink * zvec[ia[k]:ib[k]] - beta[ia[k]:ib[k]]
              maxdif <- max( abs(difbeta[ia[k]:ib[k]]) / (1.0 + abs(beta[ia[k]:ib[k]])) )
              beta[ia[k]:ib[k]] <- beta[ia[k]:ib[k]] + difbeta[ia[k]:ib[k]]
              r <- r - x[,ia[k]:ib[k]] %*% difbeta[ia[k]:ib[k]]
              ry <- r * yo
              id <- which(ry < theta)
              gr <- y
              gr[id] <- y[id] * pmax(0, ry[id]) / theta
              ctol <- max(maxdif , ctol)
              if(shrink > 0.0){
                nzgrps[i] <- nzgrps[i] + 1L
                edfs[i] <- edfs[i] + gsize[k] * shrink
              }
            } # ! k=1,ngrps
            
            # ! update intercept
            if(intercept == 1L){
              difibeta <- (sum(gr * w) / nobs) * (theta / 2)
              maxdif <- abs(difibeta) / (1.0 + abs(ibeta[i]))
              ibeta[i] <- ibeta[i] + difibeta
              r <- r - w * difibeta
              ry <- r * yo
              id <- which(ry < theta)
              gr <- y
              gr[id] <- y[id] * pmax(0, ry[id]) / theta
              ctol <- max(maxdif, ctol)
            } # ! (intercept == 1)
            
            # ! convergence check
            if(ctol < eps) break
            
          } # ! WHILE(iter < maxit) - inner
          # ! iterate until active coefficients converge ! #
          
          # ! check inactive groups in strong set ! #
          violations <- 0L
          for(k in 1:ngrps){
            if(strong[k] == 0L | active[k] == 1L) next
            grad[ia[k]:ib[k]] <- crossprod(x[,ia[k]:ib[k]], gr) / nobs
            gradnorm[k] <- sqrt(sum(grad[ia[k]:ib[k]]^2))
            if(gradnorm[k] > alpha * lambda[i] * pw[k]){
              active[k] <- 1L
              violations <- violations + 1L
            }
          } # ! k=1,ngrps
          if(violations == 0L) break
          # ! check inactive groups in strong set ! #
          
        } # ! WHILE(iter < maxit) - middle
        # ! iterate until active set converges ! # 
        
        # ! check groups in weak set ! #
        violations <- 0L
        for(k in 1:ngrps){
          if(strong[k] == 1L) next
          grad[ia[k]:ib[k]] <- crossprod(x[,ia[k]:ib[k]], gr) / nobs
          gradnorm[k] <- sqrt(sum(grad[ia[k]:ib[k]]^2))
          if(gradnorm[k] + 1e-8 > alpha * lambda[i] * pw[k]){
            strong[k] <- 1
            active[k] <- 1
            violations <- violations + 1L
          }
        }
        if(violations == 0L) break
        # ! check groups in weak set ! #
        
      } # ! WHILE(iter < maxit) - outer
      # ! iterate until strong set converges ! #
      
      # ! calculate nzcoef ! #
      for(k in 1:ngrps){
        if(active[k] == 0L) next
        for(l in 1:gsize[k]){
          if(abs(beta[ia[k] + l - 1]) > macheps){
            nzcoef[i] <- nzcoef[i] + 1L
          }
        }
      }
      # ! calculate nzcoef ! #
      
      # ! save results ! #
      betas[,i] <- beta
      iters[i] <- iter
      nzgrps[i] <- nzgrps[i] + intercept
      nzcoef[i] <- nzcoef[i] + intercept
      edfs[i] <- edfs[i] + as.numeric(intercept)
      devs[i] <- R_grpnet_svm1_dev(nobs, y, r, w^2, theta)
      # ! save results ! #
      
    }
    # ! --------------- ITERATIVE WORK --------------- ! #
    
    
    # ! --------------- POST PROCESSING --------------- ! #
    if(standardize == 1L){
      for(k in 1:ngrps){
        betas[ia[k]:ib[k],] <- betas[ia[k]:ib[k],] / xsdev[k]
      }
    }
    if(intercept == 1L){
      ibeta <- ibeta - xmean %*% betas
      if(nzgrps[1] == 1L){
        nulldev <- devs[1]
      } else {
        r <- y - w * sum(y * w) / nobs
        nulldev <- R_grpnet_svm1_dev(nobs, y, r, w^2, theta)
      }
    } else {
      nulldev <- R_grpnet_svm1_dev(nobs, y, off, w^2, theta)
    }
    names(xsdev) <- names(pw)
    pw <- xsdev
    # ! --------------- POST PROCESSING --------------- ! #
    
    
    # ! --------------- RETURN RESULTS --------------- ! #
    res <- list(nobs = nobs, 
                nvars = nvars, 
                x = x, 
                y = y, 
                w = w, 
                off = off, 
                ngrps = ngrps, 
                gsize = gsize, 
                pw = pw, 
                alpha = alpha, 
                nlam = nlam, 
                lambda = lambda, 
                lmr = lmr, 
                penid = penid, 
                gamma = gamma, 
                eps = eps, 
                maxit = maxit,
                standardize = standardize, 
                intercept = intercept, 
                ibeta = as.numeric(ibeta), 
                betas = betas, 
                iters = iters,
                nzgrps = nzgrps, 
                nzcoef = nzcoef, 
                edfs = edfs, 
                devs = devs, 
                nulldev = nulldev)
    return(res)
    # ! --------------- RETURN RESULTS --------------- ! #
    
    
  } # R_grpnet_svm1.R


R_grpnet_svm1_dev <-
  function(nobs, y, mu, wt, theta, dev = 0.0){
    muy <- (y - mu) * y / wt       # NOTE: input mu is actually r = y - mu
    for(i in 1:nobs){
      if(muy[i] > 1.0) next
      if(muy[i] > 1.0 - theta){
        dev <- dev + wt[i] * (1.0 - muy[i])^2 / (2.0 * theta)
      } else {
        dev <- dev + wt[i] * ( 1.0 - muy[i] - theta / 2.0 )
      }
    }
    return(2.0 * dev)
  } # R_grpnet_svm1_dev.R

cv.compare <-
  function(x,
           s = c("lambda.1se", "lambda.min"), 
           plot = TRUE, 
           at = 1:length(x),
           nse = 1, 
           point.col = "red", 
           line.col = "gray", 
           lwd = 2, 
           bwd = 0.02,
           labels = NULL,
           xlim = NULL,
           ylim = NULL,
           xlab = NULL,
           ylab = NULL,
           ...){
    # compare multiple cv.grpnet prediction errors
    # Nathaniel E. Helwig (helwig@umn.edu)
    # 2026-04-09
    
    
    ### xlist
    x <- as.list(x)
    nx <- length(x)
    
    ### check x class
    xcheck <- inherits(x, "cv.grpnet")
    
    ### is x cv.grpnet object?
    if(xcheck){
      
      xnames <- c("lambda.min", "lambda.1se")
      at <- 1:2
      rx <- 1
      cvm <- x$cvm[x$index]
      cvsd <- x$cvsd[x$index]
      type.measure <- x$type.measure
      family <- x$grpnet.fit$family$family
      
    } else {
      
      ### check x
      xcheck <- sapply(x, inherits, what = "cv.grpnet")
      if(any(!xcheck)) stop("Input 'x' must be a list of 'cv.grpnet' objects")
      
      ## xnames
      xnames <- labels
      if(is.null(xnames)) xnames <- paste0("m", 1:nx)
      
      ## check s
      s <- as.character(s[1])
      s <- pmatch(s, c("lambda.min", "lambda.1se"))
      if(is.na(s)) stop("Input 's' must be 'lambda.min' or 'lambda.1se'")
      s <- c("lambda.min", "lambda.1se")[s]
      sid <- ifelse(s == "lambda.min", 1, 2)
      
      ## check at
      at <- as.numeric(at)
      if(length(at) != nx) stop("Input 'at' must satisfy: length(x) == length(at)")
      rx <- range(at)
      rx <- rx[2] - rx[1]
      
      ## extract mean cv error and sd
      cvm <- cvsd <- rep(0.0, nx)
      for(i in 1:nx){
        index <- x[[i]]$index[sid]
        cvm[i] <- x[[i]]$cvm[index]
        cvsd[i] <- x[[i]]$cvsd[index]
      }
      
      ## get type.measure andfamily
      type.measure <- x[[1]]$type.measure
      family <- x[[1]]$grpnet.fit$family$family
      
    } # end if(xcheck)
    
    ### return data.frame?
    if(!plot) return(data.frame(model = xnames, cvm = cvm, cvsd = cvsd))
    
    ### check xlab and ylab
    if(is.null(xlab)) xlab <- "Models"
    if(is.null(ylab)){
      if(type.measure == "deviance"){
        titlecase <- function(x) {
          cap1 <- function(x) paste0(toupper(substr(x, 1, 1)), 
                                     tolower(substr(x, 2, nchar(x))))
          words <- strsplit(x, split = " ", fixed = TRUE)
          words <- sapply(words, cap1)
          paste(words, collapse = " ")
        }
        ylab <- paste0(titlecase(family), "Deviance")
      } else if(type.measure == "mse"){
        ylab <- "Mean Squared Error"
      } else if(type.measure == "mae"){
        ylab <- "Mean Absolute Error"
      } else if(type.measure == "class"){
        ylab <- "Misclassification Error"
      }
    }
    
    ### check xlim ylim
    if(is.null(xlim)){
      xlim <- extendrange(at)
      xlim <- xlim + c(-0.1, 0.1) * diff(xlim)
    }
    if(is.null(ylim)){
      ylim <- extendrange(c(cvm - nse*cvsd, cvm + nse*cvsd))
    }
    
    ### make plot
    plot(at, cvm, pch = 19, axes = F, xlim = xlim, ylim = ylim, 
         xlab = xlab, ylab = ylab, col = point.col, type = "n", ...)
    segments(x0 = at, y0 = cvm - nse*cvsd, y1 = cvm + nse*cvsd, lwd = lwd, col = line.col)
    segments(x0 = at - bwd*rx, y0 = cvm - nse*cvsd, x1 = at + bwd*rx, lwd = lwd, col = line.col)
    segments(x0 = at - bwd*rx, y0 = cvm + nse*cvsd, x1 = at + bwd*rx, lwd = lwd, col = line.col)
    points(at, cvm, pch = 19, col = point.col)
    axis(1, at = at, labels = xnames, ...)
    axis(2, ...)
    box()
    
  } # end cv.compare

# Some code (for startup message) is re-purposed from the R package
# "mclust" (Scrucca et al., 2016) https://cran.r-project.org/package=mclust

grpnetStartupMessage <- 
  function(){
    msg <- c(paste0("                                    _   
         __ _ _ __ _ __  _ __   ___| |_ 
        / _` | '__| '_ \\| '_ \\ / _ \\ __|
       | (_| | |  | |_) | | | |  __/ |_ 
        \\__, |_|  | .__/|_| |_|\\___|\\__|
        |___/     |_|      version ", 
                    packageVersion("grpnet"), "\n"),
             "\nType 'citation(\"grpnet\")' to cite this package.\n")
    return(msg)
  }

.onAttach <- 
  function(lib, pkg){
    msg <- grpnetStartupMessage()
    if(!interactive()) msg[1] <- paste("Package 'grpnet' version", packageVersion("grpnet"))
    packageStartupMessage(msg)      
    invisible()
  }

summary.cv.grpnet <-
  function(object, ...){
    # summarize cv.grpnet object
    # Nathaniel E. Helwig (helwig@umn.edu)
    # Updated: 2026-04-28
    
    if(!inherits(object, "cv.grpnet")) stop("Input 'object' should be of class 'cv.grpnet'.")
    method <- ifelse(is.null(object$grpnet.fit$formula), "default", "formula")
    family <- object$grpnet.fit$family$family
    if(family %in% c("multigaussian", "multinomial")){
      fit <- array(dim = c(object$grpnet.fit$nobs, length(object$grpnet.fit$ylev), 2L))
      dimnames(fit) <- list(1:object$grpnet.fit$nobs, object$grpnet.fit$ylev, c("lambda.1se", "lambda.min"))
      imp <- array(dim = c(object$grpnet.fit$ngroups - 1L, length(object$grpnet.fit$ylev), 2L))
      dimnames(imp) <- list(object$grpnet.fit$term.labels[-1], object$grpnet.fit$ylev, c("lambda.1se", "lambda.min"))
      if(method == "default"){
        fit[,,1] <- predict(object, newx = object$grpnet.fit$data$x, type = "response", s = "lambda.1se", ...)
        fit[,,2] <- predict(object, newx = object$grpnet.fit$data$x, type = "response", s = "lambda.min", ...)
        imp[,,1] <- predict(object, newx = object$grpnet.fit$data$x, type = "importance", s = "lambda.1se")
        imp[,,2] <- predict(object, newx = object$grpnet.fit$data$x, type = "importance", s = "lambda.min")
      } else {
        fit[,,1] <- predict(object, newdata = object$grpnet.fit$data, type = "response", s = "lambda.1se", ...)
        fit[,,2] <- predict(object, newdata = object$grpnet.fit$data, type = "response", s = "lambda.min", ...)
        imp[,,1] <- predict(object, newdata = object$grpnet.fit$data, type = "importance", s = "lambda.1se")
        imp[,,2] <- predict(object, newdata = object$grpnet.fit$data, type = "importance", s = "lambda.min")
      }
    } else {
      fit.type <- ifelse(family %in% c("binomail", "ordinal","svm1", "svm2"), "class", "response")
      if(method == "default"){
        fit <- cbind(predict(object, newx = object$grpnet.fit$data$x, s = "lambda.1se", type = fit.type, ...),
                     predict(object, newx = object$grpnet.fit$data$x, s = "lambda.min", type = fit.type, ...))
        imp <- cbind(predict(object, newx = object$grpnet.fit$data$x, type = "importance", s = "lambda.1se"),
                     predict(object, newx = object$grpnet.fit$data$x, type = "importance", s = "lambda.min"))
      } else {
        fit <- cbind(predict(object, newdata = object$grpnet.fit$data, s = "lambda.1se", type = fit.type, ...),
                     predict(object, newdata = object$grpnet.fit$data, s = "lambda.min", type = fit.type, ...))
        imp <- cbind(predict(object, newdata = object$grpnet.fit$data, type = "importance", s = "lambda.1se"),
                     predict(object, newdata = object$grpnet.fit$data, type = "importance", s = "lambda.min"))
      }
      colnames(fit) <- colnames(imp) <- c("lambda.1se", "lambda.min")
      if(fit.type == "class") fit <- as.data.frame(fit)
    }
    penalties <- c("LASSO", "MCP", "SCAD")
    res <- list(family = object$grpnet.fit$family$family, 
                penalty = penalties[object$grpnet.fit$args$penalty],
                nobs = object$grpnet.fit$nobs,
                ngroups = object$grpnet.fit$ngroups,
                lambda = c("lambda.1se" = object$lambda.1se, "lambda.min" = object$lambda.min),
                dev.ratio = c("lambda.1se" = object$grpnet.fit$dev.ratio[object$index[2]],
                              "lambda.min" = object$grpnet.fit$dev.ratio[object$index[1]]),
                fit = fit, 
                act = abs(imp) > 0.0,
                imp = imp)
    class(res) <- "summary.cv.grpnet"
    return(res)
    
  }  # summary.cv.grpnet

print.summary.cv.grpnet <-
  function(x, ...){
    cat("\n")
    cat("family =", x$family, "\n")
    cat("penalty =", x$penalty, "\n")
    cat("n =", x$nobs, "observations\n")
    cat("K =", x$ngroups, "groups\n")
    cat("\n% Null Deviance Explained:\n")
    print(100 * x$dev.ratio, digits = 6)
    cat("\nVariable Importance:\n")
    if(x$family %in% c("multigaussian", "multinomial")) cat("\n")
    print(100 * x$imp, digits = 6)
    if(!(x$family %in% c("multigaussian", "multinomial"))) cat("\n")
  }

R_grpnet_penalty <-
  function(znorm, penid, penone, pentwo, gamma, shrink){
    # grpnet_penalty.f90 translation to R
    # Nathaniel E. Helwig (helwig@umn.edu)
    # Updated: 2024-06-28
    
    shrink <- 1.0
    if(penid == 1L){
      shrink <- max(0, 1 - penone / znorm)
      shrink <- shrink / (1 + pentwo)
    } else if(penid == 2L){
      if(znorm <= gamma * penone * (1 + pentwo)){
        shrink <- max(0, 1 - penone / znorm)
        shrink <- shrink / (1 + pentwo - 1 / gamma)
      } else {
        shrink <- 1 / (1 + pentwo)
      }
    } else if(penid == 3L) {
      if(znorm <= (penone * (1 + pentwo) + penone)){
        shrink <- max(0, 1 - penone / znorm)
        shrink <- shrink / (1 + pentwo)
      } else if(znorm <= gamma * penone * (1 + pentwo)){
        shrink <- max(0, 1 - (penone / znorm) * (gamma / (gamma - 1)) )
        shrink <- shrink / (1 + pentwo - 1 / (gamma - 1) )
      } else {
        shrink <- 1 / (1 + pentwo)
      }
    }
    return(shrink)
    
  }

R_grpnet_ordinal <-
  function(nobs, nvars, nresp, x, y, w, off, ngrps, gsize, pw, 
           alpha, nlam, lambda, lmr, penid, gamma, eps, maxit,
           standardize, intercept, ibeta, betas, iters,
           nzgrps, nzcoef, edfs, devs, nulldev){
    # grpnet_ordinal.f90 translation to R
    # Nathaniel E. Helwig (helwig@umn.edu)
    # Updated: 2026-04-09
    
    
    # ! --------------- LOCAL DEFINITIONS --------------- ! #
    ia <- ib <- rep(0L, ngrps)
    xmean <- rep(0.0, nvars)
    xev <- rep(0.0, ngrps)
    difbeta <- rep(0.0, nvars)
    # ! --------------- LOCAL DEFINITIONS --------------- ! #
    
    
    # ! --------------- MACHINE EPSILON --------------- ! #
    macheps <- .Machine$double.eps
    # ! --------------- MACHINE EPSILON --------------- ! #
    
    
    # ! --------------- CHECK WEIGHTS --------------- ! #
    wmin <- min(w)
    wmax <- max(w)
    if(wmax > wmin){
      weighted <- 1L
      w <- nobs * w / sum(w)     # ! normalize so SUM(w) = nobs
      w <- sqrt(w)
      for(i in 1:nobs){
        x[i,] <- w[i] * x[i,]
      }
    } else {
      weighted <- 0L
      w <- rep(1.0, nobs)
    }
    # ! --------------- CHECK WEIGHTS --------------- ! #
    
    
    # ! --------------- GROUP INDICES --------------- ! #
    gid <- 0L
    for(k in 1:ngrps){
      ia[k] <- gid + 1L
      gid <- gid + gsize[k]
      ib[k] <- gid
    }
    # ! --------------- GROUP INDICES --------------- ! #
    
    
    # ! --------------- CENTER AND SCALE --------------- ! #
    if(intercept == 1L){
      for(j in 1:nvars){
        xmean[j] <- sum(x[,j]) / nobs
        x[,j] <- x[,j] - xmean[j]
      }
    }
    xsdev <- rep(1.0, ngrps)
    if(standardize == 1L){
      for(k in 1:ngrps){
        xsdev[k] <- sqrt( sum(x[,ia[k]:ib[k]]^2) / (nobs * gsize[k]) )
        if(xsdev[k] > macheps){
          x[,ia[k]:ib[k]] <- x[,ia[k]:ib[k]] / xsdev[k]
        } else {
          xsdev[k] <- 1.0
        }
      }
    }
    # ! --------------- CENTER AND SCALE --------------- ! #
    
    
    # ! --------------- GET MAX EIGENVALUE --------------- ! #
    for(k in 1:ngrps){
      if(gsize[k] == 1L){
        xev[k] <- sum(x[,ia[k]]^2) / nobs
      } else {
        xtx <- crossprod(x[,ia[k]:ib[k]]) / nobs
        xev[k] <- R_grpnet_maxeigval(A = xtx, N = gsize[k])
      }
    }
    xev <- xev / 2.0
    # ! --------------- GET MAX EIGENVALUE --------------- ! #
    
    
    # ! --------------- MISCELLANEOUS INITIALIZATIONS --------------- ! #
    maxlam <- max(lambda)
    makelambda <- 0L
    iter <- 0L
    active <- strong <- rep(0L, ngrps)
    nzgrps <- nzcoef <- rep(0L, nlam)
    ibeta <- matrix(0.0, nresp-1L, nlam)
    beta <- zvec <- grad <- rep(0.0, nvars)
    gradnorm <- rep(0.0, ngrps)
    twolam <- 0.0
    devs <- rep(0.0, nlam)
    eta <- off
    # ! --------------- MISCELLANEOUS INITIALIZATIONS --------------- ! #
    
    
    # ! --------------- SHIFT INITIALIZATION --------------- ! #
    yfreq <- ycumsum <- rep(0L, nresp)
    yfreq[1] <- ycumsum[1] <- sum(y == 1L)
    for(k in 2:nresp) {
      yfreq[k] <- sum(y == k)
      ycumsum[k] <- ycumsum[k-1] + yfreq[k]
    }
    yprob <- 1 - (ycumsum / nobs)[-nresp]
    yprob[yprob < 1e-6] <- 1e-6
    yprob[yprob > 1 - 1e-6] <- 1 - 1e-6
    shift <- log(yprob / (1 - yprob))
    # ! --------------- SHIFT INITIALIZATION --------------- ! #
    
    
    # ! --------------- GENERATE LAMBDA --------------- ! #
    if(maxlam <= macheps){
      
      makelambda <- 1L
      i <- 1L
      
      # ! find unpenalized groups ! # 
      for(k in 1:ngrps){
        if(pw[k] <= macheps){
          active[k] <- 1L
          nzgrps[i] <- nzgrps[i] + 1L
          nzcoef[i] <- nzcoef[i] + gsize[k]
        }
      }
      # ! find unpenalized groups ! #
      
      # ! iterate until active coefficients converge ! #
      if(nzgrps[i] > 0L){
        
        while(iter < maxit){
          
          # ! update iter and reset counters
          ctol <- 0.0
          iter <- iter + 1L
          
          # ! update shifts
          temp <- R_grpnet_ordinal_int(nobs, nresp, eta, y, w^2, yfreq, shift, 1.0)
          shift <- temp$theta
          ctol <- max(temp$maxdif, ctol)
          
          # ! define r
          bigtheta <- c(100, shift, -100)
          fit0 <- eta + bigtheta[y]
          fit1 <- eta + bigtheta[y+1L]
          Pi0 <- 1 / (1 + exp(-fit0))
          Pi1 <- 1 / (1 + exp(-fit1))
          r <- w * (1.0 - Pi0 - Pi1)
          
          # ! update active groups
          for(k in 1:ngrps){
            if(active[k] == 0L) next
            grad[ia[k]:ib[k]] <- crossprod(x[,ia[k]:ib[k]], r) / nobs
            zvec[ia[k]:ib[k]] <- beta[ia[k]:ib[k]] + grad[ia[k]:ib[k]] / xev[k]
            difbeta[ia[k]:ib[k]] <- zvec[ia[k]:ib[k]] - beta[ia[k]:ib[k]]
            maxdif <- max( abs(difbeta[ia[k]:ib[k]]) / (1.0 + abs(beta[ia[k]:ib[k]])) )
            beta[ia[k]:ib[k]] <- beta[ia[k]:ib[k]] + difbeta[ia[k]:ib[k]]
            eta <- eta + (x[,ia[k]:ib[k]] %*% difbeta[ia[k]:ib[k]]) / w
            fit0 <- eta + bigtheta[y]
            fit1 <- eta + bigtheta[y+1L]
            Pi0 <- 1 / (1 + exp(-fit0))
            Pi1 <- 1 / (1 + exp(-fit1))
            r <- w * (1.0 - Pi0 - Pi1)
            ctol <- max(maxdif, ctol)
          } # ! k=1,ngrps
          
          # ! convergence check
          if(ctol < eps) break
          
        } # ! WHILE(iter < maxit)
        
      } else {
        
        # ! intercept only
        while(iter < maxit){
          
          # ! update iter and reset counters
          ctol <- 0.0
          iter <- iter + 1L
          
          # ! update shifts
          temp <- R_grpnet_ordinal_int(nobs, nresp, eta, y, w^2, yfreq, shift, 1.0)
          shift <- temp$theta
          
          # ! convergence check
          ctol <- max(temp$maxdif, ctol)
          if(ctol < eps) break
          
        }  # ! WHILE(iter < maxit)
        
        # ! define r
        bigtheta <- c(100, shift, -100)
        fit0 <- eta + bigtheta[y]
        fit1 <- eta + bigtheta[y+1L]
        Pi0 <- 1 / (1 + exp(-fit0))
        Pi1 <- 1 / (1 + exp(-fit1))
        r <- w * (1.0 - Pi0 - Pi1)
        
      } # ! (nzgrps[i] > 0)
      # ! iterate until active coefficients converge ! #
      
      # ! create lambda sequence ! #
      for(k in 1:ngrps){
        if(pw[k] > macheps){
          grad[ia[k]:ib[k]] <- crossprod(x[,ia[k]:ib[k]], r) / nobs
          gradnorm[k] <- sqrt( sum(grad[ia[k]:ib[k]]^2) ) / pw[k]
        }
      }
      if(alpha > macheps){
        maxlam <- max(gradnorm / alpha)
      } else {
        maxlam <- max(gradnorm / 1e-3)
        makelambda <- 0L
      }
      minlam <- lmr * maxlam
      lambda[i] <- maxlam
      maxlam <- log(maxlam)
      minlam <- log(minlam)
      rnglam <- maxlam - minlam
      for(k in 2:nlam){
        lambda[k] <- exp( maxlam - rnglam * (k - 1) / (nlam - 1) )
      }
      # ! create lambda sequence ! #
      
      # ! calculate deviance ! #
      mu <- Pi0 - Pi1
      devs[i] <- -2.0 * sum(w^2 * log(mu))
      # ! calculate deviance ! #
      
      # ! save results ! # 
      ibeta[,i] <- shift
      betas[,i] <- beta
      iters[i] <- iter
      nzgrps[i] <- nzgrps[i] + intercept
      nzcoef[i] <- nzcoef[i] + intercept*(nresp-1)
      edfs[i] <- as.numeric(nzcoef[i])
      # ! save results ! # 
      
    }
    # ! --------------- GENERATE LAMBDA --------------- ! #
    
    
    # ! --------------- ITERATIVE WORK --------------- ! #
    for(i in 1:nlam){
      
      # ! initializations ! # 
      if(i == 1L && makelambda == 1L) next
      if(i > 1L){
        shift <- ibeta[,i-1]
        beta <- betas[,i-1]
        twolam <- alpha * (2.0 * lambda[i] - lambda[i-1])
      } else {
        grad <- crossprod(x, r) / nobs
      }
      # ! initializations ! # 
      
      # ! strong rule initialization ! #
      for(k in 1:ngrps){
        gradnorm[k] <- sqrt(sum(grad[ia[k]:ib[k]]^2))
        if(gradnorm[k] + 1e-8 > pw[k] * twolam){
          strong[k] <- 1L
        } else {
          strong[k] <- 0L
        }
      }
      # ! strong rule initialization ! #
      
      # ! iterate until strong set converges ! #
      iter <- 0L
      while(iter < maxit){
        
        # ! iterate until active set converges ! # 
        while(iter < maxit){
          
          # ! iterate until active coefficients converge ! #
          while(iter < maxit){
            
            # ! update iter and reset counters
            iter <- iter + 1L
            nzgrps[i] <- 0L
            edfs[i] <- 0.0
            ctol <- 0.0
            
            # ! update shifts
            temp <- R_grpnet_ordinal_int(nobs, nresp, eta, y, w^2, yfreq, shift, 1.0)
            shift <- temp$theta
            ctol <- max(temp$maxdif, ctol)
            
            # ! define r
            bigtheta <- c(100, shift, -100)
            fit0 <- eta + bigtheta[y]
            fit1 <- eta + bigtheta[y+1L]
            Pi0 <- 1 / (1 + exp(-fit0))
            Pi1 <- 1 / (1 + exp(-fit1))
            r <- w * (1.0 - Pi0 - Pi1)
            
            # ! update active groups
            for(k in 1:ngrps){
              if(active[k] == 0L) next
              penone <- alpha * lambda[i] * pw[k] / xev[k]
              pentwo <- (1 - alpha) * lambda[i] * pw[k] / xev[k]
              grad[ia[k]:ib[k]] <- crossprod(x[,ia[k]:ib[k]], r) / nobs
              zvec[ia[k]:ib[k]] <- beta[ia[k]:ib[k]] + grad[ia[k]:ib[k]] / xev[k]
              znorm <- sqrt(sum(zvec[ia[k]:ib[k]]^2))
              bnorm <- sqrt(sum(beta[ia[k]:ib[k]]^2))
              shrink <- R_grpnet_penalty(znorm, penid, penone, pentwo, gamma)
              if(shrink == 0.0 && bnorm == 0.0) next
              difbeta[ia[k]:ib[k]] <- shrink * zvec[ia[k]:ib[k]] - beta[ia[k]:ib[k]]
              maxdif <- max( abs(difbeta[ia[k]:ib[k]]) / (1.0 + abs(beta[ia[k]:ib[k]])) )
              beta[ia[k]:ib[k]] <- beta[ia[k]:ib[k]] + difbeta[ia[k]:ib[k]]
              eta <- eta + (x[,ia[k]:ib[k]] %*% difbeta[ia[k]:ib[k]]) / w
              fit0 <- eta + bigtheta[y]
              fit1 <- eta + bigtheta[y+1L]
              Pi0 <- 1 / (1 + exp(-fit0))
              Pi1 <- 1 / (1 + exp(-fit1))
              r <- w * (1.0 - Pi0 - Pi1)
              ctol <- max(maxdif , ctol)
              if(shrink > 0.0){
                nzgrps[i] <- nzgrps[i] + 1L
                edfs[i] <- edfs[i] + gsize[k] * shrink
              }
            } # ! k=1,ngrps
            
            # ! convergence check
            if(ctol < eps) break
            
          } # ! WHILE(iter < maxit) - inner
          # ! iterate until active coefficients converge ! #
          
          # ! check inactive groups in strong set ! #
          violations <- 0L
          for(k in 1:ngrps){
            if(strong[k] == 0L | active[k] == 1L) next
            grad[ia[k]:ib[k]] <- crossprod(x[,ia[k]:ib[k]], r) / nobs
            gradnorm[k] <- sqrt(sum(grad[ia[k]:ib[k]]^2))
            if(gradnorm[k] > alpha * lambda[i] * pw[k]){
              active[k] <- 1L
              violations <- violations + 1L
            }
          } # ! k=1,ngrps
          if(violations == 0L) break
          # ! check inactive groups in strong set ! #
          
        } # ! WHILE(iter < maxit) - middle
        # ! iterate until active set converges ! # 
        
        # ! check groups in weak set ! #
        violations <- 0L
        for(k in 1:ngrps){
          if(strong[k] == 1L) next
          grad[ia[k]:ib[k]] <- crossprod(x[,ia[k]:ib[k]], r) / nobs
          gradnorm[k] <- sqrt(sum(grad[ia[k]:ib[k]]^2))
          if(gradnorm[k] + 1e-8 > alpha * lambda[i] * pw[k]){
            strong[k] <- 1
            active[k] <- 1
            violations <- violations + 1L
          }
        }
        if(violations == 0L) break
        # ! check groups in weak set ! #
        
      } # ! WHILE(iter < maxit) - outer
      # ! iterate until strong set converges ! #
      
      # ! calculate nzcoef ! #
      for(k in 1:ngrps){
        if(active[k] == 0L) next
        for(l in 1:gsize[k]){
          if(abs(beta[ia[k] + l - 1]) > macheps){
            nzcoef[i] <- nzcoef[i] + 1L
          }
        }
      }
      # ! calculate nzcoef ! #
      
      # ! calculate deviance ! #
      mu <- Pi0 - Pi1
      devs[i] <- -sum(2 * w^2 * log(mu))
      # ! calculate deviance ! #
      
      # ! save results ! #
      ibeta[,i] <- shift
      betas[,i] <- beta
      iters[i] <- iter
      nzgrps[i] <- nzgrps[i] + intercept
      nzcoef[i] <- nzcoef[i] + intercept
      edfs[i] <- edfs[i] + as.numeric(intercept * (nresp - 1))
      # ! save results ! #
      
    }
    # ! --------------- ITERATIVE WORK --------------- ! #
    
    
    # ! --------------- POST PROCESSING --------------- ! #
    if(standardize == 1L){
      for(k in 1:ngrps){
        betas[ia[k]:ib[k],] <- betas[ia[k]:ib[k],] / xsdev[k]
      }
    }
    if(intercept == 1L){
      xmb <- as.numeric(xmean %*% betas)
      for(i in 1:nlam){
        ibeta[,i] <- ibeta[,i] - xmb[i]
      }
    } 
    if(makelambda == 1L && nzgrps[1] == 0L){
      bigtheta <- c(100, ibeta[,1], -100)
      fit0 <- bigtheta[y]
      fit1 <- bigtheta[y+1L]
      Pi0 <- 1 / (1 + exp(-fit0))
      Pi1 <- 1 / (1 + exp(-fit1))
    } else {
      shift <- log(yprob / (1 - yprob))
      eta <- rep(0.0, nobs)
      while(iter < maxit){
        ctol <- 0.0
        iter <- iter + 1L
        temp <- R_grpnet_ordinal_int(nobs, nresp, eta, y, w^2, yfreq, shift, 1.0)
        shift <- temp$theta
        ctol <- max(temp$maxdif, ctol)
        if(ctol < eps) break
      }  # ! WHILE(iter < maxit)
      bigtheta <- c(100, shift, -100)
      fit0 <- bigtheta[y]
      fit1 <- bigtheta[y+1L]
      Pi0 <- 1 / (1 + exp(-fit0))
      Pi1 <- 1 / (1 + exp(-fit1))
    }
    mu <- Pi0 - Pi1
    nulldev <- -sum(2 * w^2 * log(mu))
    names(xsdev) <- names(pw)
    pw <- xsdev
    # ! --------------- POST PROCESSING --------------- ! #
    
    
    # ! --------------- RETURN RESULTS --------------- ! #
    res <- list(nobs = nobs, 
                nvars = nvars, 
                x = x, 
                y = y, 
                w = w, 
                off = off, 
                ngrps = ngrps, 
                gsize = gsize, 
                pw = pw, 
                alpha = alpha, 
                nlam = nlam, 
                lambda = lambda, 
                lmr = lmr, 
                penid = penid, 
                gamma = gamma, 
                eps = eps, 
                maxit = maxit,
                standardize = standardize, 
                intercept = intercept, 
                ibeta = ibeta, 
                betas = betas, 
                iters = iters,
                nzgrps = nzgrps, 
                nzcoef = nzcoef, 
                edfs = edfs, 
                devs = devs, 
                nulldev = nulldev)
    return(res)
    # ! --------------- RETURN RESULTS --------------- ! #
    
    
  } # R_grpnet_binomial.R



R_grpnet_ordinal_int <-
  function(nobs, nresp, eta, y, w, yfreq, theta, maxdif){
    
    # nobs = n
    # nresp = J
    # eta = x %*% beta
    # y = integer vector (elements in 1,...,nresp)
    # w = non-negative weights
    # yfreq = integer vector of length nresp
    # theta = shift vector of length J-1
    
    
    #########***#########  GRADIENT AND HESSIAN   #########***#########
    
    # update bigtheta 
    bigtheta <- c(100, theta, -100)
    
    # initialize grad, avec, and bvec
    grad <- rep(0.0, nresp-1L)
    avec <- rep(0.0, nresp-1L)
    bvec <- rep(0.0, nresp-2L)
    
    # define shift + linear predictor (for y and y+1)
    fit0 <- eta + bigtheta[y]
    fit1 <- eta + bigtheta[y+1L]
    
    # evaluate gradient / hessian components
    Pi0 <- 1 / (1 + exp(-fit0))
    Pi1 <- 1 / (1 + exp(-fit1))
    u0 <- Pi0 * (1 - Pi0)
    u1 <- Pi1 * (1 - Pi1)
    mu <- Pi0 - Pi1
    
    # update bvec (off-diagonals)
    dtop <- exp(bigtheta[1:nresp] + bigtheta[2:(nresp+1)])
    dbot <- ( exp(bigtheta[1:nresp]) - exp(bigtheta[2:(nresp+1)]) )^2
    dtheta <- dtop / dbot
    bvec <- (yfreq[2:(nresp-1)] * dtheta[2:(nresp-1)]) / nobs
    
    # update other shift gradient and hessian components
    for(i in 1:(nresp-1)){
      j <- i + 1L
      yj <- (y == j)
      yjm1 <- (y == (j-1))
      grad[i] <- mean( w * (u0 * yj - u1 * yjm1) / mu )
      vj <- 1 / (1 + exp(-(eta + bigtheta[j])))
      vj <- (1 - vj)^2 - (1 - vj)
      va <- vj - dtheta[j-1]
      vb <- vj - dtheta[j]
      avec[i] <- - mean(w * ( yjm1 * va + yj * vb ) )
    }
    
    
    #########***#########  INVSYMTRIDIAG   #########***#########
    
    # initializations
    n <- nresp - 1L
    u <- rep(0.0, n)
    v <- rep(0.0, n)
    d <- rep(0.0, n)
    delta <- rep(0.0, n)
    
    # update d
    d[n] <- avec[n]
    for(i in (n-1):1){
      d[i] <- avec[i] - bvec[i]^2 / d[i+1]
    }
    
    # update v
    v[1] <- 1 / d[1]
    for(i in 2:n){
      v[i] <- prod(bvec[1:(i-1)] / d[1:(i-1)]) / d[i]
    }
    
    # update delta
    delta[1] <- avec[1]
    for(i in 1:(n-1)){
      delta[i+1] <- avec[i+1] - bvec[i]^2 / delta[i]
    }
    
    # update u
    u[n] <- 1 / (delta[n] * v[n])
    for(i in 1:(n-1)){
      u[i] <- prod(bvec[i:(n-1)] / delta[i:(n-1)]) / (delta[n] * v[n])
    }
    
    
    #########***#########  NEWTON-RAPHSON UPDATE   #########***#########
    
    # calculate solve(hess) %*% grad
    ivec <- rep(0.0, n)
    for(i in 1:n){
      ivec[i] <- sum( (u[pmin(i, 1:n)] * v[pmax(i, 1:n)]) * grad )
    }
    
    # update maxdif and theta
    maxdif <- max( abs(ivec) / (1.0 + abs(theta)) )
    theta <- theta + ivec
    
    # return results
    res <- list(nobs = nobs,
                nresp = nresp,
                eta = eta, 
                y = y, 
                w = w,
                yfreq = yfreq,
                theta = theta,
                maxdif = maxdif)
    return(res)
    
  } # end R_grpnet_ordinal_int

# Some code (for printing and plotting) is re-purposed from the R package
# "glmnet" (Hastie et al., 2010) https://cran.r-project.org/package=glmnet

grpnet <- 
  function(x, ...){
    UseMethod("grpnet")
  } # end grpnet

print.grpnet <-
  function(x, ...){
    cat("\nCall: ")
    print(x$call)
    cat("\n")
    df <- data.frame(x$nzgrp,
                     x$df, 
                     round(100 * x$dev.ratio, 2),
                     x$lambda)
    colnames(df) <- c("nGrp", "Df", "%Dev", "Lambda")
    intercept <- ifelse(max(abs(x$a0)) > 0, TRUE, FALSE)
    print(df, digits = 6)
    cat("\n")
  } # end print.grpnet

plot.grpnet <-
  function(x, type = c("dev.ratio", "coef", "imp", "norm", "znorm"), 
           newx, newdata, intercept = FALSE,
           color.by.group = TRUE, col = NULL, ...){
    int <- ifelse(x$args$intercept && !intercept, 1, 0)
    types <- c("dev.ratio", "coef", "imp", "norm", "znorm")
    type <- pmatch(type[1], types)
    if(is.na(type)) stop("Invalid 'type' input")
    type <- types[type]
    if(color.by.group) {
      if(is.null(col)){
        col <- 1:x$ngroups
      } else {
        if(length(col) != x$ngroups) stop("Input 'col' must be of length x$ngroups")
      }
    } else {
      if(is.null(col)) col <- rep("black", x$ngroups)
    }
    if(type == "imp" && missing(newx) && missing(newdata)){
      stop("When type = 'imp', you need to provide either 'newx' or 'newdata'.")
    }
    if(type == "dev.ratio"){
      plot(log(x$lambda), x$dev.ratio,
           xlab = expression(log(lambda)), ylab = "Explained Deviance", ...)
    } else {
      res <- predict(x, newx = newx, newdata = newdata, type = type)
      if(type == "imp"){
        colors <- col
        if(x$family$family %in% c("multigaussian", "multinomial")){
          rnames <- rownames(res[[1]])
          cnames <- colnames(res[[1]])
          for(k in 1:length(res)){
            res[[k]] <- rbind(0, res[[k]])
            rownames(res[[k]]) <- c("(Intercept)", rnames)
            colnames(res[[k]]) <- cnames
          }
          index <- (1+int):nrow(res[[1]])
          for(j in 1:length(res)){
            plot(log(x$lambda), res[[j]][1,], ylim = extendrange(sapply(res, function(x) range(x[index,], na.rm = TRUE))),
                 xlab = expression(log(lambda)), ylab = "Importance", t = "n", ...)
            legend("top", legend = x$ylev[j], bty = "n", cex = 0.8)
            for(k in index) {
              lines(log(x$lambda), res[[j]][k,], col = colors[k])
            }
          }
        } else {
          rnames <- rownames(res)
          cnames <- colnames(res)
          res <- rbind(0, res)
          rownames(res) <- c("(Intercept)", rnames)
          colnames(res) <- cnames
          index <- (1+int):nrow(res)
          plot(log(x$lambda), res[1,], ylim = extendrange(res[index,]),
               xlab = expression(log(lambda)), ylab = "Importance", t = "n", ...)
          for(k in index) {
            lines(log(x$lambda), res[k,], col = colors[k])
          }
        }
      } else if(type %in% c("norm", "znorm")){
        colors <- col
        index <- (1+int):nrow(res)
        plot(log(x$lambda), res[1,], ylim = extendrange(res[index,]),
             xlab = expression(log(lambda)), ylab = "L2 Norm", t = "n", ...)
        for(k in index) {
          lines(log(x$lambda), res[k,], col = colors[k])
        }
      } else {
        colors <- col[as.integer(as.factor(x$group))]
        if(x$family$family %in% c("multigaussian", "multinomial")){
          index <- (1+int):nrow(res[[1]])
          for(j in 1:length(res)){
            plot(log(x$lambda), res[[j]][1,], ylim = extendrange(res[[j]][index,]),
                 xlab = expression(log(lambda)), ylab = "Coefficients", t = "n", 
                 main = x$ylev[j], ...)
            #legend("top", legend = x$ylev[j], bty = "n", cex = 0.8)
            for(k in index) {
              lines(log(x$lambda), res[[j]][k,], col = colors[k])
            }
          }
        } else {
          if(x$family$family == "ordinal"){
            index <- length(x$ylev):nrow(res)
          } else{
            index <- (1+int):nrow(res)
          }
          plot(log(x$lambda), res[1,], ylim = extendrange(res[index,]),
               xlab = expression(log(lambda)), ylab = "Coefficients", t = "n", ...)
          for(k in index) {
            lines(log(x$lambda), res[k,], col = colors[k])
          }
        }
      }
    abline(h = 0)
    }
  } # end plot.grpnet

visualize.shrink <-
  function(x = seq(-5, 5, length.out = 1001), 
           penalty = c("LASSO", "MCP", "SCAD"), 
           alpha = 1, 
           lambda = 1, 
           gamma = 4, 
           fitted = FALSE,
           plot = TRUE,
           subtitle = TRUE,
           legend = TRUE,
           location = "top",
           ...){
    # plot grpnet shrinkage and selection operators
    # Nathaniel E. Helwig (helwig@umn.edu)
    # 2025-06-03
    
    
    #########***#########   INITIAL CHECKS   #########***#########
    
    # check x
    xorig <- sort(as.numeric(x))
    x <- abs(xorig)
    nobs <- length(x)
    
    # check penalty
    penalties <- c("LASSO", "MCP", "SCAD")
    penalty <- sort(unique(as.character(penalty)))
    pid <- pmatch(penalty, penalties)
    if(any(is.na(pid))) stop(paste("Input 'penalty' is invalid. Must be some subset of\n", 
                                   paste(penalties, collapse = ", ")))
    penalty <- penalties[pid]
    npen <- length(penalty)
    
    # check alpha
    alpha <- as.numeric(alpha[1])
    if(alpha < 0 | alpha > 1) stop("Input 'alpha' must satisfy:  0 <= alpha <= 1")
    
    # check lambda
    lambda <- as.numeric(lambda[1])
    if(lambda < 0) stop("Input 'lambda' must satisfy:  lambda >= 0")
    
    # check gamma
    gamma <- as.numeric(gamma[1])
    if(any(penalty == "SCAD")){
      if(gamma <= 2) stop("Input 'gamma' must satisfy: gamma > 2 (SCAD)")
    } else if(any(penalty == "MCP")){
      if(gamma <= 1) stop("Input 'gamma' must satisfy: gamma > 1 (MCP)")
    } 
    
    
    
    #########***#########   EVALUATE SHRINKAGE   #########***#########
    
    # initialize l1 and l2 penalty weights
    lambda1 <- lambda * alpha
    lambda2 <- lambda * (1-alpha)
    
    # initialize matrix to hold results
    res <- matrix(0.0, nrow = nobs, ncol = npen)
    colnames(res) <- penalty
    
    # loop through penalties
    for(j in 1:npen){
      
      if(penalty[j] == "LASSO"){
        res[,j] <- pmax(0, 1 - lambda1 / x) / (1 + lambda2)
      } else if (penalty[j] == "MCP"){
        id <- (x <= gamma * lambda1 * (1 + lambda2))
        res[id,j] <- pmax(0, 1 - lambda1 / x[id]) / (1 + lambda2 - 1/gamma)
        res[!id,j] <- 1 / (1 + lambda2)
      } else if (penalty[j] == "SCAD"){
        res[,j] <- 1 / (1 + lambda2)
        id <- which(x <= (lambda1 + lambda1 * (1 + lambda2)))
        res[id,j] <- pmax(0, 1 - lambda1 / x[id]) / (1 + lambda2)
        id <- which( (x > (lambda1 + lambda1 * (1 + lambda2))) & (x <= gamma * lambda1 * (1 + lambda2)) )
        res[id,j] <- pmax(0, 1 - (lambda1 / x[id]) * (gamma/(gamma-1)) ) / (1 + lambda2 - 1 / (gamma-1))
      }
      
    }  # end for(j in 1:npen)
    
    # fitted values instead of shrinkage factors?
    if(fitted) res <- xorig * res
    
    # return results?
    if(!plot){
      return(as.data.frame(cbind(x = xorig, res))) 
    }
    
    
    
    #########***#########   COLLECT ARGS   #########***#########
    
    # collect ...
    args <- list(...)
    
    # check args$ylim
    if(is.null(args$ylim)) args$ylim <- extendrange(res)
    
    # check args$xlim
    if(is.null(args$xlim) && fitted) args$xlim <- args$ylim
    
    # check args$lty
    if(is.null(args$lty)) args$lty <- c(1, 2, 4)
    
    # check args$lwd
    if(is.null(args$lwd)) args$lwd <- c(2, 2, 2)
    
    # check args$col
    if(is.null(args$col)) args$col <- c("black", "blue", "red")
    
    # check args$xlab
    if(is.null(args$xlab)) args$xlab <- expression(beta)
    
    # check args$ylab
    if(is.null(args$ylab)) args$ylab <- ifelse(fitted,
                                               expression(italic(S)[lambda[1] * ", " * lambda[2]](beta) %*% beta),
                                               expression(italic(S)[lambda[1] * ", " * lambda[2]](beta)))
      
    # check args$main
    if(is.null(args$main)) args$main <- ifelse(fitted,
                                               "shrunken estimators",
                                               "shrinkage operators")
    
    
    
    #########***#########   PLOT PENALTY   #########***#########
    
    # plot first penalty
    plot(xorig, res[,1], type = "l", lty = args$lty[1], col = args$col[1], 
         lwd = args$lwd[1], xlim = args$xlim, ylim = args$ylim, 
         xlab = args$xlab, ylab = args$ylab, main = args$main)
    
    if(npen > 1L){
      for(j in 2:npen){
        lines(xorig, res[,j], lty = args$lty[j], col = args$col[j], lwd = args$lwd[j])
      }
    }
    
    # fitted?
    if(fitted) abline(a = 0, b = 1, lty = 3)
    
    # subtitle?
    if(subtitle){
      subtit <- bquote(alpha == .(alpha) ~ "    " ~ gamma == .(gamma) ~ "    " ~ lambda == .(lambda))
      mtext(as.expression(subtit), side = 3)
    }
    
    # legend?
    if(legend){
      legend(location, legend = penalty, 
             col = args$col, lty = args$lty, lwd = args$lwd, bty = "n")
    }
    
    
  } # end visualize.shrink

# Some code (for arguments and outputs) is re-purposed from the R package
# "glmnet" (Hastie et al., 2010) https://cran.r-project.org/package=glmnet

grpnet.formula <-
  function(formula,
           data,
           use.rk = TRUE,
           family = c("gaussian", "multigaussian", 
                      "svm1", "svm2", "logit",
                      "binomial", "multinomial", "ordinal",
                      "poisson", "negative.binomial", 
                      "Gamma", "inverse.gaussian"),
           weights = NULL,
           offset = NULL,
           alpha = 1,
           nlambda = 100,
           lambda.min.ratio = ifelse(nobs < nvars, 0.05, 1e-4),
           lambda = NULL,
           penalty.factor = NULL,
           penalty = c("LASSO", "MCP", "SCAD"),
           gamma = 4,
           theta = 1,
           standardized = !orthogonalized,
           orthogonalized = TRUE,
           thresh = 1e-04,
           maxit = 1e05,
           proglang = c("Fortran", "R"),
           standardize.response = FALSE,
           keep.data = TRUE,
           ...){
    # group elastic net regularized regression (formula)
    # Nathaniel E. Helwig (helwig@umn.edu)
    # Updated: 2025-04-07
    
    
    ######***######   INITIAL CHECKS   ######***######
    
    ### get call
    grpnet.call <- match.call()
    
    ### check formula
    formula <- as.formula(formula)
    charform <- as.character(formula)
    if(charform[1] != "~") stop("Input 'formula' must be of the form:  y ~ x")
    
    ### model frame (modified from R's lm() function)
    mf <- match.call(expand.dots = FALSE)
    m <- match(c("formula", "data"), names(mf), 0L)
    mf <- mf[c(1L, m)]
    mf$drop.unused.levels <- TRUE
    ## need stats:: for non-standard evaluation
    mf[[1L]] <- quote(stats::model.frame)
    mf <- eval(mf, parent.frame())
    
    ### model response
    yname <- colnames(mf)[1]
    y <- model.response(mf, "any")
    
    ### model matrix
    rk.args <- NULL
    if(use.rk){
      x <- rk.model.matrix(object = formula, data = mf, ...)
      rk.args <- list(knots = attr(x, "knots"),
                      m = attr(x, "m"),
                      periodic = attr(x, "periodic"),
                      xlev = attr(x, "xlev"))
    } else {
      x <- model.matrix(object = formula, data = mf)
    }
    
    ### model terms
    terms <- attr(mf, "terms")
    xvar.class <- attr(terms, "dataClasses")[-1]
    term.labels <- attr(terms, "term.labels")
    
    ### nobs and nvars
    nobs <- nrow(x)
    nvars <- ncol(x)
    nxvars <- length(xvar.class)
    
    ### group vector
    group <- attr(x, "assign")
    gsize <- table(group)
    ngrps <- length(gsize)
    
    ### intercept?
    intercept <- ifelse(group[1] == 0, TRUE, FALSE)
    if(intercept) {
      x <- x[,-1]
      group <- group[-1]
      gsize <- gsize[-1]
      ngrps <- ngrps - 1L
      nvars <- nvars - 1L
    }
    
    ### penalty.factor
    pf <- NULL
    if(!is.null(penalty.factor)){
      pf <- sqrt(gsize)
      names(pf) <- term.labels
      penalty.factor <- as.list(penalty.factor)
      penfacname <- names(penalty.factor)
      if(is.null(penfacname)) stop("Input 'penalty.factor' must be a named list.")
      for(k in 1:length(penalty.factor)){
        id <- match(penfacname[k], term.labels)
        if(is.na(id)) warning("Input 'penalty.factor' contains the term:   '", penfacname[k], "'\nwhich does not match any of the model terms")
        pf[id] <- penalty.factor[[k]]
      }
    } # if(!is.null(penalty.factor))
    
    
    ######***######   GRPNET.DEFAULT   ######***######
    
    ### fit regularization path
    res <- grpnet.default(x = x, 
                          y = y, 
                          group = group,
                          family = family,
                          weights = weights,
                          offset = offset,
                          alpha = alpha, 
                          nlambda = nlambda,
                          lambda.min.ratio = lambda.min.ratio,
                          lambda = lambda,
                          penalty.factor = pf,
                          penalty = penalty,
                          gamma = gamma,
                          theta = theta,
                          standardized = standardized,
                          orthogonalized = orthogonalized,
                          intercept = intercept,
                          thresh = thresh,
                          maxit = maxit,
                          proglang = proglang,
                          standardize.response = standardize.response)
    
    
    ######***######   POST-PROCESSING   ######***######
    
    ### correct the call
    res$call <- grpnet.call
    
    ### add the (potential expanded) formula
    res$formula <- as.formula(paste0(yname, " ~ ", paste(term.labels, collapse = " + ")))
    
    ### add the term.labels
    if(intercept) term.labels <- c("(Intercept)", term.labels)
    res$term.labels <- term.labels
    
    ### add the rk.args
    res$rk.args <- rk.args
    
    ### add the data
    res$data <- data
    
    ### return results
    return(res)
    
  } # end grpnet.formula

rk <- 
  function(x, df = NULL, knots = NULL, m = NULL, intercept = FALSE, 
           Boundary.knots = NULL, warn.outside = TRUE, 
           periodic = FALSE, xlev = levels(x)){
    # reproducing kernel spline basis
    # Nathaniel E. Helwig (helwig@umn.edu)
    # 2025-01-21
    
    
    ######***######   NOMINAL BASIS   ######***######
    
    if(!is.null(xlev) && !is.ordered(x)){
      
      # check x
      x <- factor(x, levels = xlev)
      nlev <- length(xlev)
      n <- length(x)
      
      # check df and knots
      if(is.null(knots)){
        if(is.null(df)) {
          df <- nlev
        } else {
          df <- as.integer(df[1])
          if(df < 1L) stop("The 'df' argument must satisfy:  df >= 1")
          if(df > nlev) stop("The 'df' argument must satisfy:  df <= nlevels(x)")
        }
        knots <- xlev[seq(1, nlev, length.out = df)]
        knots <- factor(knots, levels = xlev)
      } else {
        knots <- factor(knots, levels = xlev)
        df <- length(knots)
      }
      if(df < 2L) stop("Input 'df' must be greater than one.")
      
      # evaluate basis
      mat <- outer(x, knots[-df], FUN = "==") + 0.0
      id <- which(x == knots[df])
      mat[id,] <- -1
      
      # intercept?
      if(intercept) {
        cnames <- c("(Intercept)", paste0("knot", 1:ncol(mat)))
        mat <- cbind(1, mat)
      } else {
        cnames <- paste0("knot", 1:ncol(mat))
      }
      colnames(mat) <- cnames
      
      # append attributes
      attr(mat, "df") <- df
      attr(mat, "knots") <- knots
      attr(mat, "m") <- 0L
      attr(mat, "intercept") <- intercept
      attr(mat, "Boundary.knots") <- Boundary.knots
      attr(mat, "periodic") <- periodic
      attr(mat, "xlev") <- xlev
      
      # return basis
      return(mat)
      
    } # end if(!is.null(xlev) && !is.ordered(x))
    
    
    ######***######   ORDINAL BASIS   ######***######
    
    # check m
    if(is.null(m)){
      m <- ifelse(is.null(xlev), 2L, 0L)
    } else {
      m <- as.integer(m[1])
      if(m < 0L | m > 3L) stop("Input 'm' must be 0 (ordinal), 1 (linear), 2 (cubic), or 3 (quintic)")
    }
    
    # is ordinal?
    if(m == 0L){
      
      # check x
      if(is.null(xlev)){
        x <- as.ordered(x)
        xlev <- levels(x)
      } else {
        x <- factor(x, levels = xlev, ordered = TRUE)
      }
      nlev <- length(xlev)
      n <- length(x)
      
      # check df and knots
      if(is.null(knots)){
        if(is.null(df)) {
          df <- nlev
        } else {
          df <- as.integer(df[1])
          if(df < 1L) stop("The 'df' argument must satisfy:  df >= 1")
          if(df > nlev) stop("The 'df' argument must satisfy:  df <= nlevels(x)")
        }
        knots <- xlev[seq(1, nlev, length.out = df)]
        knots <- factor(knots, levels = xlev, ordered = TRUE)
      } else {
        knots <- factor(knots, levels = xlev, ordered = TRUE)
        df <- length(knots)
      }
      
      # transform x and knots
      ix <- as.integer(x)
      ik <- as.integer(knots)
      
      # evaluate basis
      const <- (nlev - 1) * (2*nlev - 1) / (6 * nlev)
      mat <- 1 - outer(ix, ik, pmax) + const
      mat <- mat + outer(ix*(ix-1), ik*(ik-1), FUN = "+") / (2 * nlev)
      
      # evaluate penalty
      pen <- 1 - outer(ik, ik, pmax) + const
      pen <- pen + outer(ik*(ik-1), ik*(ik-1), FUN = "+") / (2 * nlev)
      
      # absorb (inverse square root of) penalty into basis
      peig <- eigen(pen, symmetric = TRUE)
      prank <- sum(peig$values > peig$values[1] * df * .Machine$double.eps)
      pisqrt <- peig$vectors[,1:prank,drop=FALSE] %*% diag(1/sqrt(peig$values[1:prank]), nrow = prank, ncol = prank)
      mat <- mat %*% pisqrt
      
      # intercept?
      if(intercept) {
        cnames <- c("(Intercept)", paste0("knot", 1:ncol(mat)))
        mat <- cbind(1, mat)
      } else {
        cnames <- paste0("knot", 1:ncol(mat))
      }
      colnames(mat) <- cnames
      
      # append attributes
      attr(mat, "df") <- df
      attr(mat, "knots") <- knots
      attr(mat, "m") <- m
      attr(mat, "intercept") <- intercept
      attr(mat, "Boundary.knots") <- Boundary.knots
      attr(mat, "periodic") <- periodic
      attr(mat, "xlev") <- xlev
      
      # return basis
      return(mat)
      
    } # end if(m == 0L)
    
    
    
    ######***######   POLYNOMIAL BASIS   ######***######
    
    # check x
    x <- as.numeric(x)
    n <- length(x)
    
    # check Boundary.knots
    if(is.null(Boundary.knots)){
      Boundary.knots <- range(x)
    } else {
      Boundary.knots <- Boundary.knots[1:2]
      if(Boundary.knots[1] >= Boundary.knots[2]) stop("Input 'Boundary.knots' must satisfy: Boundary.knots[1] < Boundary.knots[2]")
      if(warn.outside){
        if(min(x) < Boundary.knots[1] - 1e-8) warning("min(x) < Boundary.knots[1]")
        if(max(x) > Boundary.knots[2] + 1e-8) warning("max(x) > Boundary.knots[2]")
      }
    }
    
    # check df and knots
    if(is.null(knots)){
      if(is.null(df)){
        df <- 5L
      } else {
        df <- as.integer(df[1])
        if(df < 1L) stop("The 'df' argument must satisfy:  df >= 1")
        if(df > n) warning("Requested 'df' is greater than the length of 'x'")
      }
      knots <- quantile(x, probs = seq(0, 1, length.out = df))
    } else {
      knots <- as.numeric(knots)
    }
    
    # combine Boundary.knots and knots
    knots <- unique(sort(c(Boundary.knots, knots)))
    df <- length(knots)
    
    # check periodic
    if(is.null(periodic)) periodic <- FALSE
    periodic <- as.logical(periodic[1])
    
    # bernoulli polynomial function
    bernpoly <- function(x, m){
      val <- x - 1/2
      if(m == 2L){
        val <- (val^2 - 1/12) / 2
      } else if(m == 3L){
        val <- (val^3 - val / 4) / 6
      } else if(m == 4L){
        val <- (val^4 - val^2/2 + 7/240) / 24
      } else if(m == 5L){
        val <- (val^5 - (5/6) * val^3 - (7/48) * val) / 120
      } else if(m == 6L){
        val <- (val^6 - (5/4) * val^4 + (7/16) * val^2 - 31/1344) / 720
      }
      return(val)
    } # end bp
    
    # transform x and knots
    xr <- diff(Boundary.knots)
    if(xr < .Machine$double.eps) xr <- 1.0
    x <- (x - Boundary.knots[1]) / xr
    knots <- (knots - Boundary.knots[1]) / xr
    
    # evaluate basis
    mat <- matrix(0, nrow = n, ncol = df)
    if(!periodic){
      for(k in 1:m){
        mat <- mat + outer(bernpoly(x, m = k), bernpoly(knots, m = k))
      }
    }
    dif <- abs(outer(x, knots, FUN = "-"))
    mat <- mat + (-1)^(m-1) * bernpoly(dif, m = 2*m)
    
    # evaluate penalty
    pen <- matrix(0, nrow = df, ncol = df)
    if(!periodic){
      for(k in 1:m){
        pen <- pen + outer(bernpoly(knots, m = k), bernpoly(knots, m = k))
      }
    }
    dif <- abs(outer(knots, knots, FUN = "-"))
    pen <- pen + (-1)^(m-1) * bernpoly(dif, m = 2*m)
    
    # absorb (inverse square root of) penalty into basis
    peig <- eigen(pen, symmetric = TRUE)
    prank <- sum(peig$values > peig$values[1] * df * .Machine$double.eps)
    pisqrt <- peig$vectors[,1:prank,drop=FALSE] %*% diag(1/sqrt(peig$values[1:prank]), nrow = prank, ncol = prank)
    mat <- mat %*% pisqrt
    
    # intercept?
    if(intercept) {
      cnames <- c("(Intercept)", paste0("knot", 1:ncol(mat)))
      mat <- cbind(1, mat)
    } else {
      cnames <- paste0("knot", 1:ncol(mat))
    }
    colnames(mat) <- cnames
    
    # append attributes
    attr(mat, "df") <- df
    attr(mat, "knots") <- knots * xr + Boundary.knots[1]
    attr(mat, "m") <- m
    attr(mat, "intercept") <- intercept
    attr(mat, "Boundary.knots") <- Boundary.knots
    attr(mat, "periodic") <- periodic
    
    # return basis
    return(mat)
    
  } # end rk

visualize.loss <-
  function(x = seq(-3, 3, length.out = 1001),
           family = c("gaussian", "multigaussian",
                      "svm1", "svm2", "logit",
                      "binomial", "multinomial", 
                      "poisson", "negative.binomial", 
                      "Gamma", "inverse.gaussian"),
           theta = 1,
           type = c("link", "response"),
           y = NULL,
           plot = TRUE,
           add = FALSE,
           ...){
    # plot grpnet loss functions
    # Nathaniel E. Helwig (helwig@umn.edu)
    # 2025-06-03
    
    
    
    #########***#########   INITIAL CHECKS   #########***#########
    
    # check family
    family <- as.character(family[1])
    fam <- family.grpnet(family, theta = theta)
    
    # check theta
    theta <- as.numeric(theta[1])
    if(theta <= 0.0) stop("Input 'theta' must be positive")
    
    # check type
    type <- as.character(type[1])
    types <- c("link", "response")
    type <- pmatch(type, types)
    if(is.na(type)) stop("Input 'type' must be 'link' or 'response'")
    type <- types[type]
    
    # check x and y
    x <- as.numeric(x)
    if(!is.null(y)) y <- as.numeric(y[1])
    
    # check add
    add <- as.logical(add[1])
    if(!any(add == c(TRUE, FALSE))) stop ("Input 'add' must be TRUE or FALSE")
    
    # convert x to mu
    mu <- fam$linkinv(x)
    
    
    
    #########***#########   EVALUATE LOSS   #########***#########
    
    # evaluate loss
    if(family %in% c("gaussian", "multigaussian")){
      y <- ifelse(is.null(y), 0, y[1])
      loss <- (y - mu)^2
    } else if(family == "svm1"){
      y <- ifelse(is.null(y), 1, y[1])
      muy <- mu * y
      loss <- rep(0, length(mu))
      id <- (muy > 1 - theta)
      loss[id] <- pmax(1 - muy[id], 0)^2 / (2 * theta)
      loss[!id] <- 1 - muy[!id] - theta/2
    } else if(family == "svm2"){
      y <- ifelse(is.null(y), 1, y[1])
      loss <- pmax(0, 1 - mu * y)^2
    } else if(family == "logit"){
      y <- ifelse(is.null(y), 1, y[1])
      loss <- log(1 + exp(-x * y))
    } else if(family %in% c("binomial", "multinomial")){
      y <- ifelse(is.null(y), 1, y[1])
      loss <- - y * log(mu) - (1 - y) * log(1 - mu)
    } else if(family == "poisson"){
      y <- ifelse(is.null(y), 1, y[1])
      loss <- mu - y * log(mu)
    } else if(family == "negative.binomial"){
      y <- ifelse(is.null(y), 1, y[1])
      const <- lgamma(theta) - lgamma(theta + y) - theta * log(theta)
      loss <- (theta + y) * log(theta + mu) - y * log(mu) + const
    } else if(family == "Gamma"){
      y <- ifelse(is.null(y), 1, y[1])
      loss <- log(mu) + y / mu
    } else if(family == "inverse.gaussian"){
      y <- ifelse(is.null(y), 1, y[1])
      loss <- (y - mu)^2 / (mu^2 * y)
    }
    
    
    
    #########***#########   RETURN LOSS?   #########***#########
    
    if(!plot){
      df <- data.frame(eta = x, mu = mu, loss = loss)
      return(df)
    } 
    
    
    
    #########***#########   PLOT LOSS   #########***#########
    
    # collect ...
    args <- list(...)
    
    # add x and y
    if(type == "link"){
      args$x <- x
    } else {
      args$x <- mu
    }
    args$y <- loss
    
    # add xlab and ylab
    if(is.null(args$xlab)) {
      if(type == "link"){
        args$xlab <- expression(italic(eta))
      } else {
        args$xlab <- expression(italic(mu))
      }
    }
    if(is.null(args$ylab)) {
      if(type == "link"){
        args$ylab <- substitute(expression(italic(L) * "( " * eta * " | " * italic(y) * " = " * yval * " )"),
                                list(yval = y))
      } else {
        args$ylab <- substitute(expression(italic(L) * "( " * mu * " | " * italic(y) * " = " * yval * " )"),
                                list(yval = y))
      }
    }
    if(is.null(args$main)) args$main <- family
    
    # add type
    if(!add) args$type <- "l"
    
    # check args$xlim
    if(is.null(args$xlim)) {
      if(type == "link"){
        args$xlim <- extendrange(x)
      } else {
        args$xlim <- extendrange(mu)
      }
    }
    
    # check args$ylim
    if(is.null(args$ylim)) args$ylim <- extendrange(loss)
    
    # check args$lty
    if(is.null(args$lty)) args$lty <- 1L
    
    # check args$lwd
    if(is.null(args$lwd)) args$lwd <- 2L
    
    # check args$col
    if(is.null(args$col)) args$col <- "darkgray"
    
    # add lines or draw plot
    if(add){
      do.call(lines, args)
    } else {
      rm(plot)  # remove logical "plot" argument
      do.call(plot, args)
    }
    
  } # end visualize.loss

family.grpnet <-
  function(object, theta = 1){
    # prepare family for grpnet
    # Nathaniel E. Helwig (helwig@umn.edu)
    # Updated: 2025-08-15
    
    # check class of object
    if(inherits(object, "cv.grpnet")){
      return(object$grpnet.fit$family)
    } else if(inherits(object, "grpnet")){
      return(object$family)
    }
    
    # check input family object
    object <- as.character(object[1])
    if(object == "mgaussian" | object == "mvn") object <- "multigaussian"
    if(object == "hsvm") object <- "svm1"
    if(object == "sqsvm") object <- "svm2"
    if(object == "polr") object <- "ordinal"
    families <- c("gaussian", "multigaussian", "svm1", "svm2", "logit", "binomial", "multinomial", "ordinal", "poisson", "negative.binomial", "Gamma", "inverse.gaussian")
    object <- pmatch(object, families)
    if(is.na(object)) stop("'object' not recognized")
    object <- families[object]
    
    # create needed family info
    if(object %in% c("gaussian", "multigaussian")){
      
      famname <- object
      object <- gaussian()
      object$family <- famname
      
    } else if(object == "svm1"){
      
      theta <- as.numeric(theta[1])
      if(theta <= 0) stop("Input 'theta' must be positive")
      .Theta <- theta
      env <- new.env(parent = .GlobalEnv)
      assign(".Theta", theta, envir = env)
      dr <- function(y, mu, wt){
        muy <- mu * y
        vec <- rep(0.0, length(muy))
        idx <- which(muy > 1.0 - .Theta)
        vec[idx] <- (1.0 - muy[idx])^2 / (2.0 * .Theta)
        vec[!idx] <- 1.0 - muy[!idx] - .Theta / 2.0
        vec[which(muy > 1.0)] <- 0.0
        2 * wt * vec
      }
      object <- list(family = "svm1",
                     linkinv = function(eta) eta,
                     dev.resids = dr)
      
    } else if(object == "svm2"){
      
      dr <- function(y, mu, wt){
        2 * wt * pmax(0, 1 - mu * y)^2
      }
      object <- list(family = "svm2",
                     linkinv = function(eta) eta,
                     dev.resids = dr)
      
    } else if(object == "logit"){
      
      dr <- function(y, mu, wt){
        2 * wt * log(1 + exp(-mu*y)) 
      }
      object <- list(family = "logit",
                     linkinv = function(eta) {1 / (1 + exp(-eta))},
                     dev.resids = dr)
      
    } else if(object == "binomial"){
      
      dr <- function(y, mu, wt){
        ylogy <- function(y, mu){
          res <- y * log(y / mu)
          res[is.nan(res)] <- 0
          res
        }
        mu[mu < 0.000001] <- 0.000001
        mu[mu > 0.999999] <- 0.999999
        2 * wt * (ylogy(y, mu) + ylogy(1 - y, 1 - mu))
      }
      object <- list(family = "binomial",
                     linkinv = function(eta) {1 / (1 + exp(-eta))},
                     dev.resids = dr)
      
    } else if (object == "multinomial"){
      
      il <- function(eta){
        expeta <- exp(eta - apply(eta, 1, max))
        mu <- expeta / rowSums(expeta)
        mu[mu < 0.000001] <- 0.000001
        mu[mu > 0.999999] <- 0.999999
        mu
      } # end il
      dr <- function(y, mu, wt) {
        mu[mu < 0.000001] <- 0.000001
        mu[mu > 0.999999] <- 0.999999
        -2 * wt * rowSums(y * log(mu))
      }
      object <- list(family = "multinomial",
                     linkinv = il,
                     dev.resids = dr)
      
    } else if(object == "ordinal"){
      
      dr <- function(y, mu, wt){
        mu[mu < 0.000001] <- 0.000001
        mu[mu > 0.999999] <- 0.999999
        -2 * wt * log(mu)
      }
      object <- list(family = "ordinal",
                     linkinv = function(eta) {1 / (1 + exp(-eta))},
                     dev.resids = dr)
      
    } else if(object == "poisson"){
      
      il <- function(eta) pmax(exp(eta), .Machine$double.eps)
      dr <- function(y, mu, wt){
        r <- mu * wt
        p <- which(y > 0)
        if(is.matrix(mu) && ncol(mu) > 1L){
          r[p,] <- (wt * (y * log(y/mu) - (y - mu)))[p,]
        } else {
          r[p] <- (wt * (y * log(y/mu) - (y - mu)))[p]
        }
        2 * r
      }
      object <- list(family = "poisson",
                     linkinv = il,
                     dev.resids = dr)
      
    } else if(object == "negative.binomial"){
      
      theta <- as.numeric(theta[1])
      if(theta <= 0) stop("Input 'theta' must be positive")
      .Theta <- theta
      env <- new.env(parent = .GlobalEnv)
      assign(".Theta", theta, envir = env)
      
      il <- function(eta) pmax(exp(eta), .Machine$double.eps)
      dr <- function(y, mu, wt){
        2 * wt * ( y * log(pmax(1, y) / mu) - (y + .Theta) * log((y + .Theta) / (mu + .Theta)) )
      }
      object <- list(family = "negative.binomial",
                     linkinv = il,
                     dev.resids = dr)
      
    } else if(object == "Gamma"){
      
      il <- function(eta) pmax(exp(eta), .Machine$double.eps)
      dr <- function(y, mu, wt){
        -2 * wt * (log(y / mu) - (y - mu) / mu)
      }
      object <- list(family = "Gamma",
                     linkinv = il,
                     dev.resids = dr)
      
    } else if(object == "inverse.gaussian"){
      
      il <- function(eta) pmax(exp(eta), .Machine$double.eps)
      dr <- function(y, mu, wt){
        wt * ( (y - mu)^2 / (y * mu^2) )
      }
      object <- list(family = "inverse.gaussian",
                     linkinv = il,
                     dev.resids = dr)
      
    } # end if(object == "gaussian")
    
    return(object)
    
  } # family.grpnet.R

!   grpnet_multigaus.f90 - group elastic net (multivariate gaussian)
!   Nathaniel E. Helwig (helwig@umn.edu)
!   Department of Psychology and School of Statistics
!   University of Minnesota
!   Date: 2025-01-17


! INPUTS/OUTPUTS
!   nobs = number of observations (N)
!   nvars = number of variables (P)
!   nresp = number of response dimensions for y (M)
!   x = grouped predictor matrix (N,P)
!       note: (x.1, x.2, ..., x.K) where x.k is (N,P.k) with sum(P.k) = P
!   y = response matrix (N,M)
!   w = observation weight vector (N)
!       note: if min(w) = max(w), then weights are ignored
!   off = offset matrix (N,M)
!   ngrps = number of groups (K)
!   gsize = number of coefs in each group (K)
!           note: (P.1, P.2, ..., P.K) with sum(P.k) = P
!   pw = penalty weight vector (K)   (outputs xsdev)
!        note: set pw(k) = 0 to leave k-th group unpenalized
!   alpha = weight for L1 and L2 penalities (alpha = 1 for lasso/mcp/scad, alpha = 0 for ridge)
!           note: setting alpha between (0,1) gives elastic net
!   nlam = number of lambdas
!   lambda = sequence of lambda values
!            note: if max(lambda) <=0, then lambda.max is data dependent
!   lmr = lambda minimum ratio (lambda.min = lmr * lambda.max)
!         note: unless lambda is provided, lambda.max is data dependent
!   penid = penalty id: 1 = lasso, 2 = mcp, 3 = scad
!   gamma = additional hyper-parameter for mcd and scad penalties
!           note: gamma > 1 for mcp and gamma > 2 for scad
!   eps = convergence tolerance
!   maxit = maximum number of iterations
!   standardize = integer (0: no standardization, 1: yes standardization)
!   intercept = integer (0: no intercept, 1: yes intercept)
!   ibeta = output matrix of intercepts (nresp,nlam)
!   betas = output array of coefficients (nvars,nresp,nlam)
!   iters = output vector of iterations for each lambda (nlam)
!   nzgrps = number of non-zero groups for each lambda (nlam)
!   nzcoef = number of non-zero coefficients for each lambda (nlam)
!   edfs = effective degrees of freedom for each lambda (nlam)
!   devs = residual deviance for each lambda (nlam)
!   nulldev = null deviance


SUBROUTINE grpnet_multigaus(nobs, nvars, nresp, x, y, w, off, ngrps, gsize, pw, &
                            alpha, nlam, lambda, lmr, penid, gamma, eps, maxit, &
                            standardize, intercept, ibeta, betas, iters, &
                            nzgrps, nzcoef, edfs, devs, nulldev)

    IMPLICIT NONE

! --------------- ARGUMENTS --------------- !
    INTEGER nobs, nvars, nresp, ngrps, gsize(ngrps), nlam, penid, maxit
    INTEGER standardize, intercept, iters(nlam), nzgrps(nlam), nzcoef(nlam)
    DOUBLE PRECISION x(nobs, nvars), y(nobs, nresp), w(nobs), off(nobs, nresp)
    DOUBLE PRECISION pw(ngrps), alpha, lambda(nlam), lmr, gamma, eps, ibeta(nresp,nlam)
    DOUBLE PRECISION betas(nvars, nresp, nlam), edfs(nlam), devs(nlam), nulldev
! --------------- ARGUMENTS --------------- !


! --------------- LOCAL DEFINITIONS --------------- !
    INTEGER i, j, k, l, iter, violations, makelambda
    INTEGER active(ngrps), strong(ngrps), ia(ngrps), ib(ngrps), gid
    DOUBLE PRECISION wmin, wmax, rnglam, minlam, maxlam, maxdif, macheps
    DOUBLE PRECISION r(nobs, nresp), difbeta(nvars, nresp), beta(nvars, nresp)
    DOUBLE PRECISION difibeta(nresp), zvec(nvars, nresp), grad(nvars, nresp)
    DOUBLE PRECISION gradnorm(ngrps), ctol, shrink, twolam, penone, pentwo
    DOUBLE PRECISION xmean(nvars), xsdev(ngrps), xev(ngrps), znorm, bnorm
    DOUBLE PRECISION wmat(nobs, nresp)
    DOUBLE PRECISION, ALLOCATABLE :: xtx(:,:)
! --------------- LOCAL DEFINITIONS --------------- !


! --------------- MACHINE EPSILON --------------- !
    macheps = EPSILON(eps)
! --------------- MACHINE EPSILON --------------- !


! --------------- CHECK WEIGHTS --------------- !
    wmin = MINVAL(w)
    wmax = MAXVAL(w)
    wmat = 1.0D0
    IF (wmax > wmin) THEN
        w = nobs * w / SUM(w)     ! normalize so SUM(w) = nobs
        w = SQRT(w)
        wmat = SPREAD(w, 2, nresp)
        y = wmat * y
        off = wmat * off
        DO i=1,nobs
            x(i,:) = w(i) * x(i,:)
        END DO
    END IF
! --------------- CHECK WEIGHTS --------------- !
    

! --------------- GROUP INDICES --------------- !
    gid = 0
    DO k=1,ngrps
        ia(k) = gid + 1
        gid = gid + gsize(k)
        ib(k) = gid
    END DO
! --------------- GROUP INDICES --------------- !


! --------------- CENTER AND SCALE --------------- !
    IF (intercept == 1) THEN
        DO j=1,nvars
            xmean(j) = SUM(x(:,j)) / nobs
            x(:,j) = x(:,j) - xmean(j)
        END DO
    END IF
    xsdev = 1.0D0
    IF (standardize == 1) THEN
        DO k=1,ngrps
            xsdev(k) = SQRT( SUM(x(:, ia(k):ib(k))**2) / (nobs * gsize(k)) )
            IF (xsdev(k) > macheps) THEN
                x(:,ia(k):ib(k)) = x(:,ia(k):ib(k)) / xsdev(k)
            ELSE
                xsdev(k) = 1.0D0
            END IF
        END DO
    END IF
! --------------- CENTER AND SCALE --------------- !


! --------------- GET MAX EIGENVALUE --------------- !
    DO k=1,ngrps
        IF (gsize(k) == 1) THEN
            xev(k) = SUM(x(:,ia(k))**2) / nobs
        ELSE
            ALLOCATE(xtx(gsize(k), gsize(k)))
            xtx = MATMUL(TRANSPOSE(x(:,ia(k):ib(k))), x(:,ia(k):ib(k))) / nobs
            CALL grpnet_maxeigval(xtx, gsize(k), xev(k))
            DEALLOCATE(xtx)
        END IF
    END DO
! --------------- GET MAX EIGENVALUE --------------- !


! --------------- MISCELLANEOUS INITIALIZATIONS --------------- !
    maxlam = MAXVAL(lambda)
    makelambda = 0
    iter = 0
    strong = 0
    active = 0
    nzgrps = 0
    nzcoef = 0
    zvec = 0.0D0
    ibeta = 0.0D0
    beta = 0.0D0
    grad = 0.0D0
    gradnorm = 0.0D0
    twolam = 0.0D0
    devs = 0.0D0
    r = y - off
! --------------- MISCELLANEOUS INITIALIZATIONS --------------- !


! --------------- GENERATE LAMBDA --------------- !
    IF (maxlam <= macheps) THEN

        makelambda = 1
        i = 1

        ! find unpenalized groups !
        DO k=1,ngrps
            IF (pw(k) <= macheps) THEN
                active(k) = 1
                nzgrps(i) = nzgrps(i) + 1
                nzcoef(i) = nzcoef(i) + gsize(k)
            END IF
        END DO
        ! find unpenalized groups !

        ! iterate until active coefficients converge !
        IF (nzgrps(i) > 0) THEN
            DO WHILE(iter < maxit)

                ! update iter and reset counters
                ctol = 0.0D0
                iter = iter + 1

                ! update active groups
                DO k=1,ngrps
                    IF(active(k) == 0) CYCLE
                    grad(ia(k):ib(k),:) = MATMUL(TRANSPOSE(x(:,ia(k):ib(k))), r) / nobs
                    zvec(ia(k):ib(k),:) = beta(ia(k):ib(k),:) + grad(ia(k):ib(k),:) / xev(k)
                    difbeta(ia(k):ib(k),:) = zvec(ia(k):ib(k),:) - beta(ia(k):ib(k),:)
                    maxdif = MAXVAL( ABS(difbeta(ia(k):ib(k),:)) / (1.0D0 + ABS(beta(ia(k):ib(k),:))) )
                    beta(ia(k):ib(k),:) = beta(ia(k):ib(k),:) + difbeta(ia(k):ib(k),:)
                    r = r - MATMUL(x(:,ia(k):ib(k)), difbeta(ia(k):ib(k),:))
                    ctol = MAX(maxdif , ctol)
                END DO ! k=1,ngrps

                ! update intercept
                IF (intercept == 1) THEN
                    difibeta = SUM(r * wmat, 1) / nobs
                    maxdif = MAXVAL( ABS(difibeta) / (1.0D0 + ABS(ibeta(:,i))) )
                    ibeta(:,i) = ibeta(:,i) + difibeta
                    r = r - wmat * SPREAD(difibeta, 1, nobs)
                    ctol = MAX(maxdif, ctol)
                END IF ! (intercept == 1)

                ! convergence check
                IF(ctol < eps) EXIT

            END DO ! WHILE(iter < maxit)

        ELSE

            ! intercept only
            iter = 1
            IF (intercept == 1) THEN
                ibeta(:,i) = SUM(r * wmat, 1) / nobs
                r = r - wmat * SPREAD(ibeta(:,i), 1, nobs)
            END IF

        END IF ! (nzgrps(1) > 0)
        ! iterate until active coefficients converge !

        ! create lambda sequence !
        DO k=1,ngrps
            IF (pw(k) > macheps) THEN
                grad(ia(k):ib(k),:) = MATMUL(TRANSPOSE(x(:,ia(k):ib(k))), r) / nobs
                gradnorm(k) = SQRT(SUM(grad(ia(k):ib(k),:)**2)) / pw(k)
            END IF
        END DO
        IF (alpha > macheps) THEN
            maxlam = MAXVAL(gradnorm / alpha)
        ELSE
            maxlam = MAXVAL(gradnorm / 1.0E-3)
            makelambda = 0
        END IF
        minlam = lmr * maxlam
        lambda(i) = maxlam
        maxlam = LOG(maxlam)
        minlam = LOG(minlam)
        rnglam = maxlam - minlam
        DO k=2,nlam
            lambda(k) = EXP(maxlam - rnglam * (k - 1) / (nlam - 1))
        END DO
        ! create lambda sequence !

        ! save results !
        betas(:,:,i) = beta
        iters(i) = iter
        nzgrps(i) = nzgrps(i) + intercept
        nzcoef(i) = nzcoef(i) + intercept * nresp
        edfs(i) = DBLE(nzcoef(i))
        devs(i) = SUM(r**2)
        ! save results !

    END IF
! --------------- GENERATE LAMBDA --------------- !


! --------------- ITERATIVE WORK --------------- !
    DO i=1,nlam

        ! initializations !
        IF (i == 1 .AND. makelambda == 1) CYCLE
        IF (i > 1) THEN
            ibeta(:,i) = ibeta(:,i-1)
            beta = betas(:,:,i-1)
            twolam = alpha * (2.0D0 * lambda(i) - lambda(i-1))
        ELSE
            grad = MATMUL(TRANSPOSE(x), r) / nobs
        END IF
        ! initializations !

        ! strong rule initialization !
        DO k=1,ngrps
            gradnorm(k) = SQRT(SUM(grad(ia(k):ib(k),:)**2))
            IF (gradnorm(k) + 1.0E-8 > pw(k) * twolam) THEN
                strong(k) = 1
            ELSE
                strong(k) = 0
            END IF
        END DO
        ! strong rule initialization !

        ! iterate until strong set converges !
        iter = 0
        DO WHILE(iter < maxit)

            ! iterate until active set converges !
            DO WHILE(iter < maxit)

                ! iterate until active coefficients converge !
                DO WHILE(iter < maxit)

                    ! update iter and reset counters
                    iter = iter + 1
                    nzgrps(i) = 0
                    edfs(i) = 0.0D0
                    ctol = 0.0D0

                    ! update active groups
                    DO k=1,ngrps
                        IF(active(k) == 0) CYCLE
                        penone = alpha * lambda(i) * pw(k) / xev(k)
                        pentwo = (1.0D0 - alpha) * lambda(i) * pw(k) / xev(k)
                        grad(ia(k):ib(k),:) = MATMUL(TRANSPOSE(x(:,ia(k):ib(k))), r) / nobs
                        zvec(ia(k):ib(k),:) = beta(ia(k):ib(k),:) + grad(ia(k):ib(k),:) / xev(k)
                        znorm = SQRT(SUM(zvec(ia(k):ib(k),:)**2))
                        bnorm = SQRT(SUM(beta(ia(k):ib(k),:)**2))
                        CALL grpnet_penalty(znorm, penid, penone, pentwo, gamma, shrink)
                        IF(shrink == 0.0D0 .AND. bnorm == 0.0D0) CYCLE
                        difbeta(ia(k):ib(k),:) = shrink * zvec(ia(k):ib(k),:) - beta(ia(k):ib(k),:)
                        maxdif = MAXVAL( ABS(difbeta(ia(k):ib(k),:)) / (1.0D0 + ABS(beta(ia(k):ib(k),:))) )
                        beta(ia(k):ib(k),:) = beta(ia(k):ib(k),:) + difbeta(ia(k):ib(k),:)
                        r = r - MATMUL(x(:,ia(k):ib(k)), difbeta(ia(k):ib(k),:))
                        ctol = MAX(maxdif , ctol)
                        IF(shrink > 0.0D0) THEN
                            nzgrps(i) = nzgrps(i) + 1
                            edfs(i) = edfs(i) + gsize(k) * shrink
                        ENDIF
                    END DO ! k=1,ngrps

                    ! update intercept
                    IF (intercept == 1) THEN
                        difibeta = SUM(r * wmat, 1) / nobs
                        maxdif = MAXVAL( ABS(difibeta) / (1.0D0 + ABS(ibeta(:,i))) )
                        ibeta(:,i) = ibeta(:,i) + difibeta
                        r = r - wmat * SPREAD(difibeta, 1, nobs)
                        ctol = MAX(maxdif, ctol)
                    END IF ! (intercept == 1)

                    ! convergence check
                    IF(ctol < eps) EXIT

                END DO ! WHILE(iter < maxit) - inner
                ! iterate until active coefficients converge !

                ! check inactive groups in strong set !
                violations = 0
                DO k=1,ngrps
                    IF(strong(k) == 0 .OR. active(k) == 1) CYCLE
                    grad(ia(k):ib(k),:) = MATMUL(TRANSPOSE(x(:,ia(k):ib(k))), r) / nobs
                    gradnorm(k) = SQRT(SUM(grad(ia(k):ib(k),:)**2))
                    IF (gradnorm(k) > alpha * lambda(i) * pw(k)) THEN
                        active(k) = 1
                        violations = violations + 1
                    END IF
                END DO ! k=1,ngrps
                IF(violations == 0) EXIT
                ! check inactive groups in strong set !

            END DO ! WHILE(iter < maxit) - middle
            ! iterate until active set converges !

            ! check groups in weak set !
            violations = 0
            DO k=1,ngrps
                IF (strong(k) == 1) CYCLE
                grad(ia(k):ib(k),:) = MATMUL(TRANSPOSE(x(:,ia(k):ib(k))), r) / nobs
                gradnorm(k) = SQRT(SUM(grad(ia(k):ib(k),:)**2))
                IF (gradnorm(k) + 1e-8 > alpha * lambda(i) * pw(k)) THEN
                    strong(k) = 1
                    active(k) = 1
                    violations = violations + 1
                END IF
            END DO
            IF(violations == 0) EXIT
            ! check groups in weak set !

        END DO ! WHILE(iter < maxit) - outer
        ! iterate until strong set converges !

        ! calculate nzcoef !
        DO k=1,ngrps
            IF(active(k) == 0) CYCLE
            DO j=1,nresp
                DO l=1,gsize(k)
                    IF (ABS(beta(ia(k) + l - 1,j)) > macheps) THEN
                        nzcoef(i) = nzcoef(i) + 1
                    END IF
                END DO
            END DO
        END DO
        ! calculate nzcoef !

        ! save results !
        betas(:,:,i) = beta
        iters(i) = iter
        nzgrps(i) = nzgrps(i) + intercept
        nzcoef(i) = nzcoef(i) + intercept * nresp
        edfs(i) = edfs(i) + DBLE(intercept * nresp)
        devs(i) = SUM(r**2)
        ! save results !

    END DO
! --------------- ITERATIVE WORK --------------- !


! --------------- POST PROCESSING --------------- !
    IF (standardize == 1) THEN
        DO k=1,ngrps
            betas(ia(k):ib(k),:,:) = betas(ia(k):ib(k),:,:) / xsdev(k)
        END DO
    END IF
    IF (intercept == 1) THEN
        DO j=1,nresp
            ibeta(j,:) = ibeta(j,:) - MATMUL(xmean, betas(:,j,:))
        END DO
        off = wmat * SPREAD(SUM(y * wmat, 1) / nobs, 1, nobs)
        nulldev = SUM((y - off)**2)
    ELSE
        nulldev = SUM((y - off)**2)
    END IF
    pw = xsdev
! --------------- POST PROCESSING --------------- !


END SUBROUTINE

R_grpnet_maxeigval <- 
  function(A, N, MAXEV){
    A <- as.matrix(A)
    N <- as.integer(N)
    X <- rep(1 / sqrt(N), N)
    MAXEV <- 1.0
    OLDEV <- 0.0
    while(abs(MAXEV - OLDEV) >= 1e-8){
      OLDEV <- MAXEV
      X <- A %*% X
      MAXEV <- sqrt(sum(X^2))
      if(MAXEV > 0.0) X <- X / MAXEV
    }
    return(MAXEV)
  } # grpnet_maxeigval

# Some code (for printing and plotting) is re-purposed from the R package
# "glmnet" (Hastie et al., 2010) https://cran.r-project.org/package=glmnet

cv.grpnet <-
  function(x, ...){
    UseMethod("cv.grpnet")
  } # end cv.grpnet 

print.cv.grpnet <- 
  function(x, digits = max(3, getOption("digits") - 3), ...){
    # convert to Title Case
    titlecase <- function(x){
      cap1 <- function(x) paste0(toupper(substr(x, 1, 1)), tolower(substr(x, 2, nchar(x))))
      words <- strsplit(x, split = " ", fixed = TRUE)
      words <- sapply(words, cap1)
      paste(words, collapse = " ")
    }
    # enhanced measure label
    if(x$type.measure == "deviance"){
      if(x$grpnet.fit$family$family == "negative.binomial"){
        ylab <- paste("Negative Binomial", "Deviance")
      } else if(x$grpnet.fit$family$family == "inverse.gaussian"){
        ylab <- paste("Inverse Gaussian", "Deviance")
      } else {
        ylab <- paste(titlecase(x$grpnet.fit$family$family), "Deviance")
      }
    } else if(x$type.measure == "mse"){
      ylab <- "Mean Squared Error"
    } else if(x$type.measure == "mae"){
      ylab <- "Mean Absolute Error"
    } else if(x$type.measure == "class"){
      ylab <- "Misclassification Error"
    }
    # print call
    cat("\nCall:   ")
    print(x$call)
    # print measure
    cat("\nMeasure:   ", ylab, "\n\n")
    # print fit info
    df <- data.frame(Alpha = x$grpnet.fit$alpha,
                     Lambda = x$lambda[x$index], 
                     Index = x$index,
                     Measure = x$cvm[x$index],
                     SE = x$cvsd[x$index],
                     nzGroup = x$nzero[x$index],
                     nzCoef = x$grpnet.fit$nzcoef[x$index],
                     Df = x$grpnet.fit$df[x$index])
    rownames(df) <- c("min ", "1se ")
    print(df, digits = digits)
  } # end print.cv.grpnet

plot.cv.grpnet <-
  function(x, cv = TRUE, sign.lambda = 1, nzero = TRUE, ...){
    
    # check x
    if(!inherits(x, "cv.grpnet")) stop("Input 'x' must be an object of class 'cv.grpnet'")
    
    # check cv
    if(!cv) return(plot(x$grpnet.fit, ...))
    
    # check sign.lambda
    sign.lambda <- as.numeric(sign.lambda[1])
    if(!any(sign.lambda == c(-1, 1))) stop("Input 'sign.lambda' must be -1 or 1")
    xsign <- ifelse(sign.lambda < 0, "-", "")
    
    # define x coords and limits
    xcoords <- sign.lambda * log(x$lambda)
    xlimits <- range(xcoords)
    
    # enhanced y-axis label
    titlecase <- function(x){
      cap1 <- function(x) paste0(toupper(substr(x, 1, 1)), tolower(substr(x, 2, nchar(x))))
      words <- strsplit(x, split = " ", fixed = TRUE)
      words <- sapply(words, cap1)
      paste(words, collapse = " ")
    }
    if(x$type.measure == "deviance"){
      if(x$grpnet.fit$family$family == "negative.binomial"){
        ylab <- "Negative Binomial Deviance"
      } else if(x$grpnet.fit$family$family == "inverse.gaussian"){
        ylab <- "Inverse Gaussian Deviance"
      } else {
        ylab <- paste(titlecase(x$grpnet.fit$family$family), "Deviance")
      }
    } else if(x$type.measure == "mse"){
      ylab <- "Mean Squared Error"
    } else if(x$type.measure == "mae"){
      ylab <- "Mean Absolute Error"
    } else if(x$type.measure == "class"){
      ylab <- "Misclassification Error"
    }
    
    # create args
    args <- list(...)
    args$x <- xcoords
    args$y <- x$cvm
    args$type = "n"
    if(is.null(args$ylim)) args$ylim <- c(min(x$cvlo), max(x$cvup))
    if(is.null(args$pch)) args$pch <- 20
    if(is.null(args$col)) args$col <- "red"
    if(is.null(args$xlab)) args$xlab <- if(sign.lambda < 0) expression("-" * log(lambda)) else expression(log(lambda))
    if(is.null(args$ylab)) args$ylab <- ylab
    
    # set up plot
    do.call(plot, args)
    
    # add nzeros
    if(nzero) axis(side = 3, at = xcoords, labels = x$nzero, tick = FALSE, line = -1)
    
    # add error bars
    width <- 0.01 * diff(xlimits)
    segments(xcoords, x$cvlo, xcoords, x$cvup, col = "darkgray")
    segments(xcoords - width, x$cvup, xcoords + width, x$cvup, col = "darkgray")
    segments(xcoords - width, x$cvlo, xcoords + width, x$cvlo, col = "darkgray")
    
    # add points
    points(x = args$x, y = args$y, pch = args$pch, col = args$col)
    
    # add lambda.min and lambda.1se
    abline(v = sign.lambda * log(x$lambda.min), lty = 3)
    abline(v = sign.lambda * log(x$lambda.1se), lty = 3)
    
  } # end plot.cv.grpnet

# Some code (for extracting predictions) is re-purposed from the R package
# "glmnet" (Hastie et al., 2010) https://cran.r-project.org/package=glmnet

predict.cv.grpnet <-
  function(object, 
           newx,
           newdata,
           s = c("lambda.1se", "lambda.min"),
           type = c("link", "response", "class", "terms", 
                    "importance", "coefficients", "nonzero", "groups", 
                    "ncoefs", "ngroups", "norm", "znorm"),
           ...){
    # predict from a fit cv.grpnet object
    # Nathaniel E. Helwig (helwig@umn.edu)
    # Updated: 2024-05-30
    
    
    ######***######   INITIAL CHECKS   ######***######
    
    ### check object
    if(!inherits(object, "cv.grpnet")) stop("Input 'object' must be of class 'cv.grpnet'")
    
    ### check newx / newdata
    sometypes <- c("link", "response", "class", "terms", "importance")
    itype <- sometypes[pmatch(as.character(type[1]), sometypes)]
    if(itype %in% c("link", "response", "class", "terms", "importance")){
      if(is.null(object$grpnet.fit$formula)){
        newx <- as.matrix(newx)
        ncoefs <- if(is.list(object$grpnet.fit$beta)) nrow(object$grpnet.fit$beta[[1]]) else nrow(object$grpnet.fit$beta)
        if(ncol(newx) != ncoefs) stop("Input 'newx' must satisfy:  ncol(newx) == nrow(object$beta)")
      } else {
        vnames <- rownames(attr(terms(object$grpnet.fit$formula), "factors"))[-1]
        check <- match(vnames, names(newdata))
        if(any(is.na(check))) stop("Input 'data' is missing variables included in model formula")
        if(!is.null(object$grpnet.fit$rk.args)){
          newx <- rk.model.matrix(object = object$grpnet.fit$formula, data = newdata,
                                  knots = object$grpnet.fit$rk.args$knots,
                                  Boundary.knots = lapply(object$grpnet.fit$rk.args$knots, function(x) x[c(1, length(x))]),
                                  m = object$grpnet.fit$rk.args$m, 
                                  periodic = object$grpnet.fit$rk.args$periodic,
                                  xlev = object$grpnet.fit$rk.args$xlev)
        } else {
          newx <- model.matrix(object = object$grpnet.fit$formula, data = newdata)
        }
        if(colnames(newx)[1] == "(Intercept)") newx <- newx[,-1]
        ncoefs <- if(is.list(object$grpnet.fit$beta)) nrow(object$grpnet.fit$beta[[1]]) else nrow(object$grpnet.fit$beta)
        if(ncol(newx) != ncoefs) stop("Input 'newdata' produced a design matrix of the wrong dimension\n (likely due to a factor level mismatch between 'data' and 'newdata')")
        object$grpnet.fit$formula <- NULL
      }
      nobs <- nrow(newx)
      newxnames <- rownames(newx)
      if(is.null(newxnames)) newxnames <- 1:nobs
    } else {
      newx <- NULL
    }
    
    ### check s
    if(is.character(s)){
      if(!any(s[1] == c("lambda.min", "lambda.1se"))) stop("Input 's' must be a character ('lambda.min' or 'lambda.1se') or a numeric vector")
      id <- object$index[ifelse(s[1] == "lambda.min", 1, 2)]
      s <- object$lambda[id]
    } else if(is.numeric(s)){
      if(any(s < 0)) stop("Inpust 's' must be a vector of non-negative numerics")
    } else {
      stop("Input 's' must be a character ('lambda.min' or 'lambda.1se') or a numeric vector")
    } # end if(is.character(s))
    
    ### return predictions
    predict(object$grpnet.fit, newx = newx, s = s, type = type)
    
  } # end predict.cv.grpnet

rk.model.matrix <- 
  function(object, data = environment(object), ...){
    # reproducing kernel model matrix
    # Nathaniel E. Helwig (helwig@umn.edu)
    # 2024-04-10
    
    
    ######***######   INITIAL CHECKS   ######***######
    
    ### check object
    oc <- class(object)[1]
    if(!any(oc == c("formula", "terms"))){
      stop("Input 'object' should be a model formula or a terms object")
    }
    
    ### make terms
    if(oc == "formula"){
      formula <- object
      object <- terms(object, data = data)
    } else {
      formula <- object
      attributes(formula) <- NULL
      formula <- as.formula(formula)
    }
    
    ### extract information
    factors <- attr(object, "factors")
    nvars <- nrow(factors)
    nterms <- ncol(factors)
    variables <- rownames(factors)
    term.labels <- attr(object, "term.labels")
    order <- attr(object, "order")
    intercept <- attr(object, "intercept")
    response <- attr(object, "response")
    
    ### was response included?
    if(response == 1){
      variables <- variables[-1]
      nvars <- nvars - 1L
      factors <- factors[-1,,drop=FALSE]
    } else if(response > 0){
      stop("Input 'object' formula should be of the form:\n y ~ . (one response)\n   ~ . (no response)")
    }
    
    ### convert characters to factors
    for(j in 1:nvars){
      if(is.character(data[,variables[j]])){
        data[,variables[j]] <- as.factor(data[,variables[j]])
      }
    }
    
    
    ######***######   ADDITIONAL ARGUMENTS   ######***######
    
    ### collect arguments
    args <- list(...)
    
    ### was df provided?
    if(is.null(args$df)){
      args$df <- vector("list", nvars)
      names(args$df) <- variables
    } else {
      if(length(args$df) == 1L){
        df <- args$df
        args$df <- vector("list", nvars)
        names(args$df) <- variables
        args$df[1:nvars] <- df
        rm(df)
      } else {
        if(!is.list(args$df)) stop("Input 'df' should be a single integer (common df for all variables)\n or a named list giving the df for each variable.")
      }
    } # end if(is.null(args$df))
    
    ### were knots provided?
    if(is.null(args$knots)){
      args$knots <- vector("list", nvars)
      names(args$knots) <- variables
    } else {
      if(!is.list(args$knots) | is.null(names(args$knots))) {
        stop("Input 'knots' should be a named list giving the knots for each variable.")
      }
    }
    
    ### was m provided?
    if(is.null(args$m)){
      args$m <- vector("list", nvars)
      names(args$m) <- variables
    } else {
      if(length(args$m) == 1L){
        m <- args$m
        args$m <- vector("list", nvars)
        names(args$m) <- variables
        args$m[1:nvars] <- m
        rm(m)
      } else {
        if(!is.list(args$m)) stop("Input 'm' should be a single integer (common m for all variables)\n or a named list giving the penalty order for each variable.")
      }
    } # end if(is.null(args$m))
    
    ### were Boundary.knots provided?
    if(is.null(args$Boundary.knots)){
      args$Boundary.knots <- vector("list", nvars)
      names(args$Boundary.knots) <- variables
    } else {
      if(!is.list(args$Boundary.knots) | is.null(names(args$Boundary.knots))) {
        stop("Input 'Boundary.knots' should be a named list giving the boundary knots for each variable.")
      }
    }
    
    ### was periodic provided?
    if(is.null(args$periodic)){
      args$periodic <- vector("list", nvars)
      args$periodic[1:nvars] <- FALSE
      names(args$periodic) <- variables
    } else {
      if(length(args$periodic) == 1L){
        periodic <- args$periodic
        args$periodic <- vector("list", nvars)
        names(args$periodic) <- variables
        args$periodic[1:nvars] <- periodic
        rm(periodic)
      } else {
        if(!is.list(args$periodic)) stop("Input 'periodic' should be a single logical (common periodicity info for all variables)\n or a named list giving the periodicity info for each variable.")
      }
    } # end if(is.null(args$periodic))
    
    ### was xlev provided?
    if(is.null(args$xlev)){
      args$xlev <- vector("list", nvars)
      names(args$xlev) <- variables
      for(j in 1:nvars){
        temp <- levels(data[,variables[j]])
        if(!is.null(temp)) args$xlev[[j]] <- temp
      }
    } else {
      if(!is.list(args$xlev) | is.null(names(args$xlev))) {
        stop("Input 'xlev' should be a named list giving the levels for each variable.")
      }
    }
    
    
    ######***######   BUILD MODEL MATRIX   ######***######
    
    ### make marginal basis matrices
    xlist <- vector("list", nvars)
    names(xlist) <- variables
    for(j in 1:nvars){
      xlist[[j]] <- rk(x = data[,variables[j]],
                       df = args$df[[variables[j]]],
                       knots = args$knots[[variables[j]]],
                       m = args$m[[variables[j]]],
                       Boundary.knots = args$Boundary.knots[[variables[j]]],
                       periodic = args$periodic[[variables[j]]],
                       xlev = args$xlev[[variables[j]]])
      colnames(xlist[[j]]) <- paste(variables[j], colnames(xlist[[j]]), sep = ".")
    }
    df.marg <- sapply(xlist, ncol)
    nobs <- nrow(xlist[[1]])
    
    ### find number of columns of x (and g)
    df.total <- 0
    g <- NULL
    for(k in 1:nterms){
      id <- which(factors[,k] > 0)
      df.new <- prod(df.marg[id])
      df.total <- df.total + df.new
      g <- c(g, rep(k, df.new))
    }
    
    ### build model matrix
    x <- matrix(0, nrow = nobs, ncol = df.total)
    xnames <- NULL
    for(k in 1:nterms){
      id <- which(g == k)
      if(order[k] == 1L){
        x[,id] <- xlist[[term.labels[k]]]
        xnames <- c(xnames, colnames(xlist[[term.labels[k]]]))
      } else {
        tid <- which(factors[,k] > 0)
        xtemp <- xlist[[variables[tid[1]]]]
        for(j in 2:order[k]){
          xtemp <- row.kronecker(xtemp, xlist[[variables[tid[j]]]])
        }
        xnames <- c(xnames, colnames(xtemp))
        x[,id] <- xtemp
      } # end if(order[k] == 1L)
    } # end for(k in 1:nterms)
    colnames(x) <- xnames
    
    ### add intercept?
    if(intercept > 0){
      x <- cbind(1, x)
      colnames(x) <- c("(Intercept)", xnames)
      g <- c(0, g)
    }
    
    
    ######***######   RETURN RESULTS   ######***######
    
    ### attributes
    attr(x, "assign") <- g
    attr(x, "term.labels") <- term.labels
    attr(x, "knots") <- lapply(xlist, function(x) attr(x, "knots"))
    attr(x, "m") <- lapply(xlist, function(x) attr(x, "m"))
    attr(x, "periodic") <- lapply(xlist, function(x) attr(x, "periodic"))
    attr(x, "xlev") <- lapply(xlist, function(x) attr(x, "xlev"))
    
    ### return
    return(x)
    
  } # end rk.model.matrix

!   grpnet_maxeigval.f90 - maximum eigenvalue of real symmetric matrix
!   Nathaniel E. Helwig (helwig@umn.edu)
!   Department of Psychology and School of Statistics
!   University of Minnesota
!   Date: 2023-07-10
!
! INPUTS
!   A = real symmetric matrix (N x N)
!   N = number of rows/columns of A
! OUTPUT:
!   MAXEV = maximum eigenvalue of A

SUBROUTINE grpnet_maxeigval(A, N, MAXEV)
    IMPLICIT NONE
    INTEGER N
    DOUBLE PRECISION A(N,N), MAXEV, X(N), OLDEV
    X = 1.0D0 / SQRT(DBLE(N))
    MAXEV = 1.0D0
    OLDEV = 0.0D0
    DO WHILE (ABS(MAXEV - OLDEV) >= 1E-8)
        OLDEV = MAXEV
        X = MATMUL(A, X)
        MAXEV = SQRT(SUM(X**2))
        IF (MAXEV > 0.0D0) X = X / MAXEV
    END DO
END SUBROUTINE

!   grpnet_penalty.f90 - group elastic net shrinkage penalty
!   Nathaniel E. Helwig (helwig@umn.edu)
!   Department of Psychology and School of Statistics
!   University of Minnesota
!   Date: 2023-06-20
!
! INPUTS
!   znorm = Euclidean norm of least squares estimate
!   penid = penalty id: 1 = LASSO, 2 = MCP, 3 = SCAD
!   penone = L1 penalty weight
!   pentwo = L2 penalty weight
!   gamma = extra tuning parameter (for MCP and SCAD)
! OUTPUT:
!   shrink = shrinkage factor in interval (0,1)

SUBROUTINE grpnet_penalty(znorm, penid, penone, pentwo, gamma, shrink)
    IMPLICIT NONE
    INTEGER penid
    DOUBLE PRECISION znorm, penone, pentwo, gamma, shrink
    IF (penid == 1) THEN
        shrink = MAX(0.0D0, 1.0D0 - penone / znorm)
        shrink = shrink / (1.0D0 + pentwo)
    ELSE IF (penid == 2) THEN
        IF (znorm <= gamma * penone * (1.0D0 + pentwo)) THEN
            shrink = MAX(0.0D0, 1.0D0 - penone / znorm)
            shrink = shrink / (1.0D0 + pentwo - 1.0D0 / gamma)
        ELSE
            shrink = 1.0D0 / (1.0D0 + pentwo)
        END IF
    ELSE IF (penid == 3) THEN
        IF (znorm <= (penone * (1.0D0 + pentwo) + penone)) THEN
            shrink = MAX(0.0D0, 1.0D0 - penone / znorm) / (1.0D0 + pentwo)
        ELSE IF (znorm <= gamma * penone * (1.0D0 + pentwo)) THEN
            shrink = MAX(0.0D0, 1.0D0 - (penone / znorm) * (gamma / (gamma - 1.0D0)) )
            shrink = shrink / (1.0D0 + pentwo - 1.0D0 / (gamma - 1.0D0) )
        ELSE
            shrink = 1.0D0 / (1.0D0 + pentwo)
        END IF
    END IF
END SUBROUTINE

row.kronecker <- 
  function(X, Y){
    ### row-wise kronecker product
    ### Nathaniel E. Helwig (helwig@umn.edu)
    ### Mar 31, 2026
    
    # check dimensions
    nx <- nrow(X)
    ny <- nrow(Y)
    if(nx != ny) stop("Inputs 'X' and 'Y' must have the same number of rows")
    px <- ncol(X)
    py <- ncol(Y)
    pz <- px * py
    
    # check names
    xnames <- colnames(X)
    if(is.null(xnames)) xnames <- paste0("X", 1:px)
    ynames <- colnames(Y)
    if(is.null(ynames)) ynames <- paste0("Y", 1:py)
    
    # initialize and build Z
    znames <- vector(mode = "character", length = pz)
    Z <- matrix(0.0, nrow = nx, ncol = pz)
    for(j in 1:px){
      index <- 1:py + (j - 1) * py
      Z[,index] <- X[,j] * Y
      znames[index] <- paste0(xnames[j], ":", ynames)
    }
    colnames(Z) <- znames
    return(Z)
    
  }

# Some code (for arguments and outputs) is re-purposed from the R package
# "glmnet" (Hastie et al., 2010) https://cran.r-project.org/package=glmnet

cv.grpnet.formula <-
  function(formula,
           data, 
           use.rk = TRUE,
           weights = NULL,
           offset = NULL,
           alpha = c(0.01, 0.25, 0.5, 0.75, 1),
           gamma = c(3, 4, 5),
           type.measure = NULL,
           nfolds = 10, 
           foldid = NULL, 
           same.lambda = FALSE,
           parallel = FALSE, 
           cluster = NULL, 
           verbose = interactive(),
           adaptive = FALSE,
           power = 1, 
           ...){
    # k-fold cross-validation for grpnet (formula)
    # Nathaniel E. Helwig (helwig@umn.edu)
    # Updated: 2026-04-09
    
    
    ######***######   INITIAL CHECKS   ######***######
    
    ### get call
    cv.grpnet.call <- match.call()
    
    ### check formula
    formula <- as.formula(formula)
    charform <- as.character(formula)
    if(charform[1] != "~") stop("Input 'formula' must be of the form:  y ~ x")
    
    ### model frame (modified from R's lm() function)
    mf <- match.call(expand.dots = FALSE)
    m <- match(c("formula", "data"), names(mf), 0L)
    mf <- mf[c(1L, m)]
    mf$drop.unused.levels <- TRUE
    ## need stats:: for non-standard evaluation
    mf[[1L]] <- quote(stats::model.frame)
    mf <- eval(mf, parent.frame())
    
    ### model response
    yname <- colnames(mf)[1]
    y <- model.response(mf, "any")
    
    ### model matrix
    rk.args <- NULL
    if(use.rk){
      x <- rk.model.matrix(object = formula, data = mf, ...)
      rk.args <- list(knots = attr(x, "knots"),
                      m = attr(x, "m"),
                      periodic = attr(x, "periodic"),
                      xlev = attr(x, "xlev"))
    } else {
      x <- model.matrix(object = formula, data = mf)
    }
    
    ### model terms
    terms <- attr(mf, "terms")
    xvar.class <- attr(terms, "dataClasses")[-1]
    term.labels <- attr(terms, "term.labels")
    
    ### nobs and nvars
    nobs <- nrow(x)
    nvars <- ncol(x)
    nxvars <- length(xvar.class)
    
    ### group vector
    group <- attr(x, "assign")
    gsize <- table(group)
    ngrps <- length(gsize)
    
    ### intercept?
    intercept <- ifelse(group[1] == 0, TRUE, FALSE)
    if(intercept) {
      x <- x[,-1]
      group <- group[-1]
      gsize <- gsize[-1]
      ngrps <- ngrps - 1L
      nvars <- nvars - 1L
    }
    
    
    ######***######   CV.GRPNET.DEFAULT   ######***######
    
    ### collect extra arguments
    args <- list(...)
    
    ### penalty.factor given?
    if(is.null(args$penalty.factor)){
      
      res <- cv.grpnet.default(x = x, 
                               y = y,
                               group = group,
                               weights = weights,
                               offset = offset,
                               alpha = alpha,
                               gamma = gamma,
                               type.measure = type.measure,
                               nfolds = nfolds, 
                               foldid = foldid,
                               same.lambda = same.lambda,
                               parallel = parallel, 
                               cluster = cluster, 
                               verbose = verbose,
                               intercept = intercept, 
                               adaptive = adaptive,
                               power = power,
                               ...)
      
    } else {
      
      # prepare args$penalty.factor for input to cv.grpnet.default
      pf <- sqrt(gsize)
      names(pf) <- term.labels
      args$penalty.factor <- as.list(args$penalty.factor)
      penfacname <- names(args$penalty.factor)
      if(is.null(penfacname)) stop("Input 'penalty.factor' must be a named list.")
      for(k in 1:length(args$penalty.factor)){
        id <- match(penfacname[k], term.labels)
        if(is.na(id)) warning("Input 'penalty.factor' contains the term:   '", penfacname[k], "'\nwhich does not match any of the model terms")
        pf[id] <- args$penalty.factor[[k]]
      }
      args$penalty.factor <- pf
      
      # combine all arguments
      iargs <- list(x = x, 
                    y = y,
                    group = group,
                    weights = weights,
                    offset = offset,
                    alpha = alpha,
                    gamma = gamma,
                    type.measure = type.measure,
                    nfolds = nfolds, 
                    foldid = foldid,
                    same.lambda = same.lambda,
                    parallel = parallel, 
                    cluster = cluster, 
                    verbose = verbose,
                    intercept = intercept,
                    adaptive = adaptive,
                    power = power)
      allargs <- c(iargs, args)
      
      # call cv.grpnet.default
      res <- do.call(cv.grpnet.default, allargs)
      
    } # end if(is.null(args$penalty.factor))
    
    
    
    ######***######   POST-PROCESSING   ######***######
    
    ### correct the call
    res$call <- cv.grpnet.call
    
    ### add the (potential expanded) formula
    res$grpnet.fit$formula <- as.formula(paste0(yname, " ~ ", paste(term.labels, collapse = " + ")))
    
    ### add the term.labels
    if(intercept) term.labels <- c("(Intercept)", term.labels)
    res$grpnet.fit$term.labels <- term.labels
    
    ### add the rk.args
    res$grpnet.fit$rk.args <- rk.args
    
    ### add the data
    res$grpnet.fit$data <- data
    
    ### return results
    return(res)
    
  } # end cv.grpnet.formula

summary.grpnet <-
  function(object, ...){
    # summarize grpnet object
    # Nathaniel E. Helwig (helwig@umn.edu)
    # Updated: 2026-04-28
    
    if(!inherits(object, "grpnet")) stop("Input 'object' should be of class 'grpnet'.")
    method <- ifelse(is.null(object$formula), "default", "formula")
    if(method == "default"){
      fit <- predict(object, newx = object$data$x, ...)
      imp <- predict(object, newx = object$data$x, type = "importance")
    } else {
      fit <- predict(object, newdata = object$data, ...)
      imp <- predict(object, newdata = object$data, type = "importance")
    }
    if(object$family$family %in% c("multigaussian", "multinomial")){
      imp0 <- imp
      imp <- array(dim = c(nrow(imp0[[1]]), length(imp0), ncol(imp0[[1]])),
                   dimnames = list(rownames(imp0[[1]]), names(imp0), colnames(imp0[[1]])))
      for(k in 1:length(imp0)) imp[,k,] <- imp0[[k]]
    }
    penalties <- c("LASSO", "MCP", "SCAD")
    res <- list(family = object$family$family, 
                penalty = penalties[object$args$penalty],
                nobs = object$nobs,
                ngroups = object$ngroups,
                lambda = object$lambda,
                dev.ratio = object$dev.ratio,
                fit = fit, 
                act = abs(imp) > 0.0,
                imp = imp)
    class(res) <- "summary.grpnet"
    return(res)
  } # summary.grpnet

print.summary.grpnet <-
  function(x, ...){
    cat("\n")
    cat("family =", x$family, "\n")
    cat("penalty =", x$penalty, "\n")
    cat("n =", x$nobs, "observations\n")
    cat("K =", x$ngroups, "groups\n")
    nlam <- length(x$lambda)
    lamseq <- seq(1, nlam, length.out = min(5, nlam))
    cat("\n% Null Deviance Explained:\n")
    names(x$dev.ratio) <- paste0("s", 1:length(x$lambda))
    dr <- x$dev.ratio[lamseq]
    print(100 * dr, digits = 6)
    cat("\nVariable Importance:\n")
    if(x$family %in% c("multigaussian", "multinomial")){
      respvars <- dimnames(x$imp)[[2]]
      cat("\n")
      for(k in 1:length(respvars)){
        cat("  ", respvars[k], ":\n", sep = "")
        print(100 * x$imp[,k,lamseq], digits = 6)
        cat("\n")
      }
    } else {
      print(100 * x$imp[,lamseq], digits = 6)
    }
    cat("\n")
  }

R_grpnet_multinom <-
  function(nobs, nvars, nresp, x, y, w, off, ngrps, gsize, pw, 
           alpha, nlam, lambda, lmr, penid, gamma, eps, maxit,
           standardize, intercept, ibeta, betas, iters,
           nzgrps, nzcoef, edfs, devs, nulldev){
    # grpnet_multinom.f90 translation to R
    # Nathaniel E. Helwig (helwig@umn.edu)
    # Updated: 2026-04-09
    
    
    # ! --------------- LOCAL DEFINITIONS --------------- ! #
    ia <- ib <- rep(0L, ngrps)
    xmean <- rep(0.0, nvars)
    xev <- rep(0.0, ngrps)
    difbeta <- matrix(0.0, nvars, nresp)
    # ! --------------- LOCAL DEFINITIONS --------------- ! #
    
    
    # ! --------------- MACHINE EPSILON --------------- ! #
    macheps <- .Machine$double.eps
    # ! --------------- MACHINE EPSILON --------------- ! #
    
    
    # ! --------------- CHECK WEIGHTS --------------- ! #
    wmin <- min(w)
    wmax <- max(w)
    if(wmax > wmin){
      weighted <- 1L
      w <- nobs * w / sum(w)     # ! normalize so SUM(w) = nobs
      w <- sqrt(w)
      for(i in 1:nobs){
        x[i,] <- w[i] * x[i,]
      }
    } else {
      weighted <- 0L
      w <- rep(1.0, nobs)
    }
    # ! --------------- CHECK WEIGHTS --------------- ! #
    
    
    # ! --------------- GROUP INDICES --------------- ! #
    gid <- 0L
    for(k in 1:ngrps){
      ia[k] <- gid + 1L
      gid <- gid + gsize[k]
      ib[k] <- gid
    }
    # ! --------------- GROUP INDICES --------------- ! #
    
    
    # ! --------------- CENTER AND SCALE --------------- ! #
    if(intercept == 1L){
      for(j in 1:nvars){
        xmean[j] <- sum(x[,j]) / nobs
        x[,j] <- x[,j] - xmean[j]
      }
    }
    xsdev <- rep(1.0, ngrps)
    if(standardize == 1L){
      for(k in 1:ngrps){
        xsdev[k] <- sqrt( sum(x[,ia[k]:ib[k]]^2) / (nobs * gsize[k]) )
        if(xsdev[k] > macheps){
          x[,ia[k]:ib[k]] <- x[,ia[k]:ib[k]] / xsdev[k]
        } else {
          xsdev[k] <- 1.0
        }
      }
    }
    # ! --------------- CENTER AND SCALE --------------- ! #
    
    
    # ! --------------- GET MAX EIGENVALUE --------------- ! #
    for(k in 1:ngrps){
      if(gsize[k] == 1L){
        xev[k] <- sum(x[,ia[k]]^2) / nobs
      } else {
        xtx <- crossprod(x[,ia[k]:ib[k]]) / nobs
        xev[k] <- R_grpnet_maxeigval(A = xtx, N = gsize[k])
      }
    }
    xev <- xev / 4
    # ! --------------- GET MAX EIGENVALUE --------------- ! #
    
    
    # ! --------------- MISCELLANEOUS INITIALIZATIONS --------------- ! #
    maxlam <- max(lambda)
    makelambda <- 0L
    iter <- 0L
    active <- strong <- rep(0L, ngrps)
    nzgrps <- nzcoef <- rep(0L, nlam)
    ibeta <- matrix(0.0, nresp, nlam)
    beta <- zvec <- grad <- matrix(0.0, nvars, nresp)
    gradnorm <- rep(0.0, ngrps)
    twolam <- 0.0
    devs <- rep(0.0, nlam)
    eta <- off
    expeta <- exp( eta - apply(eta, 1, max) )
    mu <- expeta / rowSums(expeta)
    r <- w * (y - mu)
    # ! --------------- MISCELLANEOUS INITIALIZATIONS --------------- ! #
    
    
    # ! --------------- GENERATE LAMBDA --------------- ! #
    if(maxlam <= macheps){
      
      makelambda <- 1L
      i <- 1L
      
      # ! find unpenalized groups ! # 
      for(k in 1:ngrps){
        if(pw[k] <= macheps){
          active[k] <- 1L
          nzgrps[i] <- nzgrps[i] + 1L
          nzcoef[i] <- nzcoef[i] + gsize[k] * nresp
        }
      }
      # ! find unpenalized groups ! #
      
      # ! iterate until active coefficients converge ! #
      if(nzgrps[i] > 0L){
        
        while(iter < maxit){
          
          # ! update iter and reset counters
          ctol <- 0.0
          iter <- iter + 1L
          
          # ! update active groups
          for(k in 1:ngrps){
            if(active[k] == 0L) next
            grad[ia[k]:ib[k],] <- crossprod(x[,ia[k]:ib[k]], r) / nobs
            zvec[ia[k]:ib[k],] <- beta[ia[k]:ib[k],] + grad[ia[k]:ib[k],] / xev[k]
            zvec[ia[k]:ib[k],] <- zvec[ia[k]:ib[k],] - rowMeans(zvec[ia[k]:ib[k],,drop=FALSE])
            difbeta[ia[k]:ib[k],] <- zvec[ia[k]:ib[k],] - beta[ia[k]:ib[k],]
            maxdif <- max( abs(difbeta[ia[k]:ib[k],]) / (1.0 + abs(beta[ia[k]:ib[k],])) )
            beta[ia[k]:ib[k],] <- beta[ia[k]:ib[k],] + difbeta[ia[k]:ib[k],]
            eta <- eta + (x[,ia[k]:ib[k]] %*% difbeta[ia[k]:ib[k],]) / w
            expeta <- exp( eta - apply(eta, 1, max) )
            mu <- expeta / rowSums(expeta)
            r <- w * (y - mu)
            ctol <- max(maxdif, ctol)
          } # ! k=1,ngrps
          
          # ! update intercept
          if(intercept == 1L){
            zint <- ibeta[,i] + ( colSums(r * w) / nobs ) * 4
            zint <- zint - sum(zint) / nresp
            difibeta <- zint - ibeta[,i]
            maxdif <- abs(difibeta) / (1.0 + abs(ibeta[,i]))
            ibeta[,i] <- ibeta[,i] + difibeta
            eta <- eta + matrix(difibeta, nobs, nresp, byrow = TRUE)
            expeta <- exp( eta - apply(eta, 1, max) )
            mu <- expeta / rowSums(expeta)
            r <- w * (y - mu)
            ctol <- max(maxdif, ctol)
          } # if(intercept == 1L)
          
          # ! convergence check
          if(ctol < eps) break
          
        } # ! WHILE(iter < maxit)
        
      } else {
        
        # ! intercept only
        iter <- 1L
        if(intercept == 1L){
          ibeta[,i] <- log( colSums(y * w^2) / nobs )
          ibeta[,i] <- ibeta[,i] - sum(ibeta[,i]) / nresp
          eta <- eta + matrix(ibeta[,i], nobs, nresp, byrow = TRUE)
          expeta <- exp( eta - apply(eta, 1, max) )
          mu <- expeta / rowSums(expeta)
          r <- w * (y - mu)
        }
        
      } # ! (nzgrps[i] > 0)
      # ! iterate until active coefficients converge ! #
      
      # ! create lambda sequence ! #
      for(k in 1:ngrps){
        if(pw[k] > macheps){
          grad[ia[k]:ib[k],] <- crossprod(x[,ia[k]:ib[k]], r) / nobs
          gradnorm[k] <- sqrt( sum(grad[ia[k]:ib[k],]^2) ) / pw[k]
        }
      }
      if(alpha > macheps){
        maxlam <- max(gradnorm / alpha)
      } else {
        maxlam <- max(gradnorm / 1e-3)
        makelambda <- 0L
      }
      minlam <- lmr * maxlam
      lambda[i] <- maxlam
      maxlam <- log(maxlam)
      minlam <- log(minlam)
      rnglam <- maxlam - minlam
      for(k in 2:nlam){
        lambda[k] <- exp( maxlam - rnglam * (k - 1) / (nlam - 1) )
      }
      # ! create lambda sequence ! #
      
      # ! calculate deviance ! #
      devs[i] <- R_grpnet_multinom_dev(nobs, nresp, y, mu, w^2)
      # ! calculate deviance ! #
      
      # ! save results ! # 
      betas[,,i] <- beta
      iters[i] <- iter
      nzgrps[i] <- nzgrps[i] + intercept
      nzcoef[i] <- nzcoef[i] + intercept * nresp
      edfs[i] <- intercept * (nresp - 1)
      # ! save results ! # 
      
    }
    # ! --------------- GENERATE LAMBDA --------------- ! #
    
    
    # ! --------------- ITERATIVE WORK --------------- ! #
    for(i in 1:nlam){
      
      # ! initializations ! # 
      if(i == 1L && makelambda == 1L) next
      if(i > 1L){
        ibeta[,i] <- ibeta[,i-1]
        beta <- betas[,,i-1]
        twolam <- alpha * (2.0 * lambda[i] - lambda[i-1])
      } else {
        grad <- crossprod(x, r) / nobs
      }
      # ! initializations ! # 
      
      # ! strong rule initialization ! #
      for(k in 1:ngrps){
        gradnorm[k] <- sqrt(sum(grad[ia[k]:ib[k],]^2))
        if(gradnorm[k] + 1e-8 > pw[k] * twolam){
          strong[k] <- 1L
        } else {
          strong[k] <- 0L
        }
      }
      # ! strong rule initialization ! #
      
      # ! iterate until strong set converges ! #
      iter <- 0L
      while(iter < maxit){
        
        # ! iterate until active set converges ! # 
        while(iter < maxit){
          
          # ! iterate until active coefficients converge ! #
          while(iter < maxit){
            
            # ! update iter and reset counters
            iter <- iter + 1L
            nzgrps[i] <- 0L
            edfs[i] <- 0.0
            ctol <- 0.0
            
            # ! unweighted or weighted update?
            if(weighted == 0L){
              
              # ! update active groups
              for(k in 1:ngrps){
                if(active[k] == 0L) next
                penone <- alpha * lambda[i] * pw[k] / xev[k]
                pentwo <- (1.0 - alpha) * lambda[i] * pw[k] / xev[k]
                grad[ia[k]:ib[k],] <- crossprod(x[,ia[k]:ib[k]], r) / nobs
                zvec[ia[k]:ib[k],] <- beta[ia[k]:ib[k],] + grad[ia[k]:ib[k],] / xev[k]
                zvec[ia[k]:ib[k],] <- zvec[ia[k]:ib[k],] - rowMeans(zvec[ia[k]:ib[k],,drop=FALSE])
                znorm <- sqrt(sum(zvec[ia[k]:ib[k],]^2))
                bnorm <- sqrt(sum(beta[ia[k]:ib[k],]^2))
                shrink <- R_grpnet_penalty(znorm, penid, penone, pentwo, gamma)
                if(shrink == 0.0 && bnorm == 0.0) next
                difbeta[ia[k]:ib[k],] <- shrink * zvec[ia[k]:ib[k],] - beta[ia[k]:ib[k],]
                maxdif <- max( abs(difbeta[ia[k]:ib[k],]) / (1.0 + abs(beta[ia[k]:ib[k],])) )
                beta[ia[k]:ib[k],] <- beta[ia[k]:ib[k],] + difbeta[ia[k]:ib[k],]
                eta <- eta + x[,ia[k]:ib[k]] %*% difbeta[ia[k]:ib[k],]
                expeta <- exp( eta - apply(eta, 1, max) )
                mu <- expeta / rowSums(expeta)
                r <- y - mu
                ctol <- max(maxdif , ctol)
                if(shrink > 0.0){
                  nzgrps[i] <- nzgrps[i] + 1L
                  edfs[i] <- edfs[i] + gsize[k] * (nresp - 1) * shrink
                }
              } # ! k=1,ngrps
              
              # ! update intercept
              if(intercept == 1L){
                zint <- ibeta[,i] + ( colSums(r) / nobs ) * 4
                zint <- zint - sum(zint) / nresp
                difibeta <- zint - ibeta[,i]
                maxdif <- abs(difibeta) / (1 + abs(ibeta[,i]))
                ibeta[,i] <- ibeta[,i] + difibeta
                eta <- eta + matrix(difibeta, nobs, nresp, byrow = TRUE)
                expeta <- exp( eta - apply(eta, 1, max) )
                mu <- expeta / rowSums(expeta)
                r <- y - mu
                ctol <- max(maxdif, ctol)
              } # ! (intercept == 1)
              
            } else {
              
              # ! update active groups
              for(k in 1:ngrps){
                if(active[k] == 0L) next
                penone <- alpha * lambda[i] * pw[k] / xev[k]
                pentwo <- (1 - alpha) * lambda[i] * pw[k] / xev[k]
                grad[ia[k]:ib[k],] <- crossprod(x[,ia[k]:ib[k]], r) / nobs
                zvec[ia[k]:ib[k],] <- beta[ia[k]:ib[k],] + grad[ia[k]:ib[k],] / xev[k]
                zvec[ia[k]:ib[k],] <- zvec[ia[k]:ib[k],] - rowMeans(zvec[ia[k]:ib[k],,drop=FALSE])
                znorm <- sqrt(sum(zvec[ia[k]:ib[k],]^2))
                bnorm <- sqrt(sum(beta[ia[k]:ib[k],]^2))
                shrink <- R_grpnet_penalty(znorm, penid, penone, pentwo, gamma)
                if(shrink == 0.0 && bnorm == 0.0) next
                difbeta[ia[k]:ib[k],] <- shrink * zvec[ia[k]:ib[k],] - beta[ia[k]:ib[k],]
                maxdif <- max( abs(difbeta[ia[k]:ib[k],]) / (1.0 + abs(beta[ia[k]:ib[k],])) )
                beta[ia[k]:ib[k],] <- beta[ia[k]:ib[k],] + difbeta[ia[k]:ib[k],]
                eta <- eta + (x[,ia[k]:ib[k]] %*% difbeta[ia[k]:ib[k],]) / w
                expeta <- exp( eta - apply(eta, 1, max) )
                mu <- expeta / rowSums(expeta)
                r <- w * (y - mu)
                ctol <- max(maxdif , ctol)
                if(shrink > 0.0){
                  nzgrps[i] <- nzgrps[i] + 1L
                  edfs[i] <- edfs[i] + gsize[k] * (nresp - 1) * shrink
                }
              } # ! k=1,ngrps
              
              # ! update intercept
              if(intercept == 1L){
                zint <- ibeta[,i] + ( colSums(r * w) / nobs ) * 4
                zint <- zint - sum(zint) / nresp
                difibeta <- zint - ibeta[,i]
                maxdif <- abs(difibeta) / (1 + abs(ibeta[,i]))
                ibeta[,i] <- ibeta[,i] + difibeta
                eta <- eta + matrix(difibeta, nobs, nresp, byrow = TRUE)
                expeta <- exp( eta - apply(eta, 1, max) )
                mu <- expeta / rowSums(expeta)
                r <- w * (y - mu)
                ctol <- max(maxdif, ctol)
              } # ! (intercept == 1)
              
            } # ! IF (weighted == 0)
            
            # ! convergence check
            if(ctol < eps) break
            
          } # ! WHILE(iter < maxit) - inner
          # ! iterate until active coefficients converge ! #
          
          # ! check inactive groups in strong set ! #
          violations <- 0L
          for(k in 1:ngrps){
            if(strong[k] == 0L | active[k] == 1L) next
            grad[ia[k]:ib[k],] <- crossprod(x[,ia[k]:ib[k]], r) / nobs
            gradnorm[k] <- sqrt(sum(grad[ia[k]:ib[k],]^2))
            if(gradnorm[k] > alpha * lambda[i] * pw[k]){
              active[k] <- 1L
              violations <- violations + 1L
            }
          } # ! k=1,ngrps
          if(violations == 0L) break
          # ! check inactive groups in strong set ! #
          
        } # ! WHILE(iter < maxit) - middle
        # ! iterate until active set converges ! # 
        
        # ! check groups in weak set ! #
        violations <- 0L
        for(k in 1:ngrps){
          if(strong[k] == 1L) next
          grad[ia[k]:ib[k],] <- crossprod(x[,ia[k]:ib[k]], r) / nobs
          gradnorm[k] <- sqrt(sum(grad[ia[k]:ib[k],]^2))
          if(gradnorm[k] + 1e-8 > alpha * lambda[i] * pw[k]){
            strong[k] <- 1
            active[k] <- 1
            violations <- violations + 1L
          }
        }
        if(violations == 0L) break
        # ! check groups in weak set ! #
        
      } # ! WHILE(iter < maxit) - outer
      # ! iterate until strong set converges ! #
      
      # ! calculate nzcoef ! #
      for(k in 1:ngrps){
        if(active[k] == 0L) next
        for(j in 1:nresp){
          for(l in 1:gsize[k]){
            if(abs(beta[ia[k] + l - 1,j]) > macheps){
              nzcoef[i] <- nzcoef[i] + 1L
            }
          }
        }
      }
      # ! calculate nzcoef ! #
      
      # ! calculate deviance ! #
      devs[i] <- R_grpnet_multinom_dev(nobs, nresp, y, mu, w^2)
      # ! calculate deviance ! #
      
      # ! save results ! #
      betas[,,i] <- beta
      iters[i] <- iter
      nzgrps[i] <- nzgrps[i] + intercept
      nzcoef[i] <- nzcoef[i] + intercept * nresp
      edfs[i] <- edfs[i] + intercept * (nresp - 1)
      # ! save results ! #
      
    }
    # ! --------------- ITERATIVE WORK --------------- ! #
    
    
    # ! --------------- POST PROCESSING --------------- ! #
    if(standardize == 1L){
      for(k in 1:ngrps){
        betas[ia[k]:ib[k],,] <- betas[ia[k]:ib[k],,] / xsdev[k]
      }
    }
    if(intercept == 1L){
      for(j in 1:nresp){
        ibeta[j,] <- ibeta[j,] - xmean %*% betas[,j,]
      }
      mu <- matrix(colSums(y * w^2) / nobs, nobs, nresp, byrow = TRUE)
    } else {
      expeta <- exp( off - apply(off, 1, max) )
      mu <- expeta / rowSums(expeta)
    }
    nulldev <- R_grpnet_multinom_dev(nobs, nresp, y, mu, w^2)
    names(xsdev) <- names(pw)
    pw <- xsdev
    # ! --------------- POST PROCESSING --------------- ! #
    
    
    # ! --------------- RETURN RESULTS --------------- ! #
    res <- list(nobs = nobs, 
                nvars = nvars,
                nresp = nresp,
                x = x, 
                y = y, 
                w = w, 
                off = off, 
                ngrps = ngrps, 
                gsize = gsize, 
                pw = pw, 
                alpha = alpha, 
                nlam = nlam, 
                lambda = lambda, 
                lmr = lmr, 
                penid = penid, 
                gamma = gamma, 
                eps = eps, 
                maxit = maxit,
                standardize = standardize, 
                intercept = intercept, 
                ibeta = ibeta, 
                betas = betas, 
                iters = iters,
                nzgrps = nzgrps, 
                nzcoef = nzcoef, 
                edfs = edfs, 
                devs = devs, 
                nulldev = nulldev)
    return(res)
    # ! --------------- RETURN RESULTS --------------- ! #
    
    
  } # R_grpnet_multinom.R


R_grpnet_multinom_dev <-
  function(nobs, nresp, y, mu, wt, dev){
    dev <- 0.0
    for(i in 1:nobs){
      for(j in 1:nresp){
        if(mu[i,j] < 1e-6){
          mu[i,j] <- 1e-6
        } else if(mu[i,j] > (1 - 1e-6)){
          mu[i,j] <- 1 - 1e-6
        }
      }
      dev <- dev - 2 * wt[i] * sum( y[i,] * log(mu[i,]) )
    }
    return(dev)
  } # R_grpnet_multinom_dev.R

# Some code (for arguments and outputs) is re-purposed from the R package
# "glmnet" (Hastie et al., 2010) https://cran.r-project.org/package=glmnet

cv.grpnet.default <-
  function(x, 
           y, 
           group,
           weights = NULL,
           offset = NULL,
           alpha = c(0.01, 0.25, 0.5, 0.75, 1),
           gamma = c(3, 4, 5),
           type.measure = NULL,
           nfolds = 10, 
           foldid = NULL,
           same.lambda = FALSE,
           parallel = FALSE, 
           cluster = NULL, 
           verbose = interactive(), 
           adaptive = FALSE,
           power = 1, 
           ...){
    # k-fold cross-validation for grpnet (default)
    # Nathaniel E. Helwig (helwig@umn.edu)
    # Updated: 2026-04-09
    
    
    ######***######   INITIAL CHECKS   ######***######
    
    ### start clock
    tic <- proc.time()[3]
    
    ### get call
    cv.grpnet.call <- match.call()
    
    ### check x and y
    x <- as.matrix(x)
    nobs <- nrow(x)
    nvars <- ncol(x)
    ny <- if(is.matrix(y)) nrow(y) else length(y)
    if(ny != nobs) stop("Inputs 'x' and 'y' must satisfy:\nnrow(x) == length(y)  or  nrow(x) == nrow(y)")
    
    ### x names
    xnames <- colnames(x)
    if(is.null(xnames)) xnames <- paste0("x", 1:nvars)
    
    ### check group
    if(missing(group)) {
      group <- ingroup <- 1:nvars
      gsize <- rep(1L, nvars)
      names(gsize) <- 1:nvars
      ngrps <- nvars
    } else {
      if(length(group) != nvars) stop("Inputs 'x' and 'group' must satisfy:  ncol(x) == length(group)")
      ingroup <- group
      group <- as.factor(group)
      ngrps <- nlevels(group)
      if(ngrps > nvars) stop("Inputs 'x' and 'group' must satisfy:  ncol(x) >= nlevels(group)")
      gsize <- table(group)
      group <- as.integer(group)
    }
    
    ### check family
    args <- list(...)
    family <- args$family
    if(is.null(family)){
      family <- "gaussian"
    } else {
      family <- as.character(family[1])
      if(family == "mgaussian" | family == "mvn") family <- "multigaussian"
      if(family == "hsvm") family <- "svm1"
      if(family == "sqsvm") family <- "svm2"
      if(family == "polr") family <- "ordinal"
      families <- c("gaussian", "multigaussian", "svm1", "svm2", "logit", "binomial", "multinomial", "ordinal", "poisson", "negative.binomial", "Gamma", "inverse.gaussian")
      family <- pmatch(family, families)
      if(is.na(family)) stop("'family' not recognized")
      family <- families[family]
    }
    
    ### check weights
    if(is.null(weights)){
      noweights = TRUE
      weights <- rep(1.0, nobs)
    } else {
      noweights = FALSE
      weights <- as.numeric(weights)
      if(length(weights) != nobs) stop("Inputs 'y' and 'weights' must satisfy:  length(y) == length(weights)")
      if(any(weights < 0)) stop("Input 'weights' must be non-negative")
      weights <- weights / mean(weights)
    }
    
    ### check y
    ylev <- yfac <- NULL
    if(family == "gaussian"){
      y <- as.numeric(y)
    } else if(family == "multigaussian"){
      y <- as.matrix(y) + 0.0
      if(!is.numeric(y[1,1])) stop("Input 'y' must be a numeric matrix when family = 'multigaussian'.")
      nresp <- ncol(y)
      ylev <- colnames(y)
      if(is.null(ylev)){
        ylev <- paste0("y", 1:nresp)
        colnames(y) <- ylev
      }
      if(nresp == 1L){
        y <- as.numeric(y)
        family <- "gaussian"
      }
    } else if(family %in% c("svm1", "svm2", "logit")){
      if(is.character(y)) y <- as.factor(y)
      if(is.factor(y)){
        ylev <- levels(y)
        y <- ifelse(y == ylev[1], -1.0, 1.0)
      } else if(is.matrix(y)) {
        if(ncol(y) != 2L | any(y < 0)) stop("Input 'y' must be a matrix (# success, # failure) when family %in% c('svm1', 'svm2', 'logit')")
        ytotal <- rowSums(y)
        weights <- weights * ytotal
        wsqrt <- sqrt(weights)
        y <- 2 * (y[,1] / ytotal) - 1
      } else {
        y <- as.numeric(y)
        if(any(y < -1) | any(y > 1)) stop("Input 'y' must contain values between -1 and 1 when family %in% c('svm1', 'svm2', 'logit')" )
      }
      if(is.null(ylev)) ylev <- c(-1, 1)
      yfac <- factor(ifelse(y <= 0.0, ylev[1], ylev[2]), levels = ylev)
    } else if(family == "binomial"){
      if(is.character(y)) y <- as.factor(y)
      if(is.factor(y)){
        #if(nlevels(y) != 2L) stop("Input 'y' must be (coercible into) a factor with two levels when family = 'binomial'.")
        ylev <- levels(y)
        y <- ifelse(y == ylev[1], 0.0, 1.0)
      } else if(is.matrix(y)) {
        if(ncol(y) != 2L | any(y < 0)) stop("Input 'y' must be a matrix (# success, # failure) when family = 'binomial'")
        ytotal <- rowSums(y)
        weights <- weights * ytotal
        y <- y[,1] / ytotal
      } else {
        y <- as.numeric(y)
        if(any(y < 0) | any(y > 1)) stop("Input 'y' must contain values between 0 and 1 when family = 'binomial'" )
      }
      if(is.null(ylev)) ylev <- c(0, 1)
      yfac <- factor(ifelse(y <= 0.5, ylev[1], ylev[2]), levels = ylev)
    } else if(family == "multinomial"){
      if(is.character(y)) y <- as.factor(y)
      if(is.factor(y)){
        yrowsum <- rep(1.0, nobs)
        yfac <- y
        yint <- as.integer(y)
        ylev <- levels(y)
        nlev <- nlevels(y)
        if(nlev < 3L) stop("Input 'y' must be (coercible into) a factor with 3 or more levels.")
        y <- matrix(0.0, nrow = nobs, ncol = nlev)
        colnames(y) <- ylev
        for(i in 1:nobs) y[i,yint[i]] <- 1.0
      } else if(is.matrix(y)){
        nlev <- ncol(y)
        ylev <- colnames(y)
        if(is.null(ylev)){
          ylev <- paste0("y", 1:nlev)
          colnames(y) <- ylev
        }
        y <- t(apply(y, 1, as.integer))
        if(min(y) < 0) stop("Input 'y' must be a matrix of counts (i.e, non-negative integers)")
        yrowsum <- rowSums(y)
        #y <- y / yrowsum
        #weights <- weights * yrowsum
        yfac <- factor(ylev[unlist(apply(y, 1, which.max))], levels = ylev)
      } else {
        stop("Invalid data input: 'y' must be a factor or matrix for when family = 'multinomial'.")
      }
    } else if(family == "ordinal"){
      y <- yfac <- factor(y, ordered = TRUE)
      ylev <- levels(y)
      nlev <- nlevels(y)
      y <- as.integer(y)
      ymat <- outer(y, 1:nlev, FUN = "==") + 0.0
      intercept <- TRUE
    } else if (family == "poisson") {
      y <- as.integer(y)
      if(any(y < 0)) stop("Input 'y' must contain non-negative integers when family = 'poisson'.")
    } else if (family == "negative.binomial") {
      y <- as.integer(y)
      if(any(y < 0)) stop("Input 'y' must contain non-negative integers when family = 'negative.binomial'.")
    } else if(family == "Gamma") {
      y <- as.numeric(y)
      if(any(y <= 0)) stop("Input 'y' must contain non-negative numerics when family = 'Gamma'.")
    } else if(family == "inverse.gaussian") {
      y <- as.numeric(y)
      if(any(y <= 0)) stop("Input 'y' must contain non-negative numerics when family = 'inverse.gaussian'.")
    } # end if(family == "gaussian")
    
    ### check offset
    if(is.null(offset)){
      include.offset <- FALSE
      offset <- rep(0.0, nobs)
      if(family == "multigaussian") offset <- matrix(offset, nrow = nobs, ncol = nresp)
      if(family == "multinomial") offset <- matrix(offset, nrow = nobs, ncol = nlev)
    } else {
      include.offset <- TRUE
      if(family == "multigaussian"){
        offset <- as.matrix(offset) + 0.0
        if(nrow(offset) != nobs) stop("Inputs 'x' and 'offset' must satisfy:  nrow(x) == nrow(offset)")
        if(ncol(offset) == 1L){
          offset <- matrix(offset, nrow = nobs, ncol = nresp)
        } else {
          if(ncol(offset) != nresp) stop("Input 'offset' must be a vector of length nobs\n or a matrix of dimension nobs x ncol(y)")
        }
      } else if(family == "multinomial"){
        offset <- as.matrix(offset)
        if(nrow(offset) != nobs) stop("Inputs 'x' and 'offset' must satisfy:  nrow(x) == nrow(offset)")
        if(ncol(offset) == 1L){
          offset <- matrix(offset, nrow = nobs, ncol = nlev)
        } else {
          if(ncol(offset) != nlev) stop("Input 'offset' must be a vector of length nobs\n or a matrix of dimension nobs x nlev")
        }
      } else {
        offset <- as.numeric(offset)
        if(length(offset) != nobs) stop("Inputs 'y' and 'offset' must satisfy:  length(y) == length(offset)")
      }
    }
    
    ### check alpha
    alpha <- sort(unique(as.numeric(alpha)))
    if(any(alpha < 0 | alpha > 1)) stop("Input 'alpha' must satisfy:  0 <= alpha <= 1")
    nalpha <- length(alpha)
    
    ### get penalty
    argnames <- names(args)
    if(any(argnames == "penalty")){
      penalty <- args$penalty
    } else {
      penalty <- NULL
    }
    if(is.null(penalty)){
      penalty <- 1L
    } else {
      if(is.integer(penalty)){
        if(penalty < 1L || penalty > 3L) stop("Invalid 'penalty': must be 1 (LASSO), 2 (MCP), or 3 (SCAD)")
      } else {
        penalty <- toupper(as.character(penalty))
        penalty <- pmatch(penalty, c("LASSO", "MCP", "SCAD"))
        if(is.na(penalty)) stop("Invalid 'penalty': must be 'LASSO', 'MCP', or 'SCAD'")
      }
    }
    
    ### check gamma
    if(penalty > 1L){
      gamma <- sort(unique(as.numeric(gamma)))
      ngamma <- length(gamma)
      lower <- ifelse(penalty == 2L, 1.0, 2.0)
      if(any(gamma <= lower)) stop(paste0("Input 'gamma' must be greater than ", lower, " when penalty = '", penalty,"'"))
    } else {
      ngamma <- 1L
      gamma <- 4
    }
    
    ### check type.measure
    if(is.null(type.measure)){
      type.measure <- ifelse(family %in% c("svm1", "svm2", "logit", "binomial", "multinomial", "ordinal"), "class", "mae")
    } else {
      type.measure <- pmatch(as.character(type.measure[1]), c("deviance", "mse", "mae", "class"))
      if(is.na(type.measure)) stop("Invalid 'type.measure' argument.")
      type.measure <- c("deviance", "mse", "mae", "class")[type.measure]
      if(type.measure == "class" && !(family %in% c("svm1", "svm2", "logit", "binomial", "multinomial", "ordinal")))
        stop("Input 'type.measure' can only be set to 'class' for families:\nsvm1, svm2, logit, binomial, multinomial, ordinal")
    }
    
    ### check nfolds and foldid
    if(is.null(foldid)){
      nfolds <- as.integer(nfolds[1])
      if(nfolds < 2L | nfolds > nobs) stop("Input 'nfolds' must satisfy:  2 <= nfolds <= nrow(x)")
      if(family == "ordinal"){
        ytab <- table(yfac)
        if(min(ytab) == 1L) stop("When family = 'oridinal', response must satisfy:  min(table(as.ordered(y))) > 1")
        foldid <- rep(0L, nobs)
        for(k in 1:nlev){
          kid <- which(yfac == ylev[k])
          nkid <- length(kid)
          if(nkid < nfolds){
            foldid[kid] <- sample.int(nfolds, nkid)
          } else {
            foldid[kid] <- sample(rep(1:nfolds, length.out = nkid))
          } # end if(nkid == 1L)
        } # end for(k in 1:nlev)
      } else {
        foldid <- sample(rep(1:nfolds, length.out = nobs))
      }
    } else {
      foldid <- as.integer(foldid)
      if(length(foldid) != nobs) stop("Input 'foldid' must satisfy:  length(y) == length(foldid)")
      nfolds <- length(unique(foldid))
      if(nfolds < 2L | nfolds > nobs) stop("Input 'nfolds' must satisfy:  2 <= nfolds <= nrow(x)")
    } # end if(is.null(foldid))
    
    ### check parallel
    parallel <- as.logical(parallel[1])
    if(!any(parallel == c(TRUE, FALSE))) stop("Input 'parallel' must be TRUE or FALSE")
    
    ### check cluster
    if(parallel){
      if(nalpha == 1L) verbose <- FALSE
      if(is.null(cluster)){
        madeCluster <- TRUE
        #ncores <- parallel::detectCores()
        ncores <- 2L   # to ensure users don't accidentally spawn >2 processes
        cluster <- parallel::makeCluster(ncores)
      } else {
        madeCluster <- FALSE
        if(!any(class(cluster) == "cluster")) stop("Input 'cluster' must of an object of class 'cluster'")
      }
    } # end if(parallel)
    
    ### check verbose
    verbose <- as.logical(verbose[1])
    if(!any(verbose == c(TRUE, FALSE))) stop("Input 'verbose' must be TRUE or FALSE")
    
    ### check adaptive
    adaptive <- as.logical(adaptive[1])
    if(!any(adaptive == c(TRUE, FALSE))) stop("Input 'adaptive' must be TRUE or FALSE")
    
    ### check power
    power <- as.numeric(power[1])
    if(power <= 0) stop("Input 'power' must be a positive numeric")
    
    
    
    ######***######   ADAPTIVE OMEGA TUNING   ######***######
    
    if(adaptive){
      
      ### ridge fit
      mod <- cv.grpnet.default(x = x, 
                               y = y, 
                               group = ingroup,
                               weights = weights,
                               offset = offset,
                               alpha = 0,
                               gamma = gamma,
                               type.measure = type.measure,
                               nfolds = nfolds, 
                               foldid = foldid,
                               same.lambda = same.lambda,
                               parallel = parallel, 
                               cluster = cluster, 
                               verbose = FALSE, 
                               adaptive = FALSE,
                               power = 1, 
                               ...)
      
      ### update penalty weights
      penfac0 <- mod$grpnet.fit$args$penalty.factor
      if(mod$grpnet.fit$args$orthogonalized){
        if(mod$grpnet.fit$family$family %in% c("multigaussian", "multinomial")){
          penfac <- sqrt(rowSums(sapply(predict(mod, newx = x, type = "terms"), function(x) colSums(x^2))))
        } else {
          penfac <- sqrt(colSums(predict(mod, newx = x, type = "terms")^2))
        }
      } else if(mod$grpnet.fit$args$standardized){
        penfac <- predict(mod, newx = x, type = "znorm")[-1]
      } else {
        penfac <- predict(mod, newx = x, type = "norm")[-1]
      }
      penfac <- penfac0 / (penfac^power)
      
      ### refit with updated penalty weights
      mod <- cv.grpnet.default(x = x, 
                               y = y, 
                               group = ingroup,
                               weights = weights,
                               offset = offset,
                               alpha = alpha,
                               gamma = gamma,
                               type.measure = type.measure,
                               nfolds = nfolds, 
                               foldid = foldid,
                               same.lambda = same.lambda,
                               parallel = parallel, 
                               cluster = cluster, 
                               verbose = verbose, 
                               penalty.factor = penfac,
                               adaptive = FALSE,
                               power = 1, 
                               ...)
      
      ### return result
      attr(adaptive, "power") <- power
      mod$adaptive <- adaptive
      return(mod)
      
    } # end if(adaptive)
    
    
    
    ######***######   TUNE ALPHA AND/OR GAMMA   ######***######
    
    if(nalpha > 1L | ngamma > 1L){
      
      ## initialize to hold results
      alpha <- rev(alpha)
      tune.res <- expand.grid(alpha = alpha, gamma = gamma, cvm = 0.0)
      if(penalty == 1L) tune.res$gamma <- NULL
      
      ## loop through alpha and gamma
      counter <- 1L
      for(k in 1:ngamma){
        for(j in 1:nalpha){
          
          # print progress?
          if(verbose){
            if(penalty == 1L){
              cat("alpha = ", formatC(alpha[j], format = "f", digits = 2), ":", sep = "")
            } else {
              cat("alpha = ", formatC(alpha[j], format = "f", digits = 2), 
                  "  &  gamma = ", formatC(gamma[k], format = "f", digits = 2), ":", sep = "")
            }
          }
          
          # fit model
          res <- cv.grpnet.default(x = x, 
                                   y = y,
                                   group = ingroup,
                                   weights = weights,
                                   offset = offset,
                                   alpha = alpha[j],
                                   gamma = gamma[k],
                                   type.measure = type.measure,
                                   nfolds = nfolds, 
                                   foldid = foldid,
                                   same.lambda = same.lambda,
                                   parallel = parallel, 
                                   cluster = cluster, 
                                   verbose = FALSE,
                                   adaptive = FALSE,
                                   power = power,
                                   ...)
          
          # save results
          resmin <- res$cvm[res$index[1]]
          tune.res$cvm[counter] <- resmin
          if(verbose){
            cat("   min(cvm) = ", resmin, "\n", sep = "")
          }
          
          # update best solution?
          if((j == 1) && (k == 1)){
            best <- res
            bestmin <- best$cvm[best$index[1]]
          } else if(bestmin > resmin) {
            best <- res
            bestmin <- resmin
          }
          
          # update counter
          counter <- counter + 1L
          
        } # end for(j in 1:nalpha)
      } # end for(k in 1:ngamma)
      
      best$tune <- tune.res
      best$call <- cv.grpnet.call
      best$grpnet.fit$ylev <- ylev
      return(best)
      
    } # end if(nalpha > 1L | ngamma > 1L)
    
    
    ######***######   PRE-PROCESSING   ######***######
    
    ### collect fold ids
    fid <- vector("list", nfolds)
    for(k in 1:nfolds) fid[[k]] <- which(foldid == k)
    
    ### start progress bar
    if(verbose) pbar <- txtProgressBar(min = 0, max = nfolds + 1L, style = 3)
    
    ### generate lambda (if needed)
    grpnet.fit <- grpnet(x = x, 
                         y = y, 
                         group = group, 
                         weights = weights,
                         offset = offset, 
                         alpha = alpha,
                         gamma = gamma, 
                         ...)
    if(grpnet.fit$args$intercept){
      grpnet.fit$group <- c(rep(0, ifelse(family == "ordinal", nlev-1, 1)), ingroup)
    } else{
      grpnet.fit$group <- ingroup
    }
    grpnet.fit$ylev <- ylev
    lambda <- grpnet.fit$lambda
    nlambda <- length(lambda)
    if(verbose) setTxtProgressBar(pbar, 1)
    
    
    ######***######   K-FOLD CV   ######***######
    
    ### initialize matrix for results
    cvloss <- matrix(0.0, nrow = nlambda, ncol = nfolds)
    
    ### separate work for multigaussian and multinomial family
    if(family == "multigaussian"){
      
      # standardize.response?
      if(grpnet.fit$args$standardize.response){
        ysd <- apply(y, 2, sd)
        y <- y %*% diag(1 / ysd)
      } # end if(grpnet.fit$args$standardize.response)
      
      if(parallel){
        
        # define parcvloss function
        parcvloss.multigaussian <-
          function(testid, xmat, ymat, group, family, weights, offset, alpha, 
                   nlambda, lambda.min.ratio, lambda, penalty.factor, penalty, 
                   gamma, theta, standardized, orthogonalized, intercept, 
                   thresh, maxit, proglang, type.measure, same.lambda){
            temp <- grpnet(x = xmat[-testid,,drop=FALSE], 
                           y = ymat[-testid,,drop=FALSE], 
                           group = group, 
                           family = family,
                           weights = weights[-testid], 
                           offset = offset[-testid,],
                           alpha = alpha,
                           nlambda = nlambda,
                           lambda.min.ratio = lambda.min.ratio,
                           lambda = if(same.lambda) lambda else NULL,
                           penalty.factor = penalty.factor,
                           penalty = penalty,
                           gamma = gamma, 
                           theta = theta,
                           standardized = standardized,
                           orthogonalized = orthogonalized,
                           intercept = intercept,
                           thresh = thresh,
                           maxit = maxit,
                           proglang = proglang)
            mu <- predict(temp, newx = xmat[testid,,drop=FALSE], 
                          s = if(same.lambda) NULL else lambda)
            cvloss <- rep(0.0, nlambda)
            if(type.measure == "deviance"){
              for(i in 1:nlambda){
                cvloss[i] <- mean(temp$family$dev.resids(ymat[testid,], mu[,,i], weights[testid]))
              }
            } else if(type.measure == "mse") {
              for(i in 1:nlambda){
                cvloss[i] <- mean((ymat[testid,] - mu[,,i])^2)
              }
            } else if(type.measure == "mae"){
              for(i in 1:nlambda){
                cvloss[i] <- mean(abs(ymat[testid,] - mu[,,i]))
              }
            } 
            return(cvloss)
          } # end parcvloss
        
        # evaluate cvloss in parallel
        cvloss <- parallel::parSapply(cl = cluster, X = fid, FUN = parcvloss.multigaussian,
                                      xmat = x,
                                      ymat = y,
                                      group = group,
                                      family = family,
                                      weights = weights,
                                      offset = offset,
                                      alpha = grpnet.fit$alpha,
                                      nlambda = nlambda,
                                      lambda.min.ratio = lambda[nlambda] / lambda[1],
                                      lambda = lambda,
                                      penalty.factor = grpnet.fit$args$penalty.factor,
                                      penalty = grpnet.fit$args$penalty,
                                      gamma = grpnet.fit$args$gamma,
                                      theta = grpnet.fit$args$theta,
                                      standardized = grpnet.fit$args$standardized,
                                      orthogonalized = grpnet.fit$args$orthogonalized,
                                      intercept = grpnet.fit$args$intercept,
                                      thresh = grpnet.fit$args$thresh,
                                      maxit = grpnet.fit$args$maxit,
                                      proglang = grpnet.fit$args$proglang,
                                      type.measure = type.measure,
                                      same.lambda = same.lambda)
        
        # unvectorize
        cvloss <- matrix(cvloss, nrow = nlambda, ncol = nfolds)
        
      } else {
        
        for(k in 1:nfolds){
          if(same.lambda){
            temp <- grpnet(x = x[-fid[[k]],,drop=FALSE], 
                           y = y[-fid[[k]],,drop=FALSE], 
                           group = group, 
                           family = family,
                           weights = weights[-fid[[k]]], 
                           offset = offset[-fid[[k]],,drop=FALSE],
                           alpha = alpha,
                           nlambda = nlambda,
                           lambda.min.ratio = lambda[nlambda] / lambda[1],
                           lambda = lambda,
                           penalty.factor = grpnet.fit$args$penalty.factor,
                           penalty = grpnet.fit$args$penalty,
                           gamma = gamma,
                           theta = grpnet.fit$args$theta,
                           standardized = grpnet.fit$args$standardized,
                           orthogonalized = grpnet.fit$args$orthogonalized,
                           intercept = grpnet.fit$args$intercept,
                           thresh = grpnet.fit$args$thresh,
                           maxit = grpnet.fit$args$maxit,
                           proglang = grpnet.fit$args$proglang)
            mu <- predict(temp, newx = x[fid[[k]],,drop=FALSE], 
                          type = ifelse(type.measure == "class", "class", "response"))
          } else {
            temp <- grpnet(x = x[-fid[[k]],,drop=FALSE], 
                           y = y[-fid[[k]],,drop=FALSE], 
                           group = group, 
                           family = family,
                           weights = weights[-fid[[k]]], 
                           offset = offset[-fid[[k]],,drop=FALSE],
                           alpha = alpha,
                           nlambda = nlambda,
                           lambda.min.ratio = lambda[nlambda] / lambda[1],
                           penalty.factor = grpnet.fit$args$penalty.factor,
                           penalty = grpnet.fit$args$penalty,
                           gamma = gamma,
                           theta = grpnet.fit$args$theta,
                           standardized = grpnet.fit$args$standardized,
                           orthogonalized = grpnet.fit$args$orthogonalized,
                           intercept = grpnet.fit$args$intercept,
                           thresh = grpnet.fit$args$thresh,
                           maxit = grpnet.fit$args$maxit,
                           proglang = grpnet.fit$args$proglang)
            mu <- predict(temp, newx = x[fid[[k]],,drop=FALSE], s = lambda)
          }
          if(type.measure == "deviance"){
            for(i in 1:nlambda) {
              cvloss[i,k] <- mean(grpnet.fit$family$dev.resids(y[fid[[k]],], mu[,,i], weights[fid[[k]]]))
            }
          } else if(type.measure == "mse") {
            for(i in 1:nlambda) {
              cvloss[i,k] <- mean((y[fid[[k]],] - mu[,,i])^2)
            }
          } else if(type.measure == "mae"){
            for(i in 1:nlambda) {
              cvloss[i,k] <- mean(abs(y[fid[[k]],] - mu[,,i]))
            }
          } 
          if(verbose) setTxtProgressBar(pbar, k + 1)
        } # end for(k in 1:nfolds)
        if(verbose) close(pbar)
        
      } # end if(parallel)
      
    } else if(family == "multinomial"){
      
      if(parallel){
        
        # define parcvloss function
        parcvloss.multinomial <-
          function(testid, xmat, ymat, group, family, weights, offset, alpha, 
                   nlambda, lambda.min.ratio, lambda, penalty.factor, penalty, 
                   gamma, theta, standardized, orthogonalized, intercept, 
                   thresh, maxit, proglang, type.measure, same.lambda, yfac, yrowsum){
            temp <- grpnet(x = xmat[-testid,,drop=FALSE], 
                           y = ymat[-testid,,drop=FALSE], 
                           group = group, 
                           family = family,
                           weights = weights[-testid], 
                           offset = offset[-testid,],
                           alpha = alpha,
                           nlambda = nlambda,
                           lambda.min.ratio = lambda.min.ratio,
                           lambda = if(same.lambda) lambda else NULL,
                           penalty.factor = penalty.factor,
                           penalty = penalty,
                           gamma = gamma, 
                           theta = theta,
                           standardized = standardized,
                           orthogonalized = orthogonalized,
                           intercept = intercept,
                           thresh = thresh,
                           maxit = maxit,
                           proglang = proglang)
            temp$ylev <- levels(yfac)
            mu <- predict(temp, newx = xmat[testid,,drop=FALSE], 
                          s = if(same.lambda) NULL else lambda,
                          type = ifelse(type.measure == "class", "class", "response"))
            cvloss <- rep(0.0, nlambda)
            y01 <- ymat[testid,,drop=FALSE] / yrowsum[testid]
            if(type.measure == "deviance"){
              w01 <- weights[testid] * yrowsum[testid]
              for(i in 1:nlambda) {
                cvloss[i] <- mean(temp$family$dev.resids(y01, mu[,,i], w01))
              }
            } else if(type.measure == "mse") {
              for(i in 1:nlambda) {
                cvloss[i] <- mean((y01 - mu[,,i])^2)
              }
            } else if(type.measure == "mae"){
              for(i in 1:nlambda) {
                cvloss[i] <- mean(abs(y01 - mu[,,i]))
              }
            } else if(type.measure == "class"){
              cvloss <- 1 - colMeans(yfac[testid] == mu)
            }
            return(cvloss)
          } # end parcvloss
        
        # evaluate cvloss in parallel
        cvloss <- parallel::parSapply(cl = cluster, X = fid, FUN = parcvloss.multinomial,
                                      xmat = x,
                                      ymat = y,
                                      group = group,
                                      family = family,
                                      weights = weights,
                                      offset = offset,
                                      alpha = grpnet.fit$alpha,
                                      nlambda = nlambda,
                                      lambda.min.ratio = lambda[nlambda] / lambda[1],
                                      lambda = lambda,
                                      penalty.factor = grpnet.fit$args$penalty.factor,
                                      penalty = grpnet.fit$args$penalty,
                                      gamma = grpnet.fit$args$gamma,
                                      theta = grpnet.fit$args$theta,
                                      standardized = grpnet.fit$args$standardized,
                                      orthogonalized = grpnet.fit$args$orthogonalized,
                                      intercept = grpnet.fit$args$intercept,
                                      thresh = grpnet.fit$args$thresh,
                                      maxit = grpnet.fit$args$maxit,
                                      proglang = grpnet.fit$args$proglang,
                                      type.measure = type.measure,
                                      same.lambda = same.lambda,
                                      yfac = yfac,
                                      yrowsum = yrowsum)
        
        # unvectorize
        cvloss <- matrix(cvloss, nrow = nlambda, ncol = nfolds)
        
      } else {
        
        for(k in 1:nfolds){
          if(same.lambda){
            temp <- grpnet(x = x[-fid[[k]],,drop=FALSE], 
                           y = y[-fid[[k]],,drop=FALSE], 
                           group = group, 
                           weights = weights[-fid[[k]]], 
                           offset = offset[-fid[[k]],,drop=FALSE],
                           alpha = alpha,
                           lambda = lambda, 
                           gamma = gamma,
                           ...)
            temp$ylev <- ylev
            mu <- predict(temp, newx = x[fid[[k]],,drop=FALSE], 
                          type = ifelse(type.measure == "class", "class", "response"))
          } else {
            temp <- grpnet(x = x[-fid[[k]],,drop=FALSE], 
                           y = y[-fid[[k]],,drop=FALSE], 
                           group = group, 
                           weights = weights[-fid[[k]]], 
                           offset = offset[-fid[[k]],,drop=FALSE],
                           alpha = alpha,
                           gamma = gamma,
                           ...)
            temp$ylev <- ylev
            mu <- predict(temp, newx = x[fid[[k]],,drop=FALSE], s = lambda, 
                          type = ifelse(type.measure == "class", "class", "response"))
          }
          y01 <- y[fid[[k]],,drop=FALSE] / yrowsum[fid[[k]]]
          if(type.measure == "deviance"){
            w01 <- weights[fid[[k]]] * yrowsum[fid[[k]]]
            for(i in 1:nlambda) {
              cvloss[i,k] <- mean(grpnet.fit$family$dev.resids(y01, mu[,,i], w01))
            }
          } else if(type.measure == "mse") {
            for(i in 1:nlambda) {
              cvloss[i,k] <- mean((y01 - mu[,,i])^2)
            }
          } else if(type.measure == "mae"){
            for(i in 1:nlambda) {
              cvloss[i,k] <- mean(abs(y01 - mu[,,i]))
            }
          } else if(type.measure == "class"){
            cvloss[,k] <- 1 - colMeans(yfac[fid[[k]]] == mu)
          }
          if(verbose) setTxtProgressBar(pbar, k + 1)
        } # end for(k in 1:nfolds)
        if(verbose) close(pbar)
        
      } # end if(parallel)
      
    } else if(family == "ordinal"){
      
      if(parallel){
        
        # define parcvloss function
        parcvloss <- 
          function(testid, xmat, ymat, group, family, weights, offset, alpha, 
                   nlambda, lambda.min.ratio, lambda, penalty.factor, penalty, 
                   gamma, theta, standardized, orthogonalized, intercept, 
                   thresh, maxit, proglang, type.measure, same.lambda, yfac, yrowsum){
            temp <- grpnet(x = xmat[-testid,,drop=FALSE], 
                           y = yfac[-testid], 
                           group = group, 
                           family = family,
                           weights = weights[-testid], 
                           offset = offset[-testid],
                           alpha = alpha,
                           nlambda = nlambda,
                           lambda.min.ratio = lambda.min.ratio,
                           lambda = if(same.lambda) lambda else NULL,
                           penalty.factor = penalty.factor,
                           penalty = penalty,
                           gamma = gamma,
                           theta = theta, 
                           standardized = standardized,
                           orthogonalized = orthogonalized,
                           intercept = intercept,
                           thresh = thresh,
                           maxit = maxit,
                           proglang = proglang)
            temp$ylev <- levels(yfac)
            nlev <- length(temp$ylev)
            mu <- predict(temp, newx = xmat[testid,,drop=FALSE], 
                          s = if(same.lambda) NULL else lambda,
                          type = ifelse(type.measure == "class", "class", "response"))
            if(type.measure == "class"){
              cvloss <- 1 - colMeans(yfac[testid] == mu)
            } else {
              cvloss <- rep(0.0, nlambda)
              for(i in 1:nlambda){
                mui <- cbind(1, mu[,,i])
                mui <- cbind(mui[,1:(nlev-1)] - mui[,2:nlev], mui[,nlev])
                if(type.measure == "deviance"){
                  mu0 <- rowSums(ymat[testid,,drop=FALSE] * mui)
                  cvloss[i] <- mean(grpnet.fit$family$dev.resids(yfac[testid], mu0, weights[testid]))
                } else if(type.measure == "mse"){
                  cvloss[i] <- mean((ymat[testid,,drop=FALSE] - mui)^2)
                } else if(type.measure == "mae"){
                  cvloss[i] <- mean(abs(ymat[testid,,drop=FALSE] - mui))
                }
              } # end for(i in 1:nlambda)
            } # end if(type.measure == "class")
            return(cvloss)
          } # end parcvloss
        
        # evaluate cvloss in parallel
        cvloss <- parallel::parSapply(cl = cluster, X = fid, FUN = parcvloss,
                                      xmat = x,
                                      ymat = ymat,
                                      group = group,
                                      family = family,
                                      weights = weights,
                                      offset = offset,
                                      alpha = grpnet.fit$alpha,
                                      nlambda = nlambda,
                                      lambda.min.ratio = lambda[nlambda] / lambda[1],
                                      lambda = lambda,
                                      penalty.factor = grpnet.fit$args$penalty.factor,
                                      penalty = grpnet.fit$args$penalty,
                                      gamma = grpnet.fit$args$gamma,
                                      theta = grpnet.fit$args$theta,
                                      standardized = grpnet.fit$args$standardized,
                                      orthogonalized = grpnet.fit$args$orthogonalized,
                                      intercept = grpnet.fit$args$intercept,
                                      thresh = grpnet.fit$args$thresh,
                                      maxit = grpnet.fit$args$maxit,
                                      proglang = grpnet.fit$args$proglang,
                                      type.measure = type.measure,
                                      same.lambda = same.lambda,
                                      yfac = yfac)
        
        # unvectorize
        cvloss <- matrix(cvloss, nrow = nlambda, ncol = nfolds)
        
      } else {
        for(k in 1:nfolds){
          if(same.lambda){
            temp <- grpnet(x = x[-fid[[k]],,drop=FALSE],
                           y = y[-fid[[k]]],
                           group = group,
                           weights = weights[-fid[[k]]],
                           offset = offset[-fid[[k]]],
                           alpha = alpha,
                           lambda = lambda, 
                           gamma = gamma, 
                           ...)
            temp$ylev <- ylev    # correct levels for classification loss
            mu <- predict(temp, newx = x[fid[[k]],,drop=FALSE],
                          type = ifelse(type.measure == "class", "class", "response"))
          } else {
            temp <- grpnet(x = x[-fid[[k]],,drop=FALSE],
                           y = y[-fid[[k]]],
                           group = group,
                           weights = weights[-fid[[k]]],
                           offset = offset[-fid[[k]]], 
                           alpha = alpha,
                           gamma = gamma,
                           ...)
            temp$ylev <- ylev    # correct levels for classification loss
            mu <- predict(temp, newx = x[fid[[k]],,drop=FALSE], s = lambda,
                          type = ifelse(type.measure == "class", "class", "response"))
          } # end if(same.lambda)
          if(type.measure == "class"){
            cvloss[,k] <- 1 - colMeans(yfac[fid[[k]]] == mu)
          } else {
            for(i in 1:nlambda){
              mui <- cbind(1, mu[,,i])
              mui <- cbind(mui[,1:(nlev-1)] - mui[,2:nlev], mui[,nlev])
              if(type.measure == "deviance"){
                mu0 <- rowSums(ymat[fid[[k]],,drop=FALSE] * mui)
                cvloss[i,k] <- mean(grpnet.fit$family$dev.resids(y[fid[[k]]], mu0, weights[fid[[k]]]))
              } else if(type.measure == "mse"){
                cvloss[i,k] <- mean((ymat[fid[[k]],,drop=FALSE] - mui)^2)
              } else if(type.measure == "mae"){
                cvloss[i,k] <- mean(abs(ymat[fid[[k]],,drop=FALSE] - mui))
              }
            } # end for(i in 1:nlambda)
          } # end # end if(type.measure == "class")
          if(verbose) setTxtProgressBar(pbar, k + 1)
        } # end for(k in 1:nfolds)
        if(verbose) close(pbar)
      } # end if(parallel)
      
    } else {
      
      if(parallel){
        
        # define parcvloss function
        parcvloss <- 
          function(testid, xmat, ymat, group, family, weights, offset, alpha, 
                   nlambda, lambda.min.ratio, lambda, penalty.factor, penalty, 
                   gamma, theta, standardized, orthogonalized, intercept, 
                   thresh, maxit, proglang, type.measure, same.lambda, yfac, yrowsum){
            temp <- grpnet(x = xmat[-testid,,drop=FALSE], 
                           y = ymat[-testid], 
                           group = group, 
                           family = family,
                           weights = weights[-testid], 
                           offset = offset[-testid],
                           alpha = alpha,
                           nlambda = nlambda,
                           lambda.min.ratio = lambda.min.ratio,
                           lambda = if(same.lambda) lambda else NULL,
                           penalty.factor = penalty.factor,
                           penalty = penalty,
                           gamma = gamma,
                           theta = theta, 
                           standardized = standardized,
                           orthogonalized = orthogonalized,
                           intercept = intercept,
                           thresh = thresh,
                           maxit = maxit,
                           proglang = proglang)
            temp$ylev <- levels(yfac)
            mu <- predict(temp, newx = xmat[testid,,drop=FALSE], 
                          s = if(same.lambda) NULL else lambda,
                          type = ifelse(type.measure == "class", "class", "response"))
            if(type.measure == "deviance"){
              cvloss <- colMeans(temp$family$dev.resids(ymat[testid], mu, weights[testid]))
            } else if(type.measure == "mse") {
              cvloss <- colMeans((ymat[testid] - mu)^2)
            } else if(type.measure == "mae"){
              cvloss <- colMeans(abs(ymat[testid] - mu))
            } else if(type.measure == "class"){
              cvloss <- 1 - colMeans(yfac[testid] == mu)
            }
            return(cvloss)
          } # end parcvloss
        
        # evaluate cvloss in parallel
        cvloss <- parallel::parSapply(cl = cluster, X = fid, FUN = parcvloss,
                                      xmat = x,
                                      ymat = y,
                                      group = group,
                                      family = family,
                                      weights = weights,
                                      offset = offset,
                                      alpha = grpnet.fit$alpha,
                                      nlambda = nlambda,
                                      lambda.min.ratio = lambda[nlambda] / lambda[1],
                                      lambda = lambda,
                                      penalty.factor = grpnet.fit$args$penalty.factor,
                                      penalty = grpnet.fit$args$penalty,
                                      gamma = grpnet.fit$args$gamma,
                                      theta = grpnet.fit$args$theta,
                                      standardized = grpnet.fit$args$standardized,
                                      orthogonalized = grpnet.fit$args$orthogonalized,
                                      intercept = grpnet.fit$args$intercept,
                                      thresh = grpnet.fit$args$thresh,
                                      maxit = grpnet.fit$args$maxit,
                                      proglang = grpnet.fit$args$proglang,
                                      type.measure = type.measure,
                                      same.lambda = same.lambda,
                                      yfac = yfac)
        
        # unvectorize
        cvloss <- matrix(cvloss, nrow = nlambda, ncol = nfolds)
        
      } else {
        for(k in 1:nfolds){
          if(same.lambda){
            temp <- grpnet(x = x[-fid[[k]],,drop=FALSE],
                           y = y[-fid[[k]]],
                           group = group,
                           weights = weights[-fid[[k]]],
                           offset = offset[-fid[[k]]],
                           alpha = alpha,
                           lambda = lambda, 
                           gamma = gamma, 
                           ...)
            temp$ylev <- ylev    # correct levels for classification loss
            mu <- predict(temp, newx = x[fid[[k]],,drop=FALSE],
                          type = ifelse(type.measure == "class", "class", "response"))
          } else {
            temp <- grpnet(x = x[-fid[[k]],,drop=FALSE],
                           y = y[-fid[[k]]],
                           group = group,
                           weights = weights[-fid[[k]]],
                           offset = offset[-fid[[k]]], 
                           alpha = alpha,
                           gamma = gamma,
                           ...)
            temp$ylev <- ylev    # correct levels for classification loss
            mu <- predict(temp, newx = x[fid[[k]],,drop=FALSE], s = lambda,
                          type = ifelse(type.measure == "class", "class", "response"))
          } # end if(same.lambda)
          if(type.measure == "deviance"){
            cvloss[,k] <- colMeans(grpnet.fit$family$dev.resids(y[fid[[k]]], mu, weights[fid[[k]]]))
          } else if(type.measure == "mse") {
            cvloss[,k] <- colMeans((y[fid[[k]]] - mu)^2)
          } else if(type.measure == "mae"){
            cvloss[,k] <- colMeans(abs(y[fid[[k]]] - mu))
          } else if(type.measure == "class"){
            cvloss[,k] <- 1 - colMeans(yfac[fid[[k]]] == mu)
          }
          if(verbose) setTxtProgressBar(pbar, k + 1)
        } # end for(k in 1:nfolds)
        if(verbose) close(pbar)
      } # end if(parallel)
      
    } # end if(family == "multigaussian")
    
    
    ######***######   RESULTS   ######***######
    
    ### get mean and se
    cvm <- rowMeans(cvloss)
    cvsd <- apply(cvloss, 1, sd) / sqrt(nfolds)
    
    ### find lambda.min
    minid <- which.min(cvm)
    lambda.min <- lambda[minid]
    
    ### find lambda.1se
    min1se <- cvm[minid] + cvsd[minid]
    se1id <- which(cvm <= min1se)[1]
    lambda.1se <- lambda[se1id]
    
    ### close cluster
    if(parallel && madeCluster) parallel::stopCluster(cluster)
    
    ### correct term.labels
    if(grpnet.fit$args$intercept){
      grpnet.fit$term.labels <- c("(Intercept)", names(gsize))
    } else {
      grpnet.fit$term.labels <- names(gsize)
    }
    
    ### end clock
    toc <- proc.time()[3]
    
    ### add power attr to adaptive
    attr(adaptive, "power") <- power
    
    ### return results
    res <- list(lambda = lambda, 
                cvm = cvm, 
                cvsd = cvsd, 
                cvup = cvm + cvsd, 
                cvlo = cvm - cvsd,
                nzero = grpnet.fit$nzgrp, 
                grpnet.fit = grpnet.fit, 
                lambda.min = lambda.min,
                lambda.1se = lambda.1se,
                index = c(minid, se1id),
                type.measure = type.measure,
                adaptive = adaptive,
                call = cv.grpnet.call,
                time = toc - tic)
    class(res) <- "cv.grpnet"
    return(res)
    
  } # end cv.grpnet.default

# Some code (for extracting coefficients) is re-purposed from the R package
# "glmnet" (Hastie et al., 2010) https://cran.r-project.org/package=glmnet

coef.cv.grpnet <-
  function(object, 
           s = c("lambda.1se", "lambda.min"),
           ...){
    # coefficients from a fit cv.grpnet object
    # Nathaniel E. Helwig (helwig@umn.edu)
    # Updated: 2024-10-06
    
    
    ### check object
    if(!inherits(object, "cv.grpnet")) stop("Input 'object' must be of class 'cv.grpnet'")
    
    ### check s
    if(is.character(s)){
      if(!any(s[1] == c("lambda.min", "lambda.1se"))) stop("Input 's' must be a character ('lambda.min' or 'lambda.1se') or a numeric vector")
      id <- object$index[ifelse(s[1] == "lambda.min", 1, 2)]
      s <- object$lambda[id]
    } else if(is.numeric(s)){
      if(any(s < 0)) stop("Inpust 's' must be a vector of non-negative numerics")
    } else {
      stop("Input 's' must be a character ('lambda.min' or 'lambda.1se') or a numeric vector")
    } # end if(is.character(s))
    
    ### return predictions
    coef(object$grpnet.fit, s = s, ...)
    
  } # end coef.cv.grpnet

# Some code (for interpolating coefficients) is re-purposed from the R package
# "glmnet" (Hastie et al., 2010) https://cran.r-project.org/package=glmnet

coef.grpnet <-
  function(object, 
           s = NULL,
           ...){
    # predict from a fit grpnet object
    # Nathaniel E. Helwig (helwig@umn.edu)
    # Updated: 2025-11-07
    
    
    ######***######   INITIAL CHECKS   ######***######
    
    ### check object
    if(!inherits(object, "grpnet")) stop("Input 'object' must be of class 'grpnet'")
    
    ### check s
    if(is.null(s)){
      newlambdas <- FALSE
      s <- object$lambda
    } else {
      newlambdas <- TRUE
      s <- as.numeric(s)
      if(any(s < 0)) stop("Inpust 's' must be a vector of non-negative numerics")
    } # end if(is.null(s))
    
    ### number of fit lambdas
    nlam <- length(object$lambda)
    
    ### number of new lambdas
    ns <- length(s)
    
    
    ######***######   MULTIGAUSSIAN / MULTINOMIAL   ######***######
    
    if(object$family$family %in% c("multigaussian", "multinomial")){
      
      ### get number of response classes
      nresp <- length(object$ylev)
      
      ### coefs at object$lambda are easy...
      if(!newlambdas){
        coefs <- vector("list", nresp)
        names(coefs) <- object$ylev
        for(k in 1:nresp){
          if(nlam == 1L){
            coefs[[k]] <- matrix(c(object$a0[k], object$beta[[k]]), ncol = 1)
            rownames(coefs[[k]]) <- c("(Intercept)", rownames(object$beta[[k]]))
            colnames(coefs[[k]]) <- "s1"
          } else {
            coefs[[k]] <- rbind(object$a0[k,], object$beta[[k]])
            rownames(coefs[[k]]) <- c("(Intercept)", rownames(object$beta[[k]]))
            colnames(coefs[[k]]) <- paste0("s", 1:nlam)
          }
        }
        class(coefs) <- "coef.grpnet"
        return(coefs)
      }
      
      ### only 1 fit lambda?
      if(nlam == 1L){
        coefs <- vector("list", nresp)
        names(coefs) <- object$ylev
        for(k in 1:nresp){
          coefs[[k]] <- matrix(c(object$a0[k], object$beta[[k]]), nrow = length(object$beta[[k]]) + 1, ncol = ns)
          rownames(coefs[[k]]) <- c("(Intercept)", names(object$beta[[k]]))
          colnames(coefs[[k]]) <- rep("s1", ns)
        }
        class(coefs) <- "coef.grpnet"
        return(coefs)
      }
      
      ### min and max lambda from fit model
      lambda.max <- max(object$lambda)
      lambda.min <- min(object$lambda)
      
      ### transform lambda (and s) to the [0,1] interval
      sfrac <- (lambda.max - s) / (lambda.max - lambda.min)
      lambda <- (lambda.max - object$lambda) / (lambda.max - lambda.min)
      
      ### correct for out-of-range s values
      sfrac[sfrac > max(lambda)] <- max(lambda)
      sfrac[sfrac < min(lambda)] <- min(lambda)
      
      ### linear interpolation at sfrac
      interp <- approx(x = lambda, y = seq(lambda), xout = sfrac)$y
      
      ### L and R indices
      left <- floor(interp)
      right <- ceiling(interp)
      
      ### update sfrac w.r.t. L and R indices
      sfrac <- (sfrac - lambda[right]) / (lambda[left] - lambda[right])
      sfrac[left == right] <- 1
      sfrac[abs(lambda[left] - lambda[right]) < .Machine$double.eps] <- 1
      
      ### interpolate...
      coefs <- vector("list", nresp)
      names(coefs) <- object$ylev
      for(k in 1:nresp){
        coefs[[k]] <- rbind(object$a0[k,], object$beta[[k]])
        coefs[[k]] <- coefs[[k]][, left, drop=FALSE] %*% diag(sfrac, nrow = ns, ncol = ns) + coefs[[k]][, right, drop=FALSE] %*% diag(1 - sfrac, nrow = ns, ncol = ns)
        rownames(coefs[[k]]) <- c("(Intercept)", rownames(object$beta[[k]]))
        colnames(coefs[[k]]) <- paste0("s", 1:ns)
      }
      class(coefs) <- "coef.grpnet"
      return(coefs)
      
    } # end if(family == "multinomial")
    
    
    
    ######***######   OTHER FAMILIES   ######***######
    
    ### coefficients at object$lambda are easy...
    if(!newlambdas){
      if(nlam == 1L){
        coefs <- matrix(c(object$a0, object$beta), ncol = 1)
        if(object$family$family == "ordinal"){
          rownames(coefs) <- c(paste0("y>=", object$ylev[-1]), names(object$beta))
        } else {
          rownames(coefs) <- c("(Intercept)", names(object$beta))
        }
        colnames(coefs) <- "s1"
      } else {
        coefs <- rbind(object$a0, object$beta)
        if(object$family$family == "ordinal"){
          rownames(coefs) <- c(paste0("y>=", object$ylev[-1]), rownames(object$beta))
        } else {
          rownames(coefs) <- c("(Intercept)", rownames(object$beta))
        }
        colnames(coefs) <- colnames(object$beta)
      }
      class(coefs) <- "coef.grpnet"
      return(coefs)
    }
    
    ### only 1 fit lambda?
    if(nlam == 1L){
      coefs <- matrix(c(object$a0, object$beta), nrow = length(object$beta) + length(object$a0), ncol = ns)
      if(object$family$family == "ordinal"){
        rownames(coefs) <- c(paste0("y>=", object$ylev[-1]), names(object$beta))
      } else {
        rownames(coefs) <- c("(Intercept)", names(object$beta))
      }
      colnames(coefs) <- rep("s1", ns)
      class(coefs) <- "coef.grpnet"
      return(coefs)
    }
    
    ### min and max lambda from fit model
    lambda.max <- max(object$lambda)
    lambda.min <- min(object$lambda)
    
    ### transform lambda (and s) to the [0,1] interval
    sfrac <- (lambda.max - s) / (lambda.max - lambda.min)
    lambda <- (lambda.max - object$lambda) / (lambda.max - lambda.min)
    
    ### correct for out-of-range s values
    sfrac[sfrac > max(lambda)] <- max(lambda)
    sfrac[sfrac < min(lambda)] <- min(lambda)
    
    ### linear interpolation at sfrac
    interp <- approx(x = lambda, y = seq(lambda), xout = sfrac)$y
    
    ### L and R indices
    left <- floor(interp)
    right <- ceiling(interp)
    
    ### update sfrac w.r.t. L and R indices
    sfrac <- (sfrac - lambda[right]) / (lambda[left] - lambda[right])
    sfrac[left == right] <- 1
    sfrac[abs(lambda[left] - lambda[right]) < .Machine$double.eps] <- 1
    
    ### interpolate...
    coefs <- rbind(object$a0, object$beta)
    coefs <- coefs[, left, drop=FALSE] %*% diag(sfrac, nrow = ns, ncol = ns) + coefs[, right, drop=FALSE] %*% diag(1 - sfrac, nrow = ns, ncol = ns)
    #rownames(coefs) <- c("(Intercept)", rownames(object$beta))
    if(object$family$family == "ordinal"){
      rownames(coefs) <- c(paste0("y>=", object$ylev[-1]), rownames(object$beta))
    } else {
      rownames(coefs) <- c("(Intercept)", rownames(object$beta))
    }
    colnames(coefs) <- paste0("s", 1:ns)
    class(coefs) <- "coef.grpnet"
    return(coefs)
    
  } # end coef.grpnet

print.coef.grpnet <-
  function(x, ...){
    ndigits <- getOption("digits")
    if(is.list(x)){
      for(j in 1:length(x)){
        x[[j]] <- formatC(x[[j]], 
                          format = "f",
                          digits = ndigits,
                          zero.print = paste(rep(c(".", " "), times = c(1,  ndigits)), collapse = ""),
                          preserve.width = "common")
      }
    } else {
      x <- formatC(x, 
                   format = "f",
                   digits = ndigits,
                   zero.print = paste(rep(c(".", " "), times = c(1,  ndigits)), collapse = ""),
                   preserve.width = "common")
    }
    print.default(x, right = TRUE, quote = FALSE)
  }

# Some code (for interpolating predictions) is re-purposed from the R package
# "glmnet" (Hastie et al., 2010) https://cran.r-project.org/package=glmnet

predict.grpnet <-
  function(object, 
           newx,
           newdata,
           s = NULL,
           type = c("link", "response", "class", "terms", 
                    "importance", "coefficients", "nonzero", "groups", 
                    "ncoefs", "ngroups", "norm", "znorm"),
           ...){
    # predict from a fit grpnet object
    # Nathaniel E. Helwig (helwig@umn.edu)
    # Updated: 2026-03-31
    
    
    ######***######   INITIAL CHECKS   ######***######
    
    ### check object
    if(!inherits(object, "grpnet")) stop("Input 'object' must be of class 'grpnet'")
  
    ### check type
    thetypes <- c("link", "response", "class", "terms", "importance", "coefficients", "nonzero", "groups", "ncoefs", "ngroups", "norm", "znorm")
    family <- object$family$family
    type <- pmatch(as.character(type[1]), thetypes)
    if(is.na(type)) stop("Invalid 'type' input")
    type <- thetypes[type]
    if(type == "class" && !(family %in% c("svm1", "svm2", "logit", "binomial", "multinomial", "ordinal")))
      stop("Input 'type' can only be set to 'class' for families:\nsvm1, svm2, logit, binomial, multinomial, ordinal")
    
    
    
    ######***######   COEFFICIENTS & RELATED INFO   ######***######
    
    ### type == "coefficients"
    if(type == "coefficients"){
      return(coef.grpnet(object = object, s = s))
    }
    
    ### type == "nonzero"
    if(type == "nonzero"){
      coefs <- coef.grpnet(object = object, s = s)
      if(family %in% c("multigaussian", "multinomial")){
        return(apply(Reduce("+", lapply(coefs, function(x) abs(x) > 0)), 2, function(x) which(x > 0)))
      } else {
        return(apply(abs(coefs) > 0, 2, which))
      }
    }
    
    ### type == "ncoefs"
    if(type == "ncoefs"){
      coefs <- coef.grpnet(object = object, s = s)
      if(family %in% c("multigaussian", "multinomial")){
        return(Reduce("+", lapply(coefs, function(x) colSums(abs(x) > 0))))
      } else {
        return(colSums(abs(coefs) > 0))
      }
    }
    
    ### l2norm
    if(type %in% c("groups", "ngroups", "norm", "znorm")){
      if(family %in% c("multigaussian", "multinomial")){
        l2norm <- function(x) sum(x^2)
      } else {
        l2norm <- function(x) sqrt(sum(x^2))
      }
      grpnorm <- function(x) tapply(x, object$group, l2norm)
    }
    
    ### type == "groups"
    if(type == "groups"){
      coefs <- coef.grpnet(object = object, s = s)
      if(family %in% c("multigaussian", "multinomial")){
        norms <- sqrt(Reduce("+", lapply(coefs, function(x) apply(x, 2, grpnorm))))
      } else {
        norms <- apply(coefs, 2, grpnorm)
      }
      if(ncol(norms) > 1L){
        return(apply(norms, 2, function(x) object$term.labels[which(x > 0)]))
      } else {
        return(object$term.labels[which(norms > 0)])
      }
    }
    
    ### type == "ngroups"
    if(type == "ngroups"){
      coefs <- coef.grpnet(object = object, s = s)
      if(family %in% c("multigaussian", "multinomial")){
        return(colSums(sqrt(Reduce("+", lapply(coefs, function(x) apply(x, 2, grpnorm)))) > 0))
      } else {
        return(colSums(apply(coefs, 2, grpnorm) > 0))
      }
    }
    
    ### type == "norm" or "znorm"
    if(type %in% c("norm", "znorm")){
      coefs <- coef.grpnet(object = object, s = s)
      if(family %in% c("multigaussian", "multinomial")){
        norms <- sqrt(Reduce("+", lapply(coefs, function(x) apply(x, 2, grpnorm))))
      } else {
        norms <- apply(coefs, 2, grpnorm)
      }
      if(ncol(norms) > 1L){
        rownames(norms) <- object$term.labels
      } else {
        norms <- c(norms)
        names(norms) <- object$term.labels
      }
      if(type == "znorm"){
        return(norms * object$xsd)
      } else {
        return(norms)
      }
    }
    
    
    
    ######***######   INITIAL CHECKS (CONTINUED)   ######***######
    
    ### check newx / newdata
    if(is.null(object$formula)){
      newx <- as.matrix(newx)
      ncoefs <- if(is.list(object$beta)) nrow(object$beta[[1]]) else nrow(object$beta)
      if(ncol(newx) != ncoefs) stop("Input 'newx' must satisfy:  ncol(newx) == nrow(object$beta)")
    } else {
      vnames <- rownames(attr(terms(object$formula), "factors"))[-1]
      check <- match(vnames, names(newdata))
      if(any(is.na(check))) stop("Input 'data' is missing variables included in model formula")
      if(!is.null(object$rk.args)){
        newx <- rk.model.matrix(object = object$formula, data = newdata,
                                knots = object$rk.args$knots,
                                Boundary.knots = lapply(object$rk.args$knots, function(x) x[c(1, length(x))]),
                                m = object$rk.args$m, 
                                periodic = object$rk.args$periodic,
                                xlev = object$rk.args$xlev)
      } else {
        newx <- model.matrix(object = object$formula, data = newdata)
      }
      if(colnames(newx)[1] == "(Intercept)") newx <- newx[,-1,drop=FALSE]
      ncoefs <- if(is.list(object$beta)) nrow(as.matrix(object$beta[[1]])) else nrow(as.matrix(object$beta))
      if(ncol(newx) != ncoefs) stop("Input 'newdata' produced a design matrix of the wrong dimension\n(likely due to a factor level mismatch between 'data' and 'newdata')")
      object$formula <- NULL
    }
    nobs <- nrow(newx)
    newxnames <- rownames(newx)
    if(is.null(newxnames)) newxnames <- 1:nobs
    
    ### check s
    if(is.null(s)){
      newlambdas <- FALSE
      s <- object$lambda
    } else {
      newlambdas <- TRUE
      s <- as.numeric(s)
      if(any(s < 0)) stop("Inpust 's' must be a vector of non-negative numerics")
    } # end if(is.null(s))
    
    ### number of fit lambdas
    nlam <- length(object$lambda)
    
    ### number of new lambdas
    ns <- length(s)
    
    
    
    ######***######   TERMS   ######***######
    
    if(type == "terms"){
      
      ### get coefs
      coefs <- coef(object, s = s)
      
      ### group names
      gnames <- object$term.labels
      ngroups <- object$ngroups
      group <- object$group
      
      ### remove intercept?
      int <- 0L
      if(gnames[1] == "(Intercept)") {
        int <- 1L
        gnames <- gnames[-1]
        ngroups <- ngroups - 1L
        if(object$family$family == "ordinal"){
          int <- length(object$ylev) - 1L
        }
        group <- group[-c(1:int)]
      }
      
      ### multinomial or other?
      if(family %in% c("multigaussian", "multinomial")){
        
        ## get number of response classes
        nresp <- length(object$beta)
        
        ## initialize terms list (of arrays)
        terms <- vector("list", nresp)
        names(terms) <- object$ylev
        for(j in 1:nresp){
          terms[[j]] <- array(dim = c(nobs, ngroups, ns))
          dimnames(terms[[j]]) <- list(newxnames, gnames, paste0("s", 1:ns))
        }
        
        ## loop thru terms
        for(k in 1:ngroups){
          id <- which(group == k)
          for(j in 1:nresp){
            terms[[j]][,k,] <- newx[,id,drop=FALSE] %*% coefs[[j]][id+int,,drop=FALSE]
          }
        }
        
        ## return results
        return(lapply(terms, drop))
        
      } else {
        
        ## initialize terms array
        terms <- array(dim = c(nobs, ngroups, ns))
        dimnames(terms) <- list(newxnames, gnames, paste0("s", 1:ns))
        
        ## loop thru terms
        for(k in 1:ngroups){
          id <- which(group == k)
          terms[,k,] <- newx[,id,drop=FALSE] %*% coefs[id+int,,drop=FALSE]
        } # end for(k in 1:object$ngroups)
        
        ## return results
        return(drop(terms))
        
      } # end if(family == "multinomial")
      
    } # end if(type == "terms")
    
    
    
    ######***######   IMPORTANCE   ######***######
    
    if(type == "importance"){
      
      ### get terms
      terms <- predict.grpnet(object, newx = newx, s = s, type = "terms")
      
      ### multinomial or other?
      if(family %in% c("multigaussian", "multinomial")){
        
        ## number of terms
        nterms <- dim(terms[[1]])[2]
        termnames <- dimnames(terms[[1]])[[2]]
        
        ## get number of response classes
        nresp <- length(object$beta)
        
        ## list (of matrices) for results
        imp <- vector("list", nresp)
        names(imp) <- object$ylev
        mat <- matrix(0, nrow = nterms, ncol = ns)
        rownames(mat) <- termnames
        colnames(mat) <- paste0("s", 1:ns)
        
        ## loop thru 1:nresp
        for(j in 1:nresp){
          
          imp[[j]] <- mat
          
          if(ns == 1L){
            etai <- scale(terms[[j]], scale = FALSE)
            eta0 <- rowSums(etai)
            sst0 <- sum(eta0^2)
            imp[[j]] <- colSums(etai * eta0) / ifelse(sst0 > 0, sst0, 1)
          } else {
            for(i in 1:ns){
              etai <- scale(terms[[j]][,,i], scale = FALSE)
              eta0 <- rowSums(etai)
              sst0 <- sum(eta0^2)
              imp[[j]][,i] <- colSums(etai * eta0) / ifelse(sst0 > 0, sst0, 1)
            } 
          } # end if(ns == 1L)
          
        } # end for(j in 1:nresp)
        
        if(ns == 1L) {
          imp <- sapply(imp, c)
          colnames(imp) <- object$ylev
        }
        return(imp)
        
      } else {
        
        ## number of terms
        nterms <- dim(terms)[2]
        termnames <- dimnames(terms)[[2]]
        
        ### matrix for results
        imp <- matrix(0, nrow = nterms, ncol = ns)
        rownames(imp) <- termnames
        colnames(imp) <- paste0("s", 1:ns)
        
        ## loop thru 1:ns
        if(ns == 1L){
          etai <- scale(terms, scale = FALSE)
          eta0 <- rowSums(etai)
          sst0 <- sum(eta0^2)
          imp <- colSums(etai * eta0) / ifelse(sst0 > 0, sst0, 1)
        } else {
          for(i in 1:ns){
            etai <- scale(terms[,,i], scale = FALSE)
            eta0 <- rowSums(etai)
            sst0 <- sum(eta0^2)
            imp[,i] <- colSums(etai * eta0) / ifelse(sst0 > 0, sst0, 1)
          } 
        } # end if(ns == 1L)
        
        return(drop(imp))
        
      } # end if(family == "multinomial")
      
    } # end if(type == "importance")
    
    
    
    ######***######   MULTIGAUSSIAN / MULTINOMIAL   ######***######
    if(family %in% c("multigaussian", "multinomial")){
      
      # note: special code to return predictions for each response dim / class
      
      ### get number of response classes
      nresp <- length(object$beta)
      
      ### initialize array for predictions
      fit <- array(data = 0.0, dim = c(nobs, nresp, ns))
      dimnames(fit) <- list(newxnames, object$ylev, paste0("s", 1:ns))
      
      ### predictions at object$lambda are easy...
      if(!newlambdas){
        for(k in 1:nresp) {
          if(nlam == 1L){
            fit[,k,] <- cbind(1, newx) %*% c(object$a0[k], object$beta[[k]])
          } else {
            fit[,k,] <- cbind(1, newx) %*% rbind(object$a0[k,], object$beta[[k]])
          }
        }
        if(family == "multinomial" && type %in% c("response", "class")) {
          for(i in 1:ns) fit[,,i] <- object$family$linkinv(fit[,,i])
          if(type == "class") fit <- matrix(object$ylev[apply(fit, c(1,3), which.max)], nrow = nobs, ncol = ns)
        }
        return(drop(fit))
      }
      
      ### only 1 fit lambda?
      if(nlam == 1L){
        for(k in 1:nresp){
          fit[,k,] <- cbind(1, newx) %*% c(object$a0[k], object$beta[[k]])
        }
        if(family == "multinomial"){
          if(type == "response") {
            fit <- object$family$linkinv(fit)
          } else if(type == "class"){
            # note: family == "multinomial" is implied
            fit <- object$ylev[apply(object$family$linkinv(fit), 1, which.max)]
          }
        }
        return(fit)
      }
      
      ### min and max lambda from fit model
      lambda.max <- max(object$lambda)
      lambda.min <- min(object$lambda)
      
      ### transform lambda (and s) to the [0,1] interval
      sfrac <- (lambda.max - s) / (lambda.max - lambda.min)
      lambda <- (lambda.max - object$lambda) / (lambda.max - lambda.min)
      
      ### correct for out-of-range s values
      sfrac[sfrac > max(lambda)] <- max(lambda)
      sfrac[sfrac < min(lambda)] <- min(lambda)
      
      ### linear interpolation at sfrac
      interp <- approx(x = lambda, y = seq(lambda), xout = sfrac)$y
      
      ### L and R indices
      left <- floor(interp)
      right <- ceiling(interp)
      
      ### update sfrac w.r.t. L and R indices
      sfrac <- (sfrac - lambda[right]) / (lambda[left] - lambda[right])
      sfrac[left == right] <- 1
      sfrac[abs(lambda[left] - lambda[right]) < .Machine$double.eps] <- 1
      
      ### interpolate...
      for(k in 1:nresp){
        coefs <- rbind(object$a0[k,], object$beta[[k]])
        coefs <- coefs[, left, drop=FALSE] %*% diag(sfrac, nrow = ns, ncol = ns) + coefs[, right, drop=FALSE] %*% diag(1 - sfrac, nrow = ns, ncol = ns)
        colnames(coefs) <- paste0("s", 1:ns)
        fit[,k,] <- cbind(1, newx) %*% coefs
      }
      
      ### return fitted values
      if(family == "multinomial"){
        if(type %in% c("response", "class")) {
          for(i in 1:ns) fit[,,i] <- object$family$linkinv(fit[,,i])
          if(type == "class") fit <- matrix(object$ylev[apply(fit, c(1,3), which.max)], nrow = nobs, ncol = ns)
        }
      }
      return(drop(fit))
      
    } # end if(object$family$family == "multinomial")
    
    
    
    ######***######   ORDINAL   ######***######
    if(family == "ordinal"){
      
      nlev <- length(object$ylev)
      eta <- newx %*% object$beta
      
      ### predictions at object$lambda are easy...
      if(!newlambdas){
        if(nlam == 1L){
          fit <- matrix(eta, nrow = nobs, ncol = nlev - 1)
          rownames(fit) <- 1:nobs
          colnames(fit) <- paste0("y>=", object$ylev[-1])
          for(k in 1:(nlev - 1)) fit[,k] <- fit[,k] + object$a0[k]
          if(type == "response"){
            fit <- object$family$linkinv(fit)
          } else if(type == "class"){
            fit <- object$family$linkinv(fit)
            yc <- rep("NA", nobs)
            mu <- cbind(1, fit)
            mu <- cbind(mu[,1:(nlev-1)] - mu[,2:nlev], mu[,nlev])
            fit <- object$ylev[apply(mu, 1, which.max)]
          }
        } else {
          fit <- array(dim = c(nobs, nlev - 1, nlam))
          dimnames(fit) <- list(1:nobs, 
                                paste0("y>=", object$ylev[-1]), 
                                paste0("s", 1:length(object$lambda)))
          for(k in 1:(nlev - 1)) fit[,k,] <- eta + matrix(object$a0[k,], nrow = nobs, ncol = nlam, byrow = TRUE)
          if(type == "response"){
            fit <- object$family$linkinv(fit)
          } else if(type == "class"){
            fit <- object$family$linkinv(fit)
            yc <- matrix("NA", nrow = nobs, ncol = nlam)
            colnames(yc) <- paste0("s", 1:nlam)
            for(i in 1:nlam){
              mu <- cbind(1, fit[,,i])
              mu <- cbind(mu[,1:(nlev-1)] - mu[,2:nlev], mu[,nlev])
              yc[,i] <- object$ylev[apply(mu, 1, which.max)]
            }
            fit <- yc
          }
        } # end if(nlam == 1L)
        return(drop(fit))
      }
      
      ### only 1 fit lambda?
      if(nlam == 1L){
        fit <- matrix(eta, nrow = nobs, ncol = nlev - 1)
        rownames(fit) <- 1:nobs
        colnames(fit) <- paste0("y>=", object$ylev[-1])
        for(k in 1:(nlev - 1)) fit[,k] <- fit[,k] + object$a0[k]
        if(type == "response") {
          fit <- object$family$linkinv(fit)
        } else if(type == "class"){
          fit <- object$family$linkinv(fit)
          yc <- rep("NA", nobs)
          mu <- cbind(1, fit)
          mu <- cbind(mu[,1:(nlev-1)] - mu[,2:nlev], mu[,nlev])
          fit <- object$ylev[apply(mu, 1, which.max)]
        }
        return(drop(matrix(fit, nrow = nrow(newx), ncol = ns)))
      }
      
      ### min and max lambda from fit model
      lambda.max <- max(object$lambda)
      lambda.min <- min(object$lambda)
      
      ### transform lambda (and s) to the [0,1] interval
      sfrac <- (lambda.max - s) / (lambda.max - lambda.min)
      lambda <- (lambda.max - object$lambda) / (lambda.max - lambda.min)
      
      ### correct for out-of-range s values
      sfrac[sfrac > max(lambda)] <- max(lambda)
      sfrac[sfrac < min(lambda)] <- min(lambda)
      
      ### linear interpolation at sfrac
      interp <- approx(x = lambda, y = seq(lambda), xout = sfrac)$y
      
      ### L and R indices
      left <- floor(interp)
      right <- ceiling(interp)
      
      ### update sfrac w.r.t. L and R indices
      sfrac <- (sfrac - lambda[right]) / (lambda[left] - lambda[right])
      sfrac[left == right] <- 1
      sfrac[abs(lambda[left] - lambda[right]) < .Machine$double.eps] <- 1
      
      ### interpolate...
      coefs <- rbind(object$a0, object$beta)
      coefs <- coefs[, left, drop=FALSE] %*% diag(sfrac, nrow = ns, ncol = ns) + coefs[, right, drop=FALSE] %*% diag(1 - sfrac, nrow = ns, ncol = ns)
      colnames(coefs) <- paste0("s", 1:ns)
      
      ### return fitted values
      shiftid <- 1:(nlev - 1L)
      object$a0 <- coefs[shiftid,,drop=FALSE]
      object$beta <- coefs[-shiftid,,drop=FALSE]
      eta <- newx %*% object$beta
      fit <- array(dim = c(nobs, nlev - 1, ns))
      dimnames(fit) <- list(1:nobs, 
                            paste0("y>=", object$ylev[-1]), 
                            paste0("s", 1:ns))
      for(k in 1:(nlev - 1)) fit[,k,] <- eta + matrix(object$a0[k,], nrow = nobs, ncol = ns, byrow = TRUE)
      if(type == "response"){
        fit <- object$family$linkinv(fit)
      } else if(type == "class"){
        fit <- object$family$linkinv(fit)
        yc <- matrix("NA", nrow = nobs, ncol = ns)
        colnames(yc) <- paste0("s", 1:ns)
        for(i in 1:ns){
          mu <- cbind(1, fit[,,i])
          mu <- cbind(mu[,1:(nlev-1)] - mu[,2:nlev], mu[,nlev])
          yc[,i] <- object$ylev[apply(mu, 1, which.max)]
        }
        fit <- yc
      }
      return(drop(fit))
    } # end if(family == "ordinal")
    
    
    ######***######   OTHER FAMILIES   ######***######
    
    ### predictions at object$lambda are easy...
    if(!newlambdas){
      if(nlam == 1L){
        fit <- cbind(1, newx) %*% c(object$a0, object$beta)
      } else {
        fit <- cbind(1, newx) %*% rbind(object$a0, object$beta)
      }
      if(type == "response") {
        fit <- object$family$linkinv(fit)
      } else if(type == "class"){ 
        thresh <- ifelse(family %in% c("svm1", "svm2"), 0.0, 0.5)   # else: binomial/logit implied
        fit <- ifelse(object$family$linkinv(fit) <= thresh, object$ylev[1], object$ylev[2])
      }
      return(drop(fit))
    }
    
    ### only 1 fit lambda?
    if(nlam == 1L){
      fit <- cbind(1, newx) %*% c(object$a0, object$beta)
      if(type == "response") {
        fit <- object$family$linkinv(fit)
      } else if(type == "class"){
        thresh <- ifelse(family %in% c("svm1", "svm2"), 0.0, 0.5)   # else: binomial/logit implied
        fit <- ifelse(object$family$linkinv(fit) <= thresh, object$ylev[1], object$ylev[2])
      }
      return(drop(matrix(fit, nrow = nrow(newx), ncol = ns)))
    }
    
    ### min and max lambda from fit model
    lambda.max <- max(object$lambda)
    lambda.min <- min(object$lambda)
    
    ### transform lambda (and s) to the [0,1] interval
    sfrac <- (lambda.max - s) / (lambda.max - lambda.min)
    lambda <- (lambda.max - object$lambda) / (lambda.max - lambda.min)
    
    ### correct for out-of-range s values
    sfrac[sfrac > max(lambda)] <- max(lambda)
    sfrac[sfrac < min(lambda)] <- min(lambda)
    
    ### linear interpolation at sfrac
    interp <- approx(x = lambda, y = seq(lambda), xout = sfrac)$y
    
    ### L and R indices
    left <- floor(interp)
    right <- ceiling(interp)
    
    ### update sfrac w.r.t. L and R indices
    sfrac <- (sfrac - lambda[right]) / (lambda[left] - lambda[right])
    sfrac[left == right] <- 1
    sfrac[abs(lambda[left] - lambda[right]) < .Machine$double.eps] <- 1
    
    ### interpolate...
    coefs <- rbind(object$a0, object$beta)
    coefs <- coefs[, left, drop=FALSE] %*% diag(sfrac, nrow = ns, ncol = ns) + coefs[, right, drop=FALSE] %*% diag(1 - sfrac, nrow = ns, ncol = ns)
    colnames(coefs) <- paste0("s", 1:ns)
    
    ### return fitted values
    fit <- cbind(1, newx) %*% coefs
    if(type == "response") {
      fit <- object$family$linkinv(fit)
    } else if(type == "class"){
      thresh <- ifelse(family %in% c("svm1", "svm2"), 0.0, 0.5)   # else: binomial/logit implied
      fit <- ifelse(object$family$linkinv(fit) <= thresh, object$ylev[1], object$ylev[2])
    }
    return(drop(fit))
    
  } # end predict.grpnet

\name{predict.cv.grpnet}
\alias{predict.cv.grpnet}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Predict Method for cv.grpnet Fits
}
\description{
Obtain predictions from a cross-validated group elastic net (cv.grpnet) object.
}
\usage{
\method{predict}{cv.grpnet}(object, 
        newx,
        newdata,
        s = c("lambda.1se", "lambda.min"),
        type = c("link", "response", "class", "terms", 
                 "importance", "coefficients", "nonzero", "groups", 
                 "ncoefs", "ngroups", "norm", "znorm"),
        ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{object}{
Object of class "cv.grpnet"
}
  \item{newx}{
Matrix of new \code{x} scores for prediction (default S3 method). Must have \eqn{p} columns arranged in the same order as the \code{x} matrix used to fit the model.
}
  \item{newdata}{
Data frame of new \code{data} scores for prediction (S3 "formula" method). Must contain all variables in the \code{formula} used to fit the model.
}
  \item{s}{
Lambda value(s) at which predictions should be obtained. Can input a character ("lambda.min" or "lambda.1se") or a numeric vector. Default of "lambda.min" uses the \code{lambda} value that minimizes the mean cross-validated error. 
}
  \item{type}{
Type of prediction to return. "link" gives predictions on the link scale (\eqn{\eta}). "response" gives predictions on the mean scale (\eqn{\mu}). "class" gives predicted class labels (for "binomial" and "multinomial" families). "terms" gives the predictions for each term (group) in the model (\eqn{\eta_k}). "importance" gives the variable importance index for each term (group) in the model. "coefficients" returns the coefficients used for predictions. "nonzero" returns a list giving the indices of non-zero coefficients for each \code{s}. "groups" returns a list giving the labels of non-zero groups for each \code{s}. "ncoefs" returns the number of non-zero coefficients for each \code{s}. "ngroups" returns the number of non-zero groups for each \code{s}. "norm" returns the L2 norm of each group's (raw) coefficients for each \code{s}. "znorm" returns the L2 norm of each group's standardized coefficients for each \code{s}.
}
  \item{...}{
Additional arguments (ignored)
}
}
\details{
Predictions are calculated from the \code{\link{grpnet}} object fit to the full sample of data, which is stored as \code{object$grpnet.fit} 

See \code{\link{predict.grpnet}} for further details on the calculation of the different types of predictions.
}
\value{
Depends on three factors... \cr
1. the exponential family distribution \cr
2. the length of the input \code{s} \cr
3. the \code{type} of prediction requested

See \code{\link{predict.grpnet}} for details
}
\references{
Friedman, J.,  Hastie, T., & Tibshirani, R. (2010). Regularization paths for generalized linear models via coordinate descent. \emph{Journal of Statistical Software, 33}(1), 1-22. \doi{10.18637/jss.v033.i01}

Helwig, N. E. (2025). Versatile descent algorithms for group regularization and variable selection in generalized linear models. \emph{Journal of Computational and Graphical Statistics, 34}(1), 239-252. \doi{10.1080/10618600.2024.2362232}
}
\author{
Nathaniel E. Helwig <helwig@umn.edu>
}
\note{
Syntax is inspired by the \code{predict.cv.glmnet} function in the \bold{glmnet} package (Friedman, Hastie, & Tibshirani, 2010). 
}

%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link{cv.grpnet}} for k-fold cross-validation of \code{lambda}

\code{\link{predict.grpnet}} for predicting from \code{\link{grpnet}} objects
}
\examples{
\donttest{
######***######   family = "gaussian"   ######***######

# load data
data(auto)

# 10-fold cv (formula method, response = mpg)
set.seed(1)
mod <- cv.grpnet(mpg ~ ., data = auto)

# get fitted values at "lambda.1se"
fit.1se <- predict(mod, newdata = auto)

# get fitted values at "lambda.min"
fit.min <- predict(mod, newdata = auto, s = "lambda.min")

# compare mean absolute error for two solutions
mean(abs(auto$mpg - fit.1se))
mean(abs(auto$mpg - fit.min))



######***######   family = "multigaussian"   ######***######

# load data
data(auto)

# 10-fold cv (formula method, response = (mpg, displacement))
y <- as.matrix(auto[,c(1,3)])
set.seed(1)
mod <- cv.grpnet(y ~ ., data = auto[,-c(1,3)], family = "multigaussian",
                 standardize.response = TRUE)

# get fitted values at "lambda.1se"
fit.1se <- predict(mod, newdata = auto)

# get fitted values at "lambda.min"
fit.min <- predict(mod, newdata = auto, s = "lambda.min")

# compare mean absolute error for two solutions
mean(abs(y - fit.1se))
mean(abs(y - fit.min))



######***######   family = "svm1"   ######***######

# load data
data(auto)

# redefine origin (Domestic vs Foreign)
auto$origin <- ifelse(auto$origin == "American", "Domestic", "Foreign")

# fit model (formula method, response = origin with 2 levels)
set.seed(1)
mod <- cv.grpnet(origin ~ ., data = auto, family = "svm1")

# get predicted classes at "lambda.1se"
fit.1se <- predict(mod, newdata = auto, type = "class")

# get predicted classes at "lambda.min"
fit.min <- predict(mod, newdata = auto, type = "class", s = "lambda.min")

# compare misclassification rate for two solutions
1 - mean(auto$origin == fit.1se)
1 - mean(auto$origin == fit.min)



######***######   family = "svm2"   ######***######

# load data
data(auto)

# redefine origin (Domestic vs Foreign)
auto$origin <- ifelse(auto$origin == "American", "Domestic", "Foreign")

# fit model (formula method, response = origin with 2 levels)
set.seed(1)
mod <- cv.grpnet(origin ~ ., data = auto, family = "svm2")

# get predicted classes at "lambda.1se"
fit.1se <- predict(mod, newdata = auto, type = "class")

# get predicted classes at "lambda.min"
fit.min <- predict(mod, newdata = auto, type = "class", s = "lambda.min")

# compare misclassification rate for two solutions
1 - mean(auto$origin == fit.1se)
1 - mean(auto$origin == fit.min)



######***######   family = "logit"   ######***######

# load data
data(auto)

# redefine origin (Domestic vs Foreign)
auto$origin <- ifelse(auto$origin == "American", "Domestic", "Foreign")

# fit model (formula method, response = origin with 2 levels)
set.seed(1)
mod <- cv.grpnet(origin ~ ., data = auto, family = "logit")

# get predicted classes at "lambda.1se"
fit.1se <- predict(mod, newdata = auto, type = "class")

# get predicted classes at "lambda.min"
fit.min <- predict(mod, newdata = auto, type = "class", s = "lambda.min")

# compare misclassification rate for two solutions
1 - mean(auto$origin == fit.1se)
1 - mean(auto$origin == fit.min)



######***######   family = "binomial"   ######***######

# load data
data(auto)

# redefine origin (Domestic vs Foreign)
auto$origin <- ifelse(auto$origin == "American", "Domestic", "Foreign")

# 10-fold cv (default method, response = origin with 2 levels)
set.seed(1)
mod <- cv.grpnet(origin ~ ., data = auto, family = "binomial")

# get predicted classes at "lambda.1se"
fit.1se <- predict(mod, newdata = auto, type = "class")

# get predicted classes at "lambda.min"
fit.min <- predict(mod, newdata = auto, type = "class", s = "lambda.min")

# compare misclassification rate for two solutions
1 - mean(auto$origin == fit.1se)
1 - mean(auto$origin == fit.min)



######***######   family = "multinomial"   ######***######

# load data
data(auto)

# 10-fold cv (formula method, response = origin with 3 levels)
set.seed(1)
mod <- cv.grpnet(origin ~ ., data = auto, family = "multinomial")

# get predicted classes at "lambda.1se"
fit.1se <- predict(mod, newdata = auto, type = "class")

# get predicted classes at "lambda.min"
fit.min <- predict(mod, newdata = auto, type = "class", s = "lambda.min")

# compare misclassification rate for two solutions
1 - mean(auto$origin == fit.1se)
1 - mean(auto$origin == fit.min)



######***######   family = "ordinal"   ######***######

# load data
data(auto)

# 10-fold cv (formula method, response = cylinders with 5 levels)
set.seed(1)
mod <- cv.grpnet(cylinders ~ ., data = auto, family = "ordinal")

# get predicted classes at "lambda.1se"
fit.1se <- predict(mod, newdata = auto, type = "class")

# get predicted classes at "lambda.min"
fit.min <- predict(mod, newdata = auto, type = "class", s = "lambda.min")

# compare misclassification rate for two solutions
1 - mean(auto$cylinders == fit.1se)
1 - mean(auto$cylinders == fit.min)



######***######   family = "poisson"   ######***######

# load data
data(auto)

# 10-fold cv (formula method, response = horsepower)
set.seed(1)
mod <- cv.grpnet(horsepower ~ ., data = auto, family = "poisson")

# get fitted values at "lambda.1se"
fit.1se <- predict(mod, newdata = auto, type = "response")

# get fitted values at "lambda.min"
fit.min <- predict(mod, newdata = auto, type = "response", s = "lambda.min")

# compare mean absolute error for two solutions
mean(abs(auto$horsepower - fit.1se))
mean(abs(auto$horsepower - fit.min))



######***######   family = "negative.binomial"   ######***######

# load data
data(auto)

# 10-fold cv (formula method, response = horsepower)
set.seed(1)
mod <- cv.grpnet(horsepower ~ ., data = auto, family = "negative.binomial")

# get fitted values at "lambda.1se"
fit.1se <- predict(mod, newdata = auto, type = "response")

# get fitted values at "lambda.min"
fit.min <- predict(mod, newdata = auto, type = "response", s = "lambda.min")

# compare mean absolute error for two solutions
mean(abs(auto$horsepower - fit.1se))
mean(abs(auto$horsepower - fit.min))



######***######   family = "Gamma"   ######***######

# load data
data(auto)

# 10-fold cv (formula method, response = origin)
set.seed(1)
mod <- cv.grpnet(mpg ~ ., data = auto, family = "Gamma")

# get fitted values at "lambda.1se"
fit.1se <- predict(mod, newdata = auto, type = "response")

# get fitted values at "lambda.min"
fit.min <- predict(mod, newdata = auto, type = "response", s = "lambda.min")

# compare mean absolute error for two solutions
mean(abs(auto$mpg - fit.1se))
mean(abs(auto$mpg - fit.min))



######***######   family = "inverse.gaussian"   ######***######

# load data
data(auto)

# 10-fold cv (formula method, response = origin)
set.seed(1)
mod <- cv.grpnet(mpg ~ ., data = auto, family = "inverse.gaussian")

# get fitted values at "lambda.1se"
fit.1se <- predict(mod, newdata = auto, type = "response")

# get fitted values at "lambda.min"
fit.min <- predict(mod, newdata = auto, type = "response", s = "lambda.min")

# compare mean absolute error for two solutions
mean(abs(auto$mpg - fit.1se))
mean(abs(auto$mpg - fit.min))
}
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory (show via RShowDoc("KEYWORDS")):
\keyword{ regression }
% Use only one keyword per line.
% For non-standard keywords, use \concept instead of \keyword:
% \concept{ ~cpt1 }
% \concept{ ~cpt2 }
% Use only one concept per line.

\name{internals-grpnet}

\alias{grpnet_binomial}
\alias{grpnet_binomial_dev}
\alias{grpnet_gamma}
\alias{grpnet_gamma_dev}
\alias{grpnet_gaussian}
\alias{grpnet_invgaus}
\alias{grpnet_invgaus_dev}
\alias{grpnet_logit}
\alias{grpnet_logit_dev}
\alias{grpnet_maxeigval}
\alias{grpnet_multigaus}
\alias{grpnet_multinom}
\alias{grpnet_multinom_dev}
\alias{grpnet_negbin}
\alias{grpnet_negbin_dev}
\alias{grpnet_ordinal}
\alias{grpnet_ordinal_int}
\alias{grpnet_penalty}
\alias{grpnet_poisson}
\alias{grpnet_poisson_dev}
\alias{grpnet_svm1}
\alias{grpnet_svm1_dev}
\alias{grpnet_svm2}
\alias{grpnet_svm2_dev}

\alias{R_grpnet_binomial}
\alias{R_grpnet_binomial_dev}
\alias{R_grpnet_gamma}
\alias{R_grpnet_gamma_dev}
\alias{R_grpnet_gaussian}
\alias{R_grpnet_invgaus}
\alias{R_grpnet_invgaus_dev}
\alias{R_grpnet_logit}
\alias{R_grpnet_logit_dev}
\alias{R_grpnet_maxeigval}
\alias{R_grpnet_multigaus}
\alias{R_grpnet_multinom}
\alias{R_grpnet_multinom_dev}
\alias{R_grpnet_negbin}
\alias{R_grpnet_negbin_dev}
\alias{R_grpnet_penalty}
\alias{R_grpnet_ordinal}
\alias{R_grpnet_ordinal_int}
\alias{R_grpnet_poisson}
\alias{R_grpnet_poisson_dev}
\alias{R_grpnet_svm1}
\alias{R_grpnet_svm1_dev}
\alias{R_grpnet_svm2}
\alias{R_grpnet_svm2_dev}


%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Internal 'grpnet' Functions
}
\description{
Internal functions for 'grpnet' package, which are \code{R} translations of \code{Fortran} functions.
}
\usage{
# R translation of grpnet_binomial.f90
R_grpnet_binomial(nobs, nvars, x, y, w, off, ngrps, gsize, pw, alpha, 
                  nlam, lambda, lmr, penid, gamma, eps, maxit,
                  standardize, intercept, ibeta, betas, iters,
                  nzgrps, nzcoef, edfs, devs, nulldev)

# R translation of grpnet_gamma.f90
R_grpnet_gamma(nobs, nvars, x, y, w, off, ngrps, gsize, pw, alpha, 
               nlam, lambda, lmr, penid, gamma, eps, maxit,
               standardize, intercept, ibeta, betas, iters,
               nzgrps, nzcoef, edfs, devs, nulldev)
             
# R translation of grpnet_gaussian.f90
R_grpnet_gaussian(nobs, nvars, x, y, w, off, ngrps, gsize, pw, alpha, 
                  nlam, lambda, lmr, penid, gamma, eps, maxit,
                  standardize, intercept, ibeta, betas, iters,
                  nzgrps, nzcoef, edfs, devs, nulldev)

# R translation of grpnet_invgaus.f90
R_grpnet_invgaus(nobs, nvars, x, y, w, off, ngrps, gsize, pw, alpha, 
                 nlam, lambda, lmr, penid, gamma, eps, maxit,
                 standardize, intercept, ibeta, betas, iters,
                 nzgrps, nzcoef, edfs, devs, nulldev)
                 
# R translation of grpnet_logit.f90
R_grpnet_logit(nobs, nvars, x, y, w, off, ngrps, gsize, pw, alpha, 
               nlam, lambda, lmr, penid, gamma, eps, maxit,
               standardize, intercept, ibeta, betas, iters,
               nzgrps, nzcoef, edfs, devs, nulldev)

# R translation of grpnet_multigaus.f90
R_grpnet_multigaus(nobs, nvars, nresp, x, y, w, off, ngrps, gsize, pw, 
                   alpha, nlam, lambda, lmr, penid, gamma, eps, maxit,
                   standardize, intercept, ibeta, betas, iters,
                   nzgrps, nzcoef, edfs, devs, nulldev)

# R translation of grpnet_multinom.f90
R_grpnet_multinom(nobs, nvars, nresp, x, y, w, off, ngrps, gsize, pw,  
                  alpha, nlam, lambda, lmr, penid, gamma, eps, maxit,
                  standardize, intercept, ibeta, betas, iters,
                  nzgrps, nzcoef, edfs, devs, nulldev)
                
# R translation of grpnet_negbin.f90
R_grpnet_negbin(nobs, nvars, x, y, w, off, ngrps, gsize, pw, alpha, 
                nlam, lambda, lmr, penid, gamma, eps, maxit,
                standardize, intercept, ibeta, betas, iters,
                nzgrps, nzcoef, edfs, devs, nulldev, theta)                

# R translation of grpnet_ordinal.f90
R_grpnet_ordinal(nobs, nvars, nresp, x, y, w, off, ngrps, gsize, pw,
                 alpha, nlam, lambda, lmr, penid, gamma, eps, maxit,
                 standardize, intercept, ibeta, betas, iters,
                 nzgrps, nzcoef, edfs, devs, nulldev)
                 
# R translation of grpnet_ordinal_int.f90
R_grpnet_ordinal_int(nobs, nresp, eta, y, w, yfreq, theta, maxdif)                 

# R translation of grpnet_poisson.f90
R_grpnet_poisson(nobs, nvars, x, y, w, off, ngrps, gsize, pw, alpha, 
                 nlam, lambda, lmr, penid, gamma, eps, maxit,
                 standardize, intercept, ibeta, betas, iters,
                 nzgrps, nzcoef, edfs, devs, nulldev)
                 
# R translation of grpnet_svm1.f90
R_grpnet_svm1(nobs, nvars, x, y, w, off, ngrps, gsize, pw, alpha, 
              nlam, lambda, lmr, penid, gamma, eps, maxit,
              standardize, intercept, ibeta, betas, iters,
              nzgrps, nzcoef, edfs, devs, nulldev, theta)
              
# R translation of grpnet_svm2.f90
R_grpnet_svm2(nobs, nvars, x, y, w, off, ngrps, gsize, pw, alpha, 
              nlam, lambda, lmr, penid, gamma, eps, maxit,
              standardize, intercept, ibeta, betas, iters,
              nzgrps, nzcoef, edfs, devs, nulldev)

# R translations of deviance functions
R_grpnet_binomial_dev(nobs, y, mu, wt, dev)
R_grpnet_gamma_dev(nobs, y, mu, wt, dev)
R_grpnet_invgaus_dev(nobs, y, mu, wt, dev)
R_grpnet_logit_dev(nobs, y, mu, wt, dev)
R_grpnet_multinom_dev(nobs, nresp, y, mu, wt, dev)
R_grpnet_negbin_dev(nobs, y, mu, wt, theta, dev)
R_grpnet_poisson_dev(nobs, y, mu, wt, dev)
R_grpnet_svm1_dev(nobs, y, mu, wt, theta, dev)
R_grpnet_svm2_dev(nobs, y, mu, wt, dev)

# R translations of utility functions
R_grpnet_maxeigval(A, N, MAXEV)
R_grpnet_penalty(znorm, penid, penone, pentwo, gamma, shrink)
}

\details{
  These functions are not intended to be called by the typical user.
}

\keyword{ internal }

% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory (show via RShowDoc("KEYWORDS")):
% \keyword{ ~kwd1 }
% \keyword{ ~kwd2 }
% Use only one keyword per line.
% For non-standard keywords, use \concept instead of \keyword:
% \concept{ ~cpt1 }
% \concept{ ~cpt2 }
% Use only one concept per line.

\name{auto}
\alias{auto}
\docType{data}
\title{
Auto MPG Data Set
}
\description{
Miles per gallon and other characteristics of vehicles from the 1970s-1980s. A version of this dataset was used as the 1983 American Statistical Association Exposition dataset.
}
\usage{data("auto")}
\format{
  A data frame with 392 observations on the following 9 variables.
  \describe{
    \item{\code{mpg}}{miles per gallon (numeric vector)}
    \item{\code{cylinders}}{number of cylinders: 3,4,5,6,8 (ordered factor)}
    \item{\code{displacement}}{engine displacement in cubic inches (numeric vector)}
    \item{\code{horsepower}}{engine horsepower (integer vector)}
    \item{\code{weight}}{vehicle weight in of lbs. (integer vector)}
    \item{\code{acceleration}}{0-60 mph time in sec. (numeric vector)}
    \item{\code{model.year}}{ranging from 1970 to 1982 (integer vector)}
    \item{\code{origin}}{region of origin: American, European, Japanese (factor vector)}
  }
}
\details{
This is a modified version of the "Auto MPG Data Set" on the UCI Machine Learning Repository, which is a modified version of the "cars" dataset on StatLib. 

Compared to the version of the dataset in UCI's MLR, this version of the dataset has removed (i) the 6 rows with missing \code{horsepower} scores, and (ii) the last column giving the name of each vehicle (\code{car.name}).
}
\source{
The dataset was originally collected by Ernesto Ramos and David Donoho.

StatLib---Datasets Archive at Carnegie Mellon University
http://lib.stat.cmu.edu/datasets/cars.data

Machine Learning Repository at University of California Irvine
https://archive.ics.uci.edu/ml/datasets/Auto+MPG
}
\examples{
# load data
data(auto)

# display structure
str(auto)

# display header
head(auto)

# see 'cv.grpnet' for cross-validation examples
?cv.grpnet

# see 'grpnet' for fitting examples
?grpnet
}
\keyword{datasets}

\name{rk.model.matrix}
\alias{rk.model.matrix}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Construct Design Matrices via Reproducing Kernels
}
\description{
Creates a design (or model) matrix using the \code{\link{rk}} function to expand variables via a reproducing kernel basis. 
}
\usage{
rk.model.matrix(object, data = environment(object), ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{object}{
a \code{\link{formula}} or \code{\link{terms}} object describing the fit model
}
  \item{data}{
a data frame containing the variables referenced in \code{object}
}
  \item{...}{
additional arguments passed to the \code{\link{rk}} function, e.g., \code{df}, \code{knots}, \code{m}, etc. Arguments must be passed as a named list, see Examples.
}
}
\details{
Designed to be a more flexible alternative to the \code{\link{model.matrix}} function. The \code{\link{rk}} function is used to construct a marginal basis for each variable that appears in the input \code{object}. Tensor product interactions are formed by taking a \code{\link{row.kronecker}} product of marginal basis matrices. Interactions of any order are supported using standard formulaic conventions, see Note.
}
\value{
The design matrix corresponding to the input formula and data, which has the following attributes:

\item{assign}{an integer vector with an entry for each column in the matrix giving the term in the formula which gave rise to the column}
\item{term.labels}{a character vector containing the labels for each of the terms in the model}
\item{knots}{a named list giving the knots used for each variable in the formula}
\item{m}{a named list giving the penalty order used for each variable in the formula}
\item{periodic}{a named list giving the periodicity used for each variable in the formula}
\item{xlev}{a named list giving the factor levels used for each variable in the formula}
}
\references{
Helwig, N. E. (2017). Regression with ordered predictors via ordinal smoothing splines. Frontiers in Applied Mathematics and Statistics, 3(15), 1-13. \doi{10.3389/fams.2017.00015}

Helwig, N. E. (2021). Spectrally sparse nonparametric regression via elastic net regularized smoothers. Journal of Computational and Graphical Statistics, 30(1), 182-191. \doi{10.1080/10618600.2020.1806855}

Helwig, N. E. (2025). Versatile descent algorithms for group regularization and variable selection in generalized linear models. \emph{Journal of Computational and Graphical Statistics, 34}(1), 239-252. \doi{10.1080/10618600.2024.2362232}
}
\author{
Nathaniel E. Helwig <helwig@umn.edu>
}
\note{
For formulas of the form \code{y ~ x + z}, the constructed model matrix has the form \code{cbind(rk(x), rk(z))}, which simply concatenates the two marginal basis matrices. For formulas of the form \code{y ~ x : z}, the constructed model matrix has the form \code{row.kronecker(rk(x), rk(z))}, where \code{\link{row.kronecker}} denotes the row-wise kronecker product. The formula \code{y ~ x * z} is a shorthand for \code{y ~ x + z + x : z}, which concatenates the two previous results. Unless it is suppressed (using \code{0+}), the first column of the basis will be a column of ones named \code{(Intercept)}.
}

\seealso{
See \code{\link{rk}} for details on the reproducing kernel basis
}
\examples{
# load auto data
data(auto)

# additive effects
x <- rk.model.matrix(mpg ~ ., data = auto)
dim(x)                      # check dimensions
attr(x, "assign")           # check group assignments
attr(x, "term.labels")      # check term labels

# two-way interactions
x <- rk.model.matrix(mpg ~ . * ., data = auto)
dim(x)                      # check dimensions
attr(x, "assign")           # check group assignments
attr(x, "term.labels")      # check term labels

# specify df for horsepower, weight, and acceleration
# note: default df = 5 is used for displacement and model.year
df <- list(horsepower = 6, weight = 7, acceleration = 8)
x <- rk.model.matrix(mpg ~ ., data = auto, df = df)
sapply(attr(x, "knots"), length)   # check df

# specify knots for model.year
# note: default knots are selected for other variables
knots <- list(model.year = c(1970, 1974, 1978, 1982))
x <- rk.model.matrix(mpg ~ ., data = auto, knots = knots)
sapply(attr(x, "knots"), length)   # check df

}
\keyword{ regression }% use one of  RShowDoc("KEYWORDS")
\keyword{ smooth }% __ONLY ONE__ keyword per line

\name{StartupMessage}
\alias{StartupMessage}
\alias{grpnetStartupMessage}

\title{Startup Message for grpnet}

\description{
Prints the startup message when grpnet is loaded. Not intended to be called by the user.
}

\details{
The `grpnet' ascii start-up message was created using the taag software.
}

\references{
https://patorjk.com/software/taag/
}

\name{row.kronecker}
\alias{row.kronecker}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Row-Wise Kronecker Product
}
\description{
Calculates the row-wise Kronecker product between two matrices with the same number of rows.
}
\usage{
row.kronecker(X, Y)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{X}{
matrix of dimension \eqn{n \times p}
}
  \item{Y}{
matrix of dimension \eqn{n \times q}
}
}
\details{
Given \code{X} of dimension \code{c(n, p)} and \code{Y} of dimension \code{c(n, q)}, this function returns

\code{cbind(x[,1] * Y, x[,2] * Y, ..., x[,p] * Y)}

which is a matrix of dimension \code{c(n, p*q)}
}
\value{
Matrix of dimension \eqn{n \times pq} where each row contains the Kronecker product between the corresponding rows of \code{X} and \code{Y}.
}
\author{
Nathaniel E. Helwig <helwig@umn.edu>
}

%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
Used by the \code{\link{rk.model.matrix}} to construct basis functions for interaction terms

See \code{\link{kronecker}} for the regular kronecker product
}
\examples{
X <- matrix(c(1, 1, 2, 2), nrow = 2, ncol = 2)
Y <- matrix(1:6, nrow = 2, ncol = 3)
row.kronecker(X, Y)
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory (show via RShowDoc("KEYWORDS")):
\keyword{ array }
\keyword{ algebra }
% Use only one keyword per line.
% For non-standard keywords, use \concept instead of \keyword:
% \concept{ ~cpt1 }
% \concept{ ~cpt2 }
% Use only one concept per line.

\name{visualize.shrink}
\alias{visualize.shrink}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Plots grpnet Shrinkage Operator or its Estimator
}
\description{
Makes a plot or returns a data frame containing the group elastic net shrinkage operator (or its estimator) evaluated at a sequence of input values.
}
\usage{
visualize.shrink(x = seq(-5, 5, length.out = 1001), 
                 penalty = c("LASSO", "MCP", "SCAD"), 
                 alpha = 1, 
                 lambda = 1, 
                 gamma = 4, 
                 fitted = FALSE,
                 plot = TRUE,
                 subtitle = TRUE,
                 legend = TRUE,
                 location = "top",
                 ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{
  sequence of values at which to evaluate the penalty.
  }
  \item{penalty}{
  which penalty or penalties should be plotted?  
  }
  \item{alpha}{
  elastic net tuning parameter (between 0 and 1).  
  }
  \item{lambda}{
  overall tuning parameter (non-negative).
  }
  \item{gamma}{
  additional hyperparameter for MCP (>1) or SCAD (>2).
  }
  \item{fitted}{
  if \code{FALSE} (default), then the shrinkage operator is plotted; otherwise the shrunken estimator is plotted.
  }
  \item{plot}{
  if \code{TRUE} (default), then the result is plotted; otherwise the result is returned as a data frame.
  }
  \item{subtitle}{
  if \code{TRUE} (default), then the hyperparameter values are displayed in the subtitle. 
  }
  \item{legend}{
  if \code{TRUE} (default), then a legend is included to distinguish the different \code{penalty} types.
  }
  \item{location}{
  the legend's location; ignored if \code{legend = FALSE}.
  }
  \item{...}{
  additional arguments passed to \code{plot} function, e.g., \code{xlim}, \code{ylim}, etc.
  }
}
\details{
The updates for the group elastic net estimator have the form
\deqn{\boldsymbol\beta_{\alpha, \lambda}^{(t+1)} = S_{\lambda_1, \lambda_2}(\|\mathbf{b}_{\alpha, \lambda}^{(t+1)}\|) \mathbf{b}_{\alpha, \lambda}^{(t+1)}}
where \eqn{S_{\lambda_1, \lambda_2}(\cdot)} is a shrinkage and selection operator, and 
\deqn{\mathbf{b}_{\alpha, \lambda}^{(t+1)} = \boldsymbol\beta_{\alpha, \lambda}^{(t)} + (\delta_{(t)} \epsilon)^{-1} \mathbf{g}^{(t)}}
is the unpenalized update with \eqn{\mathbf{g}^{(t)}} denoting the current gradient.

Note that \eqn{\lambda_1 = \lambda \alpha} is the L1 tuning parameter, \eqn{\lambda_2 = \lambda (1-\alpha)} is the L2 tuning parameter, \eqn{\delta_{(t)}} is an upper-bound on the weights appearing in the Fisher information matrix, and \eqn{\epsilon} is the largest eigenvalue of the Gramm matrix \eqn{n^{-1} \mathbf{X}^\top \mathbf{X}}. 
}
\value{
If \code{plot = TRUE}, then produces a plot.

If \code{plot = FALSE}, then returns a data frame.
}
\references{
Fan, J., & Li, R. (2001). Variable selection via nonconcave penalized likelihood and its oracle properties. \emph{Journal of the American Statistical Association, 96}(456), 1348-1360. \doi{10.1198/016214501753382273}

Helwig, N. E. (2025). Versatile descent algorithms for group regularization and variable selection in generalized linear models. \emph{Journal of Computational and Graphical Statistics, 34}(1), 239-252. \doi{10.1080/10618600.2024.2362232}

Tibshirani, R. (1996). Regression and shrinkage via the Lasso. \emph{Journal of the Royal Statistical Society, Series B, 58}, 267-288. \doi{10.1111/j.2517-6161.1996.tb02080.x}

Zhang, C. H. (2010). Nearly unbiased variable selection under minimax concave penalty. \emph{The Annals of Statistics, 38}(2), 894-942. \doi{10.1214/09-AOS729}
}
\author{
Nathaniel E. Helwig <helwig@umn.edu>
}

%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link{visualize.loss}} for plotting loss functions

\code{\link{visualize.penalty}} for plotting penalty function
}
\examples{
# plot shrinkage operator
visualize.shrink()

# plot shrunken estimator
visualize.shrink(fitted = TRUE)
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory (show via RShowDoc("KEYWORDS")):
\keyword{ dplot }
\keyword{ hplot }
% Use only one keyword per line.
% For non-standard keywords, use \concept instead of \keyword:
% \concept{ ~cpt1 }
% \concept{ ~cpt2 }
% Use only one concept per line.

\name{visualize.penalty}
\alias{visualize.penalty}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Plots grpnet Penalty Function or its Derivative
}
\description{
Makes a plot or returns a data frame containing the group elastic net penalty (or its derivative) evaluated at a sequence of input values.
}
\usage{
visualize.penalty(x = seq(-5, 5, length.out = 1001), 
                  penalty = c("LASSO", "MCP", "SCAD"), 
                  alpha = 1, 
                  lambda = 1, 
                  gamma = 4, 
                  derivative = FALSE,
                  plot = TRUE,
                  subtitle = TRUE,
                  legend = TRUE,
                  location = ifelse(derivative, "bottom", "top"),
                  ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{
  sequence of values at which to evaluate the penalty.
  }
  \item{penalty}{
  which penalty or penalties should be plotted?  
  }
  \item{alpha}{
  elastic net tuning parameter (between 0 and 1).  
  }
  \item{lambda}{
  overall tuning parameter (non-negative).
  }
  \item{gamma}{
  additional hyperparameter for MCP (>1) or SCAD (>2).
  }
  \item{derivative}{
  if \code{FALSE} (default), then the penalty is plotted; otherwise the derivative of the penalty is plotted.
  }
  \item{plot}{
  if \code{TRUE} (default), then the result is plotted; otherwise the result is returned as a data frame.
  }
  \item{subtitle}{
  if \code{TRUE} (default), then the hyperparameter values are displayed in the subtitle. 
  }
  \item{legend}{
  if \code{TRUE} (default), then a legend is included to distinguish the different \code{penalty} types.
  }
  \item{location}{
  the legend's location; ignored if \code{legend = FALSE}.
  }
  \item{...}{
  additional arguments passed to \code{plot} function, e.g., \code{xlim}, \code{ylim}, etc.
  }
}
  
  
\details{
The group elastic net penalty is defined as
\deqn{P_{\alpha, \lambda}(\boldsymbol\beta) = Q_{\lambda_1}(\|\boldsymbol\beta\|) + \frac{\lambda_2}{2} \|\boldsymbol\beta\|^2}
where \eqn{Q_\lambda()} denotes the L1 penalty (LASSO, MCP, or SCAD), \eqn{\| \boldsymbol\beta \| = (\boldsymbol\beta^\top \boldsymbol\beta)^{1/2}} denotes the Euclidean norm, \eqn{\lambda_1 = \lambda \alpha} is the L1 tuning parameter,  and \eqn{\lambda_2 = \lambda (1-\alpha)} is the L2 tuning parameter. Note that \eqn{\lambda} and \eqn{\alpha} denote the \code{lambda} and \code{alpha} arguments.
}
\value{
If \code{plot = TRUE}, then produces a plot.

If \code{plot = FALSE}, then returns a data frame.
}
\references{
Fan, J., & Li, R. (2001). Variable selection via nonconcave penalized likelihood and its oracle properties. \emph{Journal of the American Statistical Association, 96}(456), 1348-1360. \doi{10.1198/016214501753382273}

Helwig, N. E. (2025). Versatile descent algorithms for group regularization and variable selection in generalized linear models. \emph{Journal of Computational and Graphical Statistics, 34}(1), 239-252. \doi{10.1080/10618600.2024.2362232}

Tibshirani, R. (1996). Regression and shrinkage via the Lasso. \emph{Journal of the Royal Statistical Society, Series B, 58}, 267-288. \doi{10.1111/j.2517-6161.1996.tb02080.x}

Zhang, C. H. (2010). Nearly unbiased variable selection under minimax concave penalty. \emph{The Annals of Statistics, 38}(2), 894-942. \doi{10.1214/09-AOS729}
}
\author{
Nathaniel E. Helwig <helwig@umn.edu>
}

%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link{visualize.loss}} for plotting loss functions

\code{\link{visualize.shrink}} for plotting shrinkage operator
}
\examples{
# plot penalty functions
visualize.penalty()

# plot penalty derivatives
visualize.penalty(derivative = TRUE)
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory (show via RShowDoc("KEYWORDS")):
\keyword{ dplot }
\keyword{ hplot }
% Use only one keyword per line.
% For non-standard keywords, use \concept instead of \keyword:
% \concept{ ~cpt1 }
% \concept{ ~cpt2 }
% Use only one concept per line.

\name{coef}
\alias{coef.cv.grpnet}
\alias{coef.grpnet}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Extract Coefficients for cv.grpnet and grpnet Fits
}
\description{
Obtain coefficients from a cross-validated group elastic net regularized GLM (cv.grpnet) or a group elastic net regularized GLM (grpnet) object.
}
\usage{
\method{coef}{cv.grpnet}(object, 
     s = c("lambda.1se", "lambda.min"),
     ...)
     
\method{coef}{grpnet}(object, 
     s = NULL,
     ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{object}{
Object of class "cv.grpnet" or "grpnet"
}
\item{s}{
Lambda value(s) at which predictions should be obtained. For "cv.grpnet" objects, default uses the 1se solution. For "grpnet" objects, default uses \code{s = object$lambda}. Interpolation is used for \code{s} values that are not included in \code{object$lambda}.  
}
\item{...}{
Additional arguments (ignored)
}
}
\details{
\emph{coef.cv.grpnet}: \cr
Returns the coefficients that are used by the \code{\link{predict.cv.grpnet}} function to form predictions from a fit \code{\link{cv.grpnet}} object. 

\emph{coef.grpnet}: \cr
Returns the coefficients that are used by the \code{\link{predict.grpnet}} function to form predictions from a fit \code{\link{grpnet}} object. 
}
\value{
For multigaussian and multinomial response variables, returns a list of length \code{length(object$ylev)}, where the \code{j}-th element is a matrix of dimension \code{c(ncoef, length(s))} giving the coefficients for \code{object$ylev[j]}.

For other response variables, returns a matrix of dimension \code{c(ncoef, length(s))}, where the \code{i}-th column gives the coefficients for \code{s[i]}.
}
\references{
Friedman, J.,  Hastie, T., & Tibshirani, R. (2010). Regularization paths for generalized linear models via coordinate descent. \emph{Journal of Statistical Software, 33}(1), 1-22. \doi{10.18637/jss.v033.i01}

Helwig, N. E. (2025). Versatile descent algorithms for group regularization and variable selection in generalized linear models. \emph{Journal of Computational and Graphical Statistics, 34}(1), 239-252. \doi{10.1080/10618600.2024.2362232}
}
\author{
Nathaniel E. Helwig <helwig@umn.edu>
}
\note{
The syntax of these functions closely mimics that of the \code{coef.cv.glmnet} and \code{coef.glmnet} functions in the \bold{glmnet} package (Friedman, Hastie, & Tibshirani, 2010).
}

%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link{print.coef.grpnet}} for printing \code{\link{coef.grpnet}} objects

\code{\link{predict.cv.grpnet}} for predicting from \code{\link{cv.grpnet}} objects

\code{\link{predict.grpnet}} for predicting from \code{\link{grpnet}} objects
}
\examples{
######***######   grpnet   ######***######

# load data
data(auto)

# fit model (formula method, response = mpg)
mod <- grpnet(mpg ~ ., data = auto)

# extract coefs for regularization path (output = 12 x 100 matrix)
coef(mod)

# extract coefs at 3 particular points (output = 12 x 3 matrix)
coef(mod, s = c(1.5, 1, 0.5))


######***######   cv.grpnet   ######***######

# load data
data(auto)

# 5-fold cv (formula method, response = mpg)
set.seed(1)
mod <- cv.grpnet(mpg ~ ., data = auto, nfolds = 5, alpha = 1)

# extract coefs for "min" solution (output = 12 x 1 matrix)
coef(mod)

# extract coefs for "1se" solution (output = 12 x 1 matrix)
coef(mod, s = "lambda.1se")

# extract coefs at 3 particular points (output = 12 x 3 matrix)
coef(mod, s = c(1.5, 1, 0.5))
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory (show via RShowDoc("KEYWORDS")):
\keyword{ regression }
% \keyword{ ~kwd2 }
% Use only one keyword per line.
% For non-standard keywords, use \concept instead of \keyword:
% \concept{ ~cpt1 }
% \concept{ ~cpt2 }
% Use only one concept per line.

\name{cv.grpnet}
\alias{cv.grpnet}
\alias{cv.grpnet.default}
\alias{cv.grpnet.formula}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Cross-Validation for grpnet
}
\description{
Implements k-fold cross-validation for \code{\link{grpnet}} to find the regularization parameters that minimize the prediction error (deviance, mean squared error, mean absolute error, or misclassification rate).
}
\usage{
cv.grpnet(x, \dots)

\method{cv.grpnet}{default}(x, 
          y, 
          group,
          weights = NULL,
          offset = NULL,
          alpha = c(0.01, 0.25, 0.5, 0.75, 1),
          gamma = c(3, 4, 5),
          type.measure = NULL,
          nfolds = 10, 
          foldid = NULL,
          same.lambda = FALSE,
          parallel = FALSE, 
          cluster = NULL, 
          verbose = interactive(), 
          adaptive = FALSE,
          power = 1,
          ...)
           
\method{cv.grpnet}{formula}(formula,
          data, 
          use.rk = TRUE,
          weights = NULL,
          offset = NULL,
          alpha = c(0.01, 0.25, 0.5, 0.75, 1),
          gamma = c(3, 4, 5),
          type.measure = NULL,
          nfolds = 10, 
          foldid = NULL, 
          same.lambda = FALSE,
          parallel = FALSE, 
          cluster = NULL, 
          verbose = interactive(), 
          adaptive = FALSE,
          power = 1,
          ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{
Model (design) matrix of dimension \code{nobs} by \code{nvars} (\eqn{n \times p}).
}
  \item{y}{
Response vector of length \eqn{n} or matrix of dimension \eqn{n \times m}. Note that matrix inputs are (i) required for multigaussian family, (ii) allowed for binomial and multinomial families (see "Binomial and multinomial" section in \code{\link{grpnet}}), and (iii) not permitted for other families.
}
  \item{group}{
Group label vector (factor, character, or integer) of length \eqn{p}. Predictors with the same label are grouped together for regularization.
}
  \item{formula}{
Model formula: a symbolic description of the model to be fitted. Uses the same syntax as \code{\link{lm}} and \code{\link{glm}}. 
}
  \item{data}{
Optional data frame containing the variables referenced in \code{formula}.
}
  \item{use.rk}{
If \code{TRUE} (default), the \code{\link{rk.model.matrix}} function is used to build the model matrix. Otherwise, the \code{\link{model.matrix}} function is used to build the model matrix. Additional arguments to the \code{\link{rk.model.matrix}} function can be passed via the \code{...} argument.
}
  \item{weights}{
Optional vector of length \eqn{n} with non-negative weights to use for weighted (penalized) likelihood estimation. Defaults to a vector of ones.
}
  \item{offset}{
Optional vector of length \eqn{n} with an a priori known term to be included in the model's linear predictor. Defaults to a vector of zeros.  
}
  \item{alpha}{
Scalar or vector specifying the elastic net tuning parameter \eqn{\alpha}. If \code{alpha} is a vector (default), then (a) the same \code{foldid} is used to compute the cross-validation error for each \eqn{\alpha}, and (b) the solution for the optimal \eqn{\alpha} is returned. 
}
\item{gamma}{
Scalar or vector specifying the penalty hyperparameter \eqn{\gamma} for MCP or SCAD. If \code{gamma} is a vector (default), then (a) the same \code{foldid} is used to compute the cross-validation error for each \eqn{\gamma}, and (b) the solution for the optimal \eqn{\gamma} is returned. 
}   
  \item{type.measure}{
Loss function for cross-validation. Options include: \code{"deviance"} for model deviance, \code{"mse"} for mean squared error, \code{"mae"} for mean absolute error, or \code{"class"} for classification error. Note that \code{"class"} is only available for binomial and multinomial families. The default is classification error (for binomial and multinomial) or mean absolute error (others).
}
  \item{nfolds}{
Number of folds for cross-validation. 
}
  \item{foldid}{
Optional vector of length \eqn{n} giving the fold identification for each observation. Must be coercible into a factor. After coersion, the \code{nfolds} argument is defined as \code{nfolds = nlevels(foldid)}.
}
  \item{same.lambda}{
Logical specfying if the same \eqn{\lambda} sequence should be used for fitting the model to each fold's data. If \code{FALSE} (default), the \eqn{\lambda} sequence is determined separately holding out each fold, and the \eqn{\lambda} sequence from the full model is used to align the predictions. If \code{TRUE}, the \eqn{\lambda} sequence from the full model is used to fit the model for each fold. The default often provides better (i.e., more stable) computational performance.
}
  \item{parallel}{
Logical specifying if sequential computing (default) or parallel computing should be used. If \code{TRUE}, the fitting for each fold is parallelized.
}
  \item{cluster}{
Optional cluster to use for parallel computing. If \code{parallel = TRUE} and \code{cluster = NULL}, then the cluster is defined \code{cluster = makeCluster(2L)}, which uses two cores. Recommended usage: \code{cluster = makeCluster(detectCores())}
}
  \item{verbose}{
Logical indicating if the fitting progress should be printed. Defaults to \code{TRUE} in interactive sessions and \code{FALSE} otherwise. 
}
\item{adaptive}{
Logical indicating if the adaptive group elastic net should be used (see Note).
}
\item{power}{
If \code{adaptive = TRUE}, then the adaptive penalty weights are defined by dividing the original penalty weights by \code{tapply(coef, group, norm, type = "F")^power}.
}
  \item{...}{
Optional additional arguments for \code{\link{grpnet}} (e.g., \code{standardize}, \code{penalty.factor}, etc.)
}
}
\details{
This function calls the \code{\link{grpnet}} function \code{nfolds+1} times: once on the full dataset to obtain the \code{lambda} sequence, and once holding out each fold's data to evaluate the prediction error. The syntax of (the default S3 method for) this function closely mimics that of the \code{cv.glmnet} function in the \bold{glmnet} package (Friedman, Hastie, & Tibshirani, 2010). 

Let \eqn{\mathbf{D}_u = \{\mathbf{y}_u, \mathbf{X}_u\}} denote the \eqn{u}-th fold's data, let \eqn{\mathbf{D}_{[u]} = \{\mathbf{y}_{[u]}, \mathbf{X}_{[u]}\}} denote the full dataset excluding the \eqn{u}-th fold's data, and let \eqn{\boldsymbol\beta_{\lambda [u]}} denote the coefficient estimates obtained from fitting the model to \eqn{\mathbf{D}_{[u]}} using the regularization parameter \eqn{\lambda}. 

The cross-validation error for the \eqn{u}-th fold is defined as
\deqn{E_u(\lambda) = C(\boldsymbol\beta_{\lambda [u]} , \mathbf{D}_u)}
where \eqn{C(\cdot , \cdot)} denotes the cross-validation loss function that is specified by \code{type.measure}. For example, the \code{"mse"} loss function is defined as
\deqn{C(\boldsymbol\beta_{\lambda [u]} , \mathbf{D}_u) = \| \mathbf{y}_u - \mathbf{X}_u \boldsymbol\beta_{\lambda [u]} \|^2}
where \eqn{\| \cdot \|} denotes the L2 norm.

The mean cross-validation error \code{cvm} is defined as
\deqn{\bar{E}(\lambda) = \frac{1}{v} \sum_{u = 1}^v E_u(\lambda) }
where \eqn{v} is the total number of folds. The standard error \code{cvsd} is defined as
\deqn{S(\lambda) = \sqrt{ \frac{1}{v (v - 1)} \sum_{u=1}^v (E_u(\lambda) - \bar{E}(\lambda))^2 } }
which is the classic definition of the standard error of the mean. 
}
\value{
\item{lambda}{regularization parameter sequence for the full data}
\item{cvm}{mean cross-validation error for each \code{lambda}}
\item{cvsd}{estimated standard error of \code{cvm}}
\item{cvup}{upper curve: \code{cvm + cvsd}}
\item{cvlo}{lower curve: \code{cvm - cvsd}}
\item{nzero}{number of non-zero groups for each \code{lambda}}
\item{grpnet.fit}{fitted grpnet object for the full data}
\item{lambda.min}{value of \code{lambda} that minimizes \code{cvm}}
\item{lambda.1se}{largest \code{lambda} such that \code{cvm} is within one \code{cvsd} from the minimum (see Note)}
\item{index}{two-element vector giving the indices of \code{lambda.min} and \code{lambda.1se} in the \code{lambda} vector, i.e., \code{c(minid, se1id)} as defined in the Note}
\item{type.measure}{loss function for cross-validation (used for plot label)}
\item{call}{matched call}
\item{time}{runtime in seconds to perform k-fold CV tuning}
\item{tune}{data frame containing the tuning results, i.e., min(cvm) for each combination of \code{alpha} and/or \code{gamma}}
}
\references{
Friedman, J.,  Hastie, T., & Tibshirani, R. (2010). Regularization paths for generalized linear models via coordinate descent. \emph{Journal of Statistical Software, 33}(1), 1-22. \doi{10.18637/jss.v033.i01}

Helwig, N. E. (2025). Versatile descent algorithms for group regularization and variable selection in generalized linear models. \emph{Journal of Computational and Graphical Statistics, 34}(1), 239-252. \doi{10.1080/10618600.2024.2362232}
}
\author{
Nathaniel E. Helwig <helwig@umn.edu>
}
\note{
When \code{adaptive = TRUE}, the adaptive group elastic net is used: \cr
(1) an initial fit with \code{alpha = 0} estimates the \code{penalty.factor} \cr
(2) a second fit using estimated \code{penalty.factor} is returned \cr

\code{lambda.1se} is defined as follows: \cr
\code{minid <- which.min(cvm)} \cr
\code{min1se <- cvm[minid] + cvsd[minid]} \cr
\code{se1id <- which(cvm <= min1se)[1]} \cr
\code{lambda.1se <- lambda[se1id]}
}

%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link{summary.cv.grpnet}} for calculating fitted values and importances

\code{\link{plot.cv.grpnet}} for plotting the cross-validation error curve

\code{\link{predict.cv.grpnet}} for predicting from \code{\link{cv.grpnet}} objects

\code{\link{grpnet}} for fitting group elastic net regularization paths
}
\examples{
\donttest{
######***######   family = "gaussian"   ######***######

# load data
data(auto)

# 10-fold cv (formula method, response = mpg)
set.seed(1)
mod <- cv.grpnet(mpg ~ ., data = auto)

# print min and 1se solution info
mod

# plot cv error curve
plot(mod)

# summarize model
summary(mod)



######***######   family = "multigaussian"   ######***######

# load data
data(auto)

# 10-fold cv (formula method, response = (mpg, displacement))
y <- as.matrix(auto[,c(1,3)])
set.seed(1)
mod <- cv.grpnet(y ~ ., data = auto[,-c(1,3)], family = "multigaussian",
                 standardize.response = TRUE)

# print min and 1se solution info
mod

# plot cv error curve
plot(mod)

# summarize model
summary(mod)



######***######   family = "svm1"   ######***######

# load data
data(auto)

# redefine origin (Domestic vs Foreign)
auto$origin <- ifelse(auto$origin == "American", "Domestic", "Foreign")

# 10-fold cv (default method, response = origin with 2 levels)
set.seed(1)
mod <- cv.grpnet(origin ~ ., data = auto, family = "svm1")

# print min and 1se solution info
mod

# plot cv error curve
plot(mod)

# summarize model
summary(mod)



######***######   family = "svm2"   ######***######

# load data
data(auto)

# redefine origin (Domestic vs Foreign)
auto$origin <- ifelse(auto$origin == "American", "Domestic", "Foreign")

# 10-fold cv (default method, response = origin with 2 levels)
set.seed(1)
mod <- cv.grpnet(origin ~ ., data = auto, family = "svm2")

# print min and 1se solution info
mod

# plot cv error curve
plot(mod)

# summarize model
summary(mod)



######***######   family = "logit"   ######***######

# load data
data(auto)

# redefine origin (Domestic vs Foreign)
auto$origin <- ifelse(auto$origin == "American", "Domestic", "Foreign")

# 10-fold cv (default method, response = origin with 2 levels)
set.seed(1)
mod <- cv.grpnet(origin ~ ., data = auto, family = "logit")

# print min and 1se solution info
mod

# plot cv error curve
plot(mod)

# summarize model
summary(mod)



######***######   family = "binomial"   ######***######

# load data
data(auto)

# redefine origin (Domestic vs Foreign)
auto$origin <- ifelse(auto$origin == "American", "Domestic", "Foreign")

# 10-fold cv (default method, response = origin with 2 levels)
set.seed(1)
mod <- cv.grpnet(origin ~ ., data = auto, family = "binomial")

# print min and 1se solution info
mod

# plot cv error curve
plot(mod)

# summarize model
summary(mod)



######***######   family = "multinomial"   ######***######

# load data
data(auto)

# 10-fold cv (formula method, response = origin with 3 levels)
set.seed(1)
mod <- cv.grpnet(origin ~ ., data = auto, family = "multinomial")

# print min and 1se solution info
mod

# plot cv error curve
plot(mod)

# summarize model
summary(mod)



######***######   family = "ordinal"   ######***######

# load data
data(auto)

# 10-fold cv (formula method, response = cylinders with 5 levels)
set.seed(1)
mod <- cv.grpnet(cylinders ~ ., data = auto, family = "ordinal")

# print min and 1se solution info
mod

# plot cv error curve
plot(mod)

# summarize model
summary(mod)



######***######   family = "poisson"   ######***######

# load data
data(auto)

# 10-fold cv (formula method, response = horsepower)
set.seed(1)
mod <- cv.grpnet(horsepower ~ ., data = auto, family = "poisson")

# print min and 1se solution info
mod

# plot cv error curve
plot(mod)

# summarize model
summary(mod)



######***######   family = "negative.binomial"   ######***######

# load data
data(auto)

# 10-fold cv (formula method, response = horsepower)
set.seed(1)
mod <- cv.grpnet(horsepower ~ ., data = auto, family = "negative.binomial")

# print min and 1se solution info
mod

# plot cv error curve
plot(mod)

# summarize model
summary(mod)



######***######   family = "Gamma"   ######***######

# load data
data(auto)

# 10-fold cv (formula method, response = origin)
set.seed(1)
mod <- cv.grpnet(mpg ~ ., data = auto, family = "Gamma")

# print min and 1se solution info
mod

# plot cv error curve
plot(mod)

# summarize model
summary(mod)



######***######   family = "inverse.gaussian"   ######***######

# load data
data(auto)

# 10-fold cv (formula method, response = origin)
set.seed(1)
mod <- cv.grpnet(mpg ~ ., data = auto, family = "inverse.gaussian")

# print min and 1se solution info
mod

# plot cv error curve
plot(mod)

# summarize model
summary(mod)
}
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory (show via RShowDoc("KEYWORDS")):
\keyword{ regression }
\keyword{ smooth }
% Use only one keyword per line.
% For non-standard keywords, use \concept instead of \keyword:
% \concept{ ~cpt1 }
% \concept{ ~cpt2 }
% Use only one concept per line.

\name{summary.cv.grpnet}
\alias{summary.cv.grpnet}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Summarize cv.grpnet Fits
}
\description{
Obtain fitted values, variable activations, variable importances from a fit group elastic net (grpnet) object.
}
\usage{
\method{summary}{cv.grpnet}(object, ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{object}{
Object of class "cv.grpnet"
}
  \item{...}{
Additional arguments passed to \code{\link{predict.cv.grpnet}} function.
}
}
\details{
Passes the training data to the \code{\link{predict.cv.grpnet}} function four times: two predictions types (\code{fit} and \code{imp}) by two lambda values (\code{lambda.1se} and \code{lambda.min}).
}
\value{
\item{family}{Name of exponential family/loss function.}
\item{penalty}{Name of L1 group penalty.}
\item{nobs}{Number of observations.}
\item{ngroups}{Number of coefficient groups.}
\item{lambda}{Vector of lambda values (1se and min).}
\item{dev.ratio}{Vector of proportions of null deviance explained (1se and min).}
\item{fit}{Fitted values obtained from \code{\link{predict.cv.grpnet}}.}
\item{act}{Variable activations obtained from \code{\link{predict.cv.grpnet}}.}
\item{imp}{Variable importances obtained from \code{\link{predict.cv.grpnet}}.}
}
\references{
Helwig, N. E. (2025). Versatile descent algorithms for group regularization and variable selection in generalized linear models. \emph{Journal of Computational and Graphical Statistics, 34}(1), 239-252. \doi{10.1080/10618600.2024.2362232}
}
\author{
Nathaniel E. Helwig <helwig@umn.edu>
}
\note{
The variables activations are defined as:  \code{act <- abs(imp) > 0.0}

For most response families, the returned components \code{fit}, \code{act}, and \code{imp} are matrices of dimension \code{K} by \code{2}.

For families with multivariate responses (i.e., \code{multigaussian} and \code{multinomial}), these components are arrays of dimension \code{K} by \code{L} by \code{2} where \code{L} is the dimension of the response.
}

%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link{cv.grpnet}} for k-fold cross-validation of \code{lambda}
}
\examples{
# see 'cv.grpnet' for summary examples
?cv.grpnet
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory (show via RShowDoc("KEYWORDS")):
% \keyword{ ~kwd1 }
% \keyword{ ~kwd2 }
% Use only one keyword per line.
% For non-standard keywords, use \concept instead of \keyword:
% \concept{ ~cpt1 }
% \concept{ ~cpt2 }
% Use only one concept per line.

\name{grpnet}
\alias{grpnet}
\alias{grpnet.default}
\alias{grpnet.formula}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Fit a Group Elastic Net Regularized GLM/GAM
}
\description{
Fits generalized linear/additive models with a group elastic net penalty using an adaptively bounded gradient descent (ABGD) algorithm (Helwig, 2025). Predictor groups can be manually input (default S3 method) or inferred from the model (S3 "formula" method). The regularization path is computed at a data-generated (default) or user-provided sequence of lambda values. }
\usage{
grpnet(x, \dots)

\method{grpnet}{default}(x, 
       y, 
       group, 
       family = c("gaussian", "multigaussian",
                  "svm1", "svm2", "logit",
                  "binomial", "multinomial", "ordinal",
                  "poisson", "negative.binomial", 
                  "Gamma", "inverse.gaussian"),
       weights = NULL, 
       offset = NULL, 
       alpha = 1, 
       nlambda = 100,
       lambda.min.ratio = ifelse(nobs < nvars, 0.05, 0.0001), 
       lambda = NULL, 
       penalty.factor = NULL,
       penalty = c("LASSO", "MCP", "SCAD"),
       gamma = 4,
       theta = 1,
       standardized = !orthogonalized,
       orthogonalized = TRUE,
       intercept = TRUE, 
       thresh = 1e-04, 
       maxit = 1e05,
       proglang = c("Fortran", "R"),
       standardize.response = FALSE,
       keep.data = TRUE,
       ...)
      
\method{grpnet}{formula}(formula,
       data, 
       use.rk = TRUE,
       family = c("gaussian", "multigaussian",
                  "svm1", "svm2", "logit",
                  "binomial", "multinomial", "ordinal",
                  "poisson", "negative.binomial", 
                  "Gamma", "inverse.gaussian"),
       weights = NULL,
       offset = NULL,
       alpha = 1,
       nlambda = 100,
       lambda.min.ratio = ifelse(nobs < nvars, 0.05, 0.0001),
       lambda = NULL,
       penalty.factor = NULL,
       penalty = c("LASSO", "MCP", "SCAD"),
       gamma = 4,
       theta = 1,
       standardized = !orthogonalized,
       orthogonalized = TRUE,
       thresh = 1e-04,
       maxit = 1e05,
       proglang = c("Fortran", "R"),
       standardize.response = FALSE,
       keep.data = TRUE,
       ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{
Model (design) matrix of dimension \code{nobs} by \code{nvars} (\eqn{n \times p}).
}
  \item{y}{
Response vector of length \eqn{n} or matrix of dimension \eqn{n \times m}. Note that matrix inputs are (i) required for multigaussian family, (ii) allowed for binomial and multinomial families (see "Binomial and multinomial" section in \code{\link{grpnet}}), and (iii) not permitted for other families.
}
  \item{group}{
Group label vector (factor, character, or integer) of length \eqn{p}. Predictors with the same label are grouped together for regularization.
}
  \item{formula}{
Model formula: a symbolic description of the model to be fitted. Uses the same syntax as \code{\link{lm}} and \code{\link{glm}}. 
}
  \item{data}{
Optional data frame containing the variables referenced in \code{formula}.
}
  \item{use.rk}{
If \code{TRUE} (default), the \code{\link{rk.model.matrix}} function is used to build the model matrix. Otherwise, the \code{\link{model.matrix}} function is used to build the model matrix. Additional arguments to the \code{\link{rk.model.matrix}} function can be passed via the \code{...} argument.
}
  \item{family}{
Character specifying the assumed distribution for the response variable. Partial matching is allowed. Options include \code{"gaussian"} (real-valued vector), \code{"multigaussian"} (real-valued matrix), \code{"svm1"} (binary response), \code{"svm2"} (binary response), \code{"logit"} (binary response), \code{"binomial"} (binary response), \code{"multinomial"} (multi-class response), \code{"poisson"} (count response), \code{"negative.binomial"} (count response), \code{"Gamma"} (positive real-valued), or \code{"inverse.gaussian"} (positive real-valued). 
}
  \item{weights}{
Optional vector of length \eqn{n} with non-negative weights to use for weighted (penalized) likelihood estimation. Defaults to a vector of ones.
}
  \item{offset}{
Optional vector of length \eqn{n} with an a priori known term to be included in the model's linear predictor. Defaults to a vector of zeros.  
}
  \item{alpha}{
Regularization hyperparameter satisfying \eqn{0 \leq \alpha \leq 1} that gives the balance between the group L1 (lasso) and group L2 (ridge) penalty. Setting \eqn{\alpha = 1} uses a group lasso penalty, setting \eqn{\alpha = 0} uses a group ridge penalty, and setting \eqn{0 < \alpha < 1} uses a group elastic net group penalty.  
}
  \item{nlambda}{
Number of \eqn{\lambda} values to use in the regularization path. Ignored if \code{lambda} is provided.
}
  \item{lambda.min.ratio}{
The proportion \eqn{0 < \pi < 1} that defines the minimum regularization parameter \eqn{\lambda_{\mathrm{min}}} as a fraction of the maximum regularization parameter \eqn{\lambda_{\mathrm{max}}} via the relationship \eqn{\lambda_{\mathrm{min}} = \pi \lambda_{\mathrm{max}}}. Ignored if \code{lambda} is provided. Note that \eqn{\lambda_{\mathrm{max}}} is defined such that all penalized effects are shrunk to zero.
}
  \item{lambda}{
Optional vector of user-supplied regularization parameter values. 
}
  \item{penalty.factor}{
Default S3 method: vector of length \eqn{K} giving the non-negative penalty weight for each predictor group. The order of the weights should correspond to the order of \code{levels(as.factor(group))}. Defaults to \eqn{\sqrt{p_k}} for all \eqn{k = 1,\ldots,K}, where \eqn{p_k} is the number of coefficients in the \eqn{k}-th group. If \code{penalty.factor[k] = 0}, then the \eqn{k}-th group is unpenalized, and the corresponding term is always included in the model. 

S3 "formula" method: named list giving the non-negative penalty weight for terms specified in the formula. Incomplete lists are allowed. Any term that is specified in \code{formula} but not in \code{penalty.factor} will be assigned the default penalty weight of \eqn{\sqrt{p_k}}. If \code{penalty.factor$z = 0}, then the variable \code{z} is unpenalized and always included in the model.
}
  \item{penalty}{
Character specifying which (group) penalty to use: LASSO , MCP, or SCAD.
}
  \item{gamma}{
Penalty hyperparameter that satisfies \eqn{\gamma > 1} for MCP and \eqn{\gamma > 2} for SCAD. Ignored for LASSO penalty.
} 
  \item{theta}{
For SVM1: additional ("smoothing") parameter, that controls the smoothing rate of the hinge loss function.
For negative binomial: additional ("size") parameter, where the variance function is defined as \eqn{V(\mu) = \mu + \mu^2/ \theta}
}
  \item{standardized}{
Logical indicating whether the predictors should be groupwise standardized. If \code{TRUE}, each column of \code{x} is mean-centered and each predictor group's design matrix is scaled to have a mean-square of one before fitting the model. Regardless of whether standardization is used, the coefficients are always returned on the original data scale.
}
\item{orthogonalized}{
Logical indicating whether the predictors should be groupwise orthogonalized. If \code{TRUE} (default), each predictor group's design matrix is orthonormalized (i.e., \eqn{\mathbf{X}_k^\top \mathbf{X}_k = n \mathbf{I}_k}) before fitting the model. Regardless of whether orthogonalization is used, the coefficients are always returned on the original data scale.
}
  \item{intercept}{
Logical indicating whether an intercept term should be included in the model. Note that the intercept is always unpenalized.   
}
  \item{thresh}{
Convergence threshold (tolerance). The algorithm is determined to have converged once the maximum relative change in the coefficients is below this threshold. See "Convergence" section.
}
  \item{maxit}{
Maximum number of iterations to allow.
}
  \item{proglang}{
Which programming language should be used to implement the ABGD algorithm? Options include \code{"Fortran"} (default) or \code{"R"}.
}
  \item{standardize.response}{
Should columns of response be standardized to have unit variance before fitting the model? Only applicable when \code{family = "multigaussian"}. Note that coefficients are returned on the original (unstandardized) scale regardless of this input.
}
  \item{keep.data}{
Should the training data be saved? Note this is needed for the \code{\link{summary.grpnet}} function to compute fitted values and importances.
}
  \item{\dots}{
Additional arguments used by the default or formula method.
}
}
\details{
Consider a generalized linear model of the form
\deqn{
g(\mu) = \mathbf{X}^\top \boldsymbol\beta
}
where \eqn{\mu = E(Y | \mathbf{X})} is the conditional expectation of the response \eqn{Y} given the predictor vector \eqn{\mathbf{X}}, the function \eqn{g(\cdot)} is a user-specified (invertible) link function, and \eqn{\boldsymbol\beta} are the unknown regression coefficients. Furthermore, suppose that the predictors are grouped, such as
\deqn{
\mathbf{X}^\top \boldsymbol\beta = \sum_{k=1}^K \mathbf{X}_k^\top \boldsymbol\beta_k
}
where \eqn{\mathbf{X} = (\mathbf{X}_1, \ldots, \mathbf{X}_K)} is the grouped predictor vector, and \eqn{\boldsymbol\beta = (\boldsymbol\beta_1, \ldots, \boldsymbol\beta_K)} is the grouped coefficient vector. 

Given \eqn{n} observations, this function finds the \eqn{\boldsymbol\beta} that minimizes
\deqn{
L(\boldsymbol\beta | \mathbf{D}) + \lambda P_\alpha(\boldsymbol\beta)
}
where \eqn{L(\boldsymbol\beta | \mathbf{D})} is the loss function with \eqn{\mathbf{D} = \{\mathbf{y}, \mathbf{X}\}} denoting the observed data, \eqn{P_\alpha(\boldsymbol\beta)} is the group elastic net penalty, and \eqn{\lambda \geq 0} is the regularization parameter. 

The loss function has the form 
\deqn{
L(\boldsymbol\beta | \mathbf{D}) = \frac{1}{n} \sum_{i=1}^n w_i \ell_i(\boldsymbol\beta | \mathbf{D}_i)
}
where \eqn{w_i > 0} are the user-supplied \code{weights}, and \eqn{\ell_i(\boldsymbol\beta | \mathbf{D}_i)} is the \eqn{i}-th observation's contribution to the loss function. Note that \eqn{\ell(\cdot) = -\log(f_Y(\cdot))} denotes the negative log-likelihood function for the given \code{family}.

The group elastic net penalty function has the form
\deqn{
P_\alpha(\boldsymbol\beta) = \alpha P_1(\boldsymbol\beta) + (1 - \alpha) P_2(\boldsymbol\beta)
}
where \eqn{\alpha \in [0,1]} is the user-specified \code{alpha} value, 
\deqn{
P_1(\boldsymbol\beta) = \sum_{k=1}^K \omega_k \| \boldsymbol\beta_k \|
}
is the group lasso penalty with \eqn{\omega_k \geq 0} denoting the \eqn{k}-th group's \code{penalty.factor}, and 
\deqn{
P_2(\boldsymbol\beta) = \frac{1}{2} \sum_{k=1}^K \omega_k \| \boldsymbol\beta_k \|^2
}
is the group ridge penalty. Note that \eqn{\| \boldsymbol\beta_k \|^2 = \boldsymbol\beta_k^\top \boldsymbol\beta_k} denotes the squared Euclidean norm. When \code{penalty \%in\% c("MCP", "SCAD")}, the group L1 penalty \eqn{P_1(\boldsymbol\beta)} is replaced by the group MCP or group SCAD penalty.

}
\value{
An object of class \code{"grpnet"} with the following elements:
\item{call }{matched call}
\item{a0 }{intercept sequence of length \code{nlambda}}
\item{beta }{coefficient matrix of dimension \code{nvars} by \code{nlambda}}
\item{alpha }{balance between the group L1 (lasso) and group L2 (ridge) penalty}
\item{lambda}{sequence of regularization parameter values}
\item{family }{exponential family defining the loss function}
\item{dev.ratio }{proportion of (null) deviance explained for each \code{lambda} (= 1 - dev / nulldev)}
\item{nulldev }{null deviance for each \code{lambda}}
\item{df }{effective degrees of freedom for each \code{lambda}}
\item{nzgrp }{number of non-zero groups for each \code{lambda}}
\item{nzcoef }{number of non-zero coefficients for each \code{lambda}}
\item{xsd }{standard deviation of x for each group}
\item{ylev }{levels of response variable (only for binomial and multinomial families)}
\item{nobs }{number of observations}
\item{group}{group label vector}
\item{ngroups }{number of groups \eqn{K}}
\item{npasses }{number of iterations for each \code{lambda}}
\item{time }{runtime in seconds to compute regularization path}
\item{offset }{logical indicating if an offset was included}
\item{args }{list of input argument values}
\item{formula }{input formula (possibly after expansion)}
\item{term.labels }{terms that appear in formula (if applicable)}
\item{rk.args }{arguments for rk.model.matrix function (if applicable)}
}
\references{
Friedman, J.,  Hastie, T., & Tibshirani, R. (2010). Regularization paths for generalized linear models via coordinate descent. \emph{Journal of Statistical Software, 33}(1), 1-22. \doi{10.18637/jss.v033.i01}

Helwig, N. E. (2025). Versatile descent algorithms for group regularization and variable selection in generalized linear models. \emph{Journal of Computational and Graphical Statistics, 34}(1), 239-252. \doi{10.1080/10618600.2024.2362232}
}
\author{
Nathaniel E. Helwig <helwig@umn.edu>
}
\note{
The syntax of (the default S3 method for) this function closely mimics that of the \code{glmnet} function in the \bold{glmnet} package (Friedman, Hastie, & Tibshirani, 2010). 
}

\section{S3 "formula" method}{
\bold{Important}: When using the S3 "formula" method, the S3 "predict" method forms the model matrix for the predictions by applying the model formula to the new data. As a result, to ensure that the corresponding S3 "predict" method works correctly, some formulaic features should be avoided. 

\bold{Polynomials}: When including polynomial terms, the \code{\link{poly}} function should be used with option \code{raw = TRUE}. Default use of the \code{\link{poly}} function (with \code{raw = FALSE}) will work for fitting the model, but will result in invalid predictions for new data. Polynomials can also be included via the \code{\link{I}} function, but this isn't recommended because the polynomials terms wouldn't be grouped together, i.e., the terms \code{x} and \code{I(x^2)} would be treated as two separate groups of size one instead of a single group of size two.

\bold{Splines}: B-splines (and other spline bases) can be included via the S3 "formula" method. However, to ensure reasonable predictions for new data, it is necessary to specify the knots directly. For example, if \code{x} is a vector with entries between zero and one, the code \code{bs(x, df = 5)} will *not* produce valid predictions for new data, but the code \code{bs(x, knots = c(0.25, 0.5, 0.75), Boundary.knots = c(0, 1))} will work as intended. Instead of attempting to integrate a call to \code{bs()} or \code{rk()} into the model formula, it is recommended that splines be included via the \code{use.rk = TRUE} argument.
}

\section{Family argument and link functions}{
Unlike the \code{\link{glm}} function, the \code{family} argument of the \code{\link{grpnet}} function \cr
* should be a character vector (not a \code{\link{family}} object) \cr
* does not allow for specification of a link function 

Currently, there is only one available link function for each \code{family}: \cr
* gaussian (identity): \eqn{\mu = \mathbf{X}^\top \boldsymbol\beta} \cr
* multigaussian (identity): \eqn{\mu = \mathbf{X}^\top \boldsymbol\beta} \cr
* svm1/svm2 (identity): \eqn{\mu = \mathbf{X}^\top \boldsymbol\beta} \cr
* binomial/logit (logit): \eqn{\log(\frac{\pi}{1 - \pi}) = \mathbf{X}^\top \boldsymbol\beta} \cr
* multinomial (symmetric): \eqn{\pi_\ell = \frac{\exp(\mathbf{X}^\top \boldsymbol\beta_\ell)}{\sum_{l = 1}^m \exp(\mathbf{X}^\top \boldsymbol\beta_l)}} \cr
* poisson (log): \eqn{\log(\mu) = \mathbf{X}^\top \boldsymbol\beta} \cr
* negative.binomial (log): \eqn{\log(\mu) = \mathbf{X}^\top \boldsymbol\beta} \cr
* Gamma (log): \eqn{\log(\mu) = \mathbf{X}^\top \boldsymbol\beta} \cr
* inverse.gaussian (log): \eqn{\log(\mu) = \mathbf{X}^\top \boldsymbol\beta} \cr
}

\section{Classification problems (svm1/svm2/logit/binomial/multinomial)}{
For \code{"svm1"}, \code{"svm2"}, and \code{"logit"} responses, three different possibilities exist for the input response: \cr
1. vector coercible into a factor with two levels \cr
2. matrix with two columns (# successes, # failures) \cr
3. numeric vector with entries between -1 and 1 \cr
In this case, the \code{weights} argument should be used specify the total number of trials.

For \code{"binomial"} responses, three different possibilities exist for the input response: \cr
1. vector coercible into a factor with two levels \cr
2. matrix with two columns (# successes, # failures) \cr
3. numeric vector with entries between 0 and 1 \cr
In this case, the \code{weights} argument should be used specify the total number of trials.

For \code{"multinomial"} responses, two different possibilities exist for the input response: \cr
1. vector coercible into a factor with more than two levels \cr 
2. matrix of integers (counts) for each category level \cr
}

\section{Convergence}{
The algorithm is determined to have converged once

\eqn{\max_j \frac{| \beta_j - \beta_j^{\mathrm{old}} |}{1 + |\beta_j^{\mathrm{old}}|}  < \epsilon }

where \eqn{j \in \{1,\ldots,p\}} and \eqn{\epsilon} is the \code{thresh} argument. 
}

%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link{summary.grpnet}} for calculating fitted values and importances

\code{\link{plot.grpnet}} for plotting the regularization path

\code{\link{predict.grpnet}} for predicting from \code{\link{grpnet}} objects

\code{\link{cv.grpnet}} for k-fold cross-validation of \code{lambda}
}
\examples{
######***######   family = "gaussian"   ######***######

# load data
data(auto)

# fit model (formula method, response = mpg)
mod <- grpnet(mpg ~ ., data = auto)

# print regularization path info
mod

# plot proportion of null deviance explained
plot(mod)

# summarize model
summary(mod)


\donttest{
######***######   family = "multigaussian"   ######***######

# load data
data(auto)

# fit model (formula method, response = (mpg, displacement))
y <- as.matrix(auto[,c(1,3)])
mod <- grpnet(y ~ ., data = auto[,-c(1,3)], family = "multigaussian",
              standardize.response = TRUE)

# print regularization path info
mod

# plot proportion of null deviance explained
plot(mod)

# summarize model
summary(mod)



######***######   family = "svm1"   ######***######

# load data
data(auto)

# redefine origin (Domestic vs Foreign)
auto$origin <- ifelse(auto$origin == "American", "Domestic", "Foreign")

# fit model (formula method, response = origin with 2 levels)
mod <- grpnet(origin ~ ., data = auto, family = "svm1")

# print regularization path info
mod

# plot proportion of null deviance explained
plot(mod)

# summarize model
summary(mod)



######***######   family = "svm2"   ######***######

# load data
data(auto)

# redefine origin (Domestic vs Foreign)
auto$origin <- ifelse(auto$origin == "American", "Domestic", "Foreign")

# fit model (formula method, response = origin with 2 levels)
mod <- grpnet(origin ~ ., data = auto, family = "svm2")

# print regularization path info
mod

# plot proportion of null deviance explained
plot(mod)

# summarize model
summary(mod)



######***######   family = "logit"   ######***######

# load data
data(auto)

# redefine origin (Domestic vs Foreign)
auto$origin <- ifelse(auto$origin == "American", "Domestic", "Foreign")

# fit model (formula method, response = origin with 2 levels)
mod <- grpnet(origin ~ ., data = auto, family = "logit")

# print regularization path info
mod

# plot proportion of null deviance explained
plot(mod)

# summarize model
summary(mod)



######***######   family = "binomial"   ######***######

# load data
data(auto)

# redefine origin (Domestic vs Foreign)
auto$origin <- ifelse(auto$origin == "American", "Domestic", "Foreign")

# fit model (formula method, response = origin with 2 levels)
mod <- grpnet(origin ~ ., data = auto, family = "binomial")

# print regularization path info
mod

# plot proportion of null deviance explained
plot(mod)

# summarize model
summary(mod)



######***######   family = "multinomial"   ######***######

# load data
data(auto)

# fit model (formula method, response = origin with 3 levels)
mod <- grpnet(origin ~ ., data = auto, family = "multinomial")

# print regularization path info
mod

# plot proportion of null deviance explained
plot(mod)

# summarize model
summary(mod)



######***######   family = "ordinal"   ######***######

# load data
data(auto)

# fit model (formula method, response = cylinders with 5 levels)
mod <- grpnet(cylinders ~ ., data = auto, family = "ordinal")

# print regularization path info
mod

# plot proportion of null deviance explained
plot(mod)

# summarize model
summary(mod)



######***######   family = "poisson"   ######***######

# load data
data(auto)

# fit model (formula method, response = horsepower)
mod <- grpnet(horsepower ~ ., data = auto, family = "poisson")

# print regularization path info
mod

# plot proportion of null deviance explained
plot(mod)

# summarize model
summary(mod)



######***######   family = "negative.binomial"   ######***######

# load data
data(auto)

# fit model (formula method, response = horsepower)
mod <- grpnet(horsepower ~ ., data = auto, family = "negative.binomial")

# print regularization path info
mod

# plot proportion of null deviance explained
plot(mod)

# summarize model
summary(mod)



######***######   family = "Gamma"   ######***######

# load data
data(auto)

# fit model (formula method, response = mpg)
mod <- grpnet(mpg ~ ., data = auto, family = "Gamma")

# print regularization path info
mod

# plot proportion of null deviance explained
plot(mod)

# summarize model
summary(mod)



######***######   family = "inverse.gaussian"   ######***######

# load data
data(auto)

# fit model (formula method, response = mpg)
mod <- grpnet(mpg ~ ., data = auto, family = "inverse.gaussian")

# print regularization path info
mod

# plot proportion of null deviance explained
plot(mod)

# summarize model
summary(mod)
}
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory (show via RShowDoc("KEYWORDS")):
\keyword{ regression }
\keyword{ smooth }
% Use only one keyword per line.
% For non-standard keywords, use \concept instead of \keyword:
% \concept{ ~cpt1 }
% \concept{ ~cpt2 }
% Use only one concept per line.

\name{family.grpnet}
\alias{family.grpnet}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Prepare 'family' Argument for grpnet
}
\description{
Takes in the \code{family} argument from \code{\link{grpnet}} and returns a list containing the information needed for fitting and/or tuning the model.
}
\usage{
family.grpnet(object, theta = 1)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{object}{
  two options: (1) an object of class "grpnet" or "cv.grpnet"; or (2) a character specifying the exponential family: \code{"gaussian"}, \code{"multigaussian"}, \code{"svm1"}, \code{"svm2"}, \code{"logit"}, \code{"binomial"}, \code{"multinomial"}, \code{"poisson"}, \code{"negative.binomial"}, \code{"Gamma"}, \code{"inverse.gaussian"}}
\item{theta}{
positive scalar that serves as an additional hyperparameter for various loss functions.

svm1: additional parameter that controls the smoothing rate for the hinge loss function (see Note below).

negative.binomial: size parameter such that the variance function is defined as \eqn{V(\mu) = \mu + \mu^2/ \theta}
}
}
\details{
There is only one available link function for each \code{family}: \cr
* gaussian (identity): \eqn{\mu = \mathbf{X}^\top \boldsymbol\beta} \cr
* multigaussian (identity): \eqn{\mu_\ell = \mathbf{X}^\top \boldsymbol\beta_\ell} \cr
* svm1/svm2 (identity): \eqn{\mu = \mathbf{X}^\top \boldsymbol\beta} \cr
* binomial/logit (logit): \eqn{\log(\frac{\pi}{1 - \pi}) = \mathbf{X}^\top \boldsymbol\beta} \cr
* multinomial (symmetric): \eqn{\pi_\ell = \frac{\exp(\mathbf{X}^\top \boldsymbol\beta_\ell)}{\sum_{l = 1}^m \exp(\mathbf{X}^\top \boldsymbol\beta_l)}} \cr
* ordinal (logit): \eqn{\log(\frac{\Pi_\ell}{1 - \Pi_\ell}) = \alpha_\ell + \mathbf{X}^\top \boldsymbol\beta} \cr
* poisson (log): \eqn{\log(\mu) = \mathbf{X}^\top \boldsymbol\beta} \cr
* negative.binomial (log): \eqn{\log(\mu) = \mathbf{X}^\top \boldsymbol\beta} \cr
* Gamma (log): \eqn{\log(\mu) = \mathbf{X}^\top \boldsymbol\beta} \cr
* inverse.gaussian (log): \eqn{\log(\mu) = \mathbf{X}^\top \boldsymbol\beta} \cr
}
\value{
List with components:
\item{family }{same as input object, i.e., character specifying the family}
\item{linkinv }{function for computing inverse of link function}
\item{dev.resids }{function for computing deviance residuals}
}
\references{
Helwig, N. E. (2025). Versatile descent algorithms for group regularization and variable selection in generalized linear models. \emph{Journal of Computational and Graphical Statistics, 34}(1), 239-252. \doi{10.1080/10618600.2024.2362232}
}
\author{
Nathaniel E. Helwig <helwig@umn.edu>
}
\note{
For \code{gaussian} family, this returns the full output produced by \code{\link[stats]{gaussian}}.

For \code{svm1} family, the quadratically smoothed hinge loss is defined as
\deqn{
\mathrm{svm1}(z) = \left\{ \begin{array}{ll}
0 & z > 1 \\
(1 - z)^2 / (2 \theta) & 1 - \theta < z \leq 1 \\
1 - z - \theta / 2 & z \leq 1 - \theta \\
\end{array} \right.
}
where \eqn{z = Y \eta} with \eqn{Y \in \{-1,1\}} denoting the response and \eqn{\eta = \mathbf{X}^\top \boldsymbol\beta} denoting the linear predictor. Note that the \code{svm1} loss function approaches the support vector machine (i.e., hinge) loss function as \eqn{\theta \rightarrow 0}.

For \code{svm2} family, the squared hinge loss is defined as
\deqn{
\mathrm{svm2}(z) = \left\{ \begin{array}{ll}
0 & z > 1 \\
(1 - z)^2 & z \leq 1 \\
\end{array} \right.
}
where \eqn{z = Y \eta} with \eqn{Y \in \{-1,1\}} denoting the response and \eqn{\eta = \mathbf{X}^\top \boldsymbol\beta} denoting the linear predictor. Note that the \code{svm1} loss function approaches the support vector machine (i.e., hinge) loss function as \eqn{\theta \rightarrow 0}.


For \code{ordinal} family, a cumulative link model is used, i.e.,
\deqn{
\log(\frac{\Pi_\ell}{1 - \Pi_\ell}) = \alpha_\ell + \mathbf{X}^\top \boldsymbol\beta
}
where \eqn{\Pi_\ell = P(Y \geq y_\ell)} is the probabilty that the response exceeds the \eqn{\ell}-th ordered response category for \eqn{\ell = 2,\ldots,m}.

}
%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link{visualize.loss}} for plotting loss functions

\code{\link{grpnet}} for fitting group elastic net regularization paths

\code{\link{cv.grpnet}} for k-fold cross-validation of \code{lambda}
}
\examples{
family.grpnet("gaussian")

family.grpnet("multigaussian")

family.grpnet("svm1", theta = 0.1)

family.grpnet("svm2")

family.grpnet("logit")

family.grpnet("binomial")

family.grpnet("multinomial")

family.grpnet("ordinal")

family.grpnet("poisson")

family.grpnet("negative.binomial", theta = 10)

family.grpnet("Gamma")

family.grpnet("inverse.gaussian")
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory (show via RShowDoc("KEYWORDS")):
\keyword{ models }
\keyword{ regression }
% \keyword{ ~kwd2 }
% Use only one keyword per line.
% For non-standard keywords, use \concept instead of \keyword:
% \concept{ ~cpt1 }
% \concept{ ~cpt2 }
% Use only one concept per line.

\name{print}
\alias{print.coef.grpnet}
\alias{print.cv.grpnet}
\alias{print.grpnet}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
S3 'print' Methods for grpnet
}
\description{
Prints some basic information about the coefficients (for \code{\link{coef.grpnet}} objects), observed cross-validation error (for \code{\link{cv.grpnet}} objects), or the computed regularization path (for \code{\link{grpnet}} objects). 
}
\usage{
\method{print}{coef.grpnet}(x, ...)

\method{print}{cv.grpnet}(x, digits = max(3, getOption("digits") - 3), ...)

\method{print}{grpnet}(x, ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{
an object of class \code{\link{coef.grpnet}}, \code{\link{cv.grpnet}}, or \code{\link{grpnet}}
}
  \item{digits}{
the number of digits to print (must be a positive integer)  
}
  \item{...}{
additional arguments for \code{\link{print}} (currently ignored)  
}
}
\details{
For \code{\link{coef.grpnet}} objects, prints the non-zero coefficients and uses "." for coefficients shrunk to zero.

For \code{\link{cv.grpnet}} objects, prints the function \code{call}, the cross-validation \code{type.measure}, and a two-row table with information about the \code{min} and \code{1se} solutions.

For \code{\link{grpnet}} objects, prints a data frame with columns \cr
* nGrp: number of non-zero groups for each \code{lambda} \cr
* Df: effective degrees of freedom for each \code{lambda} \cr
* \%Dev: percentage of null deviance explained for each \code{lambda} \cr
* Lambda: the values of \code{lambda} \cr
}

\value{
No return value (produces a printout)
}

\references{
Friedman, J.,  Hastie, T., & Tibshirani, R. (2010). Regularization paths for generalized linear models via coordinate descent. \emph{Journal of Statistical Software, 33}(1), 1-22. \doi{10.18637/jss.v033.i01}

Helwig, N. E. (2025). Versatile descent algorithms for group regularization and variable selection in generalized linear models. \emph{Journal of Computational and Graphical Statistics, 34}(1), 239-252. \doi{10.1080/10618600.2024.2362232}
}
\author{
Nathaniel E. Helwig <helwig@umn.edu>
}
\note{
Some syntax and functionality were modeled after the \code{print} functions in the \bold{glmnet} package (Friedman, Hastie, & Tibshirani, 2010).
}

%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link{coef.grpnet}} for extracting coefficients

\code{\link{cv.grpnet}} for k-fold cross-validation of \code{lambda}

\code{\link{grpnet}} for fitting grpnet regularization paths
}
\examples{
# see 'coef.grpnet' for coefficient printing examples
?coef.grpnet

# see 'cv.grpnet' for cross-validation error printing examples
?cv.grpnet

# see 'grpnet' for regularization path printing examples
?grpnet
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory (show via RShowDoc("KEYWORDS")):
\keyword{ print }
% \keyword{ ~kwd2 }
% Use only one keyword per line.
% For non-standard keywords, use \concept instead of \keyword:
% \concept{ ~cpt1 }
% \concept{ ~cpt2 }
% Use only one concept per line.

\name{visualize.loss}
\alias{visualize.loss}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Plots grpnet Loss Function on Link or Response Scale
}
\description{
Makes a plot or returns a data frame containing the specified loss function evaluated at a sequence of input values.
}
\usage{
visualize.loss(x = seq(-3, 3, length.out = 1001),
               family = c("gaussian", "multigaussian",
                          "svm1", "svm2", "logit",
                          "binomial", "multinomial", 
                          "poisson", "negative.binomial", 
                          "Gamma", "inverse.gaussian"),
               theta = 1,
               type = c("link", "response"),
               y = NULL,
               plot = TRUE,
               add = FALSE,
               ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{
  sequence of linear predictor values at which to evaluate the loss.
  }
  \item{family}{
  Character specifying the assumed distribution for the response variable. Partial matching is allowed. See options below.
  }
  \item{theta}{
  For SVM1: additional ("smoothing") parameter, that controls the smoothing rate of the hinge loss function.
For negative binomial: additional ("size") parameter, where the variance function is defined as \eqn{V(\mu) = \mu + \mu^2/ \theta}
  }
  \item{type}{
  Default of \code{type = "link"} plots \code{x} versus the loss. Change to \code{type = "response"} to plot \code{mu} versus the loss.
  }
  \item{y}{
  Response value used to compute loss. Note that the loss function is interpreted as a function of \code{x} given \code{y}.
  }
  \item{plot}{
  if \code{TRUE} (default), then the result is plotted; otherwise the result is returned as a data frame.
  }
  \item{add}{
  if \code{FALSE} (default), then the \code{\link{plot}} function is used to create a new graphic; otherwise the \code{\link{lines}} function is used to add lines to an existing graphic.
  }
  \item{...}{
  additional arguments passed to \code{plot} function, e.g., \code{xlim}, \code{ylim}, etc.
  }
}
\details{
\bold{grpnet} implements the following loss functions:
\describe{
\item{gaussian/multigaussian}{\eqn{L = (y - \mu)^2}}
\item{svm1}{\eqn{L = \left\{ \begin{array}{ll}
\frac{1}{2\theta}(1 - \mu y)_{+}^2  & 1 - \theta < \mu y \\
1 - \mu y - \theta / 2 & \mu y \leq 1 - \theta 
\end{array} \right.}}
\item{svm2}{\eqn{L = (1 - \mu y)_{+}^2}}
\item{logit}{\eqn{L = \log(1 + \exp(-\mu y))}}
\item{binomial}{\eqn{L = -y\log(\mu) - (1-y) \log(1-\mu)}}
\item{multinomial}{\eqn{L = -\sum_{l=1}^m I(y=l) \log(\mu_l)}}
\item{poisson}{\eqn{L = \mu - y \log(\mu)}}
\item{negative.binomial}{L = \eqn{(\theta + y)  \log(\theta + \mu) - y  \log(\mu) + c} \cr where \eqn{c = \log(\Gamma(\theta)) - \log(\Gamma(\theta + y)) - \theta  \log(\theta)} \cr is a constant with respect to \eqn{\mu}}
\item{Gamma}{\eqn{L = \log(\mu) + y / \mu}}
\item{inverse.gaussian}{\eqn{L = (y - \mu)^2 / (\mu^2 y)}}
}
}

\value{
If \code{plot = TRUE}, a plot is produced and nothing is returned.

If \code{plot = FALSE}, a data frame is returned with columns:
\item{eta}{linear predictor, i.e., fitted values on link scale (same as input \code{x})}
\item{mu}{expected value, i.e., fitted values on response scale (\code{family$linkinv(x)})}
\item{loss}{loss function evaluation for given \code{eta} (and \code{y})}
}
\references{
Helwig, N. E. (2025). Versatile descent algorithms for group regularization and variable selection in generalized linear models. \emph{Journal of Computational and Graphical Statistics, 34}(1), 239-252. \doi{10.1080/10618600.2024.2362232}
}
\author{
Nathaniel E. Helwig <helwig@umn.edu>
}


%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link{visualize.penalty}} for plotting penalty function

\code{\link{visualize.shrink}} for plotting shrinkage operator
}
\examples{
visualize.loss(family = "gaussian")

visualize.loss(family = "svm1", theta = 0.1)

visualize.loss(family = "svm2")

visualize.loss(family = "logit")

visualize.loss(family = "binomial")

visualize.loss(family = "poisson")

visualize.loss(family = "negative.binomial", theta = 10)

visualize.loss(family = "Gamma")

visualize.loss(family = "inverse.gaussian")
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory (show via RShowDoc("KEYWORDS")):
\keyword{ models }
\keyword{ regression }
\keyword{ dplot }
\keyword{ hplot }
% Use only one keyword per line.
% For non-standard keywords, use \concept instead of \keyword:
% \concept{ ~cpt1 }
% \concept{ ~cpt2 }
% Use only one concept per line.

\name{plot.cv.grpnet}
\alias{plot.cv.grpnet}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Plot Cross-Validation Curve for cv.grpnet Fits
}
\description{
Plots the mean cross-validation error, along with lower and upper standard deviation curves, as a function of \code{log(lambda)}.
}
\usage{
\method{plot}{cv.grpnet}(x, cv = TRUE, sign.lambda = 1, nzero = TRUE, ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{
Object of class "cv.grpnet"
}
  \item{cv}{
If \code{TRUE} (default), plots the cross-validation path as described in this helpful. Otherwise passes the \code{grpnet.fit} component to the \code{\link{plot.grpnet}} function.
}
  \item{sign.lambda}{
Default plots \code{log(lambda)} on the x-axis. Set to -1 to plot \code{-1*log(lambda)} on the x-axis instead.
}
  \item{nzero}{
Should the number of non-zero groups be printed on the top of the x-axis?
}
  \item{...}{
Additional arguments passed to the \code{\link{plot}} function.
}
}
\details{
Produces cross-validation plot only (i.e., nothing is returned).
}
\value{
No return value (produces a plot)
}
\references{
Friedman, J.,  Hastie, T., & Tibshirani, R. (2010). Regularization paths for generalized linear models via coordinate descent. \emph{Journal of Statistical Software, 33}(1), 1-22. \doi{10.18637/jss.v033.i01}

Helwig, N. E. (2025). Versatile descent algorithms for group regularization and variable selection in generalized linear models. \emph{Journal of Computational and Graphical Statistics, 34}(1), 239-252. \doi{10.1080/10618600.2024.2362232}
}
\author{
Nathaniel E. Helwig <helwig@umn.edu>
}
\note{
Syntax and functionality were modeled after the \code{plot.cv.glmnet} function in the \bold{glmnet} package (Friedman, Hastie, & Tibshirani, 2010).
}

%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link{cv.grpnet}} for k-fold cross-validation of \code{lambda}

\code{\link{plot.grpnet}} for plotting the regularization path
}
\examples{
# see 'cv.grpnet' for plotting examples
?cv.grpnet
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory (show via RShowDoc("KEYWORDS")):
\keyword{ graphs }
% Use only one keyword per line.
% For non-standard keywords, use \concept instead of \keyword:
% \concept{ ~cpt1 }
% \concept{ ~cpt2 }
% Use only one concept per line.

\name{plot.grpnet}
\alias{plot.grpnet}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Plot Regularization Path for grpnet Fits
}
\description{
Creates a profile plot of the reguarlization paths for a fit group elastic net regularized GLM (grpnet) object.
}
\usage{
\method{plot}{grpnet}(x, type = c("dev.ratio", "coef", "imp", "norm", "znorm"),
     newx, newdata, intercept = FALSE,
     color.by.group = TRUE, col = NULL, ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{
Object of class "grpnet"
}
  \item{type}{
What to plot on the Y-axis: "dev.ratio" for explained deviance, "coef" for coefficient values, "imp" for importance of each group's contribution, "norm" for L2 norm of coefficients for each group, or "znorm" for L2 norm of standardized coefficients for each group. 
}
  \item{newx}{
Matrix of new \code{x} scores for prediction (default S3 method). Ignored unless \code{type = "imp"}.
}
  \item{newdata}{
Data frame of new \code{data} scores for prediction (S3 "formula" method). Ignored unless \code{type = "imp"}.
}
  \item{intercept}{
Should the intercept be included in the plot?  
}
  \item{color.by.group}{
If \code{TRUE} (default), the coefficient paths are colored according to their group membership using the colors in \code{col}. If \code{FALSE}, all coefficient paths are plotted the same color.
}
  \item{col}{
If \code{color.by.group = TRUE}, this should be a vector of length \eqn{K} giving a color label for each group. If \code{color.by.group = FASLE}, this should be a character specifying a single (common) color. Default of \code{col = NULL} is the same as \code{col = 1:K} or \code{col = "black"}.
}
  \item{...}{
Additional arguments passed to the \code{\link{plot}} function.
}
}
\details{
Syntax and functionality were modeled after the \code{plot.glmnet} function in the \bold{glmnet} package (Friedman, Hastie, & Tibshirani, 2010).
}
\value{
Produces a profile plot showing the requested type (y-axis) as a function of \code{log(lambda)} (x-axis). 
}
\references{
Friedman, J.,  Hastie, T., & Tibshirani, R. (2010). Regularization paths for generalized linear models via coordinate descent. \emph{Journal of Statistical Software, 33}(1), 1-22. \doi{10.18637/jss.v033.i01}

Helwig, N. E. (2025). Versatile descent algorithms for group regularization and variable selection in generalized linear models. \emph{Journal of Computational and Graphical Statistics, 34}(1), 239-252. \doi{10.1080/10618600.2024.2362232}
}
\author{
Nathaniel E. Helwig <helwig@umn.edu>
}
\note{
If \code{x} is a multigaussian or multinomial model, the coefficients for each response dimension/class are plotted in a separate plot.
}

%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link{grpnet}} for fitting grpnet regularization paths

\code{\link{plot.cv.grpnet}} for plotting \code{\link{cv.grpnet}} objects
}
\examples{
# see 'grpnet' for plotting examples
?grpnet
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory (show via RShowDoc("KEYWORDS")):
\keyword{ graphs }
% Use only one keyword per line.
% For non-standard keywords, use \concept instead of \keyword:
% \concept{ ~cpt1 }
% \concept{ ~cpt2 }
% Use only one concept per line.

\name{rk}
\alias{rk}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Reproducing Kernel Basis
}
\description{
Generate a reproducing kernel basis matrix for a nominal, ordinal, or polynomial smoothing spline.
}
\usage{
rk(x, df = NULL, knots = NULL, m = NULL, intercept = FALSE, 
   Boundary.knots = NULL, warn.outside = TRUE, 
   periodic = FALSE, xlev = levels(x))
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{
  the predictor vector of length \code{n}. Can be a factor, integer, or numeric, see Note.
  }
  \item{df}{
  the degrees of freedom, i.e., number of knots to place at quantiles of \code{x}. Defaults to 5 but ignored if \code{knots} are provided.
  }
  \item{knots}{
  the breakpoints (knots) defining the spline. If \code{knots} are provided, the \code{df} is defined as \code{length(unique(c(knots, Boundary.knots)))}.
  }
  \item{m}{
  the derivative penalty order: 0 = ordinal spline, 1 = linear spline, 2 = cubic spline, 3 = quintic spline
  }
  \item{intercept}{
  should an intercept be included in the basis?  
  }
  \item{Boundary.knots}{
  the boundary points for spline basis. Defaults to \code{range(x)}.
  }
  \item{warn.outside}{
  if \code{TRUE}, a warning is provided when \code{x} values are outside of the \code{Boundary.knots}
  }
  \item{periodic}{
  should the spline basis functions be constrained to be periodic with respect to the \code{Boundary.knots}?
  }
  \item{xlev}{
  levels of \code{x} (only applicable if \code{x} is a \code{\link{factor}})
  }
}
\details{
Given a vector of function realizations \eqn{f}, suppose that \eqn{f = X \beta}, where \eqn{X} is the (unregularized) spline basis and \eqn{\beta} is the coefficient vector. Let \eqn{Q} denote the postive semi-definite penalty matrix, such that \eqn{\beta^\top Q \beta} defines the roughness penalty for the spline. See Helwig (2017) for the form of \eqn{X} and \eqn{Q} for the various types of splines.

Consider the spectral parameterization of the form \eqn{f = Z \alpha} where
\deqn{Z = X Q^{-1/2}}
is the regularized spline basis (that is returned by this function), and \eqn{\alpha = Q^{1/2} \beta} are the reparameterized coefficients. Note that \eqn{X \beta = Z \alpha} and \eqn{\beta^\top Q \beta = \alpha^\top \alpha}, so the spectral parameterization absorbs the penalty into the coefficients (see Helwig, 2021, 2024).

Syntax of this function is designed to mimic the syntax of the \code{\link[splines]{bs}} function.
}
\value{
Returns a basis function matrix of dimension \code{n} by \code{df} (plus 1 if an \code{intercept} is included) with the following attributes:
\item{df }{degrees of freedom}
\item{knots }{knots for spline basis}
\item{m }{derivative penalty order}
\item{intercept }{was an intercept included?}
\item{Boundary.knots }{boundary points of \code{x}}
\item{periodic }{is the basis periodic?}
\item{xlev }{factor levels (if applicable)}
}
\references{
Helwig, N. E. (2017). Regression with ordered predictors via ordinal smoothing splines. Frontiers in Applied Mathematics and Statistics, 3(15), 1-13. \doi{10.3389/fams.2017.00015}

Helwig, N. E. (2021). Spectrally sparse nonparametric regression via elastic net regularized smoothers. Journal of Computational and Graphical Statistics, 30(1), 182-191. \doi{10.1080/10618600.2020.1806855}

Helwig, N. E. (2025). Versatile descent algorithms for group regularization and variable selection in generalized linear models. \emph{Journal of Computational and Graphical Statistics, 34}(1), 239-252. \doi{10.1080/10618600.2024.2362232}
}
\author{
Nathaniel E. Helwig <helwig@umn.edu>
}
\note{
The (default) type of spline basis depends on the \code{\link{class}} of the input \code{x} object:

* If \code{x} is an unordered factor, then a nominal spline basis is used

* If \code{x} is an ordered factor (and \code{m = NULL}), then an ordinal spline basis is used

* If \code{x} is an integer or numeric (and \code{m = NULL}), then a cubic spline basis is used

Note that you can override the default behavior by specifying the \code{m} argument.
}

%% ~Make other sections like Warning with \section{Warning }{....} ~

\examples{

######***######   NOMINAL SPLINE BASIS   ######***######

x <- as.factor(LETTERS[1:5])
basis <- rk(x)
plot(1:5, basis[,1], t = "l", ylim = extendrange(basis))
for(j in 2:ncol(basis)){
  lines(1:5, basis[,j], col = j)
}


######***######   ORDINAL SPLINE BASIS   ######***######

x <- as.ordered(LETTERS[1:5])
basis <- rk(x)
plot(1:5, basis[,1], t = "l", ylim = extendrange(basis))
for(j in 2:ncol(basis)){
  lines(1:5, basis[,j], col = j)
}


######***######   LINEAR SPLINE BASIS   ######***######

x <- seq(0, 1, length.out = 101)
basis <- rk(x, m = 1)
plot(x, basis[,1], t = "l", ylim = extendrange(basis))
for(j in 2:ncol(basis)){
  lines(x, basis[,j], col = j)
}


######***######   CUBIC SPLINE BASIS   ######***######

x <- seq(0, 1, length.out = 101)
basis <- rk(x)
basis <- scale(basis)  # for visualization only!
plot(x, basis[,1], t = "l", ylim = extendrange(basis))
for(j in 2:ncol(basis)){
  lines(x, basis[,j], col = j)
}


######***######   QUINTIC SPLINE BASIS   ######***######

x <- seq(0, 1, length.out = 101)
basis <- rk(x, m = 3)
basis <- scale(basis)  # for visualization only!
plot(x, basis[,1], t = "l", ylim = extendrange(basis))
for(j in 2:ncol(basis)){
  lines(x, basis[,j], col = j)
}

}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory (show via RShowDoc("KEYWORDS")):
\keyword{ regression }% use one of  RShowDoc("KEYWORDS")
\keyword{ smooth }% __ONLY ONE__ keyword per line
% Use only one keyword per line.
% For non-standard keywords, use \concept instead of \keyword:
% \concept{ ~cpt1 }
% \concept{ ~cpt2 }
% Use only one concept per line.

\name{cv.compare}
\alias{cv.compare}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Compare Multiple cv.grpnet Solutions
}
\description{
Creates a plot (default) or returns a data frame (otherwise) that compares the cross-validation error for multiple \code{\link{cv.grpnet}} fits.
}
\usage{
cv.compare(x,
           s = c("lambda.1se", "lambda.min"), 
           plot = TRUE, 
           at = 1:length(x),
           nse = 1, 
           point.col = "red", 
           line.col = "gray", 
           lwd = 2, 
           bwd = 0.02,
           labels = NULL,
           xlim = NULL,
           ylim = NULL,
           xlab = NULL,
           ylab = NULL,
           ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{
a single \code{\link{cv.grpnet}} object or a list of \code{\link{cv.grpnet}} objects.
}
  \item{s}{
the tuning parameter value at which to plot results (if \code{x} is a list).
}
  \item{plot}{
switch controlling whether a plot is produced (default) versus data frame.
}
  \item{at}{
x-axis coordinates for plotting the cv error for each solution.
}
  \item{nse}{
number of standard errors to use for error bars in plot.
}
  \item{point.col}{
color for point used to plot the average of the cv error.
}
  \item{line.col}{
color for lines used to plot the standard error for the cv error.
}
  \item{lwd}{
width of lines used to plot the standard error for the cv error.  
}
  \item{bwd}{
width of standard error bars in terms of proportion of \code{range(x)}.
}
  \item{labels}{
labels for x-axis tick marks. Defaults to \code{names(x)}.
}
  \item{xlim}{
axis limits for abscissa (x-axis)
}
  \item{ylim}{
axis limits for ordinate (y-axis)
}
  \item{xlab}{
axis label for abscissa (x-axis)
}
  \item{ylab}{
axis label for ordinate (y-axis)
}
  \item{...}{
additional arguments passed to plotting functions.  
}

}
\details{
Default behavior creates a plot that displays the mean cv error +/- 1 se for each of the requested solutions. 

If the input \code{x} is a single \code{\link{cv.grpnet}} object, then the function plots the lambda.min and lambda.1se solutions. 

If the input \code{x} is a list of \code{\link{cv.grpnet}} objects, then the function plots either the lambda.min or the lambda.1se solution (controlled by \code{s} argument) for all of the input models.
}
\value{
When \code{plot = TRUE}, there is no return value (it produces a plot)

When \code{plot = FALSE}, a data.frame is returned with the mean cv error (and se) for each solution
}
\references{
Helwig, N. E. (2025). Versatile descent algorithms for group regularization and variable selection in generalized linear models. \emph{Journal of Computational and Graphical Statistics, 34}(1), 239-252. \doi{10.1080/10618600.2024.2362232}
}
\author{
Nathaniel E. Helwig <helwig@umn.edu>
}

%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link{plot.cv.grpnet}} for plotting cv error path (for all lambdas)

\code{\link{plot.grpnet}} for plotting regularization path (for single lambda)
}
\examples{
# load data
data(auto)

# LASSO penalty
set.seed(1)
mod1 <- cv.grpnet(mpg ~ ., data = auto, nfolds = 5, alpha = 1)

# MCP penalty
set.seed(1)
mod2 <- cv.grpnet(mpg ~ ., data = auto, nfolds = 5, alpha = 1, penaly = "MCP")

# SCAD penalty
set.seed(1)
mod3 <- cv.grpnet(mpg ~ ., data = auto, nfolds = 5, alpha = 1, penaly = "SCAD")

# compare lambda.min and lambda.1se for mod1
cv.compare(mod1)

# compare lambda.1se for mod1, mod2, mod3
cv.compare(x = list(mod1, mod2, mod3), labels = c("LASSO", "MCP", "SCAD"))
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory (show via RShowDoc("KEYWORDS")):
\keyword{ graphs }
% Use only one keyword per line.
% For non-standard keywords, use \concept instead of \keyword:
% \concept{ ~cpt1 }
% \concept{ ~cpt2 }
% Use only one concept per line.

\name{predict.grpnet}
\alias{predict.grpnet}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
Predict Method for grpnet Fits
}
\description{
Obtain predictions from a fit group elastic net (grpnet) object.
}
\usage{
\method{predict}{grpnet}(object, 
        newx,
        newdata,
        s = NULL,
        type = c("link", "response", "class", "terms", 
                 "importance", "coefficients", "nonzero", "groups", 
                 "ncoefs", "ngroups", "norm", "znorm"),
        ...)

}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{object}{
Object of class "grpnet"
}
  \item{newx}{
Matrix of new \code{x} scores for prediction (default S3 method). Must have \eqn{p} columns arranged in the same order as the \code{x} matrix used to fit the model. Ignored for the last six types of predictions.
}
  \item{newdata}{
Data frame of new \code{data} scores for prediction (S3 "formula" method). Must contain all variables in the \code{formula} used to fit the model. Ignored for the last six types of predictions.
}
  \item{s}{
Lambda value(s) at which predictions should be obtained. Default uses \code{s = object$lambda}. Interpolation is used for \code{s} values that are not included in \code{object$lambda}.  
}
  \item{type}{
Type of prediction to return. "link" gives predictions on the link scale (\eqn{\eta}). "response" gives predictions on the mean scale (\eqn{\mu}). "class" gives predicted class labels (for "binomial" and "multinomial" families). "terms" gives the predictions for each term (group) in the model (\eqn{\eta_k}). "importance" gives the variable importance index for each term (group) in the model. "coefficients" returns the coefficients used for predictions. "nonzero" returns a list giving the indices of non-zero coefficients for each \code{s}. "groups" returns a list giving the labels of non-zero groups for each \code{s}. "ncoefs" returns the number of non-zero coefficients for each \code{s}. "ngroups" returns the number of non-zero groups for each \code{s}. "norm" returns the L2 norm of each group's (raw) coefficients for each \code{s}. "znorm" returns the L2 norm of each group's standardized coefficients for each \code{s}.
}
  \item{...}{
Additional arguments (ignored)
}
}
\details{
When \code{type == "link"}, the predictions for each \eqn{\lambda} have the form
\deqn{\boldsymbol\eta_\lambda = \mathbf{X}_{\mathrm{new}} \boldsymbol\beta_\lambda}
where \eqn{\mathbf{X}_{\mathrm{new}}} is the argument \code{newx} (or the design matrix created from \code{newdata} by applying \code{object$formula}) and \eqn{\boldsymbol\beta_\lambda} is the coefficient vector corresponding to \eqn{\lambda}.

When \code{type == "response"}, the predictions for each \eqn{\lambda} have the form
\deqn{\boldsymbol\mu_\lambda = g^{-1}(\boldsymbol\eta_\lambda) }
where \eqn{g^{-1}(\cdot)} is the inverse link function stored in \code{object$family$linkinv}.

When \code{type == "class"}, the predictions for each \eqn{\lambda} have the form
\deqn{\mathbf{y}_\lambda = \arg\max_l \boldsymbol\mu_\lambda(l) }
where \eqn{\boldsymbol\mu_\lambda(l)} gives the predicted probability that each observation belongs to the \eqn{l}-th category (for \eqn{l = 1,\ldots,m}) using the regularization parameter \eqn{\lambda}. 

When \code{type == "terms"}, the groupwise predictions for each \eqn{\lambda} have the form
\deqn{\boldsymbol\eta_{k\lambda} = \mathbf{X}_k^{\mathrm{(new)}} \boldsymbol\beta_{k\lambda}} where \eqn{\mathbf{X}_k^{\mathrm{(new)}}} is the portion of the argument \code{newx} (or the design matrix created from \code{newdata} by applying \code{object$formula}) that corresponds to the \eqn{k}-th term/group, and \eqn{\boldsymbol\beta_{k\lambda}} are the corresponding coefficients.

When \code{type == "importance"}, the variable importance indices are defined as 
\deqn{\pi_k = \left( \boldsymbol\eta_{k\lambda}^\top \mathbf{C} \boldsymbol\eta_{0\lambda} \right) \left( \boldsymbol\eta_{0\lambda}^\top \mathbf{C} \boldsymbol\eta_{0\lambda} \right)^{-1}}
where \eqn{\mathbf{C} = (\mathbf{I}_n - \frac{1}{n} \mathbf{1}_n \mathbf{1}_n^\top)} denotes a centering matrix, and \eqn{\boldsymbol\eta_{0\lambda} = \sum_{k=1}^K \boldsymbol\eta_{k\lambda}}. Note that \eqn{\sum_{k=1}^K \pi_k = 1}, but some \eqn{\pi_k} could be negative. When they are positive, \eqn{\pi_k} gives the approximate proportion of model (explained) variation that is attributed to the \eqn{k}-th term.
}
\value{
Depends on three factors... \cr
1. the exponential family distribution \cr
2. the length of the input \code{s} \cr
3. the \code{type} of prediction requested

For most response variables, the typical output will be...
\item{*}{a matrix of dimension \code{c(newnobs, length(s))} if \code{length(s) > 1}}
\item{*}{a vector of length \code{newnobs} if length(s) == 1}

For multigaussian and multinomial response variables, the typical output will be... 
\item{*}{an array of dimension \code{c(newnobs, length(object$ylev), length(s))} if \code{type \%in\% c("link", "response")}}
\item{*}{a matrix of dimension \code{c(newobs, length(s))} if \code{type == "class"}}

Note: if \code{type == "class"}, then the output will be the same class as \code{object$ylev}. Otherwise, the output will be real-valued (or integer for the counts).

If \code{type == "terms"} and \code{!(family \%in\% c("multigaussian","multinomial"))}, the output will be...
\item{*}{an array of dimension \code{c(newnobs, nterms, length(s))} if \code{length(s) > 1}}
\item{*}{a matrix of dimension \code{c(newnobs, nterms)} if \code{length(s) == 1}}

If \code{type == "terms"} and \code{family \%in\% c("multigaussian","multinomial")}, the output will be a list of length \code{length(object$ylev)} where each element gives the terms for the corresponding response dimension/class.

If \code{type == "importance"} and \code{family != "multinomial"}, the output will be...
\item{*}{a matrix of dimension \code{c(nterms, length(s))} if \code{length(s) > 1}}
\item{*}{a vector of length \code{nterms} if \code{length(s) == 1}}

If \code{type == "importance"} and \code{family == "multinomial"}, the output will be a list of length \code{length(object$ylev)} where each element gives the importance for the corresponding response class. If \code{length(s) == 1}, the output will be simplified to matrix.

If \code{type == "coefficients"}, the output will be the same as that produced by \code{\link{coef.grpnet}}.

If \code{type == "nonzero"}, the output will be a list of length \code{length(s)} where each element is a vector of integers (indices).

If \code{type == "groups"}, the output will be a list of length \code{length(s)} where each element is a vector of characters (\code{term.labels}).

If \code{type \%in\% c("ncoefs", "ngroups")}, the output will be a vector of length \code{length(s)} where each element is an integer.

If \code{type == "norm"}, the output will be a matrix of dimension \code{c(K, length(s))}, where each cell gives the L2 norm for the corresponding group and smoothing parameter. Note that \code{K} denotes the number of groups.

}
\references{
Friedman, J.,  Hastie, T., & Tibshirani, R. (2010). Regularization paths for generalized linear models via coordinate descent. \emph{Journal of Statistical Software, 33}(1), 1-22. \doi{10.18637/jss.v033.i01}

Helwig, N. E. (2025). Versatile descent algorithms for group regularization and variable selection in generalized linear models. \emph{Journal of Computational and Graphical Statistics, 34}(1), 239-252. \doi{10.1080/10618600.2024.2362232}
}
\author{
Nathaniel E. Helwig <helwig@umn.edu>
}
\note{
Some internal code (e.g., used for the interpolation) is borrowed from the \code{predict.glmnet} function in the \bold{glmnet} package (Friedman, Hastie, & Tibshirani, 2010).
}

%% ~Make other sections like Warning with \section{Warning }{....} ~

\seealso{
\code{\link{grpnet}} for fitting grpnet regularization paths

\code{\link{predict.cv.grpnet}} for predicting from \code{\link{cv.grpnet}} objects
}
\examples{
######***######   family = "gaussian"   ######***######

# load data
data(auto)

# fit model (formula method, response = mpg)
mod <- grpnet(mpg ~ ., data = auto)

# get fitted values for regularization path (output = 392 x 100 matrix)
fit.path <- predict(mod, newdata = auto)

# get fitted values at 3 particular points (output = 392 x 3 matrix)
lam <- quantile(mod$lambda, probs = c(0.25, 0.5, 0.75))
fit.some <- predict(mod, newdata = auto, s = lam)

# compare rmse for solutions
rmse.path <- sqrt(colMeans((auto$mpg - fit.path)^2))
rmse.some <- sqrt(colMeans((auto$mpg - fit.some)^2))
plot(log(mod$lambda), rmse.path, cex = 0.5)
points(log(lam), rmse.some, pch = 0, col = "red")


\donttest{
######***######   family = "multigaussian"   ######***######

# load data
data(auto)

# fit model (formula method, response = (mpg, displacement))
y <- as.matrix(auto[,c(1,3)])
mod <- grpnet(y ~ ., data = auto[,-c(1,3)], family = "multigaussian",
              standardize.response = TRUE)

# get fitted values for regularization path (output = 392 x 2 x 100 array)
fit.path <- predict(mod, newdata = auto)

# get fitted values at 3 particular points (output = 392 x 2 x 3 array)
lam <- quantile(mod$lambda, probs = c(0.25, 0.5, 0.75))
fit.some <- predict(mod, newdata = auto, s = lam)

# compare rmse for solutions (mpg)
rmse1.path <- sqrt(colMeans((y[,1] - fit.path[,1,])^2))
rmse1.some <- sqrt(colMeans((y[,1] - fit.some[,1,])^2))
plot(log(mod$lambda), rmse1.path, cex = 0.5)
points(log(lam), rmse1.some, pch = 0, col = "red")

# compare rmse for solutions (displacement)
rmse2.path <- sqrt(colMeans((y[,2] - fit.path[,2,])^2))
rmse2.some <- sqrt(colMeans((y[,2] - fit.some[,2,])^2))
plot(log(mod$lambda), rmse2.path, cex = 0.5)
points(log(lam), rmse2.some, pch = 0, col = "red")



######***######   family = "svm1"   ######***######

# load data
data(auto)

# redefine origin (Domestic vs Foreign)
auto$origin <- ifelse(auto$origin == "American", "Domestic", "Foreign")

# fit model (formula method, response = origin with 2 levels)
mod <- grpnet(origin ~ ., data = auto, family = "svm1")

# get predicted classes for regularization path (output = 392 x 100 matrix)
fit.path <- predict(mod, newdata = auto, type = "class")

# get predicted classes at 3 particular points (output = 392 x 3 matrix)
lam <- quantile(mod$lambda, probs = c(0.25, 0.5, 0.75))
fit.some <- predict(mod, newdata = auto, type = "class", s = lam)

# compare misclassification rate for solutions
miss.path <- 1 - colMeans(auto$origin == fit.path)
miss.some <- 1 - colMeans(auto$origin == fit.some)
plot(log(mod$lambda), miss.path, cex = 0.5)
points(log(lam), miss.some, pch = 0, col = "red")



######***######   family = "svm2"   ######***######

# load data
data(auto)

# redefine origin (Domestic vs Foreign)
auto$origin <- ifelse(auto$origin == "American", "Domestic", "Foreign")

# fit model (formula method, response = origin with 2 levels)
mod <- grpnet(origin ~ ., data = auto, family = "svm2")

# get predicted classes for regularization path (output = 392 x 100 matrix)
fit.path <- predict(mod, newdata = auto, type = "class")

# get predicted classes at 3 particular points (output = 392 x 3 matrix)
lam <- quantile(mod$lambda, probs = c(0.25, 0.5, 0.75))
fit.some <- predict(mod, newdata = auto, type = "class", s = lam)

# compare misclassification rate for solutions
miss.path <- 1 - colMeans(auto$origin == fit.path)
miss.some <- 1 - colMeans(auto$origin == fit.some)
plot(log(mod$lambda), miss.path, cex = 0.5)
points(log(lam), miss.some, pch = 0, col = "red")



######***######   family = "logit"   ######***######

# load data
data(auto)

# redefine origin (Domestic vs Foreign)
auto$origin <- ifelse(auto$origin == "American", "Domestic", "Foreign")

# fit model (formula method, response = origin with 2 levels)
mod <- grpnet(origin ~ ., data = auto, family = "logit")

# get predicted classes for regularization path (output = 392 x 100 matrix)
fit.path <- predict(mod, newdata = auto, type = "class")

# get predicted classes at 3 particular points (output = 392 x 3 matrix)
lam <- quantile(mod$lambda, probs = c(0.25, 0.5, 0.75))
fit.some <- predict(mod, newdata = auto, type = "class", s = lam)

# compare misclassification rate for solutions
miss.path <- 1 - colMeans(auto$origin == fit.path)
miss.some <- 1 - colMeans(auto$origin == fit.some)
plot(log(mod$lambda), miss.path, cex = 0.5)
points(log(lam), miss.some, pch = 0, col = "red")



######***######   family = "binomial"   ######***######

# load data
data(auto)

# redefine origin (Domestic vs Foreign)
auto$origin <- ifelse(auto$origin == "American", "Domestic", "Foreign")

# fit model (formula method, response = origin with 2 levels)
mod <- grpnet(origin ~ ., data = auto, family = "binomial")

# get predicted classes for regularization path (output = 392 x 100 matrix)
fit.path <- predict(mod, newdata = auto, type = "class")

# get predicted classes at 3 particular points (output = 392 x 3 matrix)
lam <- quantile(mod$lambda, probs = c(0.25, 0.5, 0.75))
fit.some <- predict(mod, newdata = auto, type = "class", s = lam)

# compare misclassification rate for solutions
miss.path <- 1 - colMeans(auto$origin == fit.path)
miss.some <- 1 - colMeans(auto$origin == fit.some)
plot(log(mod$lambda), miss.path, cex = 0.5)
points(log(lam), miss.some, pch = 0, col = "red")



######***######   family = "multinomial"   ######***######

# load data
data(auto)

# fit model (formula method, response = origin with 3 levels)
mod <- grpnet(origin ~ ., data = auto, family = "multinomial")

# get predicted classes for regularization path (output = 392 x 100 matrix)
fit.path <- predict(mod, newdata = auto, type = "class")

# get predicted classes at 3 particular points (output = 392 x 3 matrix)
lam <- quantile(mod$lambda, probs = c(0.25, 0.5, 0.75))
fit.some <- predict(mod, newdata = auto, type = "class", s = lam)

# compare misclassification rate for solutions
miss.path <- 1 - colMeans(auto$origin == fit.path)
miss.some <- 1 - colMeans(auto$origin == fit.some)
plot(log(mod$lambda), miss.path, cex = 0.5)
points(log(lam), miss.some, pch = 0, col = "red")



######***######   family = "ordinal"   ######***######

# load data
data(auto)

# fit model (formula method, response = cylinders with 5 levels)
mod <- grpnet(cylinders ~ ., data = auto, family = "ordinal")

# get fitted values for regularization path (output = 392 x 100 matrix)
fit.path <- predict(mod, newdata = auto, type = "class")

# get fitted values at 3 particular points (output = 392 x 3 matrix)
lam <- quantile(mod$lambda, probs = c(0.25, 0.5, 0.75))
fit.some <- predict(mod, newdata = auto, type = "class", s = lam)

# compare misclassification rate for solutions
miss.path <- 1 - colMeans(auto$cylinders == fit.path)
miss.some <- 1 - colMeans(auto$cylinders == fit.some)
plot(log(mod$lambda), miss.path, cex = 0.5)
points(log(lam), miss.some, pch = 0, col = "red")



######***######   family = "poisson"   ######***######

# load data
data(auto)

# fit model (formula method, response = horsepower)
mod <- grpnet(horsepower ~ ., data = auto, family = "poisson")

# get fitted values for regularization path (output = 392 x 100 matrix)
fit.path <- predict(mod, newdata = auto, type = "response")

# get fitted values at 3 particular points (output = 392 x 3 matrix)
lam <- quantile(mod$lambda, probs = c(0.25, 0.5, 0.75))
fit.some <- predict(mod, newdata = auto, type = "response", s = lam)

# compare rmse for solutions
rmse.path <- sqrt(colMeans((auto$horsepower - fit.path)^2))
rmse.some <- sqrt(colMeans((auto$horsepower - fit.some)^2))
plot(log(mod$lambda), rmse.path, cex = 0.5)
points(log(lam), rmse.some, pch = 0, col = "red")



######***######   family = "negative.binomial"   ######***######

# load data
data(auto)

# fit model (formula method, response = horsepower)
mod <- grpnet(horsepower ~ ., data = auto, family = "negative.binomial")

# get fitted values for regularization path (output = 392 x 100 matrix)
fit.path <- predict(mod, newdata = auto, type = "response")

# get fitted values at 3 particular points (output = 392 x 3 matrix)
lam <- quantile(mod$lambda, probs = c(0.25, 0.5, 0.75))
fit.some <- predict(mod, newdata = auto, type = "response", s = lam)

# compare rmse for solutions
rmse.path <- sqrt(colMeans((auto$horsepower - fit.path)^2))
rmse.some <- sqrt(colMeans((auto$horsepower - fit.some)^2))
plot(log(mod$lambda), rmse.path, cex = 0.5)
points(log(lam), rmse.some, pch = 0, col = "red")



######***######   family = "Gamma"   ######***######

# load data
data(auto)

# fit model (formula method, response = mpg)
mod <- grpnet(mpg ~ ., data = auto, family = "Gamma")

# get fitted values for regularization path (output = 392 x 100 matrix)
fit.path <- predict(mod, newdata = auto, type = "response")

# get fitted values at 3 particular points (output = 392 x 3 matrix)
lam <- quantile(mod$lambda, probs = c(0.25, 0.5, 0.75))
fit.some <- predict(mod, newdata = auto, type = "response", s = lam)

# compare rmse for solutions
rmse.path <- sqrt(colMeans((auto$mpg - fit.path)^2))
rmse.some <- sqrt(colMeans((auto$mpg - fit.some)^2))
plot(log(mod$lambda), rmse.path, cex = 0.5)
points(log(lam), rmse.some, pch = 0, col = "red")



######***######   family = "inverse.gaussian"   ######***######

# load data
data(auto)

# fit model (formula method, response = mpg)
mod <- grpnet(mpg ~ ., data = auto, family = "inverse.gaussian")

# get fitted values for regularization path (output = 392 x 100 matrix)
fit.path <- predict(mod, newdata = auto, type = "response")

# get fitted values at 3 particular points (output = 392 x 3 matrix)
lam <- quantile(mod$lambda, probs = c(0.25, 0.5, 0.75))
fit.some <- predict(mod, newdata = auto, type = "response", s = lam)

# compare rmse for solutions
rmse.path <- sqrt(colMeans((auto$mpg - fit.path)^2))
rmse.some <- sqrt(colMeans((auto$mpg - fit.some)^2))
plot(log(mod$lambda), rmse.path, cex = 0.5)
points(log(lam), rmse.some, pch = 0, col = "red")
}
}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory (show via RShowDoc("KEYWORDS")):
\keyword{ regression }
% Use only one keyword per line.
% For non-standard keywords, use \concept instead of \keyword:
% \concept{ ~cpt1 }
% \concept{ ~cpt2 }
% Use only one concept per line.
